\section{Pattern matching}\label{pattern}

\subsection{Simple example}

\Problem{%
     The character string
$\w{x[1]..x[n]}$ is given.
Check if it contains the substring
\w{abcd}.}%
 \index{search!of a substring}%
 \index{pattern matching}

   \Solution. There are approximately \w{n} (\w{n-3}, to be
exact) positions where a substring of length $4$ may be found.
For each position, we can check whether the substring is in that
position. This would require approximately $4n$ comparisons.

However, there is a more efficient
approach. While reading the string $\w{x[1]..x[n]}$ from
left to right, we are looking for the character \w{a}. After it
appears, we look for the character \w{b} (immediately after
\w{a}), then for \w{c}, and finally \w{d}. If our expectations
are met, the substring \w{abcd} is found. If one of the letters
is not found where expected, we start from scratch looking for
\w{a} again.

     This simple algorithm can be described in different terms.
In the framework of so-called {\it finite automata}, we
 \index{finite automaton}%
 \index{automaton!finite}%
say that while scanning \w{x} from left to right
the algorithm is
in
one of the following ``states'': the initial state~(0), the state
``immediately after \w{a}''~(1), ``immediately after
\w{ab}''~(2), ``immediately after \w{abc}''~(3) and ``immediately
after \w{abcd}''~(4). When reading the next character, we change the
state according to the following rule:
$$
\begin{tabular}{|c|c|c|}
  \hline
          Current   &    Next       &   New \\
           state    &   character   &   state \\
  \hline
          0         &   \w{a}       &    1 \\
          0         &   except \w{a}&    0 \\
          1         &   \w{b}       &    2 \\
          1         &   \w{a}       &    1 \\
          1         &   except \w{a},\w{b}  &   0 \\
          2         &  \w{c}        &    3 \\
          2         &  \w{a}        &    1 \\
          2         &   except \w{a},\w{c}  &   0 \\
          3         &  \w{d}        &    4 \\
          3         &  \w{a}        &    1  \\
          3         &   except \w{a},\w{d}  & 0  \\
\hline
\end{tabular}
$$

As soon as we come to state 4, or the input string
is exhausted,
the search is complete.

%$$\diagram
% \mathrm{start}\rto
% & \squarify<1em>{0} \circled \xto `r+<0ex,-1ex>
%        `[0,0] `[0,0]^<{{}\ne a}|(.6)\tip [0,0] \rto^a
% & \squarify<1em>{1} \circled \xto `r+<0ex,1ex>
%        `[0,0]^a `[0,0]|(.6)\tip [0,0] \rto^>>b \xto `r+<0ex,-1ex>
%        `[0,-1] `[0,-1]^(.4){{}\ne a,b}|(.45)\tip [0,-1]
% & \squarify<1em>{2} \circled \xto `r+<0ex,+1ex>
%        `[0,-1]^{a} `[0,-1]|(.45)\tip [0,-1] \rto^>>c
%         \xto `r+<0ex,-1ex> `[0,-2] `[0,-2]^(.25){{}\ne a,c}|(.25)\tip [0,-2]
% & \squarify<1em>{3} \circled
%         \xto `r+<0ex,+1ex> `[0,-2]^{a} `[0,-2]|(.25)\tip [0,-2] \rto^>>d
%         \xto `r+<0ex,-1ex> `[0,-3] `[0,-3]^(.18){{}\ne a,d}|(.18)\tip [0,-3]
% & \squarify<1em>{4} \Circled
%\enddiagram$$

     The corresponding program is straightforward (we indicate the new
state even if it coincides with the old one, but
those lines may be omitted):
\begin{verbatim}
        i:=1; state:=0;
        {i is the index of the first unread character;
         state is the current state}
\end{verbatim}
\clearpage
\begin{verbatim}
        while (i <> n+1) and (state <> 4) do begin
        | if state = 0 then begin
        | | if x[i] = a then begin
        | | | state:= 1;
        | | end else begin
        | | | state:= 0;
        | | end;
        | end else if state = 1 then begin
        | | if x[i] = b then begin
        | | | state:= 2;
        | | end else if x[i] = a then begin
        | | | state:= 1;
        | | end else begin
        | | | state:= 0;
        | | end;
        | end else if state = 2 then begin
        | | if x[i] = c then begin
        | | | state:= 3;
        | | end else if x[i] = a then begin
        | | | state:= 1;
        | | end else begin
        | | | state:= 0;
        | | end;
        | end else if state = 3 then begin
        | | if x[i] = d then begin
        | | | state:= 4;
        | | end else if x[i] = a then begin
        | | | state:= 1;
        | | end else begin
        | | | state:= 0;
        | | end;
        | end;
        end;
        answer := (state = 4);
\end{verbatim}

     In other words, at any point we have information about
the maximal prefix of the pattern \w{abcd} which is a suffix
of the substring already read. (Its length is the value of
the variable \w{state}.)\es

     Let us recall the terminology used.  A {\it string\/} is
 \index{string}%
 \index{alphabet}%
 \index{prefix}%
 \index{suffix}%
 \index{substring}%
an arbitrary finite sequence of elements of a set
called an {\it alphabet}; its elements are called
{\it letters}. If
we discard some letters at the end of a string, we get another
string, which is called a {\it prefix\/} of the first string.
Any string is a prefix of itself. The {\it suffix\/} of a string
is what remains after several initial letters of a string are
discarded. Every string is a suffix of itself. A {\it
substring\/} is obtained when we discard some letters both at the
beginning and the end of a string. (In other words, substrings
are prefixes of suffixes, as well as suffixes of prefixes.)

     In terms of ``inductive functions'' (see section~\ref{indfunc})
we can describe the situation as follows: Consider a function whose
arguments are strings and whose values are either
``True'' or ``False''. The function
has value ``True'' for all strings
containing
substring
\w{abcd}. This function is not inductive, but it does have
an inductive
extension:
$$
  x \mapsto
\hbox{the length of the maximal prefix of \w{abcd} that is a
 suffix of $x$}
$$


