\subsection{Repetitions in the pattern}

\Problem{%
Can the previous algorithm be used for any other
string instead of \w{abcd}?}

     \Solution. A problem arises when the pattern contains repetitions.
For example, suppose we are looking for substring
\w{ababc}. Assume that \w{a} appears, then \w{b},
\w{a}, and \w{b} again. At this point, we are eagerly waiting for
\w{c}.
If the letter \w{d} appears instead, we should start from
scratch. However, if the letter \w{a} appears, we still have
a chance that \w{b} and \w{c} follow and the pattern is found.

Here is an illustration:
{\setlength{\tabcolsep}{.8\tabcolsep}
$$\begin{tabular}{llllllllllllll}
\w x&\w y&\w z&\w a&\w b&\w a&\w b&\w a&\w b&\w c&
                         \ldots&$\leftarrow$&input string\\
   &   &   &\w a&\w b&\w a&\w b&\w c& & & &$\leftarrow$&pattern was expected
                                                 here\\
   &   &   &   &   &\w a&\w b&\w a&\w b&\w c& &$\leftarrow$&but it is
                                                           here\\
\end{tabular}$$}
In other words,
at the point
{\setlength{\tabcolsep}{.8\tabcolsep}
$$\begin{tabular}{lllllll|lllllll}
\w x&\w y&\w z&\w a&\w b&\w a&\w b&    &    &    &
                               &$\leftarrow$&input string\\
   &   &   &\w a&\w b&\w a&\w b&\w c& & & &$\leftarrow$&pattern was expected
                                                here\\
   &   &   &   &   &\w a&\w b&\w a&\w b&\w c& &$\leftarrow$&but it is
                                                 here\\
\end{tabular}$$}
\noindent
there are two possible pattern positions to be tested.\es

However,
a finite automaton that reads the input string letter-by-letter,
changes its state according to some table, and says (after the
input string is exhausted) whether the input string
contains the given substring, is
still possible.

\Problem{%
     Construct such an automaton. Show all its
states and  the  transition
table (which determines
a
new state as a function of an old state and an input
character).}

     \Solution. As before, the current
state is the length of the maximal
prefix of the pattern that is also a suffix of the currently read
part of the string. There are six states: 0, 1~(\w{a}),
2~(\w{ab}), 3~(\w{aba}), 4~(\w{abab}), 5~(\w{ababc}). The
transition table is as follows:
%
%                               
%                                   
%         0                a             1 (a)
%         0                   a         0
%         1 (a)            b             2 (ab)
%         1 (a)            a             1 (a)
%         1 (a)               a,b       0
%         2 (ab)           a             3 (aba)
%         2 (ab)              a         0
%         3 (aba)          b             4 (abab)
%         3 (aba)          a             1 (a)
%         3 (aba)             a,b       0
%         4 (abab)         c             5 (ababc)
%         4 (abab)         a             3 (aba)
%         4 (abab)            a,c       0
$$
\begin{tabular}{|l|l|l|}
  \hline
          Current   &        Next   &   New   \\
           state    &    character  &  state \\
  \hline
          0         &   \w{a}       &    1 (\w{a})  \\
          0         &    except \w{a}&    0      \\
          1 (\w{a})     &   \w{b}       &    2 (\w{ab}) \\
          1 (\w{a})     &   \w{a}       &    1 (\w{a})  \\
          1 (\w{a})     & except \w{a},\w{b}  &  0 \\
          2 (\w{ab})    &  \w{a}        &    3  (\w{aba}) \\
          2 (\w{ab})    & except \w{a}  &    0      \\
          3 (\w{aba})   &  \w{b}        &    4  (\w{abab}) \\
          3 (\w{aba})   &  \w{a}        &    1  (\w{a})  \\
          3 (\w{aba})   &  except \w{a},\w{b}  & 0  \\
          4 (\w{abab})   &  \w{c}        &    5  (\w{ababc}) \\
          4 (\w{abab})   &  \w{a}        &    3  (\w{aba})  \\
          4 (\w{abab})   &  except \w{a},\w{c}  & 0  \\
\hline
\end{tabular}
$$
Consider the second (from below) line in this table as an
example. If the processed part is ended by \w{abab} and the next
letter is \w{a}, the new processed part is ended by \w{ababa}.
The maximal prefix of \w{ababc}, which is also a suffix
of the processed part,
is
\w{aba}.\es

\smallskip

     {\it Question\/}: as we said, the
difficulty appears because there are several possible positions
of the pattern; each position corresponds to some prefix of the
pattern that is also a suffix of the input string. The finite
automaton remembers only the longest one. What about others?

    {\it Answer}. The longest prefix-suffix $X$
determines all other prefix-suffixes. Namely, prefix-suffixes of
the processed part
are prefixes
of $X$ that are also suffixes of $X$.

\smallskip

    It is easy to write a transition table and a program for any
fixed pattern. However, we want
to write a general program that will
search for any given pattern in any given input string.
The following approach may be used.
Consider a program that has two stages.
In the first stage, it examines
the pattern and constructs a transition table for that pattern.
In the second stage, it reads the input string and behaves according
to the transition table. Such an approach is often used for more
complicated patterns (see below), but for substring search there
is more direct and efficient method called
Knuth--Morris--Pratt algorithm.
 \index{search!of a substring}%
 \index{substring!search}%
 \index{Knuth--Morris--Pratt algorithm}%
 \index{Knuth, D.E.}%
 \index{Morris, J.H.}%
 \index{Pratt V.R.}%
(A similar approach was suggested by Yu.~Matijasevich.)%
 \index{Matijasevich, Yu.V.}
We start with some auxiliary lemmas.

