\subsection{Auxiliary lemmas}

     For any string $X$, consider all the prefixes of $X$ that are
at the same time suffixes of $X$. Choose the longest one (not counting
$X$ itself), which is denoted by $l(X)$.

   For example,  $l(\w{aba})\,\,=\,\,\w{a}$, $l(\w{abab})\,\,=\,\,\w{ab}$,
$l(\w{ababa})\,\,=\w{aba}$,
$l(\w{abc})\,\,=\,\,\hbox{\rm the empty string}$.

\Problem{%
Prove that all strings $l(X)$, $l(l(X))$, $l(l(l(X)))$, etc.\
are prefixes of~$X$.}

    \Solution. Each of them is a prefix of the preceding one.
(And any prefix of a prefix of $X$ is also a prefix of $X$.)\es

    For the same reason all such strings are suffixes of $X$ as well.

\Problem{%
Prove that the sequence in the preceding problem is finite
(the last string is empty).}

   \Solution.   Each   subsequent  string  is  shorter  than  the
preceding one (since $l(Y)$ is shorter than $Y$ for any $Y$).\es

\Problem{%
    Prove that any string that is both a prefix and a  suffix
of $X$ (except for $X$ itself) is listed in
the sequence $l(X), l(l(X)),\ldots$\,.}

     \Solution. Let $Y$ be both
a prefix of $X$ and a
suffix of $X$. The string
$l(X)$ is the longest string having this property,
so $Y$ is not longer than $l(X)$. Both $Y$ and $l(X)$
are prefixes of $X$, and the shorter one is a prefix of the
longer one. Thus $Y$ is a prefix of $l(X)$.
For the same reason, $Y$ is
a suffix of $l(X)$. Using an induction argument, we assume that
the statement in question is true for all
strings
shorter than $X$. In particular, it is true for
$l(X)$. So the string
$Y$, being a prefix and a suffix
of the string
$l(X)$, is either equal to $l(X)$ or one of the
strings
$l(l(X)), l(l(l(X))),\dots$\,.\nobreak\es

\subsection{Knuth--Morris--Pratt algorithm}

    The Knuth--Morris--Pratt (KMP) algorithm takes a string
$$
     \w{X}= \w{x[1]x[2]..x[n]}
$$
as input and scans it from left to right.
The output is the sequence of nonnegative integers
$\w{L[1]}\ldots\w{L[n]}$ such that
$$
   \w{L[i]} = \hbox{\rm the length of }l(\w{x[1]..x[i]})
$$
(the function $L$ is defined in the preceding section). In other
words, \w{L[i]} is the length of the maximal prefix of
$\w{x[1]..x[i]}$ that is simultaneously a
suffix of
$\w{x[1]..x[i]}$.

\Problem{%
How can  we  use the
KMP  algorithm  to  check  whether  a  given  string  \w{A} is a
substring of a string \w{B}?}

    \Solution. Apply the
KMP algorithm to the string \w{A\#B}, where
\w{\#} is a special character that does not appear in \w{A} or
\w{B}. The string \w{A} is a substring of \w{B} if
and only if the array \w{L} (which is the output of the KMP
algorithm) contains a number equal to the length of \w{A}.\es

\Problem{%
     How do we fill the table $\w{L[1]..L[n]}$?}

    \Solution. Assume that the first \w{i} values
$\w{L[1]..L[i]}$ are already known. We read the next
input character (i.e.,~\w{x[i+1]}) and compute
\w{L[i+1]}.

\medskip

$$
\newlength{\picwidth}
\setlength{\picwidth}{.8\textwidth}
\newlength{\rightwidth}
\setbox0\hbox{\w{i+1}}
\setlength{\rightwidth}{\wd0}
\newlength{\leftwidth}
\setlength{\leftwidth}{\picwidth}
\addtolength{\leftwidth}{-\rightwidth}
\vbox{%
  \hbox to\picwidth{\w{ 1}\hfill \w{i }\w{i+1}}%
  \kern.4ex%
  \hrule
  \hbox{\strut\vrule
       \hbox to\leftwidth{\hfil processed part of \w{x}\hfil}%
      \kern-0.4pt\vrule\kern\rightwidth\vrule
  }%
  \hrule
  \kern0.1ex%
  \hbox to\picwidth{%
     $\underbrace{\hspace*{.36\picwidth}}_Z$\hfill
     $\underbrace{\hspace*{.36\picwidth}}_Z$%
  }%
}%
$$

\medskip

\noindent
How do we find \w{L[i+1]}? It is the length of the longest
prefix $Z$ of the string $\w{x[1]..x[i+1]}$ that is at
the same time a suffix of this string. Any string $Z$ having
this property (except for the empty string) is obtained from
some string $Z'$ by adding the letter \w{x[i+1]}. The string
$Z'$ is both a prefix and a suffix of the string
$\w{x[1]..x[i]}$. However, it is not the only
requirement for $Z'$; another requirement is that $Z'$ is
followed (as a prefix of $\w{x[1]..x[i]}$) by
\w{x[i+1]}.

Therefore, the string $Z$ may be found as follows. Consider all
the prefixes $Z'$ of the string $\w{x[1]}\ldots\w{x[i]}$ that
are also the suffixes of this string. Then choose the longest
one that is followed (as a prefix of $\w{x[1]}\ldots\w{x[i]}$)
by $\w{x[i+1]}$. Adding \w{x[i+1]} produces the string $Z$.

Now it is time to use the lemmas proved earlier.
Recall that all
strings that are both prefixes and suffixes may be obtained by
applying the function $l$ iteratively. Here is the program:
 %
\begin{Verbatim}
    i:=1; L[1]:= 0;
    {the table L[1]..L[i] is filled correctly}
    while i <> n do begin
    | len := L[i]
    | {len is the length of a prefix  of x[1]..x[i] that is
    |   its suffix; all longer prefixes-suffixes were
    |   tested without success}
    | while (x[len+1] <> x[i+1]) and (len > 0) do begin
    | | {this prefix does not fit also, we should apply l}
    | | len := L[len];
    | end;
    | {we either have found the longest prefix that
    |  fits our requirements (and its length is n)
    |  or have found that it does not exist (len=0)}
    | if x[len+1] = x[i+1] do begin
    | | {x[1]..x[len] is the longest prefix that fits}
    | | L[i+1] := len+1;
    | end else begin
    | | {there are no good prefixes}
    | | L[i+1] := 0;
    | end;
    | i := i+1;
    end;
\end{Verbatim}

\Problem{%
    Prove that the number of operations in the above algorithm
    is limited by $C\w{n}$ for some constant $C$.}

    \Solution. This is not obvious, because one input character
may cause many iterations in the inner loop. However, each
iteration in the inner loop decreases \w{len} by at least
\w{1}, so in this case, \w{L[i+1]} will be significantly smaller
than \w{L[i]}. On the other hand, while \w{i} is increased by
\w{1}, the value of \w{L[i]} may increase  by at most \w{1},
therefore the values of \w{i} that require many iterations in
the inner loop are rare.

     Formally, we use the inequality
$$
\w{L[i+1]} \le \w{L[i]} - \hbox{(the number of iteration at
step \w{i})} + \w{1}
$$
or
$$
 \hbox{(the number of iterations at step
  \w{i})}\le\w{L[i]}-\w{L[i+1]} + \w{1}
$$
Summing these inequalities over \w{i}, we get the
required
upper bound for the total number of iterations.\es

\Problem{%
    Imagine that we use this algorithm to determine
whether a string
\w{X} of length \w{n} is a substring of a string \w{Y} of length
\w{m}. (We explained above how to do that using a
``separator'' \w{\#}.) The algorithm runs in time
$O(\w{n}+\w{m})$ and uses memory of size $O(\w{n}+\w{m})$. Find
a way to do this using memory of size $O(\w{n})$ (which may be
significantly less if the pattern is short and the string is
long).}%
 \index{search!of a substring}
 \index{substring!search}

    \Solution. Start applying the KMP algorithm to the string
$\w{A\#B}$. Wait until the algorithm computes all the values
$\w{L[1]},\ldots,\w{L[n]}$ for the word \w{X} of length \w{n}.
All those values are stored. From then on, we keep only the value
\w{L[i]} for the current \w{i}; we only need
\w{L[i]} and the table $\w{L[1]..L[n]}$ to compute
\w{L[i+1]}.\es

    In practice, the words \w{X} and \w{Y} are usually separated, so
the scan of \w{X} and the scan of \w{Y}
should be implemented
as two
different loops. (This also makes the separator
\w{\#} unnecessary.)

\Problem{%
    Write the program discussed in the last paragraph, which
checks whether the string $\w{X}=\w{x[1]..x[n]}$ is a substring
of the string $\w{Y}=\w{y[1]..y[m]}$.}

    \Solution. First we fill the table $\w{L[1]}\ldots\w{L[n]}$
as before. Then we execute the following program:
 %
\begin{verbatim}
   j:=0; len:=0;
   {len is the length of a longest prefix  of X which is
     a suffix of y[1]..y[j]}
   while (len <> n) and (j <> m) do begin
   | while (x[len+1] <> y[j+1]) and (len > 0) do begin
   | | {this prefix does not fit}
   | | len := L[len];
   | end;
\end{verbatim}
\clearpage
\begin{Verbatim}
   | {we have found the prefix that fits or
   |     have found that it does not exist}
   | if x[len+1] = y[j+1] do begin
   | | {x[1]..x[len] is the longest prefix that fits}
   | | len := len+1;
   | end else begin
   | | {no prefixes fit}
   | | len := 0;
   | end;
   | j := j+1;
   end;
   {if len=n, X is a substring of Y;
     otherwise we reached the end of Y not finding X}
\end{Verbatim}

