\subsection{Boyer--Moore algorithm}

 \index{Boyer--Moore algorithm}
 \index{search!of a substring}
 \index{substring!search}
     This algorithm attains a goal that seems impossible
at first: In a typical situation, it reads only a tiny fraction
of all the characters of a string in which the pattern is searched.
How can this be done? The idea is rather simple. Suppose we
are searching
for the pattern \w{abcd} in a string $X$. Check
the fourth character of $X$. If it is, say, \w{e}, there is no
need to look at the first three characters, because our pattern
does not contain \w{e} and may start
only after
fourth position.

     We show below
a simplified version of the Boyer--Moore algorithm that does
not guarantee good running time in all cases.

     Let $\w{X}=\w{x[1]..x[n]}$
be the pattern we are searching
for. For any character \w{s}, we find the rightmost occurrence of
\w{s} in the string \w{X}, that is, the maximal \w{k} such that
$\w{x[k]}=\w{s}$. This information is stored in an array
\w{pos[s]}. If the character \w{s} does not appear in the
pattern at all, it is convenient to put
$\w{pos[s]}\,\w{:=}\,\w{0}$ (see below).

\Problem{%
     How do we fill the array \w{pos}?
}

     \Solution.
\begin{Verbatim}
        ...let all pos[s] be equal to 0
        for i:=1 to n do begin
        | pos[x[i]]:=i;
        end;
\end{Verbatim}

\medskip

The program searches for the pattern \w{x[1]..x[n]} in the
input string \w{y[1]..y[m]}.
When searching, store in the variable \w{last} the number of
the input
character that corresponds to the last character of the
pattern (in the current pattern position). Initially, $\w{last}
= \w{n}$ (the length of the pattern); then \w{last} increases
gradually.
\begin{verbatim}
    last:=n;
    {all previous positions of the pattern are checked}
    while last <= m do begin {the work is not finished}
    | if x[n] <> y[last] then begin
    | | {the last characters differ}
    | | last := last + (n - pos[y[last]]);
    | | {n - pos[y[last]] is the minimal shift of the
    | |    pattern that makes the character y[last]
    | |    match the corresponding character in the
    | |    pattern. If y[last] does not appear in the
    | |    pattern, the new pattern position starts
    | |    immediately after y[last]}
    | end else begin
    | | {x[n] = y[last]}
    | | check if the current position is okay, that is,
    | | if x[1]..x[n] = y[last-n+1]..y[last].
    | | If yes, inform about that.
    | | last := last+1;
    | end;
    end;
\end{verbatim}
It is recommended to start testing the condition
($\w{x[1]..x[n]} = \w{y[last-n+1]..y[last]}$) from right to left
starting from the last position (where the coincidence is already
tested). We also obtain a small optimization if we store
\w{n-pos[s]} instead of \w{pos[s]} (avoiding subtraction at each
step); \w{n-pos[s]} is the number of characters to the right of
the rightmost occurrence of character \w{s} in the pattern.

     Different versions of this algorithm exist. For example,
we may replace the line \w{last:=last+1} by
\w{last:=last+(n-u)}, where \w{u} is the position of the second
(from the right) occurrence of the character \w{x[n]} in the
pattern.

\Problem{%
     What modifications in the program are necessary?
}

    \Solution. To fill up the table \w{pos}, we use the
line
\begin{verbatim}
        for i:=1 to n-1 do...
\end{verbatim}
(all other lines remain the same);
in the main program
we replace the line
\w{last:=last+1} by
\begin{Verbatim}
        last:= last+n-pos[y[last]];
\end{Verbatim}

     This simplified version of the
Boyer--Moore algorithm sometimes
require significantly more that \w{n} operations (\w{mn} in the
worst case), so the worst-case behavior of the
Knuth--Morris--Pratt algorithm is much better.


\Problem{%
     Give an example where a pattern of length \w{n}
is not a substring of a given string of length \w{m}, but the
program above requires \w{mn} operations to determine this.}

    \Solution. Assume that the pattern is $\w{baaa..aa}$
and the string contains \w{n} letters \w{a}. Then at each step
we need \w{n} comparisons to discover that the pattern is not a
substring.\es

    The complete (not simplified) Boyer--Moore algorithm guarantees
that the number of operations does not exceed $C(\w{m}+\w{n})$
in the worst case. It uses ideas similar to those in the KMP
algorithm. Suppose we compare the pattern and the string
from right to left. Assume that we find the
coincident suffix $Z$ of the pattern, but the characters before
$Z$ in the input string and in the pattern are
different. What do we know about the input string at that point?
We have found a fragment equal to $Z$ that is preceded by a character
that differs from the character in the pattern. This information
may allow us to shift the pattern to the right several positions.
These shifts should be computed in advance for all suffixes
$Z$ of the pattern. One can prove that all
operations (the computation and use of the shift table)
can be performed in time
$C(\w m+\w n)$.

