\subsection{Rabin--Karp algorithm}

 \index{Rabin--Karp algorithm}
 \index{search!of a substring}
 \index{pattern matching}
 \index{substring!search}
     This algorithm is also based on a simple idea.  Suppose we
are looking  for  a  pattern  of  length  \w{n} in a string of length
\w{m}. Let us make a sliding window and move it along the input
string. Our goal is to check whether the substring in the window
coincides with the given pattern.

  We want to avoid character-by-character comparison and find a
faster method. Let us consider some function defined on strings of
length $n$. If this function takes on  different values when
applied to both the pattern
and
the substring in the window, we may
be sure that there is no match. Only if the function values
coincide, we have to compare strings character-by-character.

     What do we gain? It seems that we have
achieved nothing because
to compute the function value for the substring in the window,
we have to read all the characters in the window anyway. So why
not just compare them with the pattern characters?
Some gain, however,
is still possible for the following reason. When
we shift the window, the substring in it does not change
completely; a single character is
appended
on the right and
deleted on the left. If our function is well chosen, we may
compute its new value quickly, knowing its old value and the
added/deleted characters.

\Problem{%
     Find an example of such a ``well chosen'' function.
}

    \Solution. Replace all characters
in the pattern by their
codes, which are assumed to be integers. The sum of all codes is
such a function. (Indeed, after the shift, we only
have to add the
numeric value of the new character and subtract the numeric
value of the old character.)\es

\smallskip

     Given any function, most likely
there are distinct strings that are mapped to the same value.
For the same pair of strings
another function may indeed produce distinct values.
So let us have a pool of functions and
begin the algorithm by
choosing one of the functions at random. Then
an adversary who wants to choose the worst problem instance will
not know which function it is working against.

\Problem{%
     Give an example of a family of easily computable functions
(in the
sense explained above).\es
}

    \Solution. Let us choose some number $p$ (presumably prime;
see below) and some residue $x$ modulo $p$. Each string of
length $n$
is considered
as a sequence of integers (characters are
replaced by their numeric codes).
Those integers are taken to be
coefficients of a polynomial of degree $n-1$. We compute the
value of this polynomial modulo $p$ at the point $x$. This
construction provides one function of the family (for
each $p$ and $x$ we get another function). When the window is shifted
by $1$, we subtract the term of the highest degree
($x^{n-1}$ should be computed in advance), multiply by $x$, and
add the constant term.

\smallskip

     The following arguments show that the coincidence
of function
values (for different arguments) is not very
likely. Assume that $p$ is fixed and is prime. Let $X$ and $Y$
be two different words of length $n$. Then the corresponding
polynomials are different. (We assume that different characters
have different codes modulo $p$, so we need $p$ to be larger
than the size of the alphabet.) The coincidence of function
values on $X$ and $Y$ means that two different polynomials
coincide at $x$, that is, $x$ is a root of their difference.
This difference is a nonzero polynomial of degree $n-1$ and
can have at most $n-1$ roots. Therefore, if $n$ is much
smaller than $p$, the chances for the random $x$ to be a root
are negligible.

