\subsection{Automata and more complicated patterns}

   Rather than a specific string, we may search
for a string of
some type. For example, we may look for a substring of type
\w{a?b} where \w{?} denotes any single character. In other
words, we are looking for characters \w{a} and \w{b}
with
exactly
one character in between.%
 \index{pattern matching}

\Problem{%
    Construct a finite automaton that checks if the pattern
\w{a?b} is present in the input string.}
 \index{automaton!finite}%
 \index{finite automaton}

     \Solution. While reading the input string, the automaton
keeps track of whether the character \w{a} is
present at the two last positions.
The automaton has states 00,
01, 10, 11 with the following meanings:
 %
$$
\begin{tabular}{rp{.7\textwidth}}
00& no \w{a} in the last two positions\\
01& \leavevmode \w{a} is in the last position but not
           in the position immediately before it \\
10& \leavevmode \w{a} is in the position before the last one
          but not in the last position\\
11& the processed part of the input string ends with \w{aa}
\end{tabular}
$$
 %
Here is the transition table:
 %
\newcommand\bigstrut{{\large\strut}}
$$ \begin{tabular}{|c|c|c|} \hline
    Current     &   Next         &    New   \bigstrut  \\
    state       &  character     &    state \bigstrut  \\
\hline
       00       &          \w a  &              01 \bigstrut \\
       00       &      not \w  a &              00  \\
       01       &          \w a   &             11  \\
       01       &       not \w a  &             10  \\
       10       &         \w  a   &             01  \\
       10       &         \w  b   &             found  \\
       10       & not \w{a} and not \w{b}  &        00  \\
       11       &       \w    a        &        11      \\
       11       &       \w    b         &       found \\
       11       &  not \w a and not \w b   &       10 \bigstrut \\
\hline
\end{tabular}
$$
\vskip-1ex
$$\eqno\es$$



     Another widely used notation
in a pattern
is an asterisk (\w{*}), which is matched
by any string (including the empty string).
For example, the
pattern \w{ab*cd} means that we are looking for any occurrence of
\w{ab} followed by \w{cd} (the distance between \w{ab}
and \w{cd} is arbitrary).

\Problem{%
     Construct a finite automaton that checks if the input
string contains the pattern \w{ab*cd} (in the sense just
described). }

     \Solution.
$$
\begin{tabular}{|c|c|c|}
\hline
    Current     &    Next       &  New  \bigstrut   \\
    state       &  character    &  state  \bigstrut \\
\hline
    initial   &       \w a    &            \w{a}  \bigstrut  \\
    initial   &   not \w a    &            initial  \\
\w      a     & \w          b &            \w{ab}     \\
\w      a     &         \w  a &            \w{a}      \\
\w      a     & not \w a and not \w b &    initial  \\
     \w{ab}     &          \w{c}       &     \w{abc}    \\
     \w{ab}     &     not \w{c}       &     \w{ab}     \\
     \w{abc}    &          \w{d}       &      found \\
     \w{abc}    &          \w{c}       &     \w{abc}    \\
     \w{abc}    & not \w{c} and not \w{d}  &   \w{ab} \bigstrut \\
\hline
\end{tabular}
$$
\vskip-1ex
$$\eqno\es$$

    Another type of search occurs when we are looking for a
substring that belongs to a given finite set of strings.

\Problem{%
    Assume that strings $X_1,\ldots,X_k$ (patterns) and a string
$Y$ are given. Check if one of the strings $X_i$ is a substring
of the string $Y$. The number of operations should not exceed
the total length of all the strings ($X_i$ and $Y$) multiplied
by some constant which does not depend on $k$.}%
 \index{search!one of substrings}

    \Solution. The obvious approach is to check all the $X_i$
separately (using one of the algorithms given above).
However, this method does not satisfy the speed requirements
(since we have to read the string $Y$ many, in fact, $k$ times).

    Let us look at another aspect of the problem. For each
pattern $X_i$, there exists a finite automaton that tests for the
presence of $X_i$. These automata may be combined into one
automaton whose set of states is the product of the sets of
states for all the automata. This set is very large. However, most
of its elements are unreachable and may be discarded.

This idea is used below (in a modified form).

     Let us recall the Knuth--Morris--Pratt algorithm. While
reading the input string, the KMP algorithm keeps the maximal
prefix of the pattern that is a suffix of the processed part of
the input string.
Now we
need
to keep this information (the longest prefix that is a
suffix of the processed part) for all the patterns. The crucial
remark is: It is enough to keep the longest one, because all others
are uniquely determined by the longest one. Indeed, let $X$ be the
longest prefix of some pattern that is a suffix of the processed
part of the input string. Then for any pattern $P$, the longest prefix
of $P$ being a suffix of the processed part is the longest prefix of
$P$ being a suffix of $X$.

    All the patterns may be ``glued'' together to form
a tree if we ``splice'' together
 equal prefixes. For example, the set of patterns
$$
     \{\w{aaa}, \w{aab}, \w{abab}\}
$$
corresponds to the tree
%
%                      a/ *
%          a     a    / b
%       * --- * --- * --- *
%               \b     a     b
%                 \ * --- * --- *
%
{\setlength{\unitlength}{.07\textwidth}
$$\begin{picture}(10,4)
  \thicklines
  \put(1,2){\circle*{.15}}
  \put(1,2){\vector(1,0){2}}
  \put(1,2){\makebox(2,.7){\w{a}}}

  \put(3,2){\circle*{.15}}
  \put(3,2){\vector(1,0){2}}
  \put(3,2){\vector(2,-1){2}}
  \put(3,2){\makebox(2,.7){\w{a}}}
  \put(3,1){\makebox(2,0){\w{b}}}

  \put(5,2){\circle*{.15}}
  \put(5,2){\vector(1,0){2}}
  \put(5,2){\vector(2,1){2}}
  \put(5,3){\makebox(2,0){\w{a}}}

  \put(5,1){\circle*{.15}}
  \put(5,1){\vector(1,0){2}}
  \put(5,1.3){\makebox(2,.7){\w{b}}}
  \put(5,0.3){\makebox(2,.7){\w{a}}}

  \put(7,1){\circle*{.15}}
  \put(7,1){\vector(1,0){2}}
  \put(7,0.3){\makebox(2,.7){\w{b}}}

  \put(7,2){\circle*{.15}}
  \put(7,3){\circle*{.15}}
  \put(9,1){\circle*{.15}}
\end{picture}$$}%
Here is the formal definition: any prefix of any pattern is a
tree vertex; a father of a vertex is obtained by deleting the
last character.

While reading the input string, we traverse this tree. The
current position is the maximal (rightmost) vertex that is
a suffix of the processed part of the input string (that is, the
longest suffix of the processed part being a prefix of one of
the patterns).

Let us introduce a function $l$ whose arguments and values are
tree vertices, namely, $l(P) = {}$maximal tree vertex that is a
(proper) suffix of $P$. (Recall that tree vertices are strings.)
The following result will be used:

\Problem{%
    Let $P$ be a tree vertex. Prove that the set of all tree vertices
that are (proper) suffixes of $P$ is $\{l(P),
l(l(P)),\ldots\}$}

     \Solution. See the proof of the similar assertion for the
Knuth--Morris--Pratt algorithm.\es

\smallskip

    Now it is clear what the algorithm (or automaton) should do
if it is at the vertex $P$ and the next input character is
\w{z}: It should consider sequentially the vertices $P$, $l(P)$,
$l(l(P))$,\,\dots\ until it finds the vertex that has an outgoing
(to the right) edge labeled ``\w{z}''. The endpoint of that edge
is the next position of the algorithm (next state of the
automaton).

     It remains to show how to compute the values of the
function $l$ for all tree vertices. This is done as before using
the values of $l$ for shorter strings to compute the next value
of $l$. Therefore, we should consider all tree vertices in order
of increasing length. It is easy to see that this can be
done in the required time. (Please note that the constant in the
upper bound for the running time depends on the cardinality of
the alphabet.) For a discussion of
the methods used to store the tree, see
section~\ref{graphs}.\es

\smallskip

     The general question arises: Which properties of strings
can be tested using finite automata? It turns out that there is
an easily defined class of patterns that correspond to finite
automata. These patterns are called ``regular expressions''.

     {\bf Definition.}
Let $\Gamma$ be a finite alphabet.
We assume that $\Gamma$ does not contain six symbols
$\Lambda$, $\varepsilon$, \w{(}, \w{)},
\w{*} and \w{|} (these symbols will be used for constructing regular
expressions; therefore,
we should not mix them with letters from $\Gamma$).
{\it Regular expressions\/} are constructed according to the
following rules:
 \index{regular expression}%
 \index{expression!regular}
\begin{description}
\item[(a)] any letter from $\Gamma$ is a regular expression;
\item[(b)] the symbols $\Lambda$, $\varepsilon$ are regular expressions;
\item[(c)] if $A,B,C,\ldots,E$ are regular expressions,  then
          $\w{(}ABC\ldots E\w{)}$ is a regular expression;
\item[(d)] if $A,B,C,\ldots,E$ are regular expressions, then
         $\w{(}A\w{|}B\w{|}C\w{|}\ldots\w{|}E\w{)}$ is a regular
         expression;
\item[(e)] if $A$ is a regular expression, then $A\w{*}$ is a
         regular expression.
\end{description}
Each regular expression defines a set of strings (composed of
characters from $\Gamma$) according to the following rules:
\begin{description}
\item[(a)] \,A letter corresponds to a singleton whose
       element is
        a one-character string containing this letter;
\item[(b)] The symbol $\varepsilon$ corresponds to the empty set; the
    symbol $\Lambda$ corresponds to the singleton whose element
    is the empty string;
\item[(c)] the regular expression $\w{(}ABC\ldots E\w{)}$
    corresponds to the set of all strings obtained
    as follows: take a string from the set that corresponds to
    $A$, a string from the set that corresponds to $B$, to
    $C$, \ldots, and to $E$ and concatenate all those strings in the
    given order ({\it concatenation\/} of sets);
\item[(d)] the regular expression
         $\w{(}A\w{|}B\w{|}C\w{|}\ldots\w{|}E\w{)}$ corresponds to
         the union of the sets that correspond to expressions
         $A,B,C,\ldots,E$;
\item[(e)] the regular expression $A\w{*}$ corresponds to the {\it
         iteration\/} of a set corresponding to $A$, that is, to
       the set of
        all strings that may be cut into pieces in such a way that
        each piece belongs to the set corresponding to $A$.
        (In particular, the set corresponding to
         $A\w{*}$ always contains the empty string.)
\end{description}

   Sets that correspond to regular expressions are called
{\it regular\/} sets. Here are several examples:
 \index{set!regular}
 \index{regular set}
%
$$\begin{tabular}{lp{0.5\textwidth}}
{\bf Expression} &  {\bf Set}      \\[1ex]
%
\w{(a|b)*}      &      All strings composed of \w{a} and \w{b}\\
\w{(aa)*}       &      All strings of even length composed of \w{a}s,
                       including the empty string\\
\w{(}$\Lambda$\w{|a|b|aa|ab|ba|bb)}&
   all strings of length at most $2$ composed of \w{a} and \w{b}
\end{tabular}$$

\Problem{%
Find a regular expression corresponding to the set of all
strings composed of \w{a} and \w{b} that contain an even number
of \w{a}s.}

   \Solution. The expression \w{b*} defines the set of all
strings without \w{a}; the expression
$\w{(b*}\,\w{a}\,\w{b*}\,\w{a}\,\w{b*)}$ defines the set of all
words with exactly two \w{a}s. It remains to take the union of
these two sets and then to apply iteration:
 %
$$\w{(}\,\w{(b*}\,\w{a}\,\w{b*}\,\w{a}\,\w{b*)}\,|\,\w{b*}\,\w{)*}$$
 %
Another possible answer:
 %
$$\w{(}\,\w{(b*}\,\w{a}\,\w{b*}\,\w{a)}\w{*}\,\w{b*}\,\w{)}
\eqno\es$$

\Problem{%
     Write a regular expression that defines a set of
strings composed of
$\w{a},\w{b},\w{c}$ having \w{bac}
as a substring.}

     \Solution. $\w{((a|b|c)*}\,\w{bac}\,\w{(a|b|c)*)}$\es

Remark.
A more difficult problem is to write the expression for the
{\it complement\/} of this set, that is, the set of all strings
composed of $\w{a},\w{b},\w{c}$ that do {\it not\/} have
\w{bac} as a substring. This is possible, however,
as we'll see below.


\smallskip

    Now the general pattern-matching problem may be stated as follows:
check whether an input string belongs to the set corresponding
to a given regular expression.

\Problem{%
     What regular
expressions are equivalent to the patterns \w{a?b}
and \w{ab*cd} used as examples earlier?
(Please note that
the symbol \w{*} in the pattern \w{ab*cd} has a completely different
meaning compared to its use in regular expressions.)
We assume
that the alphabet is $\{\w{a},\w{b},\w{c},\w{d},\w{e}\}$.}

     \Solution.
$$ \w{((a|b|c|d|e)*}\,\w{a}\,\w{(a|b|c|d|e)}\,\w{b}\,\w{(a|b|c|d|e)*)}$$
$$ \w{((a|b|c|d|e)*}\,\w{ab}\,
\w{(a|b|c|d|e)*}\,\w{cd}\,\w{(a|b|c|d|e)*)}\eqno\es
$$

\Problem{%
     Prove that for any regular expression there exists a finite
automaton that recognizes the corresponding set of strings.}

    \Solution. To prove this, we need the notion of a {\it
nondeterministic finite automaton}.%
 \index{nondeterministic finite automaton}%
 \index{finite automaton!nondeterministic}%
 \index{automaton!finite nondeterministic}
Consider a directed graph containing several
points (vertices) and some arrows (edges) connecting those points.
Assume that some of the
edges are labeled by letters (from a given
alphabet) and some edges remain unlabeled. Assume also that two
vertices are selected; one is called the {\it initial\/} vertex
I and the other is called the {\it final} vertex F. Such a
labeled graph is called a nondeterministic finite automaton.

    Let us consider all the paths from I to F. Going along a
path, we read all
the
letters (on labeled edges). Therefore, each
path from I to F determines a string. The automaton as a
whole determines a set of strings, namely, the set of all strings
that can be read along some path from I to F. We say that these
strings are {\it accepted\/} by the automaton.

    Remark. If we draw the states of a finite automaton as
points and the transitions as labeled edges, it is clear that
finite automata are special cases of nondeterministic finite
automata. They are distinguished by the following requirements:
(a)~all edges are labeled except for the edges directed to the
final vertex; (b)~for each vertex and for each letter there is
exactly one outgoing edge labeled by this letter.

    We transform a regular expression into a finite automaton in
two stages. First, we construct a nondeterministic finite
automaton that corresponds to the same set. Then for any
nondeterministic finite automaton we construct an equivalent
deterministic finite automaton.

\Problem{%
    A regular expression is given. Construct a nondeterministic
finite automaton that corresponds to the same set.}

   \Solution. This automaton is constructed inductively,
following the definition of a regular expression. If the regular
expression is a letter or $\varepsilon$, the corresponding
automaton has one edge. If the regular expression is $\Lambda$,
the automaton has no edges at all. A union is implemented as
follows:
$${\setlength{\unitlength}{.032\textwidth}%
\thicklines
\begin{picture}(24,12)
\put(3,6){\circle*{.3}}
\put(3,6){\vector(1,0){6}}
\put(3,6){\vector(3,2){6}}
\put(3,6){\vector(3,-2){6}}
\put(1,5){\makebox{I}}
%
\put(9,2){\circle*{.3}}
\put(8,1){\framebox(8,2){}}
\put(9,1){\makebox(2,2){$\r{I}_3$}}
%
\put(9,6){\circle*{.3}}
\put(8,5){\framebox(8,2){}}
\put(9,5){\makebox(2,2){$\r{I}_2$}}
%
\put(9,10){\circle*{.3}}
\put(8,9){\framebox(8,2){}}
\put(9,9){\makebox(2,2){$\r{I}_1$}}
%
\put(15,2){\circle*{.3}}
\put(15,2){\line(3,2){6}}
\put(15,2){\vector(3,2){4}}
\put(13,1){\makebox(2,2){$\r{F}_3$}}
%
\put(15,6){\circle*{.3}}
\put(15,6){\line(1,0){6}}
\put(15,6){\vector(1,0){4}}
\put(13,5){\makebox(2,2){$\r{F}_2$}}
%
\put(15,10){\circle*{.3}}
\put(15,10){\line(3,-2){6}}
\put(15,10){\vector(3,-2){4}}
\put(13,9){\makebox(2,2){$\r{F}_1$}}
%
\put(21,6){\circle*{.3}}
\put(21,5){\makebox{F}}
\end{picture}}$$
%
Here the picture for the union of three sets is drawn. The
rectangles show the corresponding nondeterministic finite
automata; their initial and final vertices are shown. New arrows
(there are six of them) are unlabeled.

Concatenation corresponds to the following picture:
$${\setlength{\unitlength}{.032\textwidth}%
\thicklines
\begin{picture}(26,4)
\put(2,2){\circle*{.3}}
\put(2,2){\vector(1,0){2}}
\put(0,1){\makebox(2,2){I}}
%
\put(4,2){\circle*{.3}}
\put(3,1){\framebox(6,2){}}
\put(4,1){\makebox(2,2){$\r{I}_1$}}
\put(6,1){\makebox(2,2){$\r{F}_1$}}
\put(8,2){\circle*{.3}}
\put(8,2){\vector(1,0){3}}
%
\put(11,2){\circle*{.3}}
\put(10,1){\framebox(6,2){}}
\put(11,1){\makebox(2,2){$\r{I}_2$}}
\put(13,1){\makebox(2,2){$\r{F}_2$}}
\put(15,2){\circle*{.3}}
\put(15,2){\vector(1,0){3}}
%
\put(18,2){\circle*{.3}}
\put(17,1){\framebox(6,2){}}
\put(18,1){\makebox(2,2){$\r{I}_3$}}
\put(20,1){\makebox(2,2){$\r{F}_3$}}
\put(22,2){\circle*{.3}}
\put(22,2){\vector(1,0){2}}
%
\put(24,2){\circle*{.3}}
\put(24,1){\makebox(2,2){F}}
\end{picture}}$$
%
     Finally, iteration corresponds to the picture
%
$${\setlength{\unitlength}{.04\textwidth}%
\begin{picture}(8,6)
\thicklines
\put(2,2){\circle*{.3}}
\put(1,1){\framebox(6,2){}}
%
\put(2,5){\circle*{.3}}
\put(2,5){\vector(1,0){2}}
\put(0,4){\makebox(2,2){I}}
\put(2,1){\makebox(2,2){$\r{I}_1$}}
%
\put(4,5){\circle*{.3}}
\put(4,5){\vector(1,0){2}}
\put(4,5){\vector(-2,-3){2}}
\put(6,2){\vector(-2,3){2}}
\put(4,1){\makebox(2,2){$\r{F}_1$}}
%
\put(6,2){\circle*{.3}}
\put(6,5){\circle*{.3}}
\put(6,4){\makebox(2,2){F}}
%
\end{picture}}%
\eqno\es
$$

\Problem{%
     A nondeterministic finite automaton $N$ is given.
Construct an equivalent
deterministic finite automaton (or a program
with a finite number of states) that checks if an input string
$x$ is accepted by $N$ (that is, if $x$
can be read on a path from I to F).}%
 \index{determinization}

    \Solution. The states of the deterministic automaton are
{\it sets\/} of vertices of the nondeterministic automaton.
After a prefix $X$ of the input string is read, the state $s(X)$ of the
deterministic automaton is
the set of all vertices that are reachable from I
along paths carrying the string $X$ on it. In other words,
consider all paths starting from I. Each path determines a string
that can be read along it. If the string is $X$, include the
end of the path into $s(X)$.\es

\smallskip

The two-stage construction of a finite automaton corresponding
to a given regular expression, is finished.\es

\medskip

     It turns out that regular expressions, deterministic
finite automata, and nondeterministic finite automata define the
same class of sets. To prove this, it remains to solve the
following problem:

\Problem{%
     A nondeterministic finite automaton is given. Construct a
regular expression that defines the same set.}

    \Solution. Assume that the nondeterministic automaton has
vertices $1,\ldots,k$, where $1$ is its initial vertex and $k$ is
its final vertex. By $D(i,j,s)$ we denote the set of all strings
read along all the paths from $i$ to $j$ if only
$1,2,\ldots,s$ are allowed as intermediate path vertices.
By definition, the automaton itself corresponds to the
set $D(1,k,k)$.

     We prove by induction over $s$ that all sets
$D(i,j,s)$ for all $i$ and $j$ are regular. For $s=0$, this is obvious
(intermediate vertices are not permitted, therefore
each set is a finite set whose elements are strings
of length not exceeding $1$).

    Which strings are elements of $D(i,j,s+1)$? Let us consider
a path from $i$ to $j$ and mark all the steps when it enters the
$(s+1)$-th vertex. The marked steps split our path into several
paths that do not use $s+1$ as an intermediate vertex. This
argument leads to the equation
 %
$$
 D(i,j,s+1) =
   D(i,j,s)\,|\, (D(i,s+1,s)\; D(s+1,s+1,s)\w{*}\; D(s+1,j,s))
$$
 %
(here the notation for regular expressions is used for sets).
It remains to apply the induction assumption.\es

\Problem{%
     Where have you seen a similar argument?}%
 \index{Floyd algorithm}

   \Solution. In the Floyd algorithm for the shortest path (see
section~\ref{graphs}, p.~\pageref{floyd}).\nobreak\es

\Problem{%
     Prove that the class of sets corresponding to regular
expressions remains the same if we agree to use not only set
union but also complementation (and therefore set
intersection, since it can be expressed using set
union and complement).}%
 \index{regular set!complement}%
 \index{regular set!intersection}%
 \index{complement!of a regular set}%
 \index{intersection!of regular sets}

    \Solution. For the deterministic finite automata the
transition from a set to its complement is evident.\es

\smallskip

     Remark. From a practical point of view, things are not so easy.
The problem is that the transition from a nondeterministic
automaton to a deterministic one
may exponentially increase the number of states.
There are many theoretical and
practical questions concerning this problem. See the book of
Aho, Sethi, and Ullman on compilers [2].
 \index{Aho, A.V.}%
 \index{Ullman, J.D.}%
 \index{Sethi, R.}

