\section[Sets, trees, and balanced trees]%
{Sets, trees, and balanced trees}\label{trees}

\newcommand{\Tree}{{\mathrm{Tree}}}

\subsection{Set representation using trees}

\subsubsection*{Full binary trees and $T$-trees}

 \index{set!tree representation}
     Draw a point. Now draw two arrows going up-left and
up-right to two other points. From those two points also draw
two arrows, etc. The resulting tree is called a
{\it full binary tree}
(the $n$-th level has $2^{n -
1}$ points).
The initial
point (at the bottom of the tree) is called the {\it root}. Each
vertex has two
 \index{tree!full binary}
{\it sons\/} (arrows point to them), the {\it left\/}
son and the {\it right\/} son. Each vertex (except for the root)
has an unique {\it father}.

Please note that many textbooks draw trees with the root at the
top and also use words ``child'' (``parent'', ``siebling'', etc.)
instead of ``son'' (``father'', ``brother'', etc.).

     Now choose some subset of the set of all vertices of the
full binary tree. It should satisfy the following
requirement: for each vertex of the subset, its father
belongs
to the subset, too.
(Therefore, all vertices on a
path
from
the root to some vertex from the subset belong to the
subset.)
Assume that each vertex in the subset has a label that
is an element of some set $T$.
(In other words, we assume that
a mapping from the subset into the set $T$
is given.) Such a subset with labels from $T$
is called a $T$-{\it tree}. The
set of all $T$-trees is denoted by $\Tree(T)$.

   The notion of $T$-tree may be defined recursively.
Any nonempty $T$-tree is divided into three parts: the root
(which carries a label from $T$),
the left sub-tree, and the right subtree
(one or both of which
may be empty). Therefore, there is an
one-to-one
correspondence between the set of nonempty $T$-trees and the product
$T \times \Tree (T) \times \Tree (T)$.
We get the following equality:
\[
   \Tree(T) =\{\hbox{\bf empty}\}+T\times\Tree(T)\times\Tree (T).
\]
(here {\bf empty} stands for the empty tree).
\[
\setlength{\unitlength}{1em}
\begin{picture}(8,6)
\put(4,2){\circle*{0.2}}
\put(4,2){\line(-1,1){3}}
\put(4,2){\line(-1,3){1}}
\put(1,5){\line(1,0){2}}
\put(4,2){\line(1,1){3}}
\put(4,2){\line(1,3){1}}
\put(5,5){\line(1,0){2}}
\put(2,1){\makebox(4,1){\r{root}}}
\put(1,5){\makebox(2,1){\r{left}}}
\put(5,5){\makebox(2,1){\r{right}}}
\end{picture}
\]

\vspace*{-5ex}


\subsubsection*{Subtrees and height}

     Assume that some $T$-tree is fixed. For any vertex $x$,
the following objects are defined:
the {\it left subtree\/}
(the left son of $x$ and all its descendants);
the {\it right subtree\/}
(the right son of $x$ and all its descendants);
and the {\it subtree rooted at $x$}
(the vertex $x$ and all its descendants).
 \index{subtree}
The left and right subtrees of
$x$ may be empty, but the subtree rooted at
$x$ may not (it always contains the vertex $x$).
The {\it height\/}
 \index{height}
of a subtree is defined as the maximal length of the sequence
$y_1,\ldots,y_n$ of its vertices where
$y_{i+1}$ is a son of $y_i$ for all $i$, minus one.
(The height of the empty tree is $-1$ by definition;
the height of a tree containing
only the root is $0$.)

\subsubsection*{Ordered $T$-trees}

    Assume that a linear order is defined on the set $T$.
A $T$-tree is {\it ordered\/} if the following requirement is fulfilled:
 \index{ordered tree}%
 \index{tree!ordered}
for any vertex $x$, all labels in its left subtree are less than
the label at $x$ and all labels in its right subtree are greater than the
label at $x$.
$$
\setlength{\unitlength}{1.4em}
\begin{picture}(8,6)
\put(4,2){\circle*{0.2}}
\put(4,2){\line(-1,1){3}}
\put(4,2){\line(-1,3){1}}
\put(1,5){\line(1,0){2}}
\put(4,2){\line(1,1){3}}
\put(4,2){\line(1,3){1}}
\put(5,5){\line(1,0){2}}
\put(2,1){\makebox(4,1){$x$}}
\put(1.8,4){\makebox(1,1){${}<x$}}
\put(5,4){\makebox(1,1){${}>x$}}
\end{picture}
$$

\vspace*{-5ex}

   \Problem{Prove that all labels in an ordered subtree are different.}
   \Hint{Induction over the height of the tree.}

\subsubsection*{Set representation using trees}

Consider any tree as a representation of the set of
labels of its vertices. (Of course, the same set may have
different representations.)

If the set is ordered, each element can easily find its way to a
place in the tree. It starts from the root; coming to a vertex,
an element
compares itself with the label at that vertex and decides
whether to go to the left or to the right.
 %
$$
\setlength{\unitlength}{1.2em}
\begin{picture}(8,7)
\put(4,2){\vector(0,1){1.5}}
\put(3,1){\makebox(2,1){$y$}}
\put(4,4){\makebox(0,0){$x$}}
\put(3.5,4.5){\vector(-1,1){1.5}}
\put(4.5,4.5){\vector(1,1){1.5}}
\put(1,6){\makebox(2,1){$y<x$}}
\put(5,6){\makebox(2,1){$y>x$}}
\end{picture}
$$

\vspace*{-5ex}

\noindent
Using this rule, the element either finds the identical
label already
present in the tree or the place where it should stay to keep the
tree ordered.

In this section, assume that the set $T$ is a linearly ordered
set. All $T$-trees we consider are ordered.

