\subsection{Balanced trees}

     A tree is called {\it balanced\/} (or an  AVL-tree, in the honor of
 \index{tree!balanced}%
 \index{AVL-tree}%
 \index{Adelson-Velsky, G.M.}%
 \index{Landis, E.M.}%
the inventors of this algorithm, G.M.~Adelson-Velsky and
E.M.~Landis) if for any vertex, the heights of the left and the
right subtrees differ by at most $1$. (In particular, the only
son of a vertex is required to be a leaf, since the height of the
other subtree is $-1$.)

     \Problem{Find the minimal and maximal number of
vertices in a balanced tree of height $n$.}

     \Solution. The maximal number of vertices is equal to
$2^{n+1}-1$. If $m_n$ is the minimal number of vertices, then $m_{n
+ 2} = 1 + m_n + m_{n+1}$. An easy induction argument gives
$m_n = \Phi_{n+3} - 1$
 \index{Fibonacci numbers}%
(where $\Phi_n$ is the $n$-th Fibonacci number: $\Phi_1=1$,
$\Phi_2=1$, and $\Phi_{n+2} = \Phi_n + \Phi_{n+1}$).\es

     \Problem{Prove that a balanced tree with $n>1$ vertices has
height at most $C\log n$ for some constant $C$ that does not
depend on $n$.}

     \Solution. By induction over $n$, we prove that $\Phi_{n+2}
\geq a^n$ where $a$ is the larger root of the quadratic equation
$a^2=1+a$, that is, $a = (\sqrt{5} + 1)/2$. (This number is usually
called ``the golden mean''.) It remains to apply the
preceding problem.\es

\subsubsection*{Rotations}

    After an element is added or deleted, the tree may become
unbalanced, and we have to restore the balance. Therefore, we
need some tree transformations that preserve the set
of labels and the ordering requirement, but help
to balance the tree. Here are some of those
transformations:
 %
\[
\setlength{\unitlength}{.044\textwidth}
\begin{picture}(21,7)
\put(4,2){\circle*{.3}}
\put(4,2){\line(2,1){2}}
\put(3,1){\makebox(2,1){$a$}}
\put(6,3){\circle*{.3}}
\put(6,2){\makebox(1,1){$b$}}
\put(15,3){\circle*{.3}}
\put(15,3){\line(2,-1){2}}
\put(14,2){\makebox(1,1){$a$}}
\put(17,2){\circle*{.3}}
\put(16,1){\makebox(2,1){$b$}}
\put(4,2){\line(-1,1){3}}
\put(4,2){\line(-1,3){1}}
\put(1,5){\line(1,0){2}}
\put(2,4){\makebox(1,1){$P$}}
\put(6,3){\line(-1,1){3}}
\put(6,3){\line(-1,3){1}}
\put(3,6){\line(1,0){2}}
\put(4,5){\makebox(1,1){$Q$}}
\put(6,3){\line(1,1){3}}
\put(6,3){\line(1,3){1}}
\put(7,6){\line(1,0){2}}
\put(7,5){\makebox(1,1){$R$}}
\put(10,3){\makebox(1,2){$\rightarrow$}}
\put(15,3){\line(1,1){3}}
\put(15,3){\line(1,3){1}}
\put(16,6){\line(1,0){2}}
\put(16,5){\makebox(1,1){$Q$}}
\put(15,3){\line(-1,1){3}}
\put(15,3){\line(-1,3){1}}
\put(12,6){\line(1,0){2}}
\put(13,5){\makebox(1,1){$P$}}
\put(17,2){\line(1,1){3}}
\put(17,2){\line(1,3){1}}
\put(18,5){\line(1,0){2}}
\put(18,4){\makebox(1,1){$R$}}
\end{picture}
\]
     Assume that a vertex $a$ has a  right son $b$. Let $P$ be the
left subtree of $a$. Let $Q$ and $R$ be the left and right
subtrees of $b$, respectively.
The ordering requirement guarantees that $P < a
< Q < b < R$. (This means that any label in $P$ is smaller than
$a$, that $a$ is smaller than any label in $Q$, etc.) The same
condition is imposed by the ordering requirements for another
tree. The latter tree has root $b$; the left son $a$ of the root
has left subtree $P$ and right subtree $Q$; the right subtree of
the root is $R$. Therefore the first tree may be transformed to
the second one without changing the set of labels or violating
the ordering requirements. This transformation is called a {\it
small right rotation\/}. It is called ``right'' because there is
a symmetric ``left'' rotation; it is called
``small'' because there exists a
``big'' rotation, which we describe now.
 \index{rotation!left, right}
 \index{rotation!small, big}

     Let $b$ be the right son of the root vertex $a$; let $c$ be
the left son of $b$; let $P$ be the left subtree of $a$; let $Q$
and $R$ be the left and the right subtrees of $c$, respectively;
and finally,
let $S$ be the right subtree of $b$. Then $P < a < Q < c < R < b
< S$.
 %
\[
\setlength{\unitlength}{.038\textwidth}
\begin{picture}(25,8)
\put(4,1){\circle*{.3}}
\put(7,3){\circle*{.3}}
\put(5,4){\circle*{.3}}
\put(18,1){\circle*{.3}}
\put(15,3){\circle*{.3}}
\put(21,3){\circle*{.3}}
%
\put(4,1){\line(3,2){3}}
\put(7,3){\line(-2,1){2}}
\put(18,1){\line(3,2){3}}
\put(18,1){\line(-3,2){3}}
%
\put(3,0){\makebox(2,1){$a$}}
\put(7,2){\makebox(1,1){$b$}}
\put(4,3){\makebox(1,1){$c$}}
\put(17,0){\makebox(2,1){$c$}}
\put(14,2){\makebox(1,1){$a$}}
\put(21,2){\makebox(1,1){$b$}}
%
\put(4,1){\line(-1,1){3}}
\put(4,1){\line(-1,3){1}}
\put(1,4){\line(1,0){2}}
\put(2,3){\makebox(1,1){$P$}}
%
\put(5,4){\line(-1,1){3}}
\put(5,4){\line(-1,3){1}}
\put(2,7){\line(1,0){2}}
\put(3,6){\makebox(1,1){$Q$}}
%
\put(5,4){\line(1,1){3}}
\put(5,4){\line(1,3){1}}
\put(6,7){\line(1,0){2}}
\put(6,6){\makebox(1,1){$R$}}
%
\put(7,3){\line(1,1){3}}
\put(7,3){\line(1,3){1}}
\put(8,6){\line(1,0){2}}
\put(8,5){\makebox(1,1){$S$}}
%
\put(10,3){\makebox(2,2){$\rightarrow$}}
%
\put(15,3){\line(-1,1){3}}
\put(15,3){\line(-1,3){1}}
\put(12,6){\line(1,0){2}}
\put(13,5){\makebox(1,1){$P$}}
%
\put(15,3){\line(0,1){3}}
\put(15,3){\line(2,3){2}}
\put(15,6){\line(1,0){2}}
\put(15,5){\makebox(1,1){$Q$}}
%
\put(21,3){\line(0,1){3}}
\put(21,3){\line(-2,3){2}}
\put(19,6){\line(1,0){2}}
\put(20,5){\makebox(1,1){$R$}}
%
\put(21,3){\line(1,3){1}}
\put(21,3){\line(1,1){3}}
\put(22,6){\line(1,0){2}}
\put(22,5){\makebox(1,1){$S$}}
%
\end{picture}
\]
The same ordering conditions are imposed by a tree with root $c$,
its left son $a$ and right son $b$ that have the left and the
right subtrees $P$ and $Q$ (for $a$) and $R$ and $S$ (for $b$).
The corresponding transformation is called a {\it big right
rotation}. (A {\it big left rotation\/} is defined in a
symmetric way.)

\subsubsection*{How to balance a tree using rotations}

     \Problem{Suppose a tree is balanced everywhere except at the root
where the difference of heights between the left and right
subtrees equals $2$ (that is, the left and right subtrees are
balanced and their heights differs by $2$). Prove that this tree
may be transformed into a balanced tree using one of the four
transformations mentioned above and that the height remains the same
or decreases by $1$ after the transformation.}

     \Solution. Assume, for example, that the left subtree has
smaller height, which we denote by $k$. Then the height of the
right subtree is $k+2$. Denote the root of the tree by $a$.
Let $b$ be its right son (it does exist). Consider the left and
right subtrees of the vertex $b$. One of them has height $k+1$,
the other has height $k$ or $k+1$. (Its height cannot be
smaller than $k$ because the right subtree of the root is
balanced.) If the height of the left subtree of $b$ is $k+1$,
and the height of the right subtree of $b$ is $k$, a big right
rotation is needed; in all other cases, a small right rotation
suffices. Here are the three possible cases: \label{threecases}
 %
{\setlength{\unitlength}{.04\textwidth}
$$\begin{picture}(18,6)
\thinlines
\put(4,1){\line(1,1){1}}
\put(13,1){\line(-1,1){1}}
\thicklines
\put(4,0.5){\makebox(1.5,1){$a$}}
\put(4,1){\circle*{.3}}
\put(5,1.5){\makebox(1.5,1){$b$}}
\put(5,2){\circle*{.3}}
\put(13,1){\circle*{.3}}
\put(12,2){\circle*{.3}}
\put(4,1){\line(-1,1){2}}
\put(4,1){\line(-1,2){1}}
\put(2,3){\line(1,0){1}}
\put(5,2){\line(-1,1){2}}
\put(5,2){\line(-1,2){1}}
\put(3,4){\line(1,0){1}}
\put(5,2){\line(1,3){1}}
\put(5,2){\line(2,3){2}}
\put(6,5){\line(1,0){1}}
\put(8,2){\makebox(1,1){$\rightarrow$}}
\put(12,2){\line(-1,1){2}}
\put(12,2){\line(-1,2){1}}
\put(10,4){\line(1,0){1}}
\put(12,2){\line(1,1){2}}
\put(12,2){\line(1,2){1}}
\put(13,4){\line(1,0){1}}
\put(13,1){\line(2,3){2}}
\put(13,1){\line(1,1){3}}
\put(15,4){\line(1,0){1}}
\thinlines
\multiput(1,3)(0.2,0){80}{\line(1,0){.1}}
\multiput(1,4)(0.2,0){80}{\line(1,0){.1}}
\multiput(1,5)(0.2,0){80}{\line(1,0){.1}}
\end{picture}\]
\[\begin{picture}(18,5.5)
\thinlines
\put(4,1){\line(1,1){1}}
\put(13,1){\line(-1,1){1}}
\thicklines
\put(4,1){\circle*{.3}}
\put(4,0.5){\makebox(1.5,1){$a$}}
\put(5,2){\circle*{.3}}
\put(5,1.5){\makebox(1.5,1){$b$}}
\put(13,1){\circle*{.3}}
\put(12,2){\circle*{.3}}
\put(4,1){\line(-1,1){2}}
\put(4,1){\line(-1,2){1}}
\put(2,3){\line(1,0){1}}
\put(5,2){\line(-2,3){2}}
\put(5,2){\line(-1,3){1}}
\put(3,5){\line(1,0){1}}
\put(5,2){\line(1,3){1}}
\put(5,2){\line(2,3){2}}
\put(6,5){\line(1,0){1}}
\put(8,2){\makebox(1,1){$\rightarrow$}}
\put(12,2){\line(-1,1){2}}
\put(12,2){\line(-1,2){1}}
\put(10,4){\line(1,0){1}}
\put(12,2){\line(2,3){2}}
\put(12,2){\line(1,3){1}}
\put(13,5){\line(1,0){1}}
\put(13,1){\line(2,3){2}}
\put(13,1){\line(1,1){3}}
\put(15,4){\line(1,0){1}}
\thinlines
\multiput(1,3)(0.2,0){80}{\line(1,0){.1}}
\multiput(1,4)(0.2,0){80}{\line(1,0){.1}}
\multiput(1,5)(0.2,0){80}{\line(1,0){.1}}
\end{picture}\]
\[\begin{picture}(18,6)
\thinlines
\put(4,1){\line(1,1){1}}
\put(5,2){\line(-1,1){1}}
\put(13,1){\line(-1,1){1}}
\put(13,1){\line(1,1){1}}
\thicklines
\put(4,0.5){\makebox(1.5,1){$a$}}
\put(4,1){\circle*{.3}}
\put(4,3){\circle*{.3}}
\put(5,1.5){\makebox(1.5,1){$b$}}
\put(5,2){\circle*{.3}}
\put(13,1){\circle*{.3}}
\put(12,2){\circle*{.3}}
\put(14,2){\circle*{.3}}
\put(4,1){\line(-1,1){3}}
\put(4,1){\line(-2,3){2}}
\put(1,4){\line(1,0){1}}
\put(5,2){\line(2,3){2}}
\put(5,2){\line(1,1){3}}
\put(7,5){\line(1,0){1}}
\put(4,3){\line(-1,1){2}}
\put(4,3){\line(-1,2){1}}
\put(2,5){\line(1,0){1}}
\put(4,3){\line(1,1){2}}
\put(4,3){\line(1,2){1}}
\put(5,5){\line(1,0){1}}
\thinlines
\put(4,3){\line(-1,1){3}}
\put(4,3){\line(-1,2){1.5}}
\put(1,6){\line(1,0){1.5}}
\put(4,3){\line(1,1){3}}
\put(4,3){\line(1,2){1.5}}
\put(5.5,6){\line(1,0){1.5}}
\put(8,3){\makebox(1,1){$\rightarrow$}}
\thicklines
\put(12,2){\line(-1,1){3}}
\put(12,2){\line(-2,3){2}}
\put(9,5){\line(1,0){1}}
\put(12,2){\line(-1,2){1}}
\put(12,2){\line(0,1){2}}
\put(11,4){\line(1,0){1}}
\put(14,2){\line(1,1){3}}
\put(14,2){\line(2,3){2}}
\put(16,5){\line(1,0){1}}
\put(14,2){\line(1,2){1}}
\put(14,2){\line(0,1){2}}
\put(14,4){\line(1,0){1}}
\thinlines
\put(12,2){\line(-1,2){1.5}}
\put(12,2){\line(0,1){3}}
\put(10.5,5){\line(1,0){1.5}}
\put(14,2){\line(1,2){1.5}}
\put(14,2){\line(0,1){3}}
\put(14,5){\line(1,0){1.5}}
\multiput(0,4)(0.2,0){90}{\line(1,0){.1}}
\multiput(0,5)(0.2,0){90}{\line(1,0){.1}}
\multiput(0,6)(0.2,0){90}{\line(1,0){.1}}
\put(1,5){\makebox(2,1){$?$}}
\put(5,5){\makebox(2,1){$?$}}
\put(11,4){\makebox(1,1){$?$}}
\put(14,4){\makebox(1,1){$?$}}
\end{picture}\eqno\es$$
}

