    \Problem{A leaf is added to or deleted from a balanced
tree. Prove that it is possible to make the tree balanced
again using several rotations and that the number of rotations does
not exceed the tree height.}

    \Solution. We prove the more general statement:

\smallskip

    Lemma. If a subtree $Y$ of a balanced tree $X$ is replaced
by a balanced tree $Z$, and the heights of $Y$ and $Z$ differ by
$1$, then the resulting tree can be made balanced by several
rotations. The number of rotations does not exceed the height
where the change occurs (that is, where the root of $Y$ and $Z$ is
located).

\smallskip

    The addition/deletion of a leaf is a special case of the
transformation mentioned in the lemma, therefore it is enough to
prove this lemma.

    Proof of the lemma. We use induction over the height where
the change is made. If the change is made at the root, the entire
tree is replaced; in this case, the lemma is evident because the
tree $Z$ is balanced. Assume that the replaced tree $Y$ is, say,
the left subtree of some vertex $x$. Two cases are
possible:

\begin{enumerate}
   \item After replacement, the balance condition at the
vertex $x$ is still valid. (However, the balance condition at
the ancestors of $x$ may be violated because the height of the
subtree rooted at $x$ may change.) In this case, we
apply the induction hypothesis assuming that the replacement was
done at the lower level and the whole tree rooted at $x$ was
replaced.
   \item The balance condition at $x$ is no longer valid. In this
case, the height difference is $2$ (it cannot be larger because
the heights of $Y$ and $Z$ differ by at most $1$). Here two
subcases are possible:
 %
\[\setlength{\unitlength}{.04\textwidth}
\begin{picture}(20,6)
\thicklines
\put(4,1){\circle*{.3}}
\put(4,1){\line(1,2){2}}
\put(4,1){\line(1,1){4}}
\put(4,1){\line(-1,1){2}}
\put(4,1){\line(-1,2){1}}
\put(2,3){\line(1,0){1}}
\put(6,5){\line(1,0){2}}
\thinlines
\put(4,1){\line(-1,1){3}}
\put(4,1){\line(-1,2){1.5}}
\put(1,4){\line(1,0){1.5}}
\put(0,3){\makebox(1,2){$Y$}}
\put(0,2){\makebox(2,2){$Z$}}
\put(8,2){\makebox(1,2){$k$}}
\multiput(1,3)(0.2,0){35}{\line(1,0){.1}}
\multiput(1,4)(0.2,0){35}{\line(1,0){.1}}
\multiput(1,5)(0.2,0){35}{\line(1,0){.1}}
\thicklines
\put(14,1){\circle*{.3}}
\put(14,1){\line(1,2){1}}
\put(14,1){\line(1,1){2}}
\put(14,1){\line(-1,2){2}}
\put(14,1){\line(0,1){4}}
\put(15,3){\line(1,0){1}}
\put(12,5){\line(1,0){2}}
\thinlines
\put(12.5,4){\line(1,0){1.5}}
\put(13,3){\makebox(1,1){$Y$}}
\put(12,5){\makebox(2,1){$Z$}}
\put(18,2){\makebox(1,2){$k$}}
\multiput(11,3)(0.2,0){35}{\line(1,0){.1}}
\multiput(11,4)(0.2,0){35}{\line(1,0){.1}}
\multiput(11,5)(0.2,0){35}{\line(1,0){.1}}
\put(3,0){\makebox(2,0.5){\hbox{\rm(a)}}}
\put(13,0){\makebox(2,0.5){\hbox{\rm(b)}}}
\end{picture}
\]
  \begin{enumerate}
    \item The right subtree of $x$ (the one that was not
replaced) is higher. Assume that the height of the left subtree
of $x$ (i.e., $Z$) is $k$; then the height of the right subtree
is $k+2$. The height of the old left subtree of $X$ (i.e., $Y$)
was $k+1$. The subtree of the initial tree rooted at $x$ has
height $k+3$ and its height does not change after
replacement.

  By the preceding problem, a rotation can transform the subtree
rooted at $x$ into a balanced subtree of height $k+2$ or $k+3$.
While doing this, the height of the subtree rooted at $x$
(compared with its height before the transformation) did not
change or was decreased by $1$. Therefore, we apply the
induction assumption.

\item The left subtree of $x$ is higher. Let the height of the
left subtree (i.e., $Z$) be $k+2$; the right subtree has height
$k$. The old left subtree of $x$ (i.e., $Y$) was
of height
$k+1$. The
subtree rooted at $x$ (in the initial tree) has height $k+2$;
after the replacement it has height $k+3$. After a suitable
rotation (see the preceding problem), the subtree rooted at $x$
becomes balanced and its height is $k+2$ or $k+3$; therefore, the
change in height (compared with the height of the subtree of
$X$ rooted at $x$) does not exceed $1$ and the induction
assumption applies.\es
  \end{enumerate}
\end{enumerate}

   \Problem{Write addition and deletion procedures that
keep the
tree balanced. The running time should not exceed
$C\cdot(\hbox{\rm tree height})$. It is allowed to store
additional information (needed for balancing)
at the vertices of the
tree.}

    \Solution. For each vertex we keep the difference between the
heights of its right and left subtrees:
 %
\begin{eqnarray*}
 \w{diff [i]}&=&\hbox{(the height of the right subtree of \w{i})} -\\
     & &\qquad \hbox{(the height of the left subtree of \w{i}).}
\end{eqnarray*}
 %
We need four procedures that correspond to left/right,
small/big rotations. Let us first make two remarks.
 %
     (1) We want to keep the number of the tree root unchanged
during the rotation. (Otherwise it would be necessary to update
the pointer at the father vertex, which is inconvenient.)
This can be done, because the numbers of tree vertices may be
chosen independently of their content. (In our pictures, the
number is drawn near the vertex while the content is drawn
inside it.)
 %
{\newcommand{\lettbox}[3]{%
   \put(#1,#2){\framebox(1,1){$#3$}}%
  }
\[\setlength{\unitlength}{.04\textwidth}
\begin{picture}(18,7)
  \lettbox{3}{1}{a}
  \lettbox{5}{3}{b}
  \lettbox{14}{1}{b}
  \lettbox{12}{3}{a}
  \put(4,2){\line(1,1){1}}
  \put(4,2){\vector(1,1){1}}
  \put(14,2){\line(-1,1){1}}
  \put(14,2){\vector(-1,1){1}}
  \put(3,2){\line(-1,1){2}}
  \put(3,2){\line(0,1){2}}
  \put(1,4){\line(1,0){2}}
  \put(2,3){\makebox(1,1){$P$}}
  \put(5,4){\line(-1,1){2}}
  \put(5,4){\line(0,1){2}}
  \put(3,6){\line(1,0){2}}
  \put(4,5){\makebox(1,1){$Q$}}
  \put(6,4){\line(1,1){2}}
  \put(6,4){\line(0,1){2}}
  \put(6,6){\line(1,0){2}}
  \put(6,5){\makebox(1,1){$R$}}
  \put(4,1){\makebox(1,1){$\alpha$}}
  \put(6,3){\makebox(1,1){$\beta$}}
  \put(8,3){\makebox(2,1){$\rightarrow$}}
  \put(12,4){\line(-1,1){2}}
  \put(12,4){\line(0,1){2}}
  \put(10,6){\line(1,0){2}}
  \put(11,5){\makebox(1,1){$P$}}
  \put(13,4){\line(1,1){2}}
  \put(13,4){\line(0,1){2}}
  \put(13,6){\line(1,0){2}}
  \put(13,5){\makebox(1,1){$Q$}}
  \put(15,2){\line(1,1){2}}
  \put(15,2){\line(0,1){2}}
  \put(15,4){\line(1,0){2}}
  \put(15,3){\makebox(1,1){$R$}}
  \put(15,1){\makebox(1,1){$\alpha$}}
  \put(13,3){\makebox(1,1){$\beta$}}
\end{picture}
\]
\[\setlength{\unitlength}{.04\textwidth}
\begin{picture}(23,9)
  \lettbox{4}{1}{a}
  \lettbox{6}{3}{b}
  \lettbox{4}{5}{c}
  \lettbox{16}{1}{b}
  \lettbox{13}{3}{a}
  \lettbox{19}{3}{c}
  \put(5,2){\line(1,1){1}}
  \put(5,2){\vector(1,1){1}}
  \put(6,4){\line(-1,1){1}}
  \put(6,4){\vector(-1,1){1}}
  \put(16,2){\vector(-2,1){2}}
  \put(17,2){\vector(2,1){2}}
  \put(4,2){\line(-1,1){2}}
  \put(4,2){\line(0,1){2}}
  \put(2,4){\line(1,0){2}}
  \put(3,3){\makebox(1,1){$P$}}
  \put(4,6){\line(-1,1){2}}
  \put(4,6){\line(0,1){2}}
  \put(2,8){\line(1,0){2}}
  \put(3,7){\makebox(1,1){$Q$}}
  \put(5,6){\line(1,1){2}}
  \put(5,6){\line(0,1){2}}
  \put(5,8){\line(1,0){2}}
  \put(5,7){\makebox(1,1){$R$}}
  \put(7,4){\line(1,1){2}}
  \put(7,4){\line(0,1){2}}
  \put(7,6){\line(1,0){2}}
  \put(7,5){\makebox(1,1){$S$}}
  \put(5,1){\makebox(1,1){$\alpha$}}
  \put(7,3){\makebox(1,1){$\beta$}}
  \put(5,5){\makebox(1,1){$\gamma$}}
  \put(9,4){\makebox(2,1){$\rightarrow$}}
  \put(13,4){\line(-1,1){2}}
  \put(13,4){\line(0,1){2}}
  \put(11,6){\line(1,0){2}}
  \put(12,5){\makebox(1,1){$P$}}
  \put(14,4){\line(1,1){2}}
  \put(14,4){\line(0,1){2}}
  \put(14,6){\line(1,0){2}}
  \put(14,5){\makebox(1,1){$Q$}}
  \put(19,4){\line(-1,1){2}}
  \put(19,4){\line(0,1){2}}
  \put(17,6){\line(1,0){2}}
  \put(18,5){\makebox(1,1){$R$}}
  \put(20,4){\line(1,1){2}}
  \put(20,4){\line(0,1){2}}
  \put(20,6){\line(1,0){2}}
  \put(20,5){\makebox(1,1){$S$}}
  \put(17,1){\makebox(1,1){$\alpha$}}
  \put(14,3){\makebox(1,1){$\gamma$}}
  \put(20,3){\makebox(1,1){$\beta$}}
\end{picture}
\]
}
 %
     (2) After the transformation, we should update values in the
\verb|diff| array. To do this, it is enough to know the heights
of trees $P, Q,\ldots$ up to a constant (only differences are
important), so we may assume that one of the heights is equal to
$0$.

     Here are the rotation procedures:
\begin{verbatim}
procedure SR (a:integer); {small right rotation at a}
| var b: 1..n; val_a,val_b: T; h_P,h_Q,h_R: integer;
begin
| b := right [a]; {b <> null}
| val_a := val [a]; val_b := val [b];
| h_Q := 0; h_R := diff[b]; h_P := (max(h_Q,h_R)+1)-diff[a];
| val [a] := val_b; val [b] := val_a;
| right [a] := right [b] {subtree R}
| right [b] := left [b] {subtree Q}
| left [b] := left [a] {subtree P}
| left [a] := b;
| diff [b] := h_Q - h_P;
| diff [a] := h_R - (max (h_P, h_Q) + 1);
end;
\end{verbatim}
\begin{verbatim}
procedure BR(a:integer);{big right rotation at a}
| var b,c: 1..n; val_a,val_b,val_c: T;
|     h_P,h_Q,h_R,h_S: integer;
\end{verbatim}
\clearpage
\begin{verbatim}
begin
| b := right [a]; c := left [b]; {,c <> null}
| val_a := val [a]; val_b := val [b]; val_c := val [c];
| h_Q := 0; h_R := diff[c]; h_S := (max(h_Q,h_R)+1)+diff[b];
| h_P := 1 + max (h_S, h_S-diff[b]) - diff [a];
| val [a] := val_c; val [c] := val_a;
| left [b] := right [c] {subtree R}
| right [c] := left [c] {subtree Q}
| left [c] := left [a]  {subtree P}
| left [a] := c;
| diff [b] := h_S - h_R;
| diff [c] := h_Q - h_P;
| diff [a] := max (h_S, h_R) - max (h_P, h_Q);
end;
\end{verbatim}
The (small and big) left rotations are similar.\es

\smallskip

