
   The addition/deletion procedures are written as before, but now
they have to update the \w{diff} array and
restructure the tree to keep it balanced.

An auxiliary procedure with the
following pre- and postconditions
is used:
\begin{quotation}
   {\bf before:} the left and right subtrees of the vertex number
       \w{a} are balanced; the difference of heights at \w{a} is
       at most \w{2}; the \w{diff} array is filled correctly for
       the subtree rooted at \w{a};

   {\bf after:} the subtree rooted at \w{a} is now balanced; the
       \w{diff} is updated (inside that subtree); the change in
       the height of the subtree rooted at \w{a} is stored in
       \w{d} and is equal to \w{0} or \w{-1}; the remaining part
       of the tree (including the \w{diff} array) remains
       unchanged.
\end{quotation}
\begin{verbatim}
procedure balance (a: integer; var d: integer);
begin {-2 <= diff[a] <= 2}
| if diff [a] = 2 then begin
| | b := right [a];
| | if diff [b] = -1 then begin
| | | BR (a); d := -1;
| | end else if diff [b] = 0 then begin
| | | SR (a); d := 0;
| | end else begin {diff [b] = 1}
| | | SR (a); d := - 1;
| | end;
| end else if diff [a] = -2 then begin
| | b := left [a];
| | if diff [b] = 1 then begin
| | | BL (a); d := -1;
| | end else if diff [b] = 0 then begin
| | | SL (a); d := 0;
| | end else begin {diff [b] = -1}
| | | SL (a); d := - 1;
| | end;
| end else begin {-2 < diff [a] < 2, there is nothing to do}
| | d := 0;
| end;
end;
\end{verbatim}

     To restore the balance, we go downwards from a leaf to the
root. To do that, we store the path from the root to
the current vertex in
a stack. The elements of the stack are pairs
$\langle$vertex, direction of move from the vertex$\rangle$,
that is, values of type
\begin{verbatim}
      record
      | vert: 1..n; {vertex}
      | direction : (l, r); {l for left, r for right}
      end;
\end{verbatim}
The addition of an element \verb|t| is now as follows:
\begin{verbatim}
if root = null then begin
| get_free (root);
| left[root] := null; right[root] := null; diff[root] := 0;
| val[root] := t;
end else begin
| x := root; ..make the stack empty
| {invariant: it remains to add t to the nonempty subtree
|  rooted at x; the stack contains the path to x}
| while ((t < val [x]) and (left [x] <> null)) or
| |     ((t > val [x]) and (right [x] <> null)) do begin
| | if t < val [x] then begin
| | | ..add <x, l> to the stack
| | | x := left [x];
| | end else begin {t > val [x]}
| | | ..add <x, r> to the stack
| | | x := right [x];
| | end;
| end;
| if t <> val [x] then begin {t is not in the tree}
| | get_free (i); val [i] := t;
| | left [i] := null; right [i] := null; diff [i] := 0;
| | if t < val [x] then begin
| | | ..add <x, l> to the stack
| | | left [x] := i;
| | end else begin {t > val [x]}
| | | ..add <x, r> to the stack
| | | right [x] := i;
| | end;
| | d := 1;
| | {invariant: the stack contains the path to a changed
| |  subtree whose height has increased by d (= 0 or 1);
| |  this subtree is balanced; values of diff for its
| |  vertices are correct; in the remaining part of the
| |  tree everything is unchanged (including the values
| |  of diff)}
| | while (d <> 0) and the stack is non-empty do begin
| | | {d = 1}
| | | ..take a pair from stack into <v, direct>
| | | if direct = l then begin
| | | | if diff [v] = 1 then begin
| | | | | c := 0;
| | | | end else begin
| | | | | c := 1;
| | | | end;
| | | | diff [v] := diff [v] - 1;
| | | end else begin {direct = r}
| | | | if diff [v] = -1 then begin
| | | | | c := 0;
| | | | end else begin
| | | | | c := 1;
| | | | end;
| | | | diff [v] := diff [v] + 1;
| | | end;
| | | {c = the change in the height of the subtree rooted
| | |  at v (compared with the initial tree); the array
| | |  diff has correct values inside that subtree; the
| | |  balance condition at v may be violated}
| | | balance (v, d1); d := c + d1;
| | end;
| end;
end;
\end{verbatim}
It is easy to check that \verb|d| may be equal to
\verb|0| or \verb|1| (but not \verb|-1|); indeed,
if $\w{c}=\w{0}$, then
$\w{diff[v]}=\w{0}$ and balancing is not performed.

     The deletion procedure is similar. Its main part is:
\begin{verbatim}
  {invariant: the stack contains a path to the changed
   subtree whose height was changed by d (=0 or -1)
   compared with the initial tree; this subtree is
   balanced; the values of diff are correct for the
   vertices of that subtree; the remaining part of the
   tree is unchanged (including the values of diff)}
  while (d <> 0) and the stack is not empty do begin
  | {d = -1}
  | ..take a pair from the stack into <v, direct>
  | if direct = l then begin
  | | if diff [v] = -1 then begin
  | | | c := -1;
  | | end else begin
  | | | c := 0;
  | | end;
  | | diff [v] := diff [v] + 1;
  | end else begin {direct = r}
  | | if diff [v] = 1 then begin
  | | | c := -1;
  | | end else begin
  | | | c := 0;
  | | end;
  | | diff [v] := diff [v] - 1;
  | end;
  | {c = the change in the height of the subtree rooted
  |  at v (compared with the initial tree); the array diff
  |  has correct values inside that subtree; the balance
  |  condition at v may be violated}
  | balance (v, d1);
  | d := c + d1;
  end;
\end{verbatim}
It is easy to check that \verb|d| may be equal to
\verb|0| or \verb|-1| (but not \verb|-2|); indeed,
if $\w{c}=\w{-1}$, then
$\w{diff[v]}=\w{0}$ and balancing is not performed.

     Let us mention that the existence of the stack makes the
variables \verb|father| and \verb|direction| used in the deletion
procedure (see above) redundant, because now the stack top contains
the same information.

     \Problem{Prove that while the element is added,

     (a) the second case of the balancing step (see
picture on p.~\pageref{threecases}) is, in fact, impossible;

     (b) the complete balancing of the entire tree requires only
one rotation.

However, deletion may require many rotations to
restore the
balance.\es}

     Remark. Addition and deletion procedures may be
simplified if we do not want to make them similar.

\subsubsection*{Other versions of balanced trees}

     There are several other ways to represent
sets using trees.
Some of those methods also
guarantee a running time of order
$\log n$ for each operation.
 \index{tree!B-tree}%
 \index{B-tree}%
Let us sketch one of them, called {\it B-trees\/}.
(It is often used for large databases stored on a hard disk.)

     Up to now each vertex contained only one element of the
set. This element was used as a threshold that separates the
left and right subtrees. Now let the vertex store $k\ge1$ elements
of the set. The value of $k$ may be different for different
vertices and may change while adding or deleting elements (see
below). The $k$ elements stored at a vertex are used as
separators between $k+1$ subtrees (so a vertex with $k$
elements may have up to $k+1$ sons).

Assume that some number $t\ge 1$ is fixed. We consider trees
that satisfy the following requirements:
\begin{enumerate}
\item Each vertex contains  not less than $t$ and not more than
$2t$ elements. (The root is an exception; it may contain any number
of elements not exceeding $2t$.)
\item Any vertex with $k$ elements either has $k+1$ sons or does not
have any sons at all (that is, it is a {\it leaf\/}).
\item All leaves are on the same level.
\end{enumerate}

  The {\it addition\/} of an element proceeds as follows. If the
leaf where this element goes is not full (that is, contains less
than $2t$ elements), we simply add this element to that leaf. If
that leaf is full, then we have $2t+1$ elements ($2t$ old ones
and the new one).
We split them
into two leaves with $t$ elements and
the median element between them. This median element should be
added to a vertex at the
preceding
level. This is easy if that vertex
has less than $2t$ elements. If it is full,
then it is split into two vertices, a median is
found, etc. Finally, if we need to add the new element to the
root and the root is full, we split the root
into two vertices and the
tree height is increased by $1$.

  The {\it deletion\/} of an element that is placed
not at a leaf
may be reduced to the deletion of the next element of the set,
which is in a leaf. Therefore, it is enough to delete
elements from
leaves.
If the leaf becomes too small, we can borrow
some elements from
a neighboring leaf, unless it too has the minimal possible size
$t$. If both leaves have size $t$, together
they have $2t$
elements, or rather
$2t+1$ elements if we count the separator
between them. After deleting one element, the remaining $2t$
elements may be placed onto one leaf. However, the vertex of the
preceding
level may now be too small. In that case, we have to do
the same transformation at that level, etc.

    \Problem{Implement this scheme of set representation and
check that it also performs additions, deletions, and
membership tests in time $C\log n$, where $n$ is the
cardinality of the set.\es}

    \Problem{Another definition of a balanced
tree requires that
for each vertex the number of vertices in its left and right
subtrees do not differ too much. (The advantage of this
definition is that a rotation performed at some vertex does not
destroy the balance at the ancestors of that vertex.) Using this
idea, find a set representation that guarantees a
running time bound of $C\log n$ for additions, deletions and
membership tests.}

\Hint{This approach also uses small and big rotations.
The details can be found in the book of Reingold, Nievergelt, and
Deo [9].}%
 \index{Reingold, E.M.}%
 \index{Nievergelt, J.}%
 \index{Deo, N.}%

