\section{Context-free grammars}\label{grammars}
\newcommand{\summ}{\hbox{$\langle$\rm summ$\rangle$}}
\newcommand{\expr}{\hbox{$\langle$\rm expr$\rangle$}}
\newcommand{\fact}{\hbox{$\langle$\rm fact$\rangle$}}
\newcommand{\restsumm}{\hbox{$\langle$\rm restsumm$\rangle$}}
\newcommand{\restexpr}{\hbox{$\langle$\rm restexpr$\rangle$}}

\subsection{General parsing algorithm}
\label{generalparsing}

     To define a {\it context-free grammar\/}
we should:
 \index{grammar!context-free}%
 \index{context-free grammar}%
 \index{language!context-free}%
 \index{symbol!terminal}%
 \index{symbol!nonterminal}%
 \index{alphabet}%
 \index{letter}%
 \index{terminal}%
 \index{nonterminal}%
 \index{symbol!initial}%
 \index{derivation!in a grammar}%
 \index{string!generated by a grammar}%
 \index{word!generated by a grammar}%
 \index{production rule}%
 \index{axiom (of a grammar)}%
 \index{generated string}
\begin{itemize}
  \item fix a finite set $A$, called an
{\it alphabet}, whose
elements are called {\it symbols\/} or {\it letters\/};
finite sequences of symbols are called {\it strings\/} or
{\it words\/};
 \item divide all symbols in $A$ into two classes:
{\it terminal\/} symbols and {\it nonterminal\/} symbols;
 \item choose a nonterminal symbol called the
{\it initial\/} symbol, or {\it axiom};
 \item fix a finite set of {\it productions}, or {\it production
rules\/}; each production has the form
     $K \to X$,
where $K$ is some nonterminal and $X$ is a string that may contain
both terminal and nonterminal symbols.
\end{itemize}

     Assume that a context-free grammar is given (we often omit
the words ``context-free'' because we do not consider another
types of grammars). A {\it derivation\/} in this grammar is a
sequence of strings $A_0, A_1,\ldots,A_n$, where $A_0$ is
a one-letter string consisting of the initial symbol; $A_{i+1}$
is obtained from $A_i$ by replacing some nonterminal $K$ in $A_i$
by a string $X$ according to one of the production rules $K\to X$.

A string containing only terminals is {\it
generated\/} by a grammar if there exists a derivation that ends
in this string. The set
of all strings generated by some
grammar $G$ is called the {\it context-free language
generated by\/} $G$. A language (that is, a set of strings) is called
{\it context-free\/} if it is generated by some context-free
grammar.\index{context-free language}

     In this section, as well as the following one, we are
interested in the following question: A context-free grammar $G$
is given; construct an algorithm that checks if an input
string belongs to the language generated by $G$.

\medskip

     Example 1. Alphabet:\label{cf1}%
 \index{parentheses!correct expressions}
\begin{center}
\tt         ( ) [ ] E
\end{center}
(four terminals and one nonterminal \verb|E|).
Axiom: \verb|E|.
Productions:
\begin{eqnarray*}
      \mbox{\tt E} & \to & \mbox{\tt (E)}\\
      \mbox{\tt E} & \to & \mbox{\tt [E]}\\
      \mbox{\tt E} & \to & \mbox{\tt EE}\\
      \mbox{\tt E} & \to &
\end{eqnarray*}
(the last rule has the empty string on its right-hand side).

Examples of generated strings:
\begin{center}
\tt          \hspace*{10em}  \hbox{\rm(empty string)}\\
                ()\\
                ([$\,$])\\
                ()[([$\,$])]\\
\leavevmode  \hbox{\tt [()[$\,$]()[$\,$]]}
\end{center}

    Examples of strings not in the language:
\begin{center}
\tt             (\\
                )(\\
                (]\\
                ([)]
\end{center}
This grammar was considered in section~\ref{types}.
An algorithm that
checks whether an input string belongs to the
corresponding
language
was considered; that algorithm used a stack.


\medskip

     Example 2. Another grammar that generates the same
language:\label{cf2}

\smallskip

Alphabet: {\tt  ( ) [ ] T E}

\smallskip

Productions:
\begin{eqnarray*}
 \mbox{\tt E} &\to&\\
 \mbox{\tt E} &\to&\mbox{\tt TE}\\
 \mbox{\tt T} &\to&\mbox{\tt (E)}\\
 \mbox{\tt T} &\to&\mbox{\tt [E]}
\end{eqnarray*}

In all subsequent examples, the axiom will be the nonterminal
on the left-hand side of the first rule
unless stated otherwise
(in this example, the axiom is \verb|E|).

For any nonterminal $K$, consider the set of all strings
composed of terminals that can be obtained from $K$ by a
derivation. (For the axiom, this set is a language generated by
a grammar.) In a sense, each rule of the grammar is
a statement about those sets. Let us explain what
we mean using the grammar of example 2. Let $T$ and $E$ be
the sets of all strings
in the alphabet
$\{\w{(},\w{)},\w{[},\w{]}\}$
derivable from nonterminals \w{T} and \w{E},
respectively.
The
rules of the grammar correspond to the following properties:

{\hyphenpenalty=10000
\[
\begin{tabular}{lp{.6\textwidth}}
\mbox{\tt E} $\to$  &   $E$ contains an empty string\\[0.5ex]
\mbox{\tt E $\to$ TE} &
\rightskip=0pt plus 1000pt
if $A$ is in $T$ and
                $B$ is in
                $E$, then $AB$ is in $E$\\[0.5ex]
\mbox{\tt T $\to$ [E]} &
\rightskip=0pt plus 1000pt
if $A$ is in $E$, then \verb|[|$A$\verb|]| is in $T$\\[0.5ex]
\mbox{\tt T $\to$ (E)} &
\rightskip=0pt plus 1000pt
if $A$ is in $E$, then \verb|(|$A$\verb|)| is in  $T$
\end{tabular}
\]}

\noindent
These four properties of $E$ and $T$ do not determine those sets
uniquely. For example, they are still true if \ $T$ = $E$ = the
set of all strings. However, one may prove (for an arbitrary
context-free grammar) that the sets defined by the grammar are
minimal among all the sets having those properties
(``minimal'' means ``minimal up to inclusion'').

     \Problem{Give the precise
statement and proof of this claim.\es}

    \Problem{Construct a context-free grammar that generates the
following strings (and no others):

(a) $\w{0}^k\w{1}^k$ (the numbers of zeros and ones
        are equal);

(b) $\w{0}^{2k}\w{1}^k$ (the number of zeros is twice
    as large as the number of ones);

(c) $\w{0}^k\w{1}^l$ (the number of zeros $k$ is larger
        than the number of ones $l$).

(d) (communicated by M.~Sipser) all the strings $X2Y$ where $X$
and $Y$ are composed of \w{0}s and \w{1}s and $X\ne Y$.\es
}

    \Problem{Prove that there is no context-free grammar that generates
all strings of type $\w{0}^k\w{1}^k\w{2}^k$
(and no other strings).}%
 \index{language!not context-free}
    \Hint{Prove the following lemma about an arbitrary
context-free language: Any sufficiently long
string $F$ in the language
can be represented as
$F=ABCDE$
in such a way that
any string $AB^k CD^k E$ (where $B^k$ is
$B$ repeated $k$ times)
belongs to the language. To
prove this lemma, find a nonterminal that is a descendant of
itself in the ``derivation tree''.}

  A nonterminal may be considered  a ``class name'' for all
generated strings. In the next example, we use fragments of
English words as nonterminals; each fragment is considered to be one
nonterminal symbol of the grammar.

\medskip

     Example 3.\label{cf3}%
 \index{grammar!for expressions}
 \index{expression}%
 \index{summand}%
 \index{factor}
\begin{flushleft}
\qquad\qquad Terminals: \qquad \quad {\tt + * ( ) x}\\
\qquad\qquad Nonterminals: \quad \expr\ \restexpr\
   \summ\  \restsumm\ \fact\\[1ex]
\qquad\qquad Production rules:
\end{flushleft}
\begin{eqnarray*}
    \expr     &\to& \summ\ \restexpr\\
    \restexpr &\to& \hbox{\tt +}\ \expr\\
    \restexpr &\to& \\
    \summ     &\to& \fact\ \restsumm\\
    \restsumm &\to& \hbox{\tt *}\ \summ\\
    \restsumm &\to& \\
    \fact     &\to& \hbox{\tt x}\\
    \fact     &\to& \hbox{\tt(}\ \expr\ \hbox{\tt )}
\end{eqnarray*}
According to this grammar, an {\bf expr}ession is a sequence of
{\bf summ}ands separated by symbols \w{+}; a {\bf summ}and is a
sequence of {\bf fact}ors, separated by symbols \w{*}; a
{\bf fact}or is either the letter \w{x} or an {\bf expr}ession in
parentheses.

    \Problem{Give another grammar that generates the same
language.}

    Answer. Here is one possibility:
\begin{eqnarray*}
    \expr &\to&\expr\ \hbox{\tt +}\ \expr\\
    \expr &\to&\expr\ \hbox{\tt *}\ \expr\\
    \expr &\to& \hbox{\tt x}\\
    \expr &\to&  \hbox{\tt(}\ \expr\ \hbox{\tt)}
\end{eqnarray*}
This grammar is simpler, but not quite as good (see below).\es

    \Problem{An arbitrary context-free grammar is given.
Construct an algorithm that checks if an input string belongs
to the language generated by the grammar. The algorithm should
run in polynomial time: the number of operations should
not exceed $P(\hbox{input length})$ for some
polynomial $P$. (The polynomial
may depend on the grammar.)}%
   %
 \index{context-free language!polynomial decidability}%
 \index{parsing!general context-free language}%
 \index{programming!dynamic}%
 \index{dynamic programming}%
 \label{generalparsingalgorithm}

    \Solution. The required polynomial time bound
rules out any solution
based on exhaustive search. However, a polynomial
algorithm for a general context-free language exists. We give
an outline of that algorithm below. In fact, it has no practical
value, because all context-free grammars used in practice
have special properties that make more efficient
algorithms possible.

     (1) Let $K_1,\ldots,K_n$ be the nonterminals of the given
grammar.
Construct a new context-free grammar
with nonterminals
$K_1',\ldots,K_n'$. This grammar has the
following property: a string $S$ can be generated from $K_i'$ (in
the new grammar) if and only if $S$ is nonempty and can be
generated from $K_i$ in the old grammar.

     To do that, we must know which nonterminals of the given
grammar generate the empty string. Then each rule is replaced by
a set of rules obtained as follows: On the left-hand side we add
the dash, and on the right-hand side we omit some of the
nonterminals
that generate the empty string and put dashes near the
other non-terminals. For example, if the initial grammar has
the rule
 %
\begin{center}
\tt         K $\to$ L M N
\end{center}
 %
and the empty string may be generated from \w{L} and \w{N} but
not from \w{M}, the new grammar contains rules
 %
\begin{eqnarray*}
     \hbox{\tt K}' &\to&\hbox{\tt L}'\hbox{\tt M}'\hbox{\tt N}'\\
     \hbox{\tt K}' &\to& \hbox{\tt M}'\hbox{\tt N}'\\
     \hbox{\tt K}' &\to& \hbox{\tt L}'\hbox{\tt M}'\\
     \hbox{\tt K}' &\to& \hbox{\tt M}'
\end{eqnarray*}

     (2) Therefore, we have
reduced our problem to the case of a
grammar where no terminal generates an empty string. Now we
eliminate ``cycles'' of the form
 %
\begin{center} \tt
     K $\to$ L  \\
     L $\to$ M  \\
     M $\to$ N  \\
     N $\to$ K  \\
\end{center}
(each rule has one nonterminal and no terminals on the
right-hand side; nonterminals form a cycle of any length). This is
easy; we identify all the nonterminals that appear in
the same cycle.

     (3) Now the membership test for the language generated by a
grammar can be performed as follows. For any nonterminal and for
any substring of a given string, we determine whether this
substring can be generated from this nonterminal. We consider
all substrings in the order of increasing length. All
nonterminals are considered in such an order that for any rule of
the form $K \to L$, the nonterminal $L$ is considered before the
nonterminal $K$. (This is possible because there are no cycles.)
Let us explain this process by an example.

     Assume that the grammar has rules
\begin{center} \tt
        K $\to$ L \phantom{M} \phantom{N}\\
        K $\to$ M N L
\end{center}
and no other rules with \w{K} on the left-hand side. We want to
know if a given word $A$ may be derived from the nonterminal
\w{K}. This happens:
  %
\begin{itemize}
  %
\item if $A$ can be derived from \w{L};
  %
\item if $A$ can be
split into $A=BCD$ where $B$, $C$, $D$ are nonempty strings such that
$B$ can be derived from \w{M}, $C$ can be derived from \w{N}, and
$D$ can be derived from \w{L}.
  %
\end{itemize}
  %
All this
information is available because $B$, $C$, and $D$ are
shorter than $A$ and the nonterminal \w{L} is considered before the
nonterminal \w{K}.

    It is easy to see that the running time of the algorithm is
polynomial. The degree of the polynomial depends on the number
of nonterminals on the right-hand side of the grammar rules.
The degree can be made smaller if we convert the grammar into a
form where right-hand sides of rules contain not more than two
nonterminals. This can be done easily; for example, the rule
\hbox{\tt K $\to$ LMK} may be replaced by two rules \hbox{\tt K
$\to$ LN} and \hbox{\tt N $\to$ MK} where \verb|N| is a new
nonterminal.\es

    \Problem{Consider a grammar with one nonterminal symbol
\verb|K|, terminals \w{1}, \w{2}, and \w{3}, and the rules
\begin{center} \tt
     K $\to$ 0 \phantom{K} \phantom{K} \phantom{K}\\
     K $\to$ 1 K \phantom{K} \phantom{K}\\
     K $\to$ 2 K K \phantom{K}\\
     K $\to$ 3 K K K
\end{center}
How do we
check whether a given string belongs to the corresponding
language if the string is scanned from left to right? The number
of operations per character should be limited by a constant.}

  \Solution. An integer variable \verb|n| is used along with
the  invariant
relation: ``the  input string belongs to the language if and only
if the non-processed part of the input string is a concatenation of
\w{n} strings from the language''.\es

   \Problem{Repeat the previous problem for the grammar
\begin{center} \tt
     K $\to$  0  \phantom{K} \phantom{K} \phantom{K}\\
     K $\to$  K 1 \phantom{K} \phantom{K}\\
     K $\to$  K K 2 \phantom{K}\\
     K $\to$  K K K 3\es
\end{center}}

