\subsection{Recursive-descent parsing}

     Unlike the algorithm of the preceding section (which is of mostly
theoretical interest), the recursive-descent parsing algorithm is used
 \index{parsing!recursive-descent}%
 \index{recursive-descent parsing}%
quite often. However, it is not applicable to all grammars. (See
below the requirements that allow us to apply this
method.)

    The idea is as follows. For any nonterminal \w{K} we construct
a procedure \w{ReadK} (being applied to any input string $x$)
that does two things:
\begin{itemize}
  \item finds the maximal prefix $z$ of the string
$x$ that may appear as a prefix of some string derivable from \w{K};
  \item says if the string $z$ is derivable from \w{K}.
\end{itemize}

     Before we give a more detailed description of this method,
we should agree how the procedures access the input string and
how they communicate their results. We assume that the input
string is read character-by-character.
In other words, we assume that
there is a separator
between the ``already read'' (processed) part and
``the unread'' part.
(The last name should not be taken literally, because the
first symbol of the unread part may be already known to the
procedure.)

We assume that there exists a function without parameters
\begin{center} \tt
     Next: Symbol
\end{center}
which returns the first symbol of the unread part. Its values
are terminals as well as the special symbol \w{EOI} that stands for
``End Of Input''; this symbol
means that the input string is ended. (In a sense,
EOI is written after the last character of the input string.)
A call to \w{Next} does not move the separator between the
read and unread parts. There exists a special procedure \w{Move}
that ``reads'' the next character, that is,
moves the separator to the right, adding one character to the
processed part. This procedure is applicable when \w{Next<>EOI}.
Finally, we have also a Boolean variable \w{b}; its role is
described below

\smallskip

     Now we state our requirements for the procedure \w{ReadK}:
\begin{itemize}
   \item \verb|ReadK| reads the
maximal prefix $A$ of the input string (its unprocessed part)
that may appear as a prefix of some string derivable
from \verb|K|;
  \item the value of \w{b} becomes true or false depending on whether
$A$ is derivable from $K$ or is only a prefix of some derivable
string (but is not derivable itself).
\end{itemize}

     It is convenient to use the following notation:
Any string that is derivable from some nonterminal \w{K} is
called a \w{K}-{\it string}. Any string that
is a prefix of a string derivable
from \w{K} is called \w{K}-{\it prefix}.
If the two requirements for \w{ReadK} stated above are
fulfilled, we say that ``\w{ReadK} is correct for \w{K}''.

\smallskip

  Let us begin with an example. Assume that the rule
\begin{center}\tt
        K $\to$ L M
\end{center}
is the only rule of the grammar that has \verb|K| on the left-hand
side. Assume that  \verb|L|, \verb|M| are nonterminals and
\verb|ReadL|, \verb|ReadM| are correct procedures for those nonterminals.

     Consider the following procedure:
\begin{verbatim}
     procedure ReadK;
     begin
     | ReadL;
     | if b then begin
     | | ReadM;
     | end;
     end;
\end{verbatim}

    \Problem{Give an example where this procedure is not correct for
\verb|K|.}

    Answer. Assume that any string \w{000...000} is derivable from \w{L}
and that only the string \w{01} is derivable from \w{M}. Then the string
\w{00001} is derivable from \w{K}, but the procedure \w{ReadK} does not
see this.\es

     Let us give a sufficient condition for the correctness of
the procedure \w{ReadK} given above. To do that, we need some
notation. Assume that a context-free grammar is fixed and that
$N$ is some nonterminal of that grammar. Consider the $N$-string
$A$ that has a proper prefix $B$, which is also an $N$-string
(assuming such $A$ and $B$ exist). For each pair of such $A$
and $B$, consider the terminal that follows $B$ in $A$ (appears
immediately after $B$ in $A$). The set of all such symbols (for
all $A$ and $B$) is denoted by
 \index{Foll($X$)}
Foll($N$). (If no $N$-string is a proper prefix of another
$N$-string, the set Foll($N$) is empty.)

    \Problem{Find  (a) Foll({\tt E}) for the grammar given in
example 1 (see p.~\pageref{cf1}); (b) Foll({\tt  E})  and
Foll({\tt  T})  for  the  grammar  give in example 2 (see
p.~\pageref{cf2});  (c)  Foll(\summ)  and  Foll(\fact)  for  the
grammar given in example 3 (see p.~\pageref{cf3}). }

    Answer.  (a)~Foll({\tt E}) = $\{ \hbox{\tt [}, \hbox{\tt (}\}$.
(b)~Foll({\tt E}) = $\{ \hbox{\tt [}, \hbox{\tt (}\}$;
Foll({\tt T}) is empty (no \w{T}-string is a prefix of another
\w{T}-string).
(c)~Foll(\summ) = $\{\hbox{\tt *}\}$; Foll(\fact) is empty.\es

\smallskip

For any nonterminal $N$,
  \index{First($X$)}%
we denote the set of all terminals that are first characters of
nonempty $N$-strings by First($N$).
Now we are ready to give a
sufficient condition for the correctness of the procedure
\w{ReadK} in the situation explained above.

    \Problem{Prove that if Foll(\verb|L|) and First(\verb|M|) are
disjoint  sets and the set of all \w{M}-words is not empty, then
the procedure \w{ReadK} is correct for \w{K}.}

    \Solution. Consider two cases.

(1) Suppose that
after the call to \verb|ReadL| the value of \verb|b| is
false. In this case, \w{ReadL} reads the maximal \w{L}-prefix
$A$; this prefix is not an \w{L}-string. The string $A$ is a
\w{K}-prefix (here we use the fact that the set of strings
derivable from \w{M} is not empty). Will $A$ be the maximal
prefix of the input string that is at the same time a
\w{K}-prefix? The answer is ``yes''. Indeed, assume that $A$ is
not maximal and there exists a longer string $X$ that is both a
\w{K}-prefix and a prefix of the input string. Since \w{ReadL}
is correct, $X$ is not a \w{K}-prefix, and therefore, $X=BC$
where $B$ is an \w{L}-string and $C$ is a \w{M}-prefix.

If $B$ is longer than $A$, then $A$ is not the maximal prefix of
the input string that is also a \w{K}-prefix, which contradicts
the correctness of \w{ReadL}. If $B=A$, then $A$ would be an
\w{L}-string, which is not true. Therefore, $B$ is a proper
prefix of $A$, $\,C$ is not empty, and the first character of
$C$ follows the last character of $B$ in $A$. So the first
character of $C$ belongs both to Foll(\w{L}) and First(\w{M}),
which contradicts our assumption.

This contradiction shows that $A$ is a maximal prefix of the
input string that is also a \w{K}-prefix. Moreover, the argument
above shows that $A$ is not a \w{K}-string. The correctness of
the procedure \w{ReadK} is therefore established (see its code).

     (2) Assume that after the call to \verb|ReadL|, the value
of \verb|b| is true. Then the procedure \w{ReadK} reads some
string of the form $AB$ where $A$ is an \w{L}-string and $B$ is
an \w{M}-prefix. Therefore, $AB$ is a \w{K}-prefix. Let us check
that it is maximal. Assume that $C$ is a longer prefix, which is
at the same time a \w{K}-prefix. Then either $C$ is an
\w{L}-prefix (which is impossible because $A$ is the maximal
\w{L}-prefix) or $C = A'B'$, where $A'$ is an \w{L}-string and
$B'$ is an \w{M}-prefix. If $A'$ is shorter than $A$, then $B'$
is not empty and begins with a character that belongs both to
First(\verb|M|) and Foll(\verb|L|), which is impossible. If $A'$
is longer than $A$, then $A$ is not the maximal \w{L}-prefix.
Therefore, the only possibility is $A'=A$, but in this case $B$
is a prefix of $B'$, which contradicts the correctness of
\w{ReadM}. Therefore, $AB$ is the maximal prefix of the input
string that is a \w{K}-prefix.

It remains to check that the value of \w{b} returned by
\verb|ReadK| is correct. If \w{b} is true, this is evident. If
\w{b} is false, then $B$ is not an \w{M}-string, and we have to
check that $AB$ is not a \w{K}-string. Indeed, if $AB = A'B'$
where $A'$ is an \w{L}-string and $B'$ is an \w{M}-string, then
$A'$ cannot be longer than $A$ (since \verb|ReadL| reads the
maximal prefix), $A'$ cannot be equal to $A$ (since in this case
$B'$ would be equal to $B$ and could not be an \w{M}-string), and
$A'$ cannot be shorter than $A$ (since in this case the first
character of $B'$ would belong both to First(\w{M}) and
Foll(\w{L})). The correctness of \w{ReadK} is proved.\es

\smallskip

     Now we consider another special case. Assume that a context-free
grammar contains the rules
\begin{eqnarray*}
 \hbox{\tt K}&\to&\hbox{\tt L}\\
 \hbox{\tt K}&\to&\hbox{\tt M}\\
 \hbox{\tt K}&\to&\hbox{\tt N}
\end{eqnarray*}
and has no other rules with \verb|K| on the left-hand side.

    \Problem{Assume that \verb|ReadL|, \verb|ReadM|, and
\verb|ReadN| are correct (for \verb|L|, \verb|M|, and \verb|N|)
and that First(\verb|L|), First(\verb|M|), and First(\verb|N|)
are disjoint. Write a procedure \w{ReadK} that is correct for
\w{K}.}

    \Solution. Here is the procedure:
\begin{verbatim}
     procedure ReadK;
     begin
     | if (Next is in First(L)) then begin
     | | ReadL;
     | end else if (Next is in First(M)) then begin
     | | ReadM;
     | end else if (Next is in First(N)) then begin
     | | ReadN;
     | end else begin
     | | b := true or false depending on whether an
     | |      empty string is derivable from K or not
     | end;
     end;
\end{verbatim}
Let us prove that \verb|ReadK| is correct for \verb|K|. If the
symbol \verb|Next| is not in the sets First(\verb|L|),
First(\verb|M|), and First(\verb|N|), then the empty string is
the maximal prefix of the input string that is a \w{K}-prefix.
If \w{Next} belongs to one of those sets (and, therefore, does
not belong to the others), then the maximal prefix of the input
string that is a \w{K}-prefix is nonempty and the corresponding
procedure reads it.\es

    \Problem{Using the methods discussed, write a procedure
that recognizes expressions generated by the grammar of
example 3 (p.~\pageref{cf3}):
\begin{eqnarray*}
    \expr   &\to&\summ\ \restexpr\\
    \restexpr &\to& \hbox{\tt +}\ \expr\\
    \restexpr &\to& \\
    \summ   &\to& \fact\ \restsumm\\
    \restsumm &\to& \hbox{\tt *}\ \summ\\
    \restsumm &\to& \\
    \fact   &\to& \hbox{\tt x}\\
    \fact   &\to& \hbox{\tt(}\ \expr\ \hbox{\tt )}
\end{eqnarray*}
}

    \Solution. This grammar does not follow the patterns above:
among
the right-hand sides of its rules there are combinations of terminals
and nonterminals such as
\begin{center} \tt + \expr \end{center}
as well as a group of three symbols
\begin{center} \tt ( \expr\ ) \end{center}
This grammar also contains several rules with the same
left-hand side
and right-hand sides of different types, such as
\begin{eqnarray*}
    \restexpr &\to& \hbox{\tt +}\ \expr\\
    \restexpr &\to&
\end{eqnarray*}
 %
These problems are not fatal. For example, a rule of type
$\hbox{\tt K}\to \hbox{\tt L M N}$
may be replaced by two rules
$\hbox{\tt K} \to \hbox{\tt L Q}$ and
$\hbox{\tt Q} \to \hbox{\tt M N}$. The terminals on the
right-hand side may be replaced by nonterminals (the
only rule involving these nonterminals allows
to replace them
by the corresponding terminals).
If several rules have the same left-hand side and different
right-hand sides, such as
\begin{eqnarray*}
\hbox{\tt K}&\to&\hbox{\tt L M N}\\
\hbox{\tt K}&\to&\hbox{\tt P Q}\\
\hbox{\tt K}&\to&
\end{eqnarray*}
they can be replaced by rules
\begin{eqnarray*}
\hbox{\tt K}&\to&\hbox{\tt K}_1\\
\hbox{\tt K}&\to&\hbox{\tt K}_2\\
\hbox{\tt K}&\to&\hbox{\tt K}_3\\
\hbox{\tt K}_1 &\to&\hbox{\tt L M N}\\
\hbox{\tt K}_2 &\to&\hbox{\tt P Q}\\
\hbox{\tt K}_3 &\to&
\end{eqnarray*}
 %
We will not, however,
transform the grammar (example 3)
explicitly.  Instead, we imagine that this transformation is performed
(new nonterminals added), then the procedures for all nonterminals
(old and new) are written, and finally the procedures for the
new nonterminals
are eliminated (by in-line substitutions).
 %
%                               .    , 
 %
For example, for the rule
\[
   \hbox{\tt K} \to \hbox{\tt L M N}
\]
we get the procedure
\begin{verbatim}
        procedure ReadK;
        begin
        | ReadL;
        | if b then begin ReadM; end;
        | if b then begin ReadN; end;
        end;
\end{verbatim}
Its correctness is guaranteed  if (1) Foll(\verb|L|) and
First(\verb|MN|) are disjoint
(First(\w{MN}) is equal to First(\w{M}) if the empty string
is not derivable from \w{M}; otherwise,
it is equal to the union of First(\w{M})
and First(\w{N})); (2) Foll (\w{M}) and First(\w{N}) are
disjoint.

  Similarly, the rules
\begin{eqnarray*}
\hbox{\tt K}&\to&\hbox{\tt L M N}\\
\hbox{\tt K}&\to&\hbox{\tt P Q}\\
\hbox{\tt K}&\to&
\end{eqnarray*}
lead to the procedure
\begin{verbatim}
        procedure ReadK;
        begin
        | if (Next is in First(LMN)) then begin
        | | ReadL;
        | | if b then begin ReadM; end;
        | | if b then begin ReadN; end;
        | end else if (Next is in First(PQ)) then begin
        | | ReadP;
        | | if b then begin ReadQ; end;
        | end else begin
        | | b := true;
        | end;
        end;
\end{verbatim}
To prove its correctness, we require the sets
First(\w{LMN}) and First(\w{PQ}) to be
disjoint.

\smallskip

Now we apply these methods to the grammar of example 3:
\begin{verbatim}
     procedure ReadSymb (c: Symbol);
     | b := (Next = c);
     | if b then begin Move; end;
     end;

     procedure ReadExpr;
     | ReadSumm;
     | if b then begin ReadRestExpr; end;
     end;

     procedure ReadRestExpr;
     | if Next = '+' then begin
     | | ReadSymb ('+');
     | | if b then begin ReadExpr; end;
     | end else begin
\end{verbatim}
\clearpage
\begin{verbatim}
     | | b := true;
     | end;
     end;

     procedure ReadSumm;
     | ReadFact;
     | if b then begin ReadRestSumm; end;
     end;

     procedure ReadRestSumm;
     | if Next = '*' then begin
     | | ReadSymb ('*');
     | | if b then begin ReadSumm; end;
     | end else begin
     | | b := true;
     | end;
     end;

     procedure ReadFact;
     | if Next = 'x' then begin
     | | ReadSymb ('x');
     | end else if Next = '(' then begin
     | | ReadSymb ('(');
     | | if b then begin ReadExpr; end;
     | | if b then begin ReadSymb (')'); end;
     | end else begin
     | | b := false;
     | end;
     end;
\end{verbatim}
These procedures are mutually recursive, that is, some procedure uses
another one which in its turn uses the first one, etc. This is allowed in
Pascal if we use the so-called \w{forward} definitions of
the mutually recursive procedures. As usual, to prove the correctness
of recursive procedures we need to prove that (1)  each of them is correct,
assuming all calls work correctly (here our method works:
one needs only check that the corresponding sets are disjoint);
(2) the
procedure terminates. The second claim is not self-evident.
For example, if the grammar has the rule
$\hbox{\tt K}\to\hbox{\tt KK}$, then no strings are derivable from
\verb|K|, and the sets
Foll(\verb|K|) and First(\verb|K|) are empty (and therefore disjoint),
but the procedure
\begin{verbatim}
     procedure ReadK;
     begin
     | ReadK;
\end{verbatim}
\clearpage
\begin{verbatim}
     | if b then begin
     | | ReadK;
     | end;
     end;
\end{verbatim}
(written according to our guidelines) never terminates.

     In the case in question, the procedures  \verb|ReadRestExpr|,
\verb|ReadRestAdd|, and \verb|ReadFact| either terminate immediately or
decrease the length of the unprocessed part of the input string. Since
any cycle of the mutually recursive calls includes one of them,
termination is guaranteed. Our problem is solved.\es

