    \Problem{Assume that a grammar has two rules with nonterminal
\w{K} on the left-hand side:
\begin{eqnarray*}
 \hbox{\tt K}&\to&\hbox{\tt L K}\\
 \hbox{\tt K}&\to&
\end{eqnarray*}
According to these rules, any \w{K}-string is a concatenation of
several \w{L}-strings. Assume also that the sets
Foll(\verb|L|) and First(\verb|K|) (which equals First(\w{L}) in
this case) are disjoint. Assume that a procedure \w{ReadL} is correct
for \w{L}.
Write a
{\it nonrecursive\/} procedure \w{ReadK} that is correct for \w{K}.}
 % this condition is not necessary - correct the russian version!
%You may assume that the empty string is not derivable from \w{L}.}

    \Solution. As we already
know, the following recursive procedure
is correct for \w{K}:
\begin{verbatim}
     procedure ReadK;
     begin
     | if (Next is in First(L)) then begin
     | | ReadL;
     | | if b then begin ReadK; end;
     | end else begin
     | | b := true;
     | end;
     end;
\end{verbatim}
Termination is guaranteed because the
length of the unprocessed part is
decreased before the recursive call.
    This recursive procedure is equivalent to the following
nonrecursive one:
\begin{verbatim}
     procedure ReadK;
     begin
     | b := true;
     | while b and (Next is in First(L)) do begin
     | | ReadL;
     | end;
     end;
\end{verbatim}
Let us formally check this equivalence. Termination is guaranteed
both for the recursive and nonrecursive procedures. Therefore,
it is enough to check that the body of the recursive procedure becomes
equivalent to the body of the nonrecursive one if the recursive
call is replaced by the call of the nonrecursive procedure.
Let us make this replacement:
\begin{verbatim}
     if (Next is in First(L)) then begin
     | ReadL;
     | if b then begin
     | | b := true;
     | | while b and (Next is in First(L)) do begin
     | | | ReadL;
     | | end;
     | end;
     end else begin
     | b := true;
     end;
\end{verbatim}
The first command \verb|b:=true| may be deleted because at this point
\verb|b| is already true. The second command \w{b:=true} may be moved
to the beginning:
\begin{verbatim}
     b := true;
     if (Next is in First(L) then begin
     | ReadL;
     | if b then begin
     | | while b and (Next is in First(L)) do begin
     | | | ReadL;
     | | end;
     | end;
     end;
\end{verbatim}
Now the second \verb|if| may be removed (because if \verb|b| is false, the
\verb|while|-loop does nothing). We may also add the condition
\w{b} to the first \w{if}
(because \w{b} is true at that point). Thus we
get
\begin{verbatim}
     b := true;
     if b and (Next is in First(L)) then begin
     | ReadL;
     | while b and (Next is in First(L)) do begin
     | | ReadL;
     | end;
     end;
\end{verbatim}
which is equivalent to the body of the nonrecursive procedure
above (the first iteration of the loop is unfolded).\es

    \Problem{Prove the correctness of the nonrecursive procedure
shown
above
directly, without referring to the recursive version.}

    \Solution. Consider the maximal prefix of the input string that is
a \w{K}-prefix. It can be represented as a concatenation of several
nonempty strings: all are
\w{L}-strings except, maybe, the last one, which is an \w{L}-prefix.
We call those strings (including the last one) ``components''.

The invariant relation: several components are read;
\w{b} is true if and only if
the last component is an \w{L}-string.

    Let us check that this invariant relation remains true after
the next iteration. If only the last component remains, it is evident.
If several components remain, the first of the remaining components
is followed by a character that belongs to First(\w{L}) and is therefore
not in
Foll(\w{L}); so the first remaining component is a maximal
\w{L}-prefix that is also a prefix of the
of the unprocessed part.\es

  In practice a shorthand notation for grammars is used.
Namely, rules of the form
\begin{eqnarray*}
 \hbox{\tt K}&\to& \hbox{\tt L K}\\
 \hbox{\tt K}&\to&
\end{eqnarray*}
(we assume that no other rule has \w{K} on the left-hand side,
so \w{K}-strings are concatenations of \w{L}-strings) are omitted,
and \w{K} is replaced by \w{L} enclosed
in curly braces (which denotes iteration
in this case). Also, several rules with the same left-hand side
are often written as one rule where alternatives are written
one after other separated by bars.

     For example, the grammar for {\bf expr}essions given
above may be
rewritten as follows:
\begin{eqnarray*}
    \expr &\to& \summ\hbox{ \tt \{ + \summ\ \}}\\
    \summ&\to& \fact\hbox{ \tt \{ * \fact\ \}}\\
    \fact&\to& \hbox{ \tt x}\ |\ (\  \expr\ )
\end{eqnarray*}

    \Problem{Write a procedure that is correct for \expr,
following this
grammar. Use
iteration instead of recursion whenever possible.}

     \Solution.
\begin{Verbatim}
     procedure ReadSymb (c: Symbol);
     | b := (Next = c);
     | if b then begin Move; end;
     end;

     procedure ReadExpr;
     begin
     | ReadSumm;
     | while b and (Next = '+') do begin
     | | Move; ReadSumm;
     | end;
     end;

     procedure ReadSumm;
     begin
     | ReadFact;
     | while b and (Next = '*') do begin
     | | Move; ReadFact;
     | end;
     end;

     procedure ReadFact;
     begin
     | if Next = 'x' do begin
     | | Move; b := true;
     | end else if Next = '(' then begin
     | | Move; ReadExpr;
     | | if b then begin ReadSymb (')'); end;
     | end else begin
     | | b := false;
     | end;
     end;
\end{Verbatim}

\Problem{
   The assignment \w{b:=true} in the last procedure may be
omitted. Why?}

  \Solution. We may assume that all procedures are called
only
when
\w{b=true}.\nobreak\es

\subsection{Parsing algorithm for LL(1)-grammars}

     In this section, we consider one more algorithm to check if
a given string is generated by a given grammar.
This algorithm is called
 \index{parsing!LL(1)}%
 \index{LL(1)-parsing}%
LL(1)-{\it parsing}. Its main idea can be summed up
in one sentence:
we may assume that
all the production rules are applied to the leftmost
nonterminal only;
if we are lucky,
the applicable rule
is determined uniquely by the first character of the
string derivable from this nonterminal.

Now we give
the details. To begin with, we have the following



    {\bf Definition.}
A {\it leftmost derivation\/} (of a string in a grammar)
 \index{derivation!leftmost}
is a derivation where the leftmost nonterminal is replaced at each
step.

    \Problem{Each derivable word (which contains only terminals)
has the leftmost derivation.}

    \Solution. During the derivation process, different nonterminals
in a string
are replaced independently of each other. (That is why the grammar
is called ``context-free''.) In other words, if at some point of
the derivation we have the string
$\ldots K\ldots L\ldots$ where
$K$ and $L$ are nonterminals, then the substitutions  for  $K$  and
$L$ may be performed in any order.
Therefore, we can rearrange the derivation in such a way that the
left nonterminal $K$ is replaced first.\es

    \Problem{Consider the grammar with four production rules:
\begin{eqnarray*}
    && (1)\quad \hbox{\tt E} \to \\
    && (2)\quad \hbox{\tt E} \to \hbox{\tt TE}\\
    && (3)\quad \hbox{\tt T} \to \hbox{\tt (E)}\\
    && (4)\quad \hbox{\tt T} \to \hbox{\tt [E]}
\end{eqnarray*}
Find the leftmost derivation of the word
$A = \hbox{\tt [()([$\,$])]}$ and prove that it is uniquely
determined.}

    \Solution. At the first step, only the rule (2) may be applied:
\[ \hbox{\tt E} \to \hbox{\tt TE} \]
What happens with \w{T} then? Since $A$ starts with
\verb|[|, only the rule (4) can be applied:
\[ \hbox{\tt E} \to \hbox{\tt TE}\to \hbox{\tt[E]E} \]
The leftmost \verb|E| is now replaced by \verb|TE| (otherwise the
second symbol of the input string would be \verb|]|):
\[ \hbox{\tt E} \to\hbox{\tt TE}\to\hbox{\tt [E]E}
          \to\hbox{\tt [TE]E} \]
and \verb|T| is replaced  according to (3):
\[ \hbox{\tt E}\to\hbox{\tt TE}\to\hbox{\tt [E]E}
    \to\hbox{\tt [TE]E}\to\hbox{\tt [(E)E]E} \]
Now the leftmost \verb|E| should be replaced by the empty string,
otherwise the third character of the input string would be
\verb|(| or \verb|[| (other characters cannot
be the first
character of a \w{T}-string):
\[ \hbox{\tt E}\to\hbox{\tt TE}\to\hbox{\tt [E]E}
    \to\hbox{\tt [TE]E}\to\hbox{\tt [(E)E]E}\to\hbox{\tt [()E]E} \]
We continue:
\[
  \ldots\to \hbox{\tt [()TE]E}
  \to\hbox{\tt [()(E)E]E}
  \to\hbox{\tt [()(TE)E]E}
  \to\hbox{\tt [()([E]E)E]E}\to
\]
\[
  \to\hbox{\tt [()([$\,$]E)E]E}
  \to\hbox{\tt [()([$\,$])E]E}
  \to\hbox{\tt [()([$\,$])]E}
  \to\hbox{\tt [()([$\,$])]}
\]
Thus we see that the leftmost derivation is uniquely
determined.\es

\medskip

     What are the requirements for a grammar that make this
approach (finding the unique leftmost derivation) possible?
Assume that at some point the leftmost nonterminal is $K$.
In other
words, we have the string
$AKU$ where $A$ is a string containing only terminals  and
$U$ is a string that may contain both terminals and nonterminals.
Suppose the grammar has the rules
\begin{eqnarray*}
     K&\to& L\,M N\\
     K&\to& P\,Q\\
     K&\to& R
\end{eqnarray*}
and we have to choose one of them. We make the choice
based on the first symbol of the part of the input string that is
derivable from $KU$.

