     Consider the set First($LMN$) of all terminals
that are
 \index{First($X$)}%
first
symbols of nonempty strings of terminals
derivable from $LMN$. This set is equal to the union of the set
First($L$), the set First($M$) (if the empty string is derivable from $L$),
and the set First($N$) (if the empty string is derivable from both $L$
and $N$). To make the choice (based on the first character) possible,
we require that the sets
First($LMN$), First($PQ$), and First($R$) are disjoint. But this is not
the only requirement. Indeed,
it is possible, for example, that the empty string is derivable from
$LMN$, and the string derived from $U$ starts with a character in
First($PQ$). The definitions below take this problem
into account.

    A language recognized by a context-free grammar was defined
as the set of all strings {\it of terminals\/} derivable from the
initial nonterminal (axiom). We will also speak about
strings composed of terminals and nonterminals derivable from the axiom,
or from any other nonterminal, or from any string composed
of terminals and nonterminals.
So the relation ``derivable from'' can be considered as a binary
relation defined on the set of all strings composed of terminals
and nonterminals. (However, if
we say that some string is derivable and do not specify
the starting point of the derivation, we always mean that the derivation
starts from the axiom.)

     For any string $X$ composed of terminal and nonterminals,
First($X$) denotes the set of all terminals that are the first
characters of nonempty strings of terminals derivable from $X$.
If for any nonterminal there is at least one string of terminals
derivable from it, then the phrase ``of terminals'' may be omitted
in the definition. We assume in the sequel that this condition is
satisfied.

    For any nonterminal $K$, the notation Follow($K$)\label{follow}%
 \index{Follow($X$)}
is used for the set of all terminals that appear in the
derivable (from the axiom) strings immediately after $K$.
(Please do not confuse this set with Foll($K$) defined in the
preceding section.) We add the symbol \w{EOI} to Follow($K$) if
there exists a derivable string that ends with $K$.

     For each rule
\[ K \to V \]
%
(where $K$ is a nonterminal and $V$ is a string that contains
terminals and nonterminals) we define the set of {\it leading
terminals},
 \index{Lead($K\to V$)}%
 \index{terminal!leading}%
 \index{grammar!LL(1)}%
 \index{LL(1)-grammar}%
which is denoted by
Lead($K\to V$).
By definition,
Lead($K\to V$)
is equal to
First($V$) or the union of First($V$) and
Follow($K$) if the empty string is derivable from $V$.

     Definition. A context-free grammar is called an
LL(1)-{\it grammar\/} if for any two rules
$K\to V$ and $K\to W$ with the same left-hand sides, the
sets Lead($K\to V$) and Lead($K\to W$) are disjoint.

    \Problem{Is the grammar
\begin{eqnarray*}
   \hbox{\tt K}&\to&\hbox{\tt K \#}\\
   \hbox{\tt K}&\to&
\end{eqnarray*}
(derivable strings are sequences of \w{\#}'s)
an LL(1)-grammar?}

   \Solution. No, because \w{\#} is a leading terminal for both rules.
(This is true for
the second rule because \w{\#} belongs to Follow(\w{K}).)\es

    \Problem{Write an equivalent
LL(1)-grammar.}

    \Solution.
\begin{eqnarray*}
   \hbox{\tt K}&\to&\hbox{\tt \# K}\\
   \hbox{\tt K}&\to&
\end{eqnarray*}
We have
replaced a ``left-recursive'' rule by a ``right-recursive'' one.\es

\medskip

    The next problem shows that for a LL(1)-grammar, the next step
in the construction of a leftmost derivation is uniquely defined.

    \Problem{Assume that a string $X$ is
derivable in an LL(1)-grammar and  $K$ is the
leftmost nonterminal in $X$, that is, $X=AKS$ where $A$ is a string
of terminals and $S$ is a string of terminals and nonterminals.
Assume that two different rules of the grammar have $K$ on the left-hand
side, and both of them were applied
to the nonterminal $K$ selected in $X$.
Both derivations were continued and
two strings of
terminals (having prefix $A$) were obtained.
Prove that this prefix is followed by
different terminals. (Here we consider \w{EOI} as a terminal.)
}

    \Solution. Those terminals are leading terminals of two different
rules.\es

    \Problem{Prove that if a string is derivable in an
LL(1)-grammar, its leftmost derivation is unique.}

    \Solution. The preceding problem shows that at each step there
is only one possible continuation.\es

    \Problem{A grammar is called {\it left-recursive\/} grammar
if there exists a nonterminal $K$ and a string derivable from
$K$ that starts with $K$ (but is not equal to $K$).
Prove that if (1) a grammar $G$ is left-recursive;
(2) for each nonterminal $K$,
there exists a nonempty string derivable from $K$; and
(3) for each nonterminal $K$ there exists
a derivation starting from the axiom and including $K$,
then $G$
is not an LL(1)-grammar.}
 \index{grammar!left-recursive}

    \Solution. Consider the derivation of a string $KU$ from a
nonterminal $K$ where $U$ is a nonempty string. We may assume that
it is a leftmost derivation (other nonterminals may remain
untouched). Consider the derivation
\[ K \leadsto KU \leadsto KUU \leadsto \cdots\]
(here $\leadsto$ stands for several derivation steps) and
the derivation $K \leadsto A$ where $A$ is a nonempty string of
terminals. At some point these two derivations diverge; however,
both derivations may lead to a string that starts with $A$ (in the
first derivation there is still the nonterminal $K$ at the beginning,
which may be transformed to $A$). This contradicts the fact that
the next step of the leftmost derivation
is determined uniquely
by the first character of the derived string.
(This uniqueness
is valid for derivations that start from the axiom; recall that
$K$ may appear in such a derivation by assumption.)\es

\smallskip

     Therefore, the
LL(1) approach cannot be applied to left-recursive
grammars (except for trivial cases). We have to transform them to
equivalent LL(1)-grammars first (or use other parsing algorithms).

    \Problem{For any LL(1)-grammar, construct
an algorithm that checks if the input string
belongs to the language generated by the grammar. Use the
preceding results.}

    \Solution. We follow the scheme outlined above and look for a
leftmost derivation of the given string. At each point, we have an
initial part of the leftmost derivation that ends with a string
composed of terminals and nonterminals. This string has the processed
part of the input string as a prefix. Our algorithm stores the
remaining part. In other words, we keep a string \w{S} of terminals and
nonterminals with the following properties (the processed part of the input
string is denoted by \w{A}):
\begin{enumerate}
   \item the string \verb|AS| is derivable (in the grammar);
   \item any leftmost derivation of the input string includes
         the string \verb|AS|.
\end{enumerate}
These properties are denoted by ``(I)'' in the sequel.

\smallskip

    Initially, \w{A} is empty and \w{S} contains only one nonterminal (the
axiom).

\smallskip

    If at some point the string \w{S} begins with a terminal \w{t}
and $\w{t}=\w{Next}$, then we may call the procedure \w{Move} and
delete the initial terminal \w{t} from \w{S}. Indeed, this operation
leaves \w{AS} unchanged.

    If the string \w{S} starts with a terminal \w{t} and
$\w{t}\neq\w{Next}$, then the input string is not derivable at all,
because (I) implies that any (leftmost) derivation goes through the
stage \w{AS}.
(The same is true if $\w{Next}=\w{EOI}$.)

   If \w{S} is empty, the condition (I) implies that the input string
is derivable if and only if $\w{Next}=\w{EOI}$.

   The only remaining case is that \w{S} starts with some nonterminal
\w{K}. As we have already shown, all the leftmost derivations that start
with \w{S} and end with a string whose first character is \w{Next},
begin with the same production rule, that is, the production rule whose
set of leading terminals includes \w{Next}. If such a rule does not
exist, the input string is not derivable at all. If such a rule
exists, we apply it to the opening nonterminal \w{K} of the
string \w{S} and property (I) remains valid.
We arrive at the following algorithm:
\begin{verbatim}
 S := empty string;
 error := false;
 {error => input string is not derivable}
 {not error => (I)}
 while (not error) and not ((Next=EOI) and (S is empty))
 |    do begin
 | if (S starts with a terminal equal to Next) then begin
 | | Move; delete the first symbol from S;
 | end else if (S starts with a terminal different from Next)
 | |   then begin
 | | error := true;
 | end else if (S is empty) and (Next <> EOI) then begin
 | | error := true;
 | end else if (S starts with some nonterminal K and Next
 | |    belongs to the set of leading terminals for one of
 | |    the production rules for K) then begin
 | | apply this rule to K
 | end else if (S starts with some nonterminal K and Next
 | |    does not belong to the set of leading terminals
 | |    for all the production rules for K) then begin
 | | error := true;
 | end else begin
 | | {this cannot happen}
 | end;
 end;
 {the input string is derivable <=> not error}
\end{verbatim}
This algorithm always terminates. Indeed, if a terminal appears as
the first symbol in \w{S}, the algorithm stops or reads the next
input character. If nonterminals alternate as first symbols
of \w{S} in an infinite loop, then the grammar is left-recursive;
we may assume that this is not the case. (This follows from the
preceding problem; we may easily remove from the grammar
all the
nonterminals that do not appear in derivations beginning with
the axiom; the same can be done for
nonterminals from which only the empty string
is derivable.)\es

\medskip

     Remarks.
\begin{itemize}
   %
\item This algorithm uses \verb|S| as a stack (all operations
are made near its left end).
   %
\item In either of the last two cases (in the \w{if}-construct), no
input characters are read. Therefore, we can precompute the action
for all nonterminals and all possible values of \w{Next}.
Doing that, we need only one iteration per input character.
   %
\item In practice, it is convenient to have a table that lists
all actions for all pairs (input symbol, nonterminal), and a
small program that interprets this table.
\end{itemize}

\Problem{
To check if a given grammar is an LL(1)-grammar, we need to
compute Follow($T$) and First($T$) for all nonterminals $T$.
How can we do that?
}

\Solution.
If the grammar includes, say, the rule $K\to L\,M\,N$, then
($\Lambda$ denotes the empty string):
   %                                   
{\setlength{\tabcolsep}{.6\tabcolsep}
%\small
$$
\begin{tabular}{rcll}
$\r{First}\,(L)$ & $\subset$ & $\r{First}\,(K),$ & \\
$\r{First}\,(M)$ & $\subset$ & $\r{First}\,(K),$ &
   if $\Lambda$ is derivable from $L$,\\
$\r{First}\,(N)$ & $\subset$ & $\r{First}\,(K),$ &
   if $\Lambda$ is derivable both from $L$ and $M$,\\[.5ex]
%
$\r{Follow}\,(K)$ & $\subset$ & $\r{Follow}\,(N),$ & \\
$\r{Follow}\,(K)$ & $\subset$ & $\r{Follow}\,(M),$ &
   if $\Lambda$ is derivable from $N$,\\
$\r{Follow}\,(K)$ & $\subset$ & $\r{Follow}\,(L),$ &
   if $\Lambda$ is derivable both from $M$ and $N$,\\[.5ex]
%
$\r{First}\,(N)$ & $\subset$ & $\r{Follow}\,(M),$ & \\
$\r{First}\,(M)$ & $\subset$ & $\r{Follow}\,(L),$ & \\
$\r{First}\,(N)$ & $\subset$ & $\r{Follow}\,(L),$ &
   if $\Lambda$ is derivable from $M$.\\
\end{tabular}
$$%
}%
These rules (written for all productions)
allow us to generate the sets
First($T$), and thereafter Follow($T$), for all terminals and
nonterminals $T$.
As a starting point we use
$$
        \w{EOI} \in \r{Follow}\,(K)
$$
for an initial nonterminal $K$ (the axiom) and
$$
        z \in \r{First}\,(z)
$$
for any terminal $z$. We stop the generation process when the
repeated applications of the rules give no new elements of the sets
First($T$) and Follow($T$).\es

