\section{Left-to-right parsing (LR)}

     Here we consider another approach to parsing, called
LR(1)-{\it parsing} algorithm, as well as some simplified versions of it.
 \index{parsing!LR(1)}

\subsection{LR-processes}

     There are two main differences between
LR(1)-parsing and
LL(1)-parsing. First, we seek a {\it rightmost\/} derivation,
not a leftmost one. Second, we construct the derivation
from the
bottom (beginning with
the input
string) to the
top (the axiom) and not vice versa (as in LL(1)-parsing).
 \index{derivation!rightmost}%

A {\it rightmost\/} derivation is a derivation where the rightmost
nonterminal is replaced at each step.

\Problem{Prove that any derivable string of terminals has a
rightmost derivation.\nobreak\es}

     It is convenient to look at the rightmost derivation
backwards, starting from the input string.
Let us define the notion of an LR-{\it process\/}%
 \index{LR-process}
on the input string
$A$. This process involves the string $A$ and another string $S$ that
contains both terminals and nonterminals. Initially, the string $S$
is empty. The LR-process includes two types of actions:
\begin{description}
   \item[(1)] the first character of $A$ (called
the next input symbol and denoted by \w{Next}) may be moved
to the end of the string $S$ (and deleted from $A$); this
action is called a {\it shift} action;%
 \index{shift}
  \item[(2)] if the right-hand side of some production rule is
a suffix of $S$, then it can be replaced by the nonterminal
that is on the left-hand side of that rule; the string $A$
remains unchanged. This action is called a {\it reduce\/} action.
 \index{reduce}%
\end{description}

     Let us mention that the LR-process is not deterministic;
there are situations where many different actions are possible.

     We say that the
LR-process on a string $A$ is
{\it successful} if the string $A$ becomes empty and the string
$S$ contains only one nonterminal, and this
nonterminal is the initial nonterminal
(the axiom).

     \Problem{Prove that for any string $A$ (of terminals)
a successful LR-process exists if and only if $A$ is derivable
in the grammar. Find a one-to-one correspondence between
rightmost derivations and successful LR-processes.}

    \Solution.  The shift action does not change the string $SA$.
The reduce action changes $SA$ and this change is a reversed
step of a derivation. This derivation is a rightmost one because
the reduction is done at the end of $S$ and all symbols of $A$
are terminals. Therefore, each LR-process corresponds to a
rightmost derivation.

Conversely,
assume that a rightmost derivation is given. Imagine a separator
placed
after the last nonterminal in the string.
When
a production rule
is applied to that nonterminal, we may
need to move the separator
to the left (if the right-hand side of the rule applied ends with
a terminal). Splitting this move into steps (one symbol
per step) we get a process that is exactly an inverted LR-process.\es

     All changes in the string $S$ during an LR-process are made
near its right end. This is why the string $S$ is called the
{\it stack\/} of the LR-process.

     So the problem of finding the rightmost derivation
of a given string is the problem of constructing a successful
LR-process on this string. At each step we have to decide whether
we want to apply a shift or reduce action, and choose a
production rule if several reductions are possible. In the
LR(1)-algorithm,
the decision is made based on $S$ and the first symbol
of $A$. If only information about $S$ is used, it is an
LR(0) algorithm. (The exact definitions are given below.)

    Assume that a grammar is fixed. In the sequel, we assume that
for each nonterminal there exists
a string of terminals derivable from it.

     Let $\w{K}\to\w{U}$ be one of the grammar's rules (\w{K}
is a nonterminal,  \w{U} is a string of terminals and nonterminals).
We consider a set of strings (composed of both terminals and
nonterminals) called the {\it left context\/} of the
rule
$\w{K}\to\w{U}$.%
 \index{left context of the rule}
 \index{LeftCont($K \to U$)}
(Notation: $\r{LeftCont}(\w{K}\to\w{U})$.) By definition, this set
contains all the strings that
may appear as a stack content
immediately before the
reduction of \w{U} to \w{K} in a successful LR-process.

   \Problem{Reformulate this definition in terms of rightmost
derivations.}

    \Solution. Consider all rightmost derivations of the form
$$
   \langle\r{axiom}\rangle
                 \leadsto\w{XKA} \to \w{XUA},
$$
where \w{A} is a string of terminals, \w{X} is a string of
terminals and
nonterminals, and $\w{K}\to\w{U}$ is a production rule. All
strings \w{XU} that appear in those derivations form the left
context of the rule
$\w{K}\to\w{U}$. Indeed, recall that we assume that for any
nonterminal there exists a string of terminals derivable from
it; therefore, the rightmost derivation of the string
\w{XUA} may be continued until a right derivation of some string of
terminals is obtained.\es

\Problem{%
All strings from
$\r{LeftCont}(\w{K}\to\w{U})$ end with \w{U}. Prove that if we
delete this suffix \w{U}, the resulting set of strings does not
depend on which rule (for the nonterminal \w{K}) is chosen.
This set is denoted by
$\r{Left}(\w{K})$.}%
 \index{Left($K$)}

    \Solution. The preceding problem shows that $\r{Left}(\w{K})$
is the set of all strings that may appear at the left of the
rightmost nonterminal \w{K} in some rightmost derivation.\nobreak\es

\Problem{%
Prove that in the last sentence the words ``the rightmost nonterminal''
may be omitted:
$\r{Left}(\w{K})$ is the set of all strings that may appear
on the left of any occurence of \w{K}
in a rightmost derivation.}

    \Solution. The derivation may be continued and all nonterminals
on the right of \w{K} may be replaced by terminals; this replacement
does not change anything on the left of \w{K}.\es

\Problem{Let $G$ be a grammar. Construct a new grammar $G^l$
such that for any nonterminal \w{K} of $G$, the grammar $G^l$
contains a nonterminal $\langle\r{Left}\w{K}\rangle$, and all
elements of $\r{Left}(\w{K})$ (and no others) are derivable from
$\langle\r{Left}\w{K}\rangle$ in $G^l$. The terminals of $G^l$
are nonterminals and terminals of $G$. }

    \Solution. Let \w{P} be the initial nonterminal of $G$.
The new grammar $G^l$ has a rule
$$
     \langle\r{Left}\w{P}\rangle \to \qquad(\hbox{right-hand side is
      the empty string})
$$
For any production rule of the $G$, say,
$$
     \w{K} \to \w{L}\,\w{t}\,\w{M}\,\w{N}\qquad
    \hbox{(\w{L}, \w{M}, \w{N} are nonterminals, \w{t} is a terminal)}
$$
we add the following rules to $G^l$:
\begin{eqnarray*}
\langle\r{Left}\w{L}\rangle& \to &   \langle\r{Left}\w{K}\rangle\\
\langle\r{Left}\w{M}\rangle& \to &
\langle\r{Left}\w{K}\rangle\,\w{L}\,\w{t}\\
\langle\r{Left}\w{N}\rangle& \to &
             \langle\r{Left}\w{K}\rangle\,\w{L}\,\w{t}\,\w{M}
\end{eqnarray*}
etc. The meaning of the new rules may be explained as follows.
An empty string may appear on the left of \w{P}. If a string
\w{X} may appear on the left of \w{K}, then \w{X} may appear on the
left of \w{L}; at the same time
\w{XLt} may appear on the left of \w{M}, and \w{XLtM}
may appear
on the left of \w{N}. By induction over the length of a
rightmost derivation, we check that everything that may appear
on the left of some nonterminal, appears according to these
rules.\es

\Problem{%
    Why is it important in the preceding problem that we consider
only
the rightmost derivations?}

\Solution. Otherwise we must take into account transformations performed
on the left of \w{K}.\es

\Problem{%
    A context-free grammar is given. Construct an algorithm that for
any input string finds all the sets
$\r{Left}(\w{K})$ containing the string.}

     Remark (for experts only). The existence of such an
algorithm, even a finite automaton (an inductive extension with
a finite number of values, see section~\ref{indfunc}), follows
from the preceding problem. Indeed, the grammar constructed has
a special form: The right-hand sides of rules contain only one
nonterminal and it is in the leftmost position. Nevertheless, we
give an explicit construction of that automaton below.

    \Solution. By a {\it situation\/} of a given grammar we mean
 \index{situation!for a grammar}
one of its rules with some additional information; namely, one
of the positions on the right-hand side (before the first
symbol, between the first and the second symbols,\,\ldots, after
the last symbol) is marked. For example, the rule
 %
$$
  \w{K}\,\to\,\w{L}\,\w{t}\,\w{M}\,\w{N}
$$
(\w{K}, \w{L}, \w{M}, \w{N} are nonterminals, \w{t} is a terminal)
gives five situations
$$
  \w{K}\to\_\,\w{L}\,\w{t}\,\w{M}\,\w{N} \quad
  \w{K}\to\w{L}\,\_\,\w{t}\,\w{M}\,\w{N} \quad
  \w{K}\to\w{L}\,\w{t}\,\_\,\w{M}\,\w{N} \quad
  \w{K}\to\w{L}\,\w{t}\,\w{M}\,\_\,\w{N} \quad
  \w{K}\to\w{L}\,\w{t}\,\w{M}\,\w{N}\,\_
$$
(the position is indicated by the underscore sign).

    We say that a string \w{S} is {\it coherent}
\index{coherent}%
\index{string!coherent with a situation}%
with a situation
$\w{K}\to\w{U}\,\_\w{V}$ if \w{S} ends with \w{U}, that is, if
$\w{S}=\w{T}\w{U}$ for some \w{T} and, moreover, \w{T}
belongs to $\r{Left}(\w{K})$. (The meaning of this definition may
be explained as follows: the suffix \w{U} of the stack \w{S} is
ready for the future reduction of \w{UV} into \w{K}.)
Now we can give an equivalent definition of
$\r{LeftCont}(\w{K}\to\w{X})$ as the set of all strings that are
coherent with the situation $\w{K}\to\w{X}\,\_\,$, and
$\r{Left}(\w{K})$ as the set of all strings coherent with the
situation $\w{K}\to\_\,\w{X}$ (here $\w{K}\to\w{X}$ is any
production rule for nonterminal \w{K}).

  Here is an equivalent definition in terms of LR-processes: \w{S}
is coherent with the situation $\w{K}\to\w{U}\,\_\,\w{V}$ if there
exists a successful LR-process such that:
\begin{itemize}
  \item during the process, the string \w{S} appears in the stack
and \w{S} ends with \w{U};

  \item for some time \w{S} is not touched and the string \w{V}
appears on the right of \w{S};

  \item \leavevmode \w{UV} is reduced into \w{K};

  \item the LR-process continues and eventually terminates
      successfully.
\end{itemize}

\Problem{%
    Prove the equivalence of these two definitions.}

\Hint{If $\w{S} =\w{TU}$ and \w{T} belongs to
$\r{Left}(\w{K})$, then it is possible to have \w{T} on the stack,
then add
\w{U}, then \w{V}, then reduce \w{UV} to \w{K}, and finally finish
the LR-process successfully.
(Several times we use the assumption that for any nonterminal
there exists some string of terminals derivable from it;
this assumption
guarantees that we may add an
arbitrary
string to the stack.)}

\medskip

