     Our goal is to construct an algorithm that finds all \w{K}
such that the input string belongs to $\r{Left}(\w{K})$.
Consider a function that maps each string \w{S} (of terminals and
nonterminals) into the set of all situations that are coherent with \w{S}.
This set is called a  {\it state corresponding to\/} \w{S}.%
 \index{State($S$)}
We denote it by $\r{State}(\w{S})$. It is enough to show that the
function $\r{State}(\w{S})$ is inductive, that is, the value
$\r{State}(\w{SJ})$ for any terminal or nonterminal \w{J} is determined
by
$\r{State}(\w{S})$ and \w{J}. (We have seen that membership in
$\r{Left}(\w{K})$ may be expressed in terms of that function.)
Indeed, the value $\r{State}(\w{SJ})$ can be
computed according to the following
rules (1)--(3):

\begin{quote}
 (1) If the string \w{S} is coherent with the situation
$\w{K}\to\w{U}\,\_\,\w{V}$, and the string \w{V} starts with the
symbol \w{J}, that is, $\w{V}=\w{JW}$, then \w{SJ}
is coherent with the situation $\w{K}\to\w{UJ}\,\_\,\w{W}$.
\end{quote}

     This rule determines completely what situations
not starting with an underscore
are coherent with \w{SJ}. It remains to find for which nonterminals
\w{K} the string \w{SJ} belongs to
$\r{Left}(\w{K})$.
This can be
done according to the following rules:

\begin{quote}
     (2) If the situation
$\w{L}\to\w{U}\,\_\,\w{V}$ turns out to be coherent with
\w{SJ}
(according to
(1)) and \w{V} starts with a nonterminal \w{K}, then \w{SJ}
belongs to $\r{Left}(\w{K})$.

     (3) If \w{SJ} is in $\r{Left}(\w{L})$ for some \w{L},
the grammar contains a production rule
$\w{L}\to\w{V}$ and
\w{V} starts with a nonterminal \w{K}, then \w{SJ} belongs to
$\r{Left}(\w{K})$.
\end{quote}

     Please note that the rule (3) may be considered a
version of rule (2). Indeed, if the assumptions of (3)
are valid, then the situation
$\w{L}\to\_\w{V}$ is coherent with \w{SJ} and \w{V} starts with
a nonterminal \w{K}.

     The correctness of these rules becomes more or less
evident upon reflection. The only thing that requires
comment is why the rules (2) and (3) generate {\it all\/}
terminals \w{K} such that \w{SJ} belongs to
$\r{Left}(\w{K})$. Let us try to explain why.
Consider a rightmost derivation where \w{SJ} is on the left of
\w{K}. How can the nonterminal \w{K} appear in this derivation?
If the production rule that created \w{K} created a suffix of
the string \w{SJ} at the same time, then the membership of
\w{SJ} in $\r{Left}(\w{K})$ will be disclosed according to the rule (2).
On the other hand, if \w{K} was the first symbol in a string generated
by some other nonterminal \w{L}, then (because of the rule (3)) it is
enough to check that \w{SJ} belongs to
$\r{Left}(\w{L})$. It remains to apply the same argument to \w{L} and
so on.

    In terms of an LR-process, the same idea may be expressed as
follows. First, the nonterminal \w{K} may participate in several
reductions that do not touch \w{SJ} (those reductions correspond
to applications of the rule (3)\,). Then a reduction
that touches \w{SJ} is performed
(this reduction corresponds to an application
of rule (2)\,).

     It remains to determine which situations are coherent with the
empty string, that is, for which nonterminals
\w{K}, the empty string
belongs to $\r{Left}(\w{K})$. This can be done according to the
following rules:
\begin{quote}
(1) the initial nonterminal (the axiom) has this property;

(2) if \w{K} has this property,
$\w{K}\to \w{V}$ is a production rule, and
the string \w{V} starts with a nonterminal \w{L}, then \w{L}
has this property as well.\es
\end{quote}

\Problem{%
     Perform the above analysis on the grammar}
\vspace{-4mm}
\begin{eqnarray*}
     \w{E}&\to& \w{E}\;\w{+}\;\w{T}\\
     \w{E}&\to& \w{T}\\
     \w{T}&\to& \w{T}\;\w{*}\;\w{F}\\
     \w{T}&\to& \w{F}\\
     \w{F}&\to& \w{x}\\
     \w{F}&\to& \w{(}\;\w{E}\;\w{)}
\end{eqnarray*}
\label{ETF}
(which generates the same language as the grammar of
example 3, p.~\pageref{cf3}).

    \Solution. The sets State(\w{S}) for different \w{S}
are shown in the table, p.~\pageref{firsttable}.
\begin{figure}[htbp]
{%\scriptsize
\label{firsttable}
$$
\renewcommand{\arraystretch}{1.3}
\begin{tabular}{|l|l|}
\hline
String \w{S} & $\r{State}(\w{S})$\\
\hline
empty & $\w{E}\to\_\,\w{E+T}\quad
          \w{E}\to\_\w{T}\quad
          \w{T}\to\_\w{T*F}$\\
       & $\quad \w{T}\to\_\w{F}\quad
          \w{F}\to\_\,\w{x}\quad
          \w{F}\to\_\w{(E)}$\\
\w{E}  & $\w{E}\to\w{E}\,\_\w{+T}$\\
\w{T}  & $\w{E}\to\w{T}\_\quad
         \w{T}\to\w{T}\_\w{*F}$\\
\w{F}  & $\w{T}\to\w{F}\_$\\
\w{x}  & $\w{F}\to\w{x}\,\_$\\
\w{(}  & $\w{F}\to\w{(}\_\,\w{E)}\quad
          \w{E}\to\_\,\w{E+T}\quad
          \w{E}\to\_\w{T}$\\
 &$\quad  \w{T}\to\_\w{T*F}\quad
          \w{T}\to\_\,\w{F}\quad
          \w{F}\to\_\w{x}\quad
          \w{F}\to\_\,\w{(E)}$\\
\w{E+} & $\w{E}\to\w{E+}\_\w{T}\quad
          \w{T}\to\_\w{T*F}\quad
          \w{T}\to\_\,\w{F}$\\
 &$\quad  \w{F}\to\_\,\w{x}\quad
          \w{F}\to\_\,\w{(E)}$\\
\w{T*} & $\w{T}\to\w{T*}\_\,\w{F}\quad
          \w{F}\to\_\,\w{x}\quad
          \w{F}\to\_\,\w{(E)}$\\
\w{(E} & $\w{F}\to\w{(E}\,\_\w{)}\quad
          \w{E}\to\w{E}\,\_\w{+T}$\\
\w{(T} & = \w{T}\\
\w{(F} & = \w{F}\\
\w{(x} & = \w{x}\\
\w{((} & = \w{(}\\
\w{E+T}& $\w{E}\to\w{E+T}\_\quad
          \w{T}\to\w{T}\_\w{*F}$\\
\w{E+F}& = \w{F}\\
\w{E+x}& = \w{x}\\
\w{E+(}& = \w{(}\\
\w{T*F}& $\w{T}\to\w{T*F}\_$\\
\w{T*x}& = \w{x}\\
\w{T*(}& = \w{(}\\
\w{(E)}& $\w{F}\to\w{(E)}\_$\\
\w{(E+}& = \w{E+}\\
\w{E+T*}& = \w{T*}\\
\hline
\end{tabular}
$$
\hbox to \textwidth{\hss \hbox{State(\w{S}), problem 14.1.10}\hss}
\par}
\end{figure}
The equals sign means that the sets of situations that are values
of the function $\r{State}(\w{S})$ of the strings connected by the
the equals sign are equal.

Here is the rule to find $\r{State}(\w{SJ})$, provided we know
$\r{State}(\w{S})$ and \w{J} (here \w{S} is a string of terminals and
nonterminals, and \w{J} is a terminal or nonterminal):
\begin{quote}
Find $\r{State}(\w{S})$ in the right column;
consider
the corresponding
string \w{T} in the left column; append the symbol \w{J} to the end
and find the set corresponding to the string \w{J}.
(If the string \w{J} is not in the table, then
$\r{State}(\w{SJ})$
is empty.)\es
\end{quote}

\subsection{LR(0)-grammars}

     Recall that our goal is to find a derivation for a given
string. In other words, we seek a successful LR-process on this
string. In all cases where our methods are applicable, such a
successful LR-process (for a given string) is unique. We find it
stepwise. At any point, we find the only possible next step. To
ensure that only one step is possible, we need to put some
requirements on the grammar. In this section we consider the
simplest case, the so-called LR(0)-{\it grammars}.

     As we already know:
\begin{quote}
     (1) The reduction according to the rule
     $\w{K}\to\w{U}$ with stack \w{S} may appear  in a successful
     LR-process if and only if
     \w{S} belongs to $\r{LeftCont}(\w{K}\to\w{U})$
     or, equivalently, if \w{S} is coherent with situation
     $\w{K}\to\w{U}\_$.
\end{quote}
     A similar statement about shift is as follows:
\begin{quote}
     (2) A shift with next symbol \w{a} and stack \w{S}
     may appear in a successful LR-process if and only if
     \w{S} is coherent with some situation of type
     $\w{K}\to\w{U}\_\w{aV}$.
\end{quote}

\Problem{%
Prove the above claim.}

\Hint{Assume that a shift occurs and a new terminal \w{a}
is added to the stack \w{S}. Consider the first reduction that
includes this terminal.}

    Assume that some grammar is fixed. Consider an arbitrary
string \w{S} of terminals and nonterminals. If the set
$\r{State}(\w{S})$ contains a situation where the underscore sign
is followed by a terminal, we say that the string
\w{S} {\it allows a shift}. If the set $\r{State}(\w{S})$ contains a situation
where the underscore sign is the last symbol, we say that the string
\w{S} {\it allows a reduction\/} (according to the corresponding
rule). We say that there is a {\it shift/reduce\/} conflict for the
string \w{S}
 \index{conflict!shift/reduce}
 \index{conflict!reduce/reduce}
if both shift and reduction are allowed.
We say that there is a {\it reduce/reduce\/} conflict
for \w{S} if the string \w{S} allows a reduction
according to two different
rules.

