     The algorithm that checks whether a given string is
derivable in a LR(0)-grammar uses only some of the
information
available.
Indeed, for each state it knows in advance which
action (shift or reduction --- and which reduction) is the only
possible one. More elaborate algorithms
can make a choice
using the next input symbol
as well as the stack content.
Looking at the set State(\w{S}), it is easy to say for which
input symbols a shift is possible. (It is possible for all
terminals that follow the underscore in situations coherent
with \w{S}.) The more difficult problem is: How do we use the next
input symbol to decide if a reduction is possible?

There are two methods: the first is simpler, the second is more
powerful. The grammars for which the first method is applicable are
called SLR(1)-grammars (S for Simple). The second method uses all
available information; these grammars
are called LR(1)-grammars. (There is also an intermediate class
of grammars called LALR(1)-grammars, discussed below.)

\subsection{SLR(1)-grammars}

     Recall that for any nonterminal \w{K} we have defined (see
p.~\pageref{follow}) the set $\r{Follow}(\w{K})$ of
terminals that may follow $\w{K}$ in strings derivable from the
initial nonterminal; this set also includes  the symbol \w{EOI} if
\w{K} may appear at the end of a derivable string.

\Problem{%
     Prove that if at some point of the LR-process the last
symbol of the stack \w{S} is \w{K} and the process can be
finished successfully, then \w{Next} belongs to
$\r{Follow}(\w{K})$.}

   \Solution. This fact is an immediate consequence of the
definition (recall the correspondence between rightmost
derivations and successful LR-processes).\es

\medskip

     Assume that some grammar is fixed. Consider a string \w{S}
of terminals and nonterminals, and a terminal \w{x}. If the set
$\r{State}(\w{S})$ contains a situation where the underscore is
followed by a terminal \w{x}, we say that the pair
$\langle\w{S},\w{x}\rangle$ {\it allows a shift}. If the set
$\r{State}(\w{S})$ contains a situation $\w{K}\to\w{U}\_\,$
where \w{x} belongs to $\r{Follow}(\w{K})$, we say that the pair
$\langle \w{S},\w{x}\rangle$ SLR(1)-{\it allows reduction\/}
(according to
the
rule $\w{K}\to\w{U}$). We say that for the pair
$\langle \w{S},\w{x}\rangle$ there is a SLR(1)-{\it conflict of
type shift/reduce}, if both shift and reduction are allowed. We
say that
 \index{conflict!shift/reduce}%
 \index{conflict!reduce/reduce}%
for the pair $\langle \w{S},\w{x}\rangle$ there is a SLR(1)-{\it
conflict of type reduce/reduce} if reductions according to
different rules are allowed.

      The grammar is called a SLR(1)-{\it grammar\/} if it has no
 \index{grammar!SLR(1)}%
 \index{SLR(1)-grammar}%
SLR(1)-conflicts of type shift/reduce and reduce/reduce for all
pairs $\langle \w{S},\w{x}\rangle$.

\Problem{%
    Assume that a SLR(1)-grammar is given. Prove that each
string has at most one rightmost derivation. Give an algorithm
to check whether a given string is derivable in the grammar.}

     \Solution. We
repeat the argument used for
LR(0)-grammars. The
difference is that the choice of the next action depends on the
next input symbol (\w{Next}).\es

\Problem{%
      Check if the grammar shown above on p.~\pageref{ETF}
(having nonterminals \w{E}, \w{T} and \w{F}) is an
SLR(1)-grammar.}

     \Solution. Yes; both conflicts that prevent it from being
a LR(0)-grammar are resolved when we take the next input
symbol into account. Indeed, for both \w{T} and \w{E+T} a
shift is possible only when $\w{Next}=\w{*}$, and the symbol
\w{*} belongs neither to $\r{Follow}(\w{E}) =
\{\w{EOI},\w{+},\w{)}\}$ nor to $\r{Follow}(\w{T}) =
\{\w{EOI},\w{+},\w{*},\w{)}\}$. Therefore, reduction is
impossible when $\w{Next}=\w{*}$.\es


\subsection{LR(1)-grammars, LALR(1)-grammars}

     The SLR(1) approach still does not use all available
information to decide if reduction is possible. It checks
(for a given rule)
whether reduction is possible with a given stack
content
and {\it
separately\/} checks whether reduction
is possible with a given input symbol \w{Next}. However,
these tests are not independent. It may happen that both checks
give a positive answer, but nevertheless the reduction for the
given \w{S} {\it and\/} the given \w{Next} is impossible. The
LR(1)-approach is free of this deficiency.

     The LR(1)-approach is as follows: All our definitions and
statements are modified to take into account what symbol is on
the right of the replaced nonterminal while using a production
rule. In other words, we carefully inspect the next symbol when
reduction is performed.

     Let $\w{K}\to\w{U}$ be one of the production rules, and let
\w{t} be a terminal or a special symbol \w{EOI} (which is
assumed to be at the end of the input string). We define the
set
 \index{LeftCont($K\to U$, $t$)}
$\r{LeftCont}(\w{K}\to\w{U},\w{t})$ as the set of all strings
that may be the stack content
immediately before the reduction \w{U} to
\w{K} during a successful LR-process, assuming that $\w{Next} =
\w{t}$ at the time of reduction.

    All strings in $\r{LeftCont}(\w{K}\to\w{U})$ have suffix
\w{U}. If we discard this suffix, we obtain the set of all strings
that appear in the rightmost derivations immediately before the
nonterminal \w{K} followed by \w{t}. This set (which
does not depend on the specific rule $\w{K}\to\w{U}$, but only
on the nonterminal \w{K}) is denoted by
 \index{Left($K$, $t$)}
$\r{Left}(\w{K},\w{t})$.

\Problem{%
     Write a grammar
whose nonterminals
generate the sets
     $\r{Left}(\w{K},\w{t})$
for all nonterminals \w{K} of the given grammar.}

     \Solution. Nonterminals are symbols
$\langle\w{LeftK}\:\:\w{t}\rangle$ for any nonterminal \w{K} and any terminal
\w{t} (and also for $\w{t}=\w{EOI}$). Its production rules are as follows:
Let \w{P} be the initial nonterminal (the axiom) of the given grammar.
Then our new grammar has the rule
$$  \langle \w{LeftP}\:\:\w{EOI}\rangle\to \qquad
    \hbox{(the right-hand side is the empty string).}$$
Each rule of the given grammar produces several rules of the
new one. For example, if the given grammar has a rule
$$
    \w{K}\to\w{L}\,\w{u}\,\w{M}\,\w{N}
$$
(\w{L}, \w{M}, \w{N} are nonterminals, \w{u} is a terminal),
then the new grammar has rules
 %
$$ \langle\w{LeftL}\:\w{u}\rangle \to
   \langle\w{LeftK}\:\w{x}\rangle$$
for all terminals \w{x};
 %
$$ \langle\w{LeftM}\:\w{s}\rangle \to
   \langle\w{LeftK}\:\w{y}\rangle\,\w{L}\,\w{u}$$
for any \w{s} that may appear as a first character in a string
derivable from \w{N}, and for any \w{y}, as well as for all pairs
$\w{s} = \w{y}$, if the empty string is derivable from \w{N}); and
 %
$$ \langle\w{LeftN}\:\w{s}\rangle \to
   \langle\w{LeftK}\:\w{s}\rangle\,\w{L}\,\w{u}\,\w{M}$$
for any terminal \w{s}.\es

\Problem{%
     How
should we
modify the definition of a situation?}

     \Solution. Now a situation is defined as a pair
$$[
\hbox{\rm situation in the old sense},
\hbox{\rm terminal or \w{EOI}}
]\eqno\es$$

\Problem{%
     How to modify the definition of a string coherent with a
situation?}

    \Solution. The string \w{S} (of terminals and nonterminals)
is coherent with the situation $[\w{K}\to\w{U}\_\w{V},\,\w{t}]$
(here \w{t} is a terminal or \w{EOI}) if \w{U} is a suffix of
\w{S}, that is, if $\w{S}=\w{TU}$, and, moreover, \w{T} belongs to
$\r{Left}(\w{K},\w{t})$.\es

\Problem{%
Show how to compute inductively the set
$\r{State}(\w{S})$ of all situations coherent with a given
string \w{S}.}

     Answer.

\begin{quote}
  %
(1) If a string \w{S} is coherent with a situation
$[\w{K}\to\w{U}\_\w{V},\,\w{t}]$ and the first character in
\w{V} is \w{J}, that is, $\w{V}=\w{JW}$, then \w{SJ} is coherent
with the situation $[\w{K}\to\w{UJ}\_\w{W},\,\w{t}]$.
 %
\end{quote}

\noindent
     This rule determines completely which situations that do
not start with underscore are coherent with \w{SJ}. It remains
to find out for which nonterminals \w{K} and terminals \w{t} the
string \w{SJ} belongs to $\r{Left}(\w{K},\w{t})$. This is done
according to the following two rules:

\begin{quote}
     (2) If the situation $[\w{L}\to\w{U}\_\w{V},\w{t}]$ is
coherent with \w{SJ} (according to (1)) and \w{V} starts with a
nonterminal \w{K}, then \w{SJ} belongs to $\r{Left}(\w{K},\w{s})$
for any terminal \w{s} that may appear as a first symbol in a
string derivable from $\w{V}\setminus\w{K}$ (the string \w{V}
without the first symbol \w{K}), as well as for $\w{s}=\w{t}$,
if the empty string is derivable from $\w{V}\setminus\w{K}$.

\smallskip

    (3) If \w{SJ} is in $\r{Left}(\w{L},\w{t})$ for some \w{L}
and \w{t}, and $\w{L}\to\w{V}$ is a production rule, and \w{V}
starts with a nonterminal \w{K}, then \w{SJ} belongs to
$\r{Left}(\w{K},\w{s})$ for any nonterminal \w{s} that may
appear as a first symbol in a string derivable from
$\w{V}\setminus\w{K}$, as well as for $\w{s}=\w{t}$, if the
empty string is derivable from $\w{V}\setminus\w{K}$.\es
 %
\end{quote}


