\Problem{%
      Give the definition of
the
shift/reduce and shift/shift
conflicts in the LR(1)-case.}

     Solution. Assume that a grammar is fixed. Let \w{S} be an
arbitrary string of terminals and nonterminals. If the set
$\r{State}(\w{S})$ contains a situation where the underscore
sign is followed by a terminal \w{t}, we say that the pair
$\langle \w{S},\w{t}\rangle$ {\it allows a shift}.
(This definition is
the same as in the SLR(1)-case; we ignore the second components of
pairs in $\r{State}(\w{S})$.)

     If $\r{State}(\w{S})$ contains a situation whose first
component ends with the underscore sign and the second component
is a terminal \w{t}, we say that the pair $\langle
\w{S},\w{t}\rangle$ LR(1)-{\it allows a reduction\/} (via
the corresponding rule). We say that there is a LR(1)-{\it conflict
of type shift/reduce\/} for a pair $\langle \w{S},\w{t}\rangle$ if
this pair allows both shift and reduction.
 \index{conflict!shift/reduce}%
 \index{conflict!reduce/reduce}%
We say that there is a
LR(1)-{\it conflict of type reduce/reduce\/}
for a pair $\langle \w{S},\w{t}\rangle$ if this pair allows
reductions according to different rules.\es

    The grammar is called a LR(1)-{\it grammar}, if there are
 \index{grammar!LR(1)}%
 \index{LR(1)-grammar}%
no LR(1)-conflicts of type shift/reduce and reduce/reduce for
all pairs $\langle \w{S},\w{t}\rangle$.

\Problem{%
For any LR(1)-grammar,
construct an algorithm that checks if a given string is
derivable in the grammar.}

    Solution. As before, at each stage of the LR-process we can
determine which action is the only possible one.\es

\medskip

     It is useful to understand how the notions of
LR(0)-coherence and LR(1)-coherence are related. (It is used
below, when LALR(1)-grammars are considered.)

\Problem{%
     Find and prove the connection between the notions of
LR(0)-coherence and LR(1)-coherence.}

  \Solution. Assume that a grammar is fixed. The string \w{S} of
terminals and nonterminals is LR(0)-coherent with situation
$\w{K}\to\w{U}\_\w{V}$ if and only if it is LR(1)-coherent with
the pair $[\w{K}\to\w{U}\_\w{V},\w{t}]$ for some terminal \w{t}
(or for $\w{t}=\w{EOI}$). In other words, $\r{Left}(\w{K})$ is
the union of the sets $\r{Left}(\w{K},\w{t})$ for all \w{t}. (In
the latter form, the statement is almost obvious.)\es

    Remark. Thus the function $\r{State}(\w{S})$ in the
LR(1)-sense is an extension of the function $\r{State}(\w{S})$
in the LR(0)-sense: $\r{State}_{\mathrm LR(0)}(\w{S})$ is obtained
from $\r{State}_{\mathrm LR(1)}(\w{S})$
when
we discard the  second
component of all pairs.

    Now we give a definition of a LALR(1)-grammar. Assume
that a context-free grammar is fixed, \w{S} is a string of
terminals and nonterminals, and \w{t} is a terminal (or \w{EOI}). We
say that the pair $\langle\w{S},\w{t}\rangle$ LALR(1)-allows
reduction (according to some production rule) if there is
another string $\w{S}_1$ with $\r{State}_{\mathrm
LR(0)}(\w{S}_0) = \r{State}_{\mathrm LR(0)}(\w{S}_1)$ such that
the pair $\langle\w{S}_1,\w{t}\rangle$ LR(1)-allows reduction
according to that rule. Thereafter, the conflicts are defined in
a natural way and a grammar is called a
 \index{grammar!LALR(1)}%
 \index{LALR(1)-grammar}%
LALR(1)-{\it grammar} if there are no conflicts.

\Problem{%
    Prove that every  SLR(1)-grammar is a
LALR(1)-grammar and every  LALR(1)-grammar is a
LR(1)-grammar.}

    \Hint{This is an easy consequence of the definitions.}


\Problem{%
    Find an  algorithm  that  checks  if  an  input  string  is
derivable  in  an LALR(1)-grammar. This algorithm
should keep less information in
the stack than the corresponding LR(1)-algorithm.}

    \Hint{It is sufficient to store the sets
$\r{State}_{\mathrm LR(0)}(\w{S})$ in the stack,
because the
LALR(1)-possibility of reduction is determined by those sets.
(Therefore, the only difference with SLR(1)-algorithm is in the
table of possible reductions.)}

\Problem{%
    Give an example of an LALR(1)-grammar that is not a
SLR(1)-grammar.\es}

\Problem{%
    Give an example of an LR(1)-grammar that is not a
LALR(1)-grammar.\es}

\subsection{General remarks about parsing algorithms}

Practical applications of the methods described is a
delicate matter. (For example, we need to store the tables as
compactly as possible.) Sometimes a given grammar is not
an LL(1)-grammar but still is an LR(1)-grammar. Often the given grammar
can be transformed into an equivalent
LL(1)-grammar.
It is not clear which of these two approaches is more practical.
The following general rule may be useful.
If you design the language,
keep it simple and do not use the same symbols for different
purposes. Then usually it is easy to write an LL(1)-grammar or
a recursive-descent parser. However, if the language is already
defined by an LR(1)-grammar that is not LL(1), it is better
not to change the grammar, just write a LR(1)-parser. To do
this, you may use tools for automatic parser generation such
as \w{yacc} (UNIX) and \w{bison} (GNU).

    Much useful information about the theoretical and practical
aspects of parsing can be found in the well-written book of
Alfred~V.~Aho, Ravi~Sethi, and Jeffrey~D.~Ullman on compilers [2].
 \index{Aho, A.V.}%
 \index{Sethi, R.}%
 \index{Ullman, J.D.}

