\section{Variables, expressions, assignments}\label{start}

     \subsection{Problems without arrays}

     \Problem{Consider two integer variables
\w{a} and \w{b}.
Write a program block that exchanges the values of
\w{a} and \w{b} (i.e., the value of \w{a} becomes the value of
\w{b} and vice versa).}%
\index{value exchange}


     \Solution. We use an auxiliary integer variable \w{t}.
\begin{Verbatim}
        t := a;
        a := b;
        b := t;
\end{Verbatim}
If we try to eliminate this auxiliary variable by writing
\begin{verbatim}
        a := b;
        b := a;
\end{verbatim}
we get an incorrect program (the value of \w{a} is lost after the first
assignment).

    \Problem{Solve the preceding problem without an auxiliary variable.
(Assume all variables accept arbitrary integer values.)}

    \Solution. (By \w{a0} and \w{b0} we denote the initial values of
\w{a} and  \w{b}.)
\begin{Verbatim}
        a := a + b; {a = a0 + b0, b = b0}
        b := a - b; {a = a0 + b0, b = a0}
        a := a - b; {a = b0, b = a0}
\end{Verbatim}

     \Problem{Let \w{a} be an integer and \w{n} be a nonnegative
integer. Compute $\w{a}^\w{n}$.
In other words, we ask for a program that does not change
the values of \w{a} and \w{n} and assigns
the value $\w{a}^\w{n}$
to another variable
(say, \w{b}).
(The program may use other variables as well.)}%
\index{power!computation}

    \Solution. Consider an integer variable \w{k}, whose range is
$\w{0..n}$. (We maintain the property:
\w{b} = $\w{a}^{\w{k}}$.)
\begin{verbatim}
        k := 0; b := 1;
        {b = a^k}
        while k <> n do begin
        | k := k + 1;
        | b := b * a;
        end;
\end{verbatim}

\clearpage

Another solution:
\begin{Verbatim}
        k := n; b := 1;
        {a^n = b * (a^k)}
        while k <> 0 do begin
        | k := k - 1;
        | b := b * a;
        end;
\end{Verbatim}

     \Problem{Solve the preceding problem with the additional
requirement that the number of execution
steps should be
of order
$\log \w{n}$ (i.e., it should not exceed
$C\log \w{n}$ for some constant $C$).}%
\index{power!quick computation}

    \Solution. Let us make some changes in the second solution
of the preceding problem:
\begin{verbatim}
        k := n;  b := 1;  c := a;
        {a^n = b * (c^k)}
        while k <> 0 do begin
        | if k mod 2 = 0 then begin
        | | k := k div 2;
        | | c := c*c;
        | end else begin
        | | k := k - 1;
        | | b := b * c;
        | end;
        end;
\end{verbatim}
%Every other iteration
%(at least) the first alternative (\w{then}) is
%chosen (if \w{k} is odd, after \w{k:=k-1}, the value of \w{k} becomes
%even). Therefore, after two iterations of the \verb|while|-loop, the
%value of \w{k} becomes two times less (or even smaller).
In both cases (even \w{k} and odd \w{k}) the value of \w{k} decreases;
if \w{k} is even, it is divided by \w{2}; if \w{k} is odd, after
\w{k := k - 1} it becomes even and is divided by \w{2} during the
next iteration. Therefore, after any two iterations \w{k} becomes
twice smaller (or even less). \es

     \Problem{Two nonnegative integers \w{a} and \w{b} are given.
Compute the product \w{a*b} (only
\w{+}, \w{-}, \w{=}, \w{<>} are allowed).}

    \Solution.
\begin{Verbatim}
        k := 0; c := 0;
        {invariant relation: c = a * k}
        while k <> b do begin
        | k := k + 1;
        | c := c + a;
        end;
        {c = a * k and k = b, therefore, c = a * b}
\end{Verbatim}

     \Problem{Two nonnegative integers \w{a} and
\w{b} are given. Compute
$\w{a}+\w{b}$. Only assignments of the form
\begin{eqnarray*}
 \langle\r{variable1}\rangle&:=&\langle\r{variable2}\rangle;\\
  \langle\r{variable}\rangle&:=&\langle\r{number}\rangle;   \\
  \langle\r{variable1}\rangle&:=&\langle\r{variable2}\rangle + 1;
\end{eqnarray*}
are allowed.
}

\Hint{Use the
         invariant relation \w{c = a + k}}

   \Problem{A nonnegative integer \w{a} and positive integer \w{d}
are given. Compute the quotient \w{q} and the remainder \w{r} when
\w{a} is divided by \w{d}. Do not use the operations
\w{div} or \w{mod}.}%
\index{division}\index{remainder}

    \Solution. By definition,
$\w{a}=\w{q*d}+\w{r}$ and $\w{0} \le \w{r} <\w{d}$.
\begin{Verbatim}
        {a >= 0; d > 0}
        r := a; q := 0;
        {invariant relation: a = q * d + r, 0 <= r}
        while not (r < d) do begin
        | {r >= d}
        | r := r - d; {r >= 0}
        | q := q + 1;
        end;
\end{Verbatim}

\Problem{For a given nonnegative integer
$\w{n}$,  compute $\w{n}!$
($\w{n}!$ is the product $1\cdot 2 \cdot3 \cdots n$; we assume that
$\w{0}!=\w{1}$).\es}%
\index{factorial}

\Problem{The Fibonacci sequence is defined as follows:
$\w{a}_{\w{0}}=\w{0}$, $\w{a}_{\w{1}}=\w{1}$,
$\w{a}_{\w{k}}= \w{a}_{\w{k-1}} +\w{a}_{\w{k-2}}$ for
$\w{k}\ge\w{2}$.
For a given $\w{n}$,
compute $\w{a}_{\w{n}}$.\es
}%
\index{Fibonacci numbers}

\Problem{Repeat the preceding
problem with the additional requirement that the number
of operations should be proportional to
 $\log \w{n}$. (Use only integer variables.)
}%
\index{Fibonacci numbers!fast computation}

    \Hint{Any pair of consecutive Fibonacci numbers is the product of
the matrix
\[ \left[
\begin{array}{cc}
          \w{1}&\w{1}\\
          \w{1}&\w{0}
\end{array}
\right]\]
and the preceding pair. Therefore, it is enough to compute the $\w{n}$-th
power of this matrix. It can be done in
$C\log \w{n}$ steps in the same manner as problem 1.1.4
   (for integers).}

{\samepage
\Problem{%
     For a nonnegative integer
\w{n}, compute
\[ \frac{\w{1}}{\w{0}!} + \frac{\w{1}}{\w{1}!}+
                  \cdots+\frac{\w{1}}{\w{n}!}. \]
}
\ess
}

\Problem{%
Repeat the preceding
problem with the additional requirement that the number of
steps (i.e., the number of assignments performed during the execution)
should be of order $\w{n}$
(i.e., not greater than $C\w{n}$ for
some constant $C$).
}

    \Solution. The invariant relation:
$\w{sum} = \w{1}/\w{1}!+\cdots+\w{1}/\w{k}!$,
$\w{last}
= \w{1}/\w{k}!$ (it is important not to compute $\w{k}!$ each time
from scratch).\es

\Problem{%
     Two nonnegative integers \w{a} and \w{b} are not both zero.
Compute \w{GCD(a,b)}, the greatest common divisor of
\w{a} and \w{b}.
}%
\index{GCD}%
\index{greatest common divisor}

     \Solution\ (version 1).
\begin{verbatim}
        if a > b then begin
        | k := a;
        end else begin
        | k := b;
        end;
        {k = max (a,b)}
        {invariant relation: no numbers greater than k are
          common divisors}
        while not ((a mod k = 0) and (b mod k = 0)) do begin
        | k := k - 1;
        end;
        {k is a common divisor, all larger k are not}
\end{verbatim}
       (version 2  --- Euclid's algorithm).%
\index{Euclid's algorithm}
We assume that \w{GCD(0,0)=0}. Then \w{GCD(a,b)} =
\w{GCD(a-b,b)} = \w{GCD(a,b-a)} with
\w{GCD(a,0)} = \w{GCD(0,a)} =
\w{a} for all $\w{a},\w{b}\ge \w{0}$.
This property allows us to decrease \w{a} and \w{b} without
changing
\w{GCD(a,b)}.
 %
\begin{Verbatim}
         m := a; n := b;
        {invariant relation: GCD(a,b) = GCD(m,n); m,n >= 0 }
        while not ((m=0) or (n=0)) do begin
        | if m >= n then begin
        | | m := m - n;
        | end else begin
        | | n := n - m;
        | end;
        end;
        {m = 0 or n = 0}
        if m = 0 then begin
        | k := n;
        end else begin {n = 0}
        | k := m;
        end;
\end{Verbatim}

\Problem{%
      Write down a modified version of Euclid's algorithm
using the identities \w{GCD(a,b)} = \w{GCD(a mod b, b)}
for $\w{a}\ge\w{b}$ and
 \w{GCD(a,b)} = \w{GCD(a, b mod a)} for
$\w{b}\ge\w{a}$.\es
}%
\index{Euclid's algorithm}

\Problem{%
     Nonnegative integers \w{a} and \w{b} are given,
at least one of which is not zero.
Find \w{d} = \w{GCD(a,b)} and integers \w{x} and \w{y} such that
$\w{d} = \w{a*x} + \w{b*y}$.
}%
\index{Euclid's algorithm}%
\index{Diophantine equation}%

     \Solution. Add the
auxiliary variables
\w{p}, \w{q}, \w{r}, \w{s}
to Euclid's algorithm
and add the
requirements \w{m = p*a+q*b} and \w{n = r*a+s*b} to the
invariant relation:
\begin{Verbatim}
        m:=a; n:=b; p:=1; q:=0; r:=0; s:=1;
        {invariant relation:
            GCD(a,b) = GCD(m,n);
            m,n >= 0
            m = p*a + q*b;
            n = r*a + s*b.}
        while not ((m=0) or (n=0)) do begin
        | if m >= n then begin
        | | m := m - n;
        | | p := p - r;
        | | q := q - s;
        | end else begin
        | | n := n - m;
        | | r := r - p;
        | | s := s - q;
        | end;
        end;
        if m = 0 then begin
        | k := n; x := r; y := s;
        end else begin
        | k := m; x := p; y := q;
        end;
\end{Verbatim}

\Problem{%
     Solve the preceding problem using the
\w{mod} operator.\es}%
\index{Euclid's algorithm}%

\Problem{%
      (E.~Dijkstra) Let us add three variables
\w{u}, \w{v}, \w{z}
to Euclid's algorithm:}%
\index{Dijkstra, E.W.}%
\index{LCM}%
\index{least common multiple}%
% verbatim inside definition is dangerous
{\samepage
\begin{verbatim}
      m := a; n := b; u := b; v := a;
     {invariant relation: GCD (a,b) = GCD (m,n); m,n >= 0 }
     while not ((m=0) or (n=0)) do begin
     | if m >= n then begin
     | | m := m - n; v := v + u;
     | end else begin
     | | n := n - m; u := u + v;
     | end;
     end;
\end{verbatim}
}
\clearpage
\begin{verbatim}
     if m = 0 then begin
     | z:= v;
     end else begin {n=0}
     | z:= u;
     end;
\end{verbatim}
Prove that after execution the value of
\w{z} is twice as large as the least common multiple
of \w{a} and \w{b}, that is,
$\w{z} = \w{2}\cdot\w{LCM(a,b)}$.

    \Solution. Look at the value of
$\w{m}\cdot\w{u} + \w{n}\cdot\w{v}$, which remains unchanged during
program
execution. Initially it is equal to
$\w{2}\w{a}\w{b}$; therefore, this expression has the same value
at the end. Now apply the identity
$\w{GCD} (\w{a},\w{b}) \cdot \w{LCM} (\w{a}, \w{b}) = \w{a}\w{b}$.\es

\Problem{%
     Write a version of Euclid's algorithm using the
identities
  $$
 \w{GCD}(\w{2}\w{a}, \w{2}\w{b})=\w{2}\cdot\w{GCD}(\w{a},\w{b}); \quad
 \w{GCD}(\w{2}\w{a},\w{b})=\w{GCD}(\w{a},\w{b}) \quad
 \r{for odd \w{b}}
  $$
The algorithm should avoid division (\w{div} and \w{mod} operations);
only division by $2$ and the test ``to be even'' are allowed.
(The number of operations should be of order
$\log\w{k}$ if both numbers do not exceed \w{k}.)
}%
\index{Euclid's algorithm!binary}%

    \Solution.
\begin{verbatim}
m:=a; n:=b; d:=1;
{GCD(a,b) = d * GCD(m,n)}
while not ((m=0) or (n=0)) do begin
| if (m mod 2 = 0) and (n mod 2 = 0) then begin
| | d:= d*2; m:= m div 2; n:= n div 2;
| end else if (m mod 2 = 0) and (n mod 2 = 1) then begin
| | m:= m div 2;
| end else if(m mod 2 = 1) and (n mod 2 = 0) then begin
| | n:= n div 2;
| end else if (m mod 2=1) and (n mod 2=1) then begin
| | if m >=n then begin
| | | m:= m-n;
| | end else begin {m < n}
| | | n:= n-m;
| | end;
| end;
end;
{m=0 => answer=d*n; n=0 => answer=d*m}
\end{verbatim}
%
If both numbers \w{m} and \w{n} do not exceed \w{k},
the number of operations does not exceed $C\log\w{k}$;
indeed, each other operation makes at least one of the
numbers \w{m} and \w{n} twice smaller.\es

\Problem{%
Modify the solution of the preceding problem to find
\w{x} and \w{y} such that
$\w{a}\w{x}+\w{b}\w{y}=\w{GCD}(\w{a},\w{b})$.
}%
\index{Diophantine equation}%
\index{Zvonkin, D.}%

    \Solution. (The idea was communicated by
D.~Zvonkin.) Assume that both \w{a} and \w{b} are even. In this
case we divide both of them by $2$;
the values of \w{x} and \w{y} we are
looking for remain unchanged.
Therefore, without loss of generality, we may
assume that at least one of the numbers
\w{a} and \w{b} is odd. (This property will remain true.)

     As before, we wish to maintain the numbers
$\w{p},\w{q},\w{r},\w{s}$ such that
\begin{eqnarray*}
     \w{m}&=&\w{a}\w{p}+\w{b}\w{q}\\
     \w{n}&=&\w{a}\w{r} + \w{b}\w{s}
\end{eqnarray*}
The problem, however, is that if we divide
\w{m} by \w{2} (say),  then
we should at the same time divide
\w{p} and \w{q} by \w{2}. In this case \w{p} and \w{q}
are no longer integers
but become finite binary fractions, that is, numbers of the
type $r/2^s$.
Such a number can be represented by a pair
$\langle r,s\rangle$.
As a result, we get \w{d} as a linear combination of \w{a} and \w{b}
with coefficients being finite binary fractions. In other words,
we have
\[
      \w{2}^\w{i} \w{d} = \w{a}\w{x} + \w{b}\w{y}
\]
for some integers $\w{x},\w{y}$ and nonnegative integer
$\w{i}$. What should we do if
$\w{i} > \w{0}$? If both \w{x} and \w{y} are even, we may
divide them by $2$ (and decrease \w{i} by $1$).
If not, we apply the transformations:
\begin{eqnarray*}
     \w{x}&:=&\w{x} + \w{b}  \\
     \w{y}&:=&\w{y} - \w{a}
\end{eqnarray*}
(this transformation leaves
$\w{a}\w{x}+\w{b}\w{y}$ unchanged).
Let us see why this works.
Recall that one of the numbers \w{a} and \w{b} is odd (according
to our assumption). Let \w{a} be odd.
If \w{y} is even, then \w{x} is even as well (otherwise
$\w{a}\w{x}+\w{b}\w{y}$ is odd); this case is considered above.
If \w{a} and \w{y} are
odd, then \w{y} becomes even after
executing the statement
$\w{y}:=\w{y}-\w{a}$.\es

\Problem{%
     Write a program that prints the squares of the
natural numbers
$\w{0},\w{1},\ldots,\w{n}$.
}

   \Solution.
\begin{Verbatim}
        k:=0;
        writeln (k*k);
        {invariant relation: k<=n, all the squares
           up to (k*k) are printed}
        while not (k=n) do begin
        | k := k + 1;
        | writeln (k*k);
        end;
\end{Verbatim}

\Problem{%
Repeat the preceding
problem, but only
addition and subtraction are
allowed. The number of steps should be of order
\w{n}.
}

    \Solution. We use the variable \verb|k_square|,
and maintain the relation  $\hbox{\verb|k_square|}= \w{k}^2$:
\begin{Verbatim}
        k := 0; k_square := 0;
        writeln (k_square);
        while not (k = n) do begin
        | k := k + 1;
        | {k_square = (k-1) * (k-1) = k*k - 2*k + 1}
        | k_square := k_square + k + k - 1;
        | writeln (k_square);
        end;
\end{Verbatim}

     Remark. We can avoid subtraction by the following trick:
\begin{verbatim}
        while not (k = n) do begin
        | k_square := k_square + k;
        | {k_square = k*k + k}
        | k := k + 1;
        | {k_square = (k-1)*(k-1)+(k-1)=k*k-k}
        | k_square := k_square + k;
        end;
\end{verbatim}

