\Problem{%
     Write a program that prints the factorization of a given
integer $\w{n}>\w{0}$. (In other words, it should print prime
numbers whose product is equal to
\w{n}; if $\w{n} =
\w{1}$, nothing should be printed.)}%
\index{factorization}%
\index{prime factors}%

    \Solution\ (version 1).
\begin{verbatim}
   k := n;
   {invariant relation: the product of k and all numbers
    printed is equal to n; only prime numbers are printed}
   while not (k = 1) do begin
   | t := 2;
   | {invariant relation: k has no divisors in (1,t)}
   | while k mod t <> 0 do begin
   | | t := t + 1;
   | end;
   | {t is the smallest divisor of k greater than 1;
   |  therefore, t is prime}
   | writeln (t);
   | k:=k div t;
   end;
\end{verbatim}

{\samepage
     (version 2)
\begin{Verbatim}
     k := n; t := 2;
     {the product of k and all number printed is equal
      to n; only prime numbers are printed;
      k has no divisors in (1,t)}
     while not (k = 1) do begin
     | if k mod t = 0  then begin
     | | {k is a multiple of t and has no divisors
     | |   less than t; therefore, t is prime}
     | | k := k div t;
     | | writeln (t);
     | end else begin
     | | {k is not a multiple of t}
     | | t := t+1;
     | end;
     end;
\end{Verbatim}
}

\Problem{%
     Solve the preceding problem taking into account the
following fact:
any composite number $N$ has a factor not exceeding
$\sqrt{N}$.}

    \Solution. In version 2 of the above solution,
replace \w{t:=t+1} by
\begin{Verbatim}
         if t*t > k then begin
         | t:=k;
         end else begin
         | t:=t+1;
         end;
\end{Verbatim}

\Problem{%
     Check whether a given number
$\w{n} > \w{1}$ is prime.\es}

\Problem{%
     (This problem requires some algebra)
A Gaussian integer
$\w{n} + \w{m}i \in
{\Bbb Z}[i]$
is given. (a) Check whether it is a prime element in
${\Bbb Z}[i]$; (b) print its factorization as
a product
of prime factors in ${\Bbb Z}[i]$.\es}%
\index{Gaussian integers}%

\Problem{%
     Assume the command \w{write(i)} is allowed for
$\w{i}=\w{0}$, \w{1}, \w{2},$\ldots$,\w{9}. Write a program that prints
the decimal representation of a given positive integer \w{n}.
}%
\index{decimal number!printing}%

     \Solution.
\begin{verbatim}
    base:=1;
    {base is an integer power of 10 not exceeding n}
    while 10 * base <= n do begin
    | base:= base * 10;
    end;
    {base is a maximal power of 10 not exceeding n}
    k:=n;
    {invariant relation: it remains to print k with the same
     number of digits as in base; base = 100..00}
    while base <> 1 do begin
    | write(k div base);
    | k:= k mod base;
    | base:= base div 10;
    end;
    {base=1; it remains to write one digit k}
    write(k);
\end{verbatim}
Please note that this program assumes that $\w{n}>0$.\es

\medskip

(A typical mistake while solving this problem is that the
numbers with
zeros in the middle are printed incorrectly. The invariant relation
mentioned above allows the case $\w{k}<\w{base}$; in this case,
the decimal representation of
\w{k} begins with zero.)

\Problem{%
     Write a program that prints the decimal representation
in reverse. (For $\w{n}=\w{173}$, the program should print \w{371}.)}

     \Solution.
\begin{Verbatim}
      k:= n;
      {invariant relation: it remains to print k reversed}
      while k <> 0 do begin
      | write (k mod 10);
      | k:= k div 10;
      end;
\end{Verbatim}

\Problem{%
     A nonnegative integer \w{n} is given. Count all the solutions
of the inequality
$\w{x}^2 + \w{y}^2 < \w{n}$ where \w{x} and \w{y} are nonnegative
integers. The program should not use operations with real numbers
(such as $\sqrt{\phantom{x}}$, etc.).}%
\index{integer points in a circle}%

     \Solution.
\begin{verbatim}
     a := 0; s := 0;
     {invariant relation: s = the number of all pairs
      <x,y> such that x*x + y*y < n and x < a}
     while a*a < n do begin
     | ...
     | {t = the number of nonnegative integers y such that
     |  a*a + y*y < n (for fixed a)}
     | a := a + 1;
     | s := s + t;
     end;
     {a*a >= n, therefore s is the total number of solutions}
\end{verbatim}

     Here the ellipsis represents part of the program that is
still to be written. Here it is:
\begin{Verbatim}
        b := 0; t:= 0;
        {invariant relation: t is the number of integers y
          such that a*a + y*y < n and 0<=y<b }
        while a*a + b*b < n do begin
        | b := b + 1;
        | t := t + 1;
        end;
        {a*a + b*b >= n,  so t is the number of nonnegative
           integers y such that a*a + y*y < n}
\end{Verbatim}

\Problem{%
     The same problem with the additional restriction that the total
number of operations should be of order $\sqrt{\w{n}}$.
(The previous solution requires about \w{n} operations.)}

     \Solution. We have to count all the integer grid points
in the first quadrant that lie inside the circle of radius
$\sqrt{\w{n}}$. The set in question (call it
\w{X}) is a union of columns of points having
width $\w{1}$ and non-increasing height.
\medskip
$$
{\def\bulbox{\hbox to 1.5em{\hss
         \vbox to 1.5em{\vss \hbox{$\bullet$} \vss}%
         \hss}}%
 \def\empbox{\hbox to 1.5em{\hss
         \vbox to 1.5em{\vss \hbox{} \vss}%
         \hss}}%
   \vbox{\hbox{\bulbox}%
      \hbox{\bulbox\bulbox\bulbox}%
      \hbox{\bulbox\bulbox\bulbox}%
      \hbox{\bulbox\bulbox\bulbox\bulbox}%
   }%
}%
$$
The idea is to trace the boundary of this set, which resembles a
staircase that goes down as we move from left to right.
The current position is \w{<a,b>}. We use one more variable
\w{s} and maintain the following invariant relation:
\begin{quote}\leavevmode
  \w{<a,b>} is on the top of \w{a}-th column;   \\
  \w{s} is the number of points in the preceding columns.
\end{quote}
     Formally,
\begin{itemize}
\item \ \w{b} is minimal among all $\w{b}\ge\w{0}$ such that
    \w{<a,b>} is not in \w{X};
\item \ \w{s} is the number of all pairs $\w{<x,y>}$ of nonnegative
   integers such that $\w{x}<\w{a}$ and $\w{<x,y>}\in\w{X}$.
\end{itemize}

These conditions will be denoted by \w{(I)}.
{\samepage
\begin{verbatim}
 a := 0; b := 0;
 while <0,b> is in X do begin
 | b := b + 1;
 end;
\end{verbatim}
}
\begin{verbatim}
 {a = 0, b is minimal among all b >= 0
  such that <a,b> is not in X }
 s := 0;
 {invariant relation: (I)}
 while not (b = 0) do begin
 | s := s + b;
 | {s is the number of points in columns 0..a}
 | a := a + 1;
 | {point <a,b> is outside X, it should be moved down to
 |    restore (I) unless (I) is already true}
 | while (b <> 0) and (<a, b-1> is not in X) do begin
 | | b := b - 1;
 | end;
 end;
 {(I), b = 0, therefore the a-th column and all subsequent
   columns are empty; s is the required number}
\end{verbatim}
An estimate for the number of steps is evident. First  we  move
up performing
not more than $\sqrt{\w{n}}$ steps. Then we move
right
and down in
not  more  than  $\sqrt{\w{n}}$  steps  in  each
direction.\es

\Problem{%
     Nonnegative integers \w{n} and \w{k} are given, with
$\w{n}>\w{1}$.
Print \w{k} digits of the decimal representation of the
number $\w{1}/\w{n}$. (If two decimal representations exist,
such as $0.499\ldots=0.500\ldots$, print the latter.)
The program should use integer variables only.
}%
\index{decimal number!printing}%

     \Solution. Moving the decimal point
of the number
$\w{1}/\w{n}$,~$\:\w{k}$
positions to the right, we get the number $\w{10}^\w{k}/\w{n}$.
We wish to print its integer part, that is, we must compute
$\w{10}^\w{k}\w{ div }\w{n}$. We do not want to compute $\w{10}^\w{k}$
because of the possibility of
integer overflow. Instead, we perform ordinary division.
Here is the program:

\begin{Verbatim}
   m := 0;
   r := 1;
   {m digits of 1/n  are printed; it remains to print
     k - m digits of the decimal expansion of r/n}
   while m <> k do begin
   | write ( (10 * r) div n);
   | r := (10 * r) mod n;
   | m := m + 1;
   end;
\end{Verbatim}
