
\Problem{%
     A natural number $\w{n} > \w{1}$ is given. Find the length
of the period of the decimal number
$\w{1}/\w{n}$.}%
\index{period of a decimal fraction}
\index{decimal fraction!period}%

     \Solution.
The period of a decimal fraction is equal to the period
of the sequence of ``remainders'' \w{r}
(see the solution of the preceding problem). [Prove this fact;
do not forget to prove that the period of the fraction
cannot be less than the period of the sequence of remainders.]
In the sequence of remainders
all terms that form the period are
distinct and the length of the nonperiodic initial segment
does not exceed \w{n}. Therefore, it is enough to find the
$(\w{n}+\w{1})$-th term of the sequence, and then to find the minimal
\w{k} such that the $(\w{n}+\w{1}+\w{k})$-th term is equal to the
$(\w{n}+\w{1})$-th term.
\begin{Verbatim}
  m := 0;
  r := 1;
  {r/n = what remains from 1/n after the decimal point
    is moved m positions to the right and the integral
    part is discarded}
  while m <> n+1 do begin
  | r := (10 * r) mod n;
  | m := m + 1;
  end;
  c := r;
  {c = (n+1)-th term of the sequence of remainders}
  r := (10 * r) mod n;
  k := 1;
  {r = (n+k+1)-th term of the same sequence}
  while r <> c do begin
  | r := (10 * r) mod n;
  | k := k + 1;
  end;
\end{Verbatim}

\Problem{%
(R.W.~Floyd,\index{Floyd, R.W.}
communicated by Yu.V.~Matijasevich\index{Matijasevich, Yu.V.})
A function $\w{f}:\{\w{1..N}\} \to\{\w{1..N}\}$
is given. Find the period of the sequence
$\w{1},\w{f(1)},\w{f(f(1))},\ldots\ $ The number of operations
should be proportional to the length of the smallest initial
segment that includes the period (this length may be
significantly less than \w{N}).}

     \Solution. After discarding the initial segment, we have a periodic
sequence, and all terms in the period are different.
\begin{verbatim}
  {Notation: f[n,1]=f(f(...f(1)...)) (n times)}
  k := 1;
  a := f(1);
  b := f(f(1));
  {a = f[k,1]; b = f[2k,1]}
  while a <> b do begin
  | k:=k+1; a:=f(a); b:=f(f(b));
  end;
  {a = f[k,1] = f[2k,1]; f[k,1] is in the periodic part}
  m := 1;  b := f(a);
  {b = f[k+m,1]; f[k,1],...,f[k+m-1,1] are different}
\end{verbatim}
\eject
\begin{Verbatim}
  while a <> b do begin
  | m:=m+1; b:=f(b);
  end;
  {period = m}
\end{Verbatim}

\Problem{%
     (E.~Dijkstra). A function \w{f} whose arguments and values are
non-negative integers is defined as follows:
$\w{f}(\w{0}) = \w{0}$,
$\w{f}(\w{1})=\w{1}$, $\w{f}(\w{2n}) = \w{f}(\w{n})$,
$\w{f}(\w{2n}+\w{1}) = \w{f}(\w{n}) + \w{f}(\w{n}+\w{1})$.
Write a program that computes $\w{f}(\w{n})$ for a given
\w{n}; the number of operations should be of order
$\log\w{n}$.}%
\index{Dijkstra, E.W.}%

     \Solution.
\begin{Verbatim}
  k := n; a := 1; b := 0;
  {invariant relation: 0<=k, f(n) = a*f(k) + b*f(k+1)}
  while k <> 0 do begin
  | if k mod 2 = 0  then begin
  | | l := k div 2;
  | | {k=2l, f(k)=f(l), f(k+1) = f(2l+1) = f(l) + f(l+1),
  | |  f (n) = a*f(k) + b*f(k+1) = (a+b)*f(l) + b*f(l+1)}
  | | a := a + b; k := l;
  | end else begin
  | | l := k div 2;
  | | {k = 2l + 1, f(k) = f(l) + f(l+1),
  | |  f(k+1) = f(2l+2) = f(l+1),
  | |  f(n) = a*f(k) + b*f(k+1) = a*f(l) + (a+b)*f(l+1)}
  | | b := a + b; k := l;
  | end;
  end;
  {k = 0, f(n) = a * f(0) + b * f(1) = b,
   b is the answer}
\end{Verbatim}

\Problem{%
     The same problem for $\w{f}(\w{0}) = \w{13}$,
$\w{f}(\w{1}) = \w{17}$,  $\w{f}(\w{2n}) =
     \w{43}\,\w{f}(\w{n}) + \w{57}\,\w{f}(\w{n}+\w{1})$
                and
$\w{f}(\w{2n}+\w{1}) = \w{91}\,\w{f}(\w{n}) +
     \w{179}\,\w{f}(\w{n}+\w{1})$ for $\w{n}\ge\w{1}$. }

\Hint{The program stores \w{k}, \w{a}, \w{b}, \w{c}
such that $\w{f}(\w{n})=\w{a}\cdot\w{f}(\w{k})+
\w{b}\cdot\w{f}(\w{k+1})+
\w{c}\cdot\w{f}(\w{k+2})$.}

\Problem{%
     Two nonnegative integers \w{a} and \w{b} are given,
with
$\w{b}>\w{0}$.
Find \w{a mod b} and \w{a div b} using only integer
variables and avoiding explicit
\w{div} and \w{mod} operations (the only exception:
an even number may be divided by $2$).
The number of operations should not exceed
$C_1\log(\w{a}/\w{b})+C_2$ for some constants
$C_1$ and $C_2$.
}%
\index{division!fast}%
\clearpage
     \Solution.
\begin{Verbatim}
  b1 := b;
  while b1 <= a do begin
  | b1 := b1 * 2;
  end;
  {b1 > a, b1 = b * (integer power of 2)}
  q:=0; r:=a;
  {invariant relation: q, r are quotient and remainder when
   a is divided by b1; b1 = b * (some integer power of 2)}
  while b1 <> b do begin
  | b1 := b1 div 2 ; q := q * 2;
  | { a = b1 * q + r, 0 <= r < 2 * b1}
  | if r >= b1 then begin
  | | r := r - b1;
  | | q := q + 1;
  | end;
  end;
  {q, r are quotient and remainder when a is divided by b}
\end{Verbatim}

