     \subsection{Arrays}

     We assume in the sequel that $\w{x},\w{y},\w{z}$
are defined as
\w{array[1..n]} \w{of} \w{integer}
(here \w{n} is a fixed positive integer constant) unless
otherwise stated.%
\index{array}%

\Problem{%
     Fill the array \w{x} with zeros. (Write a program
fragment whose execution guarantees that all values
\w{x[1]..x[n]} are zero independent of the initial value
of \w{x}.)}

     \Solution.
\begin{Verbatim}
          i := 0;
          {invariant relation: x[1],..,x[i] = 0}
          while i <> n do begin
          | i := i + 1;
          | {x[1]..x[i-1] = 0}
          | x[i] := 0;
          end;
\end{Verbatim}

\Problem{%
     Count the number of zeros in an array \w{x}. (Write a program
fragment that does not change the value of \w{x} and guarantees that
the
integer variable \w{k} contains the number of zeros
among \w{x[1]..x[n]}.)}

     \Solution.
\begin{Verbatim}
    ...
    {invariant: k = number of zeros among x[1]..x[i] }
    ...
\end{Verbatim}

\Problem{%
     Not using assignment statement for arrays, write a program
that is equivalent to the statement \w{x:=y}.}

     \Solution.
\begin{Verbatim}
  i := 0;
  {invariant: y is unchanged, x[t]=y[t] for all t<=i}
  while i <> n do begin
  | i := i + 1;
  | x[i] := y[i];
  end;
\end{Verbatim}

\Problem{%
     Find the maximum value among \w{x[1]..x[n]}.}

     \Solution.
\begin{Verbatim}
    i := 1; max := x[1];
    {invariant relation: max = maximum(x[1]..x[i])}
    while i <> n do begin
    | i := i + 1;
    | {max = maximum(x[1]..x[i-1])}
    | if x[i] > max then begin
    | | max := x[i];
    | end;
    end;
\end{Verbatim}

\Problem{%
     An array \w{x:\ array[1..n]} \w{of} \w{integer} is given such that
$\w{x[1]}\le\w{x[2]}\le\cdots\le\w{x[n]}$. Find the number of different
elements among \w{x[1]..x[n]}.}%
\index{number!of different elements}%

    \Solution\ (version 1).
\begin{verbatim}
  i := 1; k := 1;
  {invariant relation: k = the number of
   different elements among x[1]..x[i]}
\end{verbatim}

\begin{verbatim}
  while i <> n do begin
  | i := i + 1;
  | if x[i] <> x[i-1] then begin
  | | k := k + 1;
  | end;
  end;
\end{verbatim}

     (version 2) The number in question is one unit larger
than the number of
\w{i} in \w{1..n-1} such that \w{x[i]} is not equal to
\w{x[i+1]}, plus one.
\clearpage
\begin{Verbatim}
  k := 1;
  for i := 1 to n-1 do begin
  | if x[i]<> x[i+1] then begin
  | | k := k + 1;
  | end;
  end;
\end{Verbatim}

\Problem{%
     An array \w{x:\ array[1..n]} \w{of} \w{integer} is given.
Compute the number of different elements among \w{x[1]..x[n]}.
(The number of operations should be of order
$\w{n}^2$.)\es}%
\index{number!of different elements}%

\Problem{%
     The same problem with an additional requirement: the number of
operations should be of order
$\w{n}\,\log\w{n}$. }

\Hint{See chapter \ref{sorting} on sorting.}

\Problem{%
     The same problem where all elements are integers in
\w{1..k} and the number of operations should be of order
$\w{n}+\w{k}$.\es}

\Problem{%
     (Communicated by A.L.~Brudno.) A rectangular field
$\w{m}\times\w{n}$ contains $\w{mn}$ squares.
Some squares are marked as black. It is known that black squares are
grouped into several disjoint rectangles that are at least one apart
from each other.
Assuming that the colors of squares are represented as
$$
     \w{array [1..m] of array [1..n] of Boolean;}
$$
count the number of rectangles. The number of operations should be
of order $\w{mn}$.}%
\index{Brudno, A.L.}%

     \Solution. The number of rectangles is equal to the number of their
upper left corners. It is easy to check whether a square is in the upper
left corner. Just check the color of the cell as well as the
colors of its
upper and left neighbors. (Don't forget the case when the cell is on
the left or upper boundary of a given $\w{m}\times\w{n}$ rectangle.)\es

\Problem{%
     An array \w{x[1]..x[n]} is given. Without using
other arrays,
put its elements in reverse order.}

     \Solution. We should exchange
\w{x[i]} and \w{x[n+1-i]} for all
\w{i} such that $\w{i}<\w{n}+\w{1}-\w{i}$,
i.e., $\w{2}\w{i} < \w{n} + \w{1}$ $\Leftrightarrow$
$\w{2}\w{i}\le \w{n}$ $\Leftrightarrow$  $\w{i} \le \w{n div 2}$ :
\begin{Verbatim}
  for i := 1 to n div 2 do begin
  | ...exchange x[i] and x[n+1-i];
  end;
\end{Verbatim}

\Problem{%
     (From D.~Gries' book [6]) An array \w{x[1]..x[m+n]}
is considered as a concatenation of two
segments: a prefix
\w{x[1]..x[m]} of length \w{m} and a suffix
\w{x[m+1]..x[m+n]} of length
\w{n}. Without using
other arrays, exchange these prefix and suffix segments.
(The number of operations should be of order
$\w{m}+\w{n}$.)}%
\index{Gries, D.}%
\index{exchange}%

     \Solution. (version 1) Reverse
the prefix segment
(see the preceding problem),
then the suffix segment, and finally the
whole array.

     (version 2, A.G.~Kushnirenko)%
\index{Kushnirenko, A.G.}
Imagine that the array is written down along a circle. Then
the required transformation is a rotation. Recall that rotation may
be represented as the composition of two axial symmetries. Each
symmetry can be performed by exchanges without extra memory.

     (version 3) Consider the more general problem: Exchange two
adjacent segments \w{x[p+1]..x[q]} and \w{x[q+1]..x[r]} in an
array. Assume that the length of the left segment (called $A$ in
the sequel) does not exceed the length of the right segment
(called $B$). Split $B$ into two segments $B_1$ and $B_2$, where
$B_1$ is an initial segment of $B$ of the same length as $A$.
(So, $B=B_1+B_2$, where $+$ stands for concatenation.) We need to
transform $A+B_1+B_2$ into $B_1+B_2+A$. We can easily exchange
$A$ and $B_1$ because they have equal lengths. After that we
get $B_1 + A + B_2$ and it remains to exchange $A$ and $B_2$.
Therefore, we have reduced our problem to a similar problem for
shorter segments. Here is the outline of the program:

\begin{verbatim}
  p := 0; q := m; r := m + n;
  {invariant relation: it remains
    to exchange x[p+1..q], x[q+1..r]}
  while (p <> q) and (q <> r) do begin
  | {both segments are nonempty}
  | if (q - p) <= (r - q) then begin
  | | ..exchange x[p+1]..x[q] and x[q+1]..x[q+(q-p)]
  | | pnew := q; qnew := q + (q - p);
  | | p := pnew; q := qnew;
  | end else begin
  | | ..exchange x[q-(r-q)+1]..x[q] and x[q+1]..x[r]
  | | qnew := q - (r - q); rnew := q;
  | | q := qnew; r := rnew;
  | end;
  end;
\end{verbatim}
The number of operations may be estimated as follows.
At each step the part of the array that should be processed
becomes shorter by the length of $A$. The number of operations
required is also proportional to the
length of $A$.\es

\Problem{%
     An array \w{a:\ array[0..n]}
\w{of} \w{integer}
contains the coefficients of a polynomial of degree \w{n}.
Compute the value of this polynomial at the point \w{x},
that is, $\w{a[n]}\,\w{x}^\w{n}+\cdots+\w{a[1]}\,\w{x}+\w{a[0]}$.}%
\index{polynomial!value}%
\index{Horner's rule}%

     \Solution. (The algorithm described below is called Horner's rule)
\begin{Verbatim}
  k := 0; y := a[n];
  {invariant relation: 0 <= k <= n,
   y= a[n]*(x ** k)+...+a[n-1]*(x ** (k-1))+...+
                     + a[n-k]*(x ** 0)}
  while k <> n do begin
  | k := k + 1;
  | y := y * x + a [n-k];
  end;
\end{Verbatim}

\Problem{%
      (Requires some calculus; communicated by A.G.~Kushnirenko)\
Extend Horner's rule to compute not only the value of a polynomial
at some point, but also
the value of the
derivative of the same polynomial at the same point.}%
\index{Kushnirenko, A.G.}%
\index{polynomial!derivative}%

     \Solution. When a new coefficient is added, the polynomial
changes from $P(x)$ to $Q(x)=xP(x)+c$. The derivative $Q'(x)$
is equal to $xP'(x) + P(x)$. Therefore we can easily compute
$Q(x)$ and $Q'(x)$ if we know $x$, $c$, $P(x)$ and $P'(x)$.\es

This solution has a unexpected
feature: we do not need to know in advance the degree of the
polynomial. If we add this requirement and ask to compute the
value of the derivative only (not mentioning the polynomial itself),
we get a rather confusing problem.

\medskip

There is a general statement about the computation of
derivatives:

\Problem{%
(W.~Baur, V.~Strassen) \
Assume that a ``straight-line'' program (containing only
assignment statements) computes the value of some polynomial
$P(x_1,\ldots,x_n)$ given the variables
$x_1,\ldots,x_n$.
We assume that the right-hand sides of the assignment
statements
are expressions that contain only addition, multiplication, constants,
variables
$x_1,\ldots,x_n$ and the variables that appear on the left-hand side
of previous assignment statements.
Prove that there exists a program of the same type that computes
all $n$ derivatives $\partial P/\partial x_1,\ldots,
\partial P/\partial x_n$,
where the number of arithmetic
operations is only $C$ times larger than in the
original program. (Here the constant $C$ does not depend on
$n$.)}%
\index{Baur, W.}%
\index{Strassen, V.}

\Hint{We may assume that each assignment consists of addition,
multiplication
by a constant, or multiplication of two variables.
Use induction on the number of statements, applying
the inductive assumption to the program obtained
by deleting the {\it first} command of the program.
}

\Problem{%
     Two arrays
\w{a:\  array[0..k]} \w{of} \w{integer} and \w{b:\ array[0..l]}
\w{of} \w{integer}
contain the coefficients of two polynomials of degrees
\w{k} and \w{l} respectively.
Put into \w{c:\ array[0..m]} \w{of} \w{integer}
the coefficients of their product.
(Here $\w{k},\w{l},\w{m}$ are nonnegative integers such that
$\w{m}=\w{k}+\w{l}$; the array element indexed by
\w{i} contains the coefficient of
$\w{x}^\w{i}$.) }%
\index{polynomial!multiplication}%
\index{multiplication!of polynomials}%

     \Solution.
\begin{verbatim}
          for i:=0 to m do begin
          | c[i]:=0;
          end;
\end{verbatim}
\clearpage
\begin{Verbatim}
          for i:=0 to k do begin
          | for j:=0 to l do begin
          | | c[i+j] := c[i+j] + a[i]*b[j];
          | end;
          end;
\end{Verbatim}

\Problem{%
     The multiplication algorithm for polynomials given above uses about
$n^2$ operations to compute the product of two polynomials of degree
$n$. Find an (asymptotically)
more effective algorithm that uses only
$O(n^{\log 4/\log 3})$ operations.}%
\index{polynomial!multiplication}%
\index{fast multiplication}%
\index{multiplication!fast}%

    \Hint{Suppose we want to multiply two polynomials
of degree $2k$. Represent these polynomials as
$$
         A(x)\,x^k + B(x) \quad  \mbox{and} \quad   C(x)\,x^k + D(x)
$$
where $A,B,C,D$ are polynomials of degree $k$.
The product in question is equal to
$$
       A(x)C(x)\,x^{2k}  +  (A(x)D(x)+B(x)C(x))\,x^k  + B(x)D(x).
$$
The natural way to compute $AC$, $AD+BC$, $BD$ requires four
multiplications of degree $k$ polynomials. However, the
following trick requires only three multiplications:
compute
$AC$, $BD$ and $(A+B)(C+D)$, then use the identity
$AD+BC=(A+B)(C+D)-AC-BD$.}

