
\Problem{%
     Two arrays \w{x:\ array[1..k]} \w{of}
\w{integer} and
\w{y:\ array[1..l]} \w{of} \w{integer} are sorted
($\w{x[1]}<\cdots<\w{x[k]}$,
$\w{y[1]}<\cdots<\w{y[l]}$).
Find the number of common elements in both arrays, that is, the
number of integers \w{t} such that
$\w{t} = \w{x[i]} = \w{y[j]}$ for some \w{i} and \w{j}.
(The number of operations should be
of order $\w{k}+\w{l}$.)}%
\index{number!of common elements}%

     \Solution.
{\samepage
\begin{verbatim}
  k1:=0; l1:=0; n:=0;
  {invariant relation: 0<=k1<=k; 0<=l1<=l;
   the number in question is n plus the number of common
   elements in x[k1+1],...,x[k] and y[l1+1],...,y[l]}
  while (k1 <> k) and (l1 <> l) do begin
  | if x[k1+1] < y[l1+1] then begin
  | | k1 := k1 + 1;
  | end else if x[k1+1] > y[l1+1] then begin
  | | l1 := l1 + 1;
  | end else begin {x[k1+1] = y[l1+1]}
  | | k1 := k1 + 1;
  | | l1 := l1 + 1;
  | | n := n + 1;
  | end;
  end;
\end{verbatim}
}
\begin{verbatim}
  {k1 = k or l1 = l; therefore, one of the sets
   mentioned in the invariant relation is empty
   and n is the number in question}
\end{verbatim}

Remark. In the last alternative it is enough to increase only one of
the variables \w{k1} and \w{l1} (though the symmetry would be
broken if we did that).\es

\Problem{%
     Solve the preceding problem with the assumption that
$\w{x[1]}\le\cdots\le\w{x[k]}$ and $\w{y[1]}\le\cdots\le\w{y[l]}$
(arrays are nondecreasing but not necessarily increasing).}

     \Solution. In the third alternative of the previous solution,
when increasing \w{k1} and \w{l1} by \w{1}, we decreased (by
\w{1}) the number of common elements in
$\w{x[k1+1]}\ldots\w{x[k]}$ and $\w{x[l1+1]}\ldots\w{x[l]}$.
For nondecreasing arrays, this is not enough since the same element
may appear many times.
A more complicated procedure is
required:
\begin{verbatim}
          ...
          end else begin {x[k1+1] = y[l1+1]}
          | t := x [k1+1];
          | while (k1<k) and (x[k1+1]=t) do begin
          | | k1 := k1 + 1;
          | end;
          | while (l1<l) and (x[l1+1]=t) do begin
          | | l1 := l1 + 1;
          | end;
          | n := n+1;
          end;
\end{verbatim}

     Remark. This program has a bug, however.
If in the condition
$$
           \w{(k1<k) and (x[k1+1]=t)}
$$
(or in the similar second condition) the first expression (\w{k1<k})
is false, the second one is meaningless (index out of bounds)
and an error may occur. Some versions of Pascal use ``short circuit
evaluation'' of Boolean expressions: when evaluating
\w{A and B}%
\index{error!index out of bounds}
the evaluation of \w{B} is ''short circuited'' when
\w{A} is false.
(This is the default behaviour of Turbo Pascal 5.0,
\index{Turbo Pascal}%
but not
3.0.) In this case, the
problem disappears.

     Rather than rely on implementation-dependent
features \index{Pascal}
\index{Pascal}%
(short-circuit evaluation is not prescribed by the Pascal's author,
N.~Wirth),\index{Wirth, N.}
we can do the following. Introduce an additional variable
\verb|b: Boolean| and write:
\begin{verbatim}
        if k1 < k  then b := (x[k1+1]=t)  else  b:=false;
        {b = (k1<k) and (x[k1+1] = t)}
        while  b  do  begin
        | k1 := k1+1;
        | if k1 < k then b := (x[k1+1]=t)  else  b:=false;
        end;
\end{verbatim}
Another possibility (which is shorter, but less symmetric):
\begin{verbatim}
        end else begin {x[k1+1] = y[l1+1]}
        | if k1 + 1 = k then begin
        | | k1 := k1 + 1;
        | | n := n + 1;
        | end else if x[k1+1] = x [k1+2] then begin
        | | k1 := k1 + 1;
        | end else begin
        | | k1 := k1 + 1;
        | | n := n + 1;
        | end;
        end;
\end{verbatim}

    Alternatively, we can increase the constant in the array declaration
and reserve a spare memory location.\es

\Problem{%
     Two arrays \w{x:\ array[1..k]} \w{of}
\w{integer} and \w{y:\ array[1..l]} \w{of} \w{integer} satisfying
$\w{x[1]}\le\cdots\le\w{x[k]}$,
$\w{y[1]}\le\cdots\le\w{y[l]}$ are given.
Find the number of different elements among
$\w{x[1]},\ldots,\w{x[k]},\w{y[1]},\ldots,\w{y[l]}$. (The number
of operations should be of order
$\w{k}+\w{l}$.)\es}

\Problem{%
     Two arrays $\w{x[1]}\le\cdots\le\w{x[k]}$ and
$\w{y[1]}\le\cdots\le\w{y[l]}$ are given. Merge them into one array
$\w{z[1]}\le\cdots\le\w{z[m]}$ ($\w{m}=\w{k}+\w{l}$). Any element should
appear is \w{z} as many times as it appears in
\w{x} and \w{y} together.
The number of operations should be of order
\w{m}.}%
\index{merge!of sorted arrays}%

     \Solution.
\begin{verbatim}
 k1 := 0; l1 := 0;
 {invariant relation: the answer is the concatenation
   of z[1]..z[k1+l1] and the merge of
   x[k1+1]..x[k] and y[l1+1]..y[l]}
 while (k1 <> k) or (l1 <> l) do begin
 | if k1 = k then begin
 | | {l1 < l}
 | | l1 := l1 + 1;
 | | z[k1+l1] := y[l1];
 | end else if l1 = l then begin
 | | {k1 < k}
 | | k1 := k1 + 1;
 | | z[k1+l1] := x[k1];
 | end else if x[k1+1] <= y[l1+1] then begin
 | | k1 := k1 + 1;
 | | z[k1+l1] := x[k1];
 | end else if x[k1+1] >= y[l1+1] then begin
 | | l1 := l1 + 1;
 | | z[k1+l1] := y[l1];
 | end else begin
 | | {this cannot happen}
 | end;
 end;
 {k1 = k, l1 = l, arrays are merged}
\end{verbatim}
This process can be illustrated as follows. Assume we have
two piles of cards with a word on each card, and each pile is
alphabetically sorted. We merge them into one pile as follows.
At every
step we compare the first cards of both piles and take the one
which is alphabetically first.
If one pile is already empty,
we take the remaining
cards from the other
pile.\es

\Problem{%
     Two arrays $\w{x[1]}\le\cdots\le\w{x[k]}$ and
$\w{y[1]}\le\cdots\le\w{y[l]}$ are given. Find their
``intersection'', i.e., an array
$\w{z[1]}\le\cdots\le\w{z[m]}$ that contains their common
elements. The multiplicity of each element in \w{z}
should be equal to the smaller of its multiplicities in
\w{x} and \w{y}.
The number of operations should be of order
$\w{k}+\w{l}$.\es}%
\index{intersection!of sorted arrays}%

\Problem{%
     Two arrays $\w{x[1]}\le\cdots\le\w{x[k]}$  and
$\w{y[1]}\le\cdots\le\w{y[l]}$
and a number
\w{q} are given. Find \w{i} and \w{j}
such that $\w{x[i]}+\w{y[j]}$ is as close to \w{q} as possible.
(The number of operations should be of order
\w{k+l}. You may use a fixed number of auxiliary
integer variables;
the arrays \w{x} and \w{y} are read-only.)}%
\index{nearest sum}%

    \Hint{We need to find the minimal distance between
$\w{x[1]}\le\cdots\le\w{x[k]}$ and
$\w{q}-\w{y[l]}\le\cdots\le\w{q}-\w{y[1]}$. This is easily done
while merging these numbers into one (imaginary)
array.}

\Problem{%
     (from D.~Gries' book [6]) There is a number that is present in
all three nondecreasing arrays
$\w{x[1]}\le\cdots\le\w{x[p]}$,
$\w{y[1]}\le\cdots\le\w{y[q]}$,
$\w{z[1]}\le\cdots\le\w{z[r]}$. Find this number (or one of them,
if there is more than one). The number of operations should be of order
$\w{p}+\w{q}+\w{r}$.}%
\index{Gries, D.}%
\index{common element (in sorted arrays)}%

     \Solution.
\begin{Verbatim}
  p1:=1; q1=1; r1:=1;
  {invariant relation: x[p1]..x[p], y[q1]..y[q],
   z[r1]..z[r] have an element in common}
  while not ((x[p1]=y[q1]) and (y[q1]=z[r1])) do begin
  | if x[p1]<y[q1] then begin
  | | p1:=p1+1;
  | end else if y[q1]<z[r1] then begin
  | | q1:=q1+1;
  | end else if z[r1]<x[p1] then begin
  | | r1:=r1+1;
  | end else begin
  | | {this cannot happen}
  | end;
  end;
  {x[p1] = y[q1] = z[r1]}
  writeln (x[p1]);
\end{Verbatim}

\Problem{%
   Repeat the previous problem assuming that
we do not know in advance if such a
common element exist. Determine whether or not it exists
and locate it if it does.\es}

\Problem{%
     The array \w{a[1..n]} consists of arrays
\w{[1..m]} of integers:
$$
\w{a: array [1..n] of array [1..m] of integer;}
$$
$$
\w{a[1][1]}\le\cdots\le\w{a[1][m]},\ \ldots, \
\w{a[n][1]}\le\cdots\le\w{a[n][m]}.
$$
It is known that there is a common
number present in all
\w{a[i]}
(that is,
there exists an \w{x} such that for all \w{i} in \w{1..n}
there exists a \w{j} in \w{1..m} such that $\w{a[i][j]}=\w{x}$).
Find such a number \w{x}. }

     \Solution. We use an array $\w{b[1]..b[n]}$ whose elements
mark
the start of the ``non-scanned'' portions of arrays
$\w{a[1]},\ldots,\w{a[n]}$.
\begin{verbatim}
 for k:=1 to n do begin
 |  b[k]:=1;
 end;
 eq := true;
 for k := 2 to n do begin
 | eq := eq and (a[1][b[1]] = a[k][b[k]]);
 end;
 {invariant relation: non-scanned parts have nonempty
  intersection, i.e., there is x such that for any i in
  [1..n] there is j in [b[i]..m] such that a[i][j] = ;
  eq <=> first non-scanned elements are all equal}
 while not eq do begin
 | s := 1; k := 1;
 | {a[s][b[s]] is minimal among a[1][b[1]]..a[k][b[k]]}
 | while k <> n do begin
 | | k := k + 1;
 | | if a[k][b[k]] < a[s][b[s]] then begin
 | | | s := k;
 | | end;
 | end;
 | {a[s][b[s]] is minimal among a[1][b[1]]..a[n][b[n]]}
 | b [s] := b [s] + 1;
\end{verbatim}
\clearpage
\begin{Verbatim}
 | for k := 2 to n do begin
 | | eq := eq and (a[1][b[1]] = a[k][b[k]]);
 | end;
 end;
 writeln (a[1][b[1]]);
\end{Verbatim}

\Problem{%
      Our solution of the preceding problem requires
$\w{m}\w{n}^2$ operations. Find an algorithm that needs only
$O(\w{mn})$ operations
(i.e., not more than $C\w{mn}$ operations for some $C$).}

    \Hint{We have to break the symmetry and choose one of the rows
as a ``principal'' row.
We move along the principal row maintaining the
following relation: in all other rows the maximal
element not exceeding the
current element of the principal row is located.}

