\Problem{%
     (Binary search)\label{binsearch} An array
$\w{x[1]}\le\cdots\le\w{x[n]}$ of integers and an integer \w{a}
are given. Determine if
\w{a} is present in \w{x}, that is,
if there exists an
\w{i} in \w{1..n} such that $\w{x[i]}=\w{a}$.
(The number of operations should be of order
$\log\w{n}$.) }%
\index{search!binary}%
\index{binary search}%

     \Solution. (We assume that $\w{n}>\w{0}$.)
\begin{verbatim}
  l := 1; r := n+1;
  {r > l, if a is present, it is present among x[l]..x[r-1]}
  while r - l <> 1 do begin
  | m := l + (r-l) div 2 ;
  | {l < m < r }
  | if x[m] <= a then begin
  | | l := m;
  | end else begin {x[m] > a}
  | | r := m;
  | end;
  end;
\end{verbatim}
(Check that the invariant relation is maintained even if
 $\w{x[m]} = \w{a}$.)

     At each step the difference
 $\w{r}-\w{l}$ is halved, so we get the required
bound for the number of operations.

Program can be simplified using the equality
$$
\w{l} + \w{(r-l)}\,\w{div}\,\w{2} =
     (\w{2l} + (\w{r}-\w{l}))\,\w{div}\, \w{2} =
     (\w{r}+\w{l})\,\w{div}\,\w{2}.
\eqno\es
$$

  Remark. It is very important that the array
\w{x[1]..x[n]} is sorted; otherwise we have to test
all \w{n} elements \w{x[1]..x[n]} to be sure that
a given element is not in the array (``sequential search'').

\Problem{%
     (From D.~Gries' book [6]) An array \w{x:~array[1..n]} \w{of}
\w{array[1..m]} \w{of} \w{integer} is sorted both row-wise
and column-wise:
$$
     \w{x[i][j]} \le  \w{x[i][j+1]},
$$
$$
     \w{x[i][j]} \le \w{x[i+1][j]},
$$
Determine if a given number \w{a} is present among the array
elements
\w{x[i][j]}. }%
\index{Gries, D.}%

     \Solution. Represent \w{x} as a rectangular
matrix. Choose a rectangle that contains \w{a}
(assuming that \w{a} is present at all)
and then make this rectangle smaller and smaller.
This rectangle contains \w{x[i][j]} such that
$\w{1}\le\w{i}\le\w{l}$ and $\w{k}\le\w{j}\le\w{m}$.
%
%               1                     k         m
%              -----------------------------------
%             1|                     |***********|
%              |                     |***********|
%              |                     |***********|
%             l|                     |***********|
%              |---------------------------------|
%              |                                 |
%             n|                                 |
%              -----------------------------------
%
\setlength{\unitlength}{1.2em}
\[
\begin{picture}(11,8)
\thinlines
\put(1,1){\line(1,0){9}}
\put(1,7){\line(1,0){9}}
\put(1,1){\line(0,1){6}}
\put(10,1){\line(0,1){6}}
\put(6,1){\dashbox{0.1}(4,6){}}
\put(1,3){\dashbox{0.1}(9,4){}}
\put(0,1){\makebox(1,1){\w{n}}}
\put(0,3){\makebox(1,1){\w{l}}}
\put(0,6){\makebox(1,1){\w{1}}}
\put(1,7){\makebox(1,1){\w{1}}}
\put(6,7){\makebox(1,1){\w{k}}}
\put(9,7){\makebox(1,1){\w{m}}}
\put(6,3){\framebox(4,4){?}}
\end{picture}
\]
(The rectangle is empty if $\w{l} = \w{0}$ or
$\w{k}=\w{m}+\w{1}$.)
\begin{verbatim}
  l:=n; k:=1;
  {l>=0, k<=m+1, if a is present at all, it is present
   inside the rectangle}
  while (l > 0) and (k < m+1) and (x[l][k] <> a) do begin
  | if x[l][k] < a then begin
  | | k := k + 1; {left column cannot contain a, delete it}
  | end else begin {x[l][k] > a}
  | | l := l - 1; {last row cannot contain a, delete it}
  | end;
  end;
  {x[l][k] = a or the rectangle is empty}
  answer:= (l > 0) and (k < m+1) ;
\end{verbatim}

     Remark. Here the same error as in problem 1.2.18
appears:
\w{x[l][k]} may be undefined. (We leave its correction
to the reader.)\es
\index{error!index out of bounds}%

\Problem{%
     (Moscow programming contest) A nondecreasing integer
array
$\w{a[1]}\le\w{a[2]}\le\cdots\le\w{a[n]}$ contains positive
numbers only. Find the minimal positive integer that cannot
be represented as a sum of several elements of this array
(no element may be used not more than once). The number of
operations should be of order \w{n}.}

     \Solution. Assume all numbers that can be represented
as sums of subsets of $\{\w{a[1]},\ldots,\w{a[k]}\}$ form the set
$\{\w{1},\w{2},\ldots,\w{N}\}$. If $\w{a[k+1]} > \w{N+1}$, then
$\w{N+1}$ is the smallest number that cannot be represented as
the sum of some subset of $\{\w{a[1]},\ldots,\w{a[n]}\}$. If
$\w{a[k+1]}\le\w{N+1}$, then all numbers that can be represented
as sums of subsets of $\{\w{a[1]}\ldots\w{a[k+1]}\}$ form the set
$\{\w{1},\w{2},\ldots,\w{N+a[k+1]}\}$.
\begin{verbatim}
  k := 0; N := 0;
  {invariant relation: all the numbers that can be
   represented as sums of subsets of {a[1],..,a[k]},
   form the set {1,2,...,N}}
  while (k <> n) and (a[k+1] <= N+1) do begin
  | N := N + a[k+1];
  | k := k + 1;
  end;
  {(k = n) or (a[k+1] > N+1); the answer is N+1
   in both cases}
  writeln (N+1);
\end{verbatim}
(Error: when the first condition in the \w{while}-construct
is false, the second is undefined.)\es
\index{error!index out of bounds}%

\Problem{%
     (Requires some algebra) An integer array
$\w{a[1]..a[n]}$ contains some permutation of
$\w{1..n}$  (each of numbers $\w{1..n}$ appears
exactly once).
\par
     (a) Determine if the permutation is even.
\par
     (b) Without using
other arrays, replace the permutation by its inverse
permutation (i.e.,
if $\w{a[i]}=\w{j}$ was true before execution, then
$\w{a[j]}=\w{i}$ is true after execution). }%
(In both (a) and (b), the number of operations should be of order
\w{n}.)
\index{permutation!even}%
\index{permutation!inverse}%
\index{even permutation}%
\index{inverse permutation}%

     \Hint{(a) The number of cycles determines whether
a permutation is even or odd. To mark an already counted cycle, we can
(for example) change the sign of its elements.
(b) The inverse permutation is computed cycle by cycle.}

\Problem{%
     An array \w{a[1..n]} and a threshold \w{b} are given.
Rearrange the elements of the array in such a way that all
elements on the left of some boundary do not exceed \w{b} whereas
all elements on the right of the boundary are greater than or
equal to \w{b}. The number of operations should be proportional
to \w{n}.}

     \Solution.
\begin{Verbatim}
        l:=0; r:=n;
        {invariant relation: a[1]..a[l]<=b; a[r+1]..a[n]>=b}
        while l <> r do begin
        | if a[l+1] <= b then begin
        | | l:=l+1;
        | end else if a[r] >=b then begin
        | | r:=r-1;
        | end else begin {a[l+1]>b; a[r]<b}
        | | ..exchange a[l+1] and a[r]
        | | l:=l+1; r:=r-1;
        | end;
        end;
\end{Verbatim}

\Problem{%
     Repeat the previous
problem with the additional restriction that the
elements smaller than \w{b} should precede elements
equal to \w{b} which themselves should precede
elements greater than \w{b}.
}\label{separation}

     \Solution. We need three boundaries to divide our segment
into four parts. The first part contains elements smaller than \w{b};
the second part contains only elements equal
to \w{b}; the third part may contain anything;
and the fourth part contains only elements greater than \w{b}.
(We can get a more symmetric solution using a fourth boundary, but
that's not important.) At each step we consider the left
element of the third part (just to the right of the second boundary).
\begin{Verbatim}
   l:=0; m:=0; r:=n;
   {invariant relation:
      a[1..l]<b; a[l+1..m]=b; a[r+1]..a[n]>b}
   while m <> r do begin
   | if a[m+1]=b then begin
   | | m:=m+1;
   | end else if a[m+1]>b then begin
   | | ..exchange a[m+1] and a[r]
   | | r:=r-1;
   | end else begin {a[m+1]<b}
   | | ..exchange a[m+1] and a[l+1]
   | | l:=l+1; m:=m+1;
   end;
\end{Verbatim}

\Problem{%
     (This version of the preceding problem is called the
``Dutch flag''
problem in E.~Dijkstra's book [4].) The array
contains \w{n} elements; each element is equal to
\w{0}, \w{1}, or \w{2}. Sort the array
if the only allowed operation
(besides reading its elements) is the exchange of two elements
of the array. The number of operations should be proportional
to \w{n}.\es
}%
\index{Dijkstra, E.W.}%
\index{Dutch flag}%

\Problem{%
     An array \w{a[1..n]} and a number $\w{m}\le\w{n}$ are given.
For any segment formed by \w{m} adjacent elements (there are
$\w{n}-\w{m}+\w{1}$ segments of this type) compute its sum.
The total number of operations should be of order
\w{n}.}

     \Solution. When moving the segment to the right, add one element
and subtract another.\es

\Problem{%
     A square matrix \w{a[1..n][1..n]} and a number
$\w{m}\le\w{n}$ are given.
For any $\w{m}\times\w{m}$-subsquare, compute the sum of
its elements. The total number of operations should be of
order $\w{n}^2$. }

     \Solution. First compute the sum for all horizontal rectangles
of size $\w{m}\times\w{1}$.
(When such a rectangle is shifted to the right, one element is added
and one is subtracted.) After computing all these sums, we
compute the sums for squares. (When a square is shifted down, one rectangle
is added and another rectangle is subtracted.)\es

\Problem{%
  An array $\w{a[1]..a[n]}$ contains all integers
in \w{[0..n]} except one. Find this omitted integer
with fixed additional
memory. Number of operations should be proportional to \w{n}.
}

\Hint{Add all the numbers.}

