\subsection[Inductive functions]%
{Inductive functions (following A.G.~Kushnirenko)}\label{indfunc}

\index{function!inductive}%
\index{inductive function}%
\index{Kushnirenko, A.G.}%
     Let \w{M} be a set. Let \w{f} be a function whose arguments are
finite sequences of elements of \w{M} and whose values belong to some
other set \w{N}. The function \w{f} is called
{\em inductive\/} if its value on the sequence
$\w{x[1]..x[n]}$ is uniquely determined by its value on the
sequence $\w{x[1]..x[n-1]}$ and by \w{x[n]}, that is, if
there is a function $\w{F} : \w{N}\times\w{M} \to
\w{N}$ such that
$$
   \w{f}(\langle\w{x[1]},\ldots,\w{x[n]}\rangle) =
      \w{F}(\, \w{f}(\langle\w{x[1]},\ldots,\w{x[n-1]}\rangle),\w{x[n]}).
$$
For example, the sum $\w{x[1]}+\cdots+\w{x[n]}$ is an inductive
function (it is enough to know the sum
$\w{x[1]}+\cdots+\w{x[n-1]}$ and the value of \w{x[n]} to compute
$\w{x[1]}+\cdots+\w{x[n]}$). At the same time, the
average value is not an
inductive function; if we know \w{x[n]} and the average of
$\w{x[1]},\ldots,\w{x[n-1]}$, but have no information about
\w{n}, we cannot compute the average of
$\w{x[1]},\ldots,\w{x[n]}$.

An inductive function can be computed as follows:
\begin{verbatim}
  k := 0; f := f0;
  {invariant relation:
    f is a value of the function on <x[1],...,x[k]>}
  while  k<>n do begin
  | k := k + 1;
  | f := F (f, x[k]);
  end;
\end{verbatim}

     Here \w{f0} is the value of the function on the empty sequence
(sequence of length \w{0}). If \w{f} is defined only on nonempty
sequences, the first line should be replaced by
$$ \w{k:=1; f:=f(<x[1]>);}$$

%Inductive extensions.

     If a given function
\w{f} is not inductive, it is instructive to look for its
{\em inductive extension}.
\index{inductive extension}%
\index{extension, inductive}%
By an inductive extension of \w{f} we mean an inductive function
\w{g}  whose values determine uniquely the values of
\w{f} (i.e., there exists a function
\w{t} such that
$$
 \w{f} (\langle\w{x[1]}\ldots\w{x[n]}\rangle)
   = \w{t} (\w{g}(\langle\w{x[1]}\ldots\w{x[n]}\rangle))
$$
for all $\langle\w{x[1]}\ldots\w{x[n]}\rangle$). One can prove
that there exists a minimal extension \w{F} among all inductive
extensions of a given function \w{f}. Here the word ``minimal'' means
that for any other inductive extension \w{g}
the values of \w{F} are determined uniquely by the values of \w{g},
that is, $\w{F}(\w{x})=\w{u}(\w{g}(\w{x}))$ for some function \w{u}.

\Problem{%
     Find an inductive extension for the following functions:
\par
  (a) the average value of a sequence of real numbers;
\par
  (b) the number of elements in a sequence that are equal to its maximal
element;
\par
  (c)  the second largest element of the sequence (second
from the top after
the sequence is sorted in nondescending order);
\par
  (d) the maximal number of
consecutive
equal elements;
\par
  (e) the maximal length of a monotone (nonincreasing or nondecreasing)
fragment composed of consecutive elements of a sequence;
\par
  (f) the number of groups of ones separated by zeros (in a
0-1-sequence).}

     \Solution.

(a) As we have seen, the average value is not an inductive function.
   However, the average value is a ratio of two inductive functions.
   The first one is the sum of all the terms; the second one is the
   number of terms. Therefore, the combination
   $\langle$the sum of all elements; the length$\rangle$
   is an inductive extension.

(b) $\langle$the maximal element; the number of elements equal to
     the maximal element$\rangle$;

(c) $\langle$the maximal element; the second maximal element$\rangle$;

(d) $\langle$the maximal number of adjacent equal elements;
     the maximal number of adjacent equal elements at the end of
     the
     sequence;
     the last element$\rangle$;

(e) $\langle$the maximal length of a monotone fragment; the
      maximal length
      of a nondecreasing fragment at the end of the sequence;
      maximal length of a nonincreasing fragment at the end of the
      sequence; the last term of the sequence$\rangle$;

(f) $\langle$the number of \w{1}-groups; the last term$\rangle$.\es

\Problem{%
     (Communicated by D.V.~Varsonofiev)
Two sequences
$\w{x[1]..x[n]}$ and $\w{y[1]..y[k]}$ of integers are
given. Determine if the second sequence is a subsequence of the
first one, that is, if it is possible to delete some terms of the
first sequence to obtain the second one. The number of operations
should be of order $\w{n}+\w{k}$. }%
\index{Varsonofiev, D.V.}%
\index{subsequence!test}%

       \Solution. (version 1) Reduce the problem to the
same problem involving shorter sequences.
\begin{verbatim}
  n1:=n;
  k1:=k;
  {invariant relation: the answer is TRUE <=>
    it is possible to get y[1]..y[k1] out of x[1]..x[n1]}
  while (n1 > 0) and (k1 > 0) do begin
  | if x[n1] = y[k1] then begin
  | | n1 := n1 - 1;
  | | k1 := k1 - 1;
  | end else begin
  | | n1 := n1 - 1;
  | end;
  end;
  {n1 = 0 or k1 = 0; if k1 = 0, the answer is positive;
   if k1<>0 (and n1 = 0), the answer is negative}
  answer := (k1 = 0);
\end{verbatim}

     We use the following fact:  If $\w{x[n1]}=\w{y[k1]}$
and $\w{y[1]..y[k1]}$ is a subsequence of
$\w{x[1]..x[n1]}$, then $\w{y[1]..y[k1-1]}$ is a
subsequence of $\w{x[1]..x[n1-1]}$.

     (version 2) The function
$\langle\w{x[1]..x[n1]}\rangle$ $\mapsto$ [the maximal
\w{k1} such that $\w{y[1]..y[k1]}$ is a
subsequence of $\w{x[1]..x[n1]}$] is inductive.\es

\Problem{%
     Two sequences $\w{x[1]..x[n]}$ and
$\w{y[1]..y[k]}$ of integers are given. Find the maximal
length of a sequence that is a subsequence of both given
sequences. The number of operations should be of order
$\w{n}\cdot\w{k}$.
}%
\index{subsequence!common}%

     Solution (communicated by M.N.~Weinzweig and A.M.~Dimentman).
\index{Weinzweig, M.N.}%
\index{Dimentman, A.M.}%
Denote the maximal length of a common
subsequence of sequences
$\w{x[1]..x[p]}$ and $\w{y[1]..y[q]}$
by $\w{f}(\w{p},\w{q})$.
Then
\begin{eqnarray*}
\w{x[p]}\ne\w{y[q]} &\Rightarrow&
    \w{f}(\w{p},\w{q}) = \cmr{max}\,(\w{f}(\w{p},\w{q}\!-\!\w{1}),
                        \w{f}(\w{p}\!-\!\w{1},\w{q}));\\
\w{x[p]}=\w{y[q]} &\Rightarrow&
    \w{f}(\w{p},\w{q}) = \cmr{max}\,(\w{f}(\w{p},\w{q}\!-\!\w{1}),
                        \w{f}(\w{p}\!-\!\w{1},\w{q}),
               \w{f}(\w{p}\!-\!\w{1},\w{q}\!-\!\w{1})\!+\!\w{1})
\end{eqnarray*}
(In the second case, the maximum of three numbers is in fact equal
to  the third number because
$\w{f}(\w{p}-\w{1},\w{q}-\w{1})+\w{1} \ge \w{f}(\w{p},\w{q}-\w{1}),
\w{f}(\w{p}-\w{1},\w{q})$.)

     Therefore we can construct a table of \w{f}-values.
This table is of size
$\w{n}\cdot\w{k}$. We can even proceed using only
\w{k} (or \w{n}) memory locations
if we compute (for $\w{p}=\w{1},\w{2},\ldots$) the
array
$\langle\w{f(p,0)},\ldots,\w{f(p,k)}\rangle$
(it is an inductive function of \w{p}).\nobreak\es

\Problem{%
\hspace{-.3em}(from D.~Gries' book [6]) A sequence of integers
$\w{x[1]},\ldots,\w{x[n]}$ is given. Find the maximum length
of an increasing subsequence. (The number of operations should
be of order
$\w{n}\,\log\w{n}$). }%
\index{Gries, D.}%
\index{subsequence!increasing}%

     \Solution. The function in question is not inductive. However,
it has the following inductive extension: it consists of the
maximal length of the increasing subsequences (denoted by \w{k} in
the sequel) and the numbers
$\w{u[1]},\ldots,\w{u[k]}$, where $\w{u[i]}$ is the minimal last
term of all increasing subsequences of length
\w{i}. Evidently, $\w{u[1]}\le\cdots\le\w{u[k]}$. When a new term is
appended to \w{x}, the values of \w{u} and \w{k} should be updated.
\begin{verbatim}
     n1 := 1; k := 1; u[1] := x[1];
     {invariant: k and u satisfy the description above}
     while n1 <> n do begin
     | n1 := n1 + 1;
     | ...
     | {i is the maximal number in 1..k such that
     |   u[i] < x[n1]; i=0 if there is no such numbers}
\end{verbatim}
\clearpage
\begin{verbatim}
     | if i = k then begin
     | | k := k + 1;
     | | u[k+1] := x[n1];
     | end else begin {i < k, u[i] < x[n1] <= u[i+1] }
     | | u[i+1] := x[n1];
     | end;
     end;
\end{verbatim}
The omitted
fragment employs binary search (see 1.2.27, p.~\ref{binsearch})
In the invariant
relation we assume that $\w{u[0]}=-\infty$ and
$\w{u[k+1]} =+\infty$. The goal is $\w{u[i]} <
\w{x[n1]}\le \w{u[i+1]}$.
\begin{verbatim}
  i:=0; j:=k+1;
  {u[i] < x[n1] <= u[j], j > i}
  while (j - i) <> 1 do begin
  | s := i + (j-i) div 2;    {i < s < j}
  | if x[n1] <= u[s] then begin
  | | j := s;
  | end else begin {u[s] < x[n1]}
  | | i := s;
  | end;
  end;
  {u[i] < x[n1] <= u[j], j-i = 1}
\end{verbatim}

     Remark. We get a simpler (but not minimal) inductive
extension if, for any \w{i}, we keep the maximal length of an
increasing sequence whose last term is
\w{x[i]}. This extension leads to an algorithm requiring
$\w{n}^2$ operations.\es

\Problem{%
What changes are needed in the solution of the previous
problem if we are looking for a   maximal {\it nondecreasing\/}
sequence?\es}

