\section{Generation of combinatorial objects}\label{generation}

    In this section, we deal with problems that require us
to generate
all the elements of some finite set
one-by-one.

\subsection{Sequences}

    \Problem{Print all the sequences of length \verb|k| composed of
the numbers
\verb|1..n|.}%
\index{sequences!generation}

     \Solution. Let us print them in alphabetic order
(a sequence \verb|a| precedes sequence \verb|b| if
for some \verb|s| their initial segments of length
\verb|s| are equal and the
\verb|(s+1)|-th term of \w{a} is smaller.)
The first sequence in this ordering is
\verb|<1,1,...,1>|; the last one is
\verb|<n,n,...,n>|.
We use an array
\verb|x[1]..x[k]| to store the last sequence printed.
\begin{verbatim}
        ..make x[1]...x[k] equal to 1
        ..print x
        ..make last[1]...last[k] make equal to n
        {all sequences up to x (including x) are printed}
        while x <> last do begin
        | ...x := the successor of x
        | ...print x
        end;
\end{verbatim}

     Let us explain how to get the successor of
\verb|x|. By definition, the successor should have the same
first \w{s} terms and larger \w{(s+1)}-th term.
This is possible only if
$\w{x[s+1]}<\w{n}$.
To get the immediate successor, we find the maximal \w{s}
with this property and increase the corresponding element
by \w{1}.
In other words, we move along the sequence
from right to left and find the rightmost term that is smaller
than \w{n} (it does exist, because \w{x<>last} by assumption).
Then we increase it by \w{1} and make all the subsequent
terms equal to \w{1}.
\begin{verbatim}
        p := k;
        while not (x[p] < n) do begin
        | p := p-1;
        end;
        {x[p] < n, x[p+1] =...= x[k] = n}
        x[p] := x[p] + 1;
        for i := p+1 to k do begin
        | x[i] := 1;
        end;
\end{verbatim}

     Remark. If we use the numbers $\w{0..n-1}$ instead of
$\w{1..n}$, then finding the successor corresponds to
adding \w{1} in \w{n}-ary notation.\es

    \Problem{The program above uses comparisons for arrays
(\verb|x <> last|). Eliminate this step by using a
Boolean variable
\verb|l| and adding the requirement}
$$
\w{l}\Leftrightarrow \w{x} = \w{last}
$$
to the invariant relation.\es

    \Problem{Print all subsets of the set \verb|{1..k}|.}%
\index{subsets!generation}%

     \Solution. These subsets are in one-to-one correspondence with
all sequences of \w{0}s and \w{1}s  of length \w{k}.\es

    \Problem{Print all sequences of length \w{k}
of positive integers such that the
\w{i}-th term does not exceed \w{i}
for all \w{i}.\es}

