    \subsection{Permutations}

    \Problem{Print all permutations of \verb|1..n| (i.e.,
all sequences of length \verb|n| that contain all the numbers
$\w{1}\ldots\w{n}$).}%
 \index{permutations!generation}%

     \Solution. We store the current
permutation in an array \verb|x[1]..x[n]|.
Permutations are printed in
lexicographic order. The first permutation (in this
order) is
$\langle \w{1}\,\w{2..n}\rangle$.
The last one is
$\langle\w{n..2}\,\w{1}\rangle$.
How do we find the next permutation
(in the lexicographic order)?
When is it possible to
increase the
\w{k}-th term in a permutation without
changing
all preceding terms?
The answer is: When the term
is smaller than one of the next terms
(i.e., terms with larger
indices).
Therefore, to find the next permutation we should find the
maximum \w{k}
for which it is possible, that is, a
\w{k} such that
\[
    \w{x[k] < x[k+1] > } \cdots \w{ > x[n]}
\]
Next we increase \w{x[k]} but keep the increase
as small as possible. This
means that we must find the minimal
number among
\verb|x[k+1]..x[n]| that is larger than \w{x[k]}.
After we exchange \verb|x[k]| with the number found, we have to
rearrange \w{x[k+1]..x[n]} to make the permutation as small as
possible. To achieve this goal, we put \w{x[k+1]..x[n]}
in increasing order. (Fortunately, they are already arranged
in {\it decreasing\/} order.)

     Here's how to get the next permutation:
{\samepage
\begin{verbatim}
 {<x[1]...x[n]> <> <n...2,1>}
 k:=n-1;
 {after x[k] terms go in decreasing order: x[k+1]>...>x[n]}
 while x[k] > x[k+1] do begin
 | k:=k-1;
 end;
 {x[k] < x[k+1] > ... >  x[n]}
 t:=k+1;
 {t <=n, all terms x[k+1] > ... > x[t] are bigger than x[k]}
 while (t < n) and (x[t+1] > x[k]) do begin
 | t:=t+1;
 end;
\end{verbatim}
}
\begin{verbatim}
 {x[k+1] > ... > x[t] > x[k] > x[t+1] > ... > x[n]}
 ..exchange x[k] and x[t]
 {x[k+1] > ... > x[n]}
 ..put x[k+1]...x[n] in reversed order
\end{verbatim}

Remark.  This program suffers from the usual problem:
\verb|x[t+1]| is undefined when $\w{t}=\w{n}$.\es

