  \subsection{Subsets}

  \Problem{Generate all subsets of the set
\verb|{1..n}| having \w{k} elements.}%
 \index{subsets!of given cardinality, generation}%

     \Solution. Each subset may be represented by a bit string
of length \w{n} that contains exactly \w{k} \w{1}s. (We'll consider
another representation later.) We generate these bit strings in
alphabetic order. There is a natural way to do this: Generate
all bit strings and select those that contain exactly \w{k}
\w{1}s. However, this solution is considered
as inefficient, because
bit strings with \w{k} \w{1}s form a tiny fraction of all
bit strings of length \w{n}.
In the program below the generation of
each subsequent string requires not more than
{\tt $C\cdot$n} operations (for some constant $C$).

     When is it possible to increase the
\w{s}-th term of a bit string with \w{k} \w{1}s without
changing the
preceding terms? If \verb|x[s]| is changed from
\verb|0| to \verb|1|, we should replace
\verb|1| by \verb|0| somewhere to keep the total number of \w{1}s fixed.
Therefore, it is necessary to have \w{1}s on the right of
\w{x[s]}.

     Conclusion: If we want to find the next bit string with
\w{k} \w{1}s, we need
\w{x[s]} to be the rightmost \w{0} that has some
\w{1}s on the right.
In this case we have \verb|x[s+1]=1| (otherwise, \verb|x[s]| is not the
rightmost one). Therefore, we should look for the maximal
\verb|s| such that
\verb|x[s]=0| and \verb|x[s+1]=1|:
%
%                 ______________________
%              x |________|0|1...1|0...0|
%                          s

\smallskip

\[
\hbox{\tt x}\,
{\setlength{\tabcolsep}{.2em}
\begin{tabular}{|c|c|c|c|}
\hline
\hspace*{5em}\strut& \hbox{%
\vtop to 0pt{\hbox{\tt 0}%
    \vspace*{.9ex}
    \hbox to .5em{\hss$\uparrow$\hss}%
    \hbox to .5em{\hss \tt s\hss}%
    \vss
}}%
& \hbox{\tt 1..1} & \hbox{\tt 0..0}\\
\hline
\end{tabular}}
\]

\vspace*{2ex}
\noindent
The term \verb|x[s+1]| may be followed by several \w{1}s and then
several \w{0}s. After we replace
\verb|x[s]| by \verb|1| the next terms should be chosen to get the
alphabetically first string, that is, \w{0}s should precede \w{1}s.
Here is what we get:
\begin{quote}
  first string:  $\underbrace{\w{0}\ldots\w{0}}_{n-k}
                  \underbrace{\w{1}\ldots\w{1}}_{k}    $
\end{quote}
\begin{quote}
  last string: $\underbrace{\w{1}\ldots\w{1}}_{k}
                  \underbrace{\w{0}\ldots\w{0}}_{n-k}  $
\end{quote}
\begin{quote}
  How to find the next string after \verb|x[1]..x[n]|
  (assuming it exists):
\end{quote}
{\samepage
\begin{verbatim}
     s := n - 1;
     while not ((x[s]=0) and (x[s+1]=1)) do begin
     | s := s - 1;
     end;
\end{verbatim}
}
\begin{Verbatim}
     {x[s] should be changed from 0 to 1}
     num:=0;
     for k := s to n do begin
     | num := num + x[k];
     end;
     {num is the number of 1s among x[s]...x[n], the number
      of 0s is (length - number of 1s), that is, (n-s+1)-num}
     x[s]:=1;
     for k := s+1 to n-num+1 do begin
     | x[k] := 0;
     end;
     {it remains to put num-1 1s at the end}
     for k := n-num+2 to n do begin
     | x[k]:=1;
     end;
\end{Verbatim}

\medskip

    We can also represent a subset by a list of its elements.
To obtain the unique representation we require
that elements should be
listed in increasing order.
Now we come to the following problem:

   \Problem{Generate (in alphabetic order)
all increasing sequences of length \w{k}
consisting of the numbers \verb|1..n|.
(Example: for \verb|n=5|, \verb|k=2| we get
$\langle${\tt 12 13 14 15 23 24 25 34
35 45}$\rangle$.)}%
 \index{monotone sequences!generation}%

    \Solution. The first sequence is
$\langle\w{1}\,\w{2..k}\rangle$;
the last one is  $\langle\w{(n-k+1)..(n-1)}\,\w{n}\rangle$.
When is it possible to increase
the \verb|s|-th element of the sequence?
Answer: If it is less than
\verb|n-k+s|.
After the \w{s}-th element is increased,
all subsequent elements should
form an arithmetic sequence with difference
\w{1}.
Here is the algorithm:
\begin{Verbatim}
        s:=n;
        while not (x[s] < n-k+s) do begin
        | s:=s-1;
        end;
        {s-th element should be increased};
        x[s] := x[s]+1;
        for i := s+1 to n do begin
        | x[i] := x[i-1]+1;
        end;
\end{Verbatim}

  \Problem{Suppose we represent subsets of \w{1..n}
of cardinality \w{k} by {\it decreasing\/} sequences of
length \w{k}.
(Example:
$\langle${\tt 21 31 32 41 42 43 51 52 53 54}$\rangle$.)
How do we generate these sequences in alphabetical order?}
 %
   \Hint{Find the maximal \verb|s| such that
\verb|x[s+1]+1 < x[s]|.
(If it does not exist, let \verb|s=0|.)
Now increase \verb|x[s+1]| by \verb|1|
and let all subsequent elements be as small as possible
(\verb|x[t]=k+1-t| for \w{t>s}).}

   \Problem{Solve the two preceding problems if alphabetic
order is replaced by reversed alphabetic order.\es}

    \Problem{Generate all injective mappings of
the set \verb|{1..k}| into
\verb|{1..n}|
(assume that $\w{k}\le\w{n}$). A mapping is injective if
no two elements of \w{1..k} are mapped to the same
element of \w{1..n}.
Generation of each mapping should require no more that
$C\cdot\w{k}$ operations.}
  %
    \Hint{This problem can be reduced to generation of permutations
and generation of subsets.}

