\subsection{Partitions}

  \Problem{Generate all partitions of a given positive integer
\w{n}, that is, all the representations of \w{n} as a sum of positive
integers. We do not take the order of the summands into account.
(Example: For \verb|n=4|,
partitions are \verb|1+1+1+1|, \verb|2+1+1|, \verb|2+2|, \verb|3+1|
and
\verb|4|.)}%
 \index{partitions!generation}%

     \Solution. Let us agree that (i) the summands are written in
nonincreasing order; and (ii) the partitions are generated
in alphabetic
order. We store a partition in the initial part of an array
\verb|x[1]..x[n]|; the length of the partition is \w{k}, and
the
summands are
\w{x[1]..x[k]}.
At the beginning, $\w{k}=\w{n}$ and all
$\w{x[1]}\ldots\w{x[n]}$ are equal to
$1$. At the end,
$\w{x[1]}=\w{n}$ and $\w{k}=\w{1}$.

     When can we increase \verb|x[s]| leaving all preceding elements
unchanged? This is possible only if
$\w{x[s-1]}>\w{x[s]}$ or if $\w{s}=\w{1}$. Moreover,
\w{x[s]} may not be the last element of the partition (because
an increase in \w{x[s]}
should be compensated by a decrease in the
subsequent elements). After \w{x[s]} is increased, all subsequent
elements should be chosen as small as possible.
\begin{Verbatim}
        s := k - 1;
        while not ((s=1) or (x[s-1] > x[s])) do begin
        | s := s-1;
        end;
        {x[s] should be increased}
        x[s] := x[s] + 1;
        sum := 0;
        for i := s+1 to k do begin
        | sum := sum + x[i];
        end;
        {sum = the sum of terms after x[s]}
        for i := 1 to sum-1 do begin
        | x [s+i] := 1;
        end;
        k := s+sum-1;
\end{Verbatim}

 \Problem{In this problem,
partitions are still represented as nonincreasing sequences,
but now we want to generate them in reversed alphabetic
order (e.g., for \verb|n=4|, we would generate
{\tt 4, 3+1, 2+2,
2+1+1, 1+1+1+1}).}
  \Hint{The rightmost term that may be decreased is the
rightmost term not equal to \w{1}. Find it and decrease it by
\w{1}. All subsequent terms should be taken as large as possible
(equal to the selected term when possible; the last one may be smaller).}

   \Problem{Partitions are represented as nondecreasing sequences;
generate them in alphabetic order.
For example,
when $\w{n}=\w{4}$, we would generate
{\tt 1+1+1+1, 1+1+2, 1+3, 2+2, 4}.}
  \Hint{The last term \w{x[k]} cannot be increased, but
the term \w{x[k-1]}
can. (Of course, the last one should be decreased to maintain
the
sum.) If the sequence is no longer nondecreasing, we
combine two terms into one. If the sequence is still nondecreasing,
then \w{x[k]} should be split into several terms equal to \w{x[k-1]}
(except for the last one, which may be larger).}

 \Problem{Partitions are represented as nondecreasing sequences.
Generate them in reversed alphabetic order.
(For $\w{n}=\w{4}$ we have {\tt 4, 2+2, 1+3, 1+1+2,
1+1+1+1}.)}
  %
    \Hint{The element \verb|x[s]| can be decreased only if
\verb|s=1| or \verb|x[s-1]<x[s]|. If
\verb|x[s]| is not the last term, these conditions are sufficient.
If it is the last one, then we must also have
$\hbox{\tt x[s-1]}\le\lfloor\hbox{\tt x[s]/2}\rfloor$
or \verb|s=1|. (Here $\lfloor\alpha\rfloor$ stands for the integer
part of $\alpha$, that is, the greatest
integer not exceeding $\alpha$.)}

