\subsection{Gray codes and similar problems}

     Sometimes it is useful to generate objects in an order such
that the next object is only a small modification of the preceding one.
In this section, we consider several problems of this type.%
 \index{code!Gray}%
 \index{Gray code}%

  Consider $2^n$ strings of length $n$ containing only $0$s and
$1$'s.
  \Problem{%
Prove that it is possible to list all of them in an order such that
two neighboring strings differ only in one bit.}

  \Solution. Use induction on $n$. Assume that $x_1,\ldots,x_k$ is such
a sequence of $n$-bit
strings (here $k=2^n$; for any $i$, strings $x_i$ and $x_{i+1}$
differ only in one bit). Then the  following  sequence  includes
all $(n+1)$-bit strings and satisfies the desired
condition:
 $$
   0x_1, 0x_2,\ldots, 0x_k, 1x_k, 1x_{k-1},\ldots,1x_1
 $$

In geometric terms, the problem states that we can traverse the
$n$-dimensional Boolean cube visiting each vertex exactly once.
The solutions considers $n$-dimensional Boolean cube as composed
of two $n-1$-dimensional Boolean cubes; we traverse one of them
(using the inductive assumption) and then switch to another one.\es

\medskip

We'll return to this problem later.

  \Problem{Generate all sequences of length \verb|n| composed of
the numbers \verb|1..k| in such an order
that neighboring sequences differ
only in one place, and the numbers at this place differ by
\w{1}.}

     \Solution. Consider a rectangular chess board of width \w{n} and
height \w{k}. Place a piece in each column of the chess board.
The position is represented  by a
sequence of \w{n} integers (each between \w{1} and \w{k}); the
\w{i}-th
number represents a position of the piece in the
\w{i}-th column.
At each piece we draw a small arrow that points up or down.
Initially, all the pieces are in the first row and all the arrows
point up.
We move
pieces according to the following rule: Find the rightmost piece
that can be moved in the direction of the arrow on it, and move it.
At the same time all the pieces on the right (they cannot move
in the direction of their arrows) are turned over.

     It is evident that at each step only one piece is moving, therefore
only one term in the corresponding sequence is changed by \w{1}. Let us
prove by induction on \w{n} that all sequences of length \w{n} composed
of the
numbers \w{1..k} will appear. The case \w{n=1} is evident, so
assume that
$\w{n}>\w{1}$. Divide all moves into two categories.
The first category is formed by moves where the last (rightmost) piece
is moving. The second category is formed by moves where the moving
piece is not the last one. In this case the rightmost piece is near the
border and is turned over. Therefore, each move of the second category is
followed by \verb|k-1| moves of the first category; during this period
the rightmost piece visits all the cells. Let us forget now about the
rightmost piece. Then the first \w{n-1} pieces are moving according to
the prescribed rules. Therefore, by the
induction assumption, all sequences
of length \verb|n-1| appear exactly once. The movements of the last piece
make \w{k} sequences of length \w{n} out of each sequence of length
\w{n-1}.

    The program keeps an array \verb|x[1]..x[n]| (positions of pieces)
and an array  \verb|d[1]..d[n]| composed of numbers
\verb|+1| and  \verb|-1| (\verb|+1| denotes up-arrow;
\verb|-1| denotes down-arrow).

\smallskip

Initial state: $\w{x[1]}=\cdots\w{x[n]}=\w{1}$;
               $\w{d[1]}=\cdots=\w{d[n]}=\w{1}$.

\smallskip

The following algorithm produces the next position according to the
description above. At the same time, it checks whether the next position
exists; the answer is stored in a Boolean variable \w{p}.

\begin{verbatim}
  {if possible, make a move and let p := true;
   otherwise, p := false }
  i := n;
\end{verbatim}
\clearpage
\begin{verbatim}
  while (i > 1) and
  | (((d[i]=1) and (x[i]=n)) or ((d[i]=-1) and (x[i]=1)))
  |   do begin
  | i:=i-1;
  end;
  if (d[i]=1 and x[i]=n) or (d[i]=-1 and x[i]=1)
  |    then begin
  | p:=false;
  end else begin
  | p:=true;
  | x[i] := x[i] + d[i];
  | for j := i+1 to n do begin
  | | d[j] := - d[j];
  | end;
  end;
\end{verbatim}

     Remark. For the case $\w{k}=2$ there is another solution that
uses the binary system. (It is this solution that is
usually associated with the name
``Gray code''.)

     Let us write down all the numbers $0,\ldots,2^n-1$ in binary
notation. For example, for $n=3$ we have:
\[
         000\quad 001\quad 010\quad 011
         \quad 100\quad 101\quad 110\quad 111
\]
Each number is transformed according to the following rule:
each digit (except the first one) is replaced by its sum (modulo $2$)
with the
preceding (untransformed) digit.
In other words, the number with binary digits
     $a_1,a_2,\ldots,a_n$  is transformed into the number with binary
digits
     $a_1, a_1 + a_2, a_2 + a_3,\ldots,a_{n-1} + a_n$
(addition modulo 2). For $n=3$, we get the following list:
\[
            000\quad 001\quad 011\quad 010\quad 110
      \quad 111\quad 101\quad 100
\]

    It is easy to check that the transformation described (which
can
be applied to any sequence of $n$ binary digits, giving another
sequence of the same length) is invertible. Therefore, the list
obtained contains all sequences of length $n$.

    On the other hand, adding $1$ to a number in binary notation
means replacement of the
suffix
$011...1$ by $100...0$. This change leads to a change of exactly
one digit after the transformation is applied.\es

\bigskip

    {\bf An application of Gray codes}. Assume that some mechanical
device has a rotating drum and we wish to get information about the
position of this drum. If we make  half the drum white,
the remaining half black, and use a light sensor, we can measure
the position of the drum up to $180^{\circ}$.%
 \index{angle detector}%

\bigskip
\clearpage
     Drum cover:

%                    0       1
%            -> |_|_|_|_|*|*|*|*| <- (       ).

\begin{center}
\leavevmode
    \phantom{$\rightarrow$\qquad}%
    \hbox to 4em{\hss0\hss}\hbox to4em{\hss1\hss}%
    \phantom{\qquad$\leftarrow$\quad glue together}\\
\end{center}
\begin{center}
\leavevmode \raisebox{.3ex}{$\rightarrow$\qquad}%
     \boxit{\hspace*{4em}\rule{4em}{1em}}\qquad
     \raisebox{.3ex}{$\leftarrow$\quad glue together}
\end{center}

     If we make another track with black and white parts, and use a
second light sensor, we can measure the position angle up to
$90^\circ$:

%                  0   0   1   1
%                  0   1   0   1
%                _ _ _ _
%               |_|_|_|_|*|*|*|*|
%               |_|_|*|*|_|_|*|*|
%
\begin{center}
\nointerlineskip
\leavevmode
    \hbox to 2em{\hss0\hss}%
    \hbox to2em{\hss0\hss}%
    \hbox to2em{\hss1\hss}%
    \hbox to2em{\hss1\hss} \\[.5ex]
\leavevmode
    \hbox to 2em{\hss0\hss}%
    \hbox to2em{\hss1\hss}%
    \hbox to2em{\hss0\hss}%
    \hbox to2em{\hss1\hss} \\[1.5ex]
\leavevmode \boxit{\hspace*{4em}\rule{4em}{1em}}\\
\leavevmode \boxit{\hspace*{2em}\rule{2em}{1em}%
                   \hspace*{2em}\rule{2em}{1em}}
\end{center}

\noindent
With a third track,
%                0 0 0 0 1 1 1 1
%                0 0 1 1 0 0 1 1
%                0 1 0 1 0 1 0 1
%                _ _ _ _
%               |_|_|_|_|*|*|*|*|
%               |_|_|*|*|_|_|*|*|
%               |_|*|_|*|_|*|_|*|
\begin{center}
\nointerlineskip
\leavevmode
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss1\hss} \\[.5ex]
\leavevmode
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss1\hss} \\[.5ex]
\leavevmode
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss1\hss} \\[1.5ex]
\leavevmode \boxit{\hspace*{4em}\rule{4em}{1em}}\\
\leavevmode \boxit{\hspace*{2em}\rule{2em}{1em}%
                   \hspace*{2em}\rule{2em}{1em}}\\
\leavevmode \boxit{\hspace*{1em}\rule{1em}{1em}%
                   \hspace*{1em}\rule{1em}{1em}%
                   \hspace*{1em}\rule{1em}{1em}%
                   \hspace*{1em}\rule{1em}{1em}}
\end{center}

\noindent
the precision becomes $45^\circ$, etc. However, there is a problem
with this scheme. When two light sensors change their state
from black to white, this change may not
happen at exactly the same time,
and for a while the data are senseless.

We can use Gray codes to overcome this difficulty:
we arrange the black and white sectors in such a way that
only one track changes color each time. (This is also true
for the last change after a complete rotation is performed.)

%                0 0 0 0 1 1 1 1
%                0 0 1 1 1 1 0 0
%                0 1 1 0 0 1 1 0
%                _ _ _ _
%               |_|_|_|_|*|*|*|*|
%               |_|_|*|*|*|*|_|_|
%               |_|*|*|_|_|*|*|_|
\begin{center}
\nointerlineskip
\leavevmode
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss1\hss} \\[.5ex]
\leavevmode
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss0\hss} \\[.5ex]
\leavevmode
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss0\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss1\hss}%
    \hbox to1em{\hss0\hss} \\[1.5ex]
\leavevmode \boxit{\hspace*{4em}\rule{4em}{1em}}\\
\leavevmode \boxit{\hspace*{2em}\rule{4em}{1em}%
                   \hspace*{2em}}\\
\leavevmode \boxit{\hspace*{1em}\rule{2em}{1em}%
                   \hspace*{2em}%
                   \rule{2em}{1em}%
                   \hspace*{1em}}
\end{center}

\noindent
     The above
formula allows us
to convert the sensor
data into the corresponding
rotation angle easily.

    \Problem{Generate all permutations of the
numbers \verb|1..n| in such a way
that each permutation
is obtained from the preceding one by an exchange
(transposition)
of two adjacent numbers. For example, for \verb|n=3|, one of the possible
answers is}
\begin{center}
\tt
3.2 1 $\to$ 2 3.1 $\to$ 2.1 3 $\to$ 1 2.3 $\to$
1.3 2 $\to$ 3 1 2
\end{center}
(the dots indicate which numbers are exchanged at each step)%
 \index{permutations!generation}%

     \Solution. Put the set of all permutations into
one-to-one correspondence
with another set. This latter
set contains all
sequences \verb|y[1]..y[n]| of nonnegative integers such that
$\w{y[1]}\le \w{0}$, \dots,
$\w{y[n]}\le \w{n-1}$.
It has the same cardinality as the set of all permutations. The
one-to-one correspondence is established
as follows:
Each permutation corresponds to the sequence
\verb|y[1]..y[n]|, where \verb|y[i]| is the number of \w{j}'s
such that both (a) $\w{j}<\w{i}$ and (b) \w{j} is located to the left of
\w{i} in this permutation. Why is it an one-to-one correspondence?
Any permutation of \w{1..n} can be obtained from a permutation of
\w{1..n-1} by inserting \w{n} into one of the \w{n} places (before the
first term, between the first and the second terms,\ldots, after the
last term). What means this insertion for the corresponding sequence
of integers? A number that ranges from \w{0} to \w{n-1} is appended to
the end while the other terms remain unchanged.

This one-to-one correspondence can be explained by the following
metaphor. Consider \w{n} cards with numbers \w{1..n} written on
the cards, and a growing pile made of the cards. Initially the
pile has only one card with number \w{1} written on it. At the
next step we add the card with number \w{2}. There are two
possible positions for that card (either before the first card
or after it). Then we add the card with number \w{3} on it;
there are three posiible positions, etc. After we add the last
card (there are \w{n} possible positions), we get a permutation
of the numbers \w{1..n}. This permutation is determined by
positions chosen at steps \w{1..n}; if we denote by \w{y[i]} the
number of cards before the inserted card at step \w{i},
we get the one-to-one correspondence defined above.

We make one more remark about this correspondence. Assume that we
increase or decrease \w{y[i]} by \w{1} for some \w{i} (leaving the
other \w{y[j]}'s unchanged). Assume also that all subsequent
\w{y[j]}'s have maximal or minimal values. In this case
two adjacent numbers in our permutation are exchanged. Namely, an
increase in \w{y[i]} means that \w{i} is exchanged with its
right neighbor, while a
decrease means that \w{i} is exchanged with
its left neighbor.

    Now recall how we generated all sequences of numbers \w{1..k}
in such a way that each sequence differs from the preceding sequence
in
one and only one
place by using $\w{n}\times\w{k}$ rectangle.
Now replace it
by a board that resembles a staircase (the
\w{i}-th column is a rectangle
of width \w{1} and height \w{i}). Moving pieces according to the rules
described above (using arrows on pieces), we traverse all the
sequences, and the property mentioned above (that the
\w{i}-th term changes
only if all subsequent terms are maximal or minimal) holds.

   To implement this scheme we need to modify the permutation
according to the changes on the board. An obvious approach is to
search for a given number \w{i} at each step. We can save
ourselves
some
work if we keep (in addition to the permutation itself)
the function
\begin{center}
       {\tt i} $\mapsto$ position of \w{i} in the permutation
\end{center}
that is, the inverse mapping,%
 \index{permutation!inverse}
and update both the permutation and its inverse.
Here is the program:
{\samepage
\begin{verbatim}
 program test;
 | const n = ...;
 | var
 |   x: array [1..n] of 1..n; {permutation}
 |   inv_x: array [1..n] of 1..n; {inverse permutation}
 |   y: array [1..n] of integer; {y[i] < i}
 |   d: array [1..n] of -1..1; {arrows}
 |   b: Boolean;
 |
 | procedure print_x;
 | | var i: integer;
 | begin
 | | for i := 1 to n do begin
 | | | write (x[i], ' ');
 | | end;
 | | writeln;
 | end;
 |
 | procedure set_first; {first: y[i]=0 for all i}
 | | var i : integer;
 | begin
 | | for i := 1 to n do begin
 | | | x[i] := n + 1 - i;
 | | | inv_x[i] := n + 1 - i;
 | | | y[i] := 0;
 | | | d[i] := 1;
 | | end;
 | end;
 |
 | procedure move (var done : Boolean);
 | | var i, j, pos1, pos2, val1, val2, tmp : integer;
 | begin
 | | i := n;
 | | while (i > 1) and (((d[i]=1) and (y[i]=i-1)) or
 | | |          ((d[i]=-1) and (y[i]=0))) do begin
 | | | i := i-1;
 | | end;
 | | done := (i > 1);
 | | {simplification: the first term cannot be changed}
 | | if done then begin
 | | | y[i] := y[i] + d[i];
\end{verbatim}
}
\begin{Verbatim}
 | | | for j := i+1 to n do begin
 | | | | d[j] := -d[j];
 | | | end;
 | | | pos1 := inv_x[i];
 | | | val1 := i;
 | | | pos2 := pos1 + d[i];
 | | | val2 := x[pos2];
 | | | {pos1, pos2 are positions of elements to be
 | | |   exchanged; val1, val2 are its values; val2 < val1}
 | | | tmp := x[pos1];
 | | | x[pos1] := x[pos2];
 | | | x[pos2] := tmp;
 | | | tmp := inv_x[val1];
 | | | inv_x[val1] := inv_x[val2];
 | | | inv_x[val2] := tmp;
 | | end;
 | end;
 |
 begin
 | set_first;
 | print_x;
 | b := true;
 | {all permutations up to the current one (including it)
 |  are printed;
 |  if b is false, the current one is the last one}
 | while b do begin
 | | move (b);
 | | if b then print_x;
 | end;
 end.
\end{Verbatim}

