\subsection{Some remarks}

     Let us review the approach we've been using. We
introduce some order on the objects to be generated
and write a procedure that
obtains the next object
(in this order).
In the Gray code problems, we were forced to maintain
some additional
information (directions of arrows). Finally, when generating
permutations in such a way that only two numbers are exchanged at a time,
we establish a one-to-one correspondence
between the set to be generated and some other (presumable simpler)
set. There are some cases where this trick is useful. In this section,
we
consider several problems of this type
connected with the so-called Catalan numbers.
 \index{Catalan numbers}%

   \Problem{Generate all sequences of length \verb|2n|,
composed of \w{1}s and \w{-1}s satisfying the
following conditions: (a) the sum of all terms is \w{0};
(b) the sum of any prefix is nonnegative, that is, the
number of \w{-1}s does not exceed the number of \w{1}s.
(The number of such sequences is called the
{\it Catalan number\/}; see
the formula for Catalan numbers on
p.~\pageref{catalan}, problem 2.7.3.)
}

     \Solution. Represent \w{1} by a vector \verb|(1,1)| and
represent
\w{-1} by \verb|(1,-1)|. In terms of vectors,
we are looking for all
paths from \verb|(0,0)| to \verb|(2n,0)| that never go below the
$x$-axis.

     Let us generate the sequences in alphabetic order
(assuming that \verb|-1| precedes \verb|1|). The first sequence
is the ``zig-zag''
\begin{center}
\tt     1, -1, 1, -1, \dots
\end{center}
The last sequence will be the sequence
\begin{center}
\tt     1, 1, 1,$\,$\dots, 1, -1, -1,$\,$\dots, -1.
\end{center}

But how do we generate
the next sequence? It should coincide with the
current sequence
up to some point where they differ and \verb|-1| is
replaced by \w{1}. This place should be as close to the end as
possible. But there is a restriction; \verb|-1| may be replaced
by \verb|1| only if there is \w{1} on the right of it (which can
be replaced by \w{-1}). After we replace \w{-1} by \w{1}, we
are faced with
the following problem: A prefix of the sequence is
fixed; find the minimal sequence with that prefix. The solution:
extend the given prefix step by step;
at each step append \w{-1} if possible (the sum must be
nonnegative); otherwise, append \w{1}. Here is the resulting
program:
\begin{Verbatim}
    ...
    type array2n = array [1..2n] of integer;
    ...
    procedure get_next (var a: array2n; var last: Boolean);
    | {a is replaced by the next sequence if it exists
    |  (and last:=false), otherwise last:=true}
    | var k, i, sum: integer;
    begin
    | k:=2*n;
    | {invariant: a[k+1..2n] contains only -1s}
    | while a[k] = -1 do begin k:=k-1; end;
    | {k is maximal among all k such that a[k]=1}
    | while (k>0) and (a[k] = 1) do begin k:=k-1; end;
    | {a[k] is the rightmost -1 preceding some 1;
    |  k=0 if there is no -1 on the left of 1}
    | if k = 0 then begin
    | | last := true;
    | end else begin
    | | last := false;
    | | i:=0; sum:=0;
    | | {sum = a[1]+...+a[i]}
    | | while i<>k do begin
    | | | i:=i+1; sum:= sum+a[i];
    | | end;
    | | {sum = a[1]+...+a[k], a[k]=-1}
    | |  a[k]:= 1; sum:= sum+2;
    | | {all a[1]..a[k] have their final values,
    | |   sum=a[1]+...+a[k]}
    | | while k <> 2*n do begin
    | | | k:=k+1;
    | | | if sum > 0 then begin
    | | | | a[k]:=-1
    | | | end else begin
    | | | | a[k]:=1;
    | | | end;
    | | | sum:= sum+a[k];
    | | end;
    | | {k=2n, sum=a[1]+...a[2n]=0}
    | end;
    end;
\end{Verbatim}

  \Problem{Find all possible ways to compute the product of \w{n}
factors. (The order of the
factors remains unchanged.) Each multiplication should be
indicated by parentheses. For example, for $\w{n}=\w{4}$, the following
five expressions should be generated:
\begin{center} \tt
((ab)c)d, (a(bc))d,
(ab)(cd), a((bc)d), a(b(cd)).
\end{center}}%
 \index{product!non-associative}%
 \index{parentheses}%
    \Hint{Each order of operations corresponds to a sequence
of commands of the stack calculator described on
p.~\pageref{stackcalc}.
}

  \Problem{There are \verb|2n| points
on a circle numbered (along the circle)
by the numbers
$\w{1..2n}$. Generate all possible
ways to draw \w{n}
non-intersecting
segments having those \w{2n} points as endpoints.\es}%
 \index{chords of a circle}%

  \Problem{Generate all ways to cut a convex
polygon with
\verb|n| vertices into
triangles using \verb|n-2| diagonals.\es}%
 \index{triangulation of a polygon}%
 \index{polygon, triangulation}%

  (We will discuss polygon triangulations in the section~\ref{delrec} on
dynamic programming on p.~\pageref{triangulation}.)

