\subsection{Counting}

In this section we considered several methods that may be
used to generate all the elements of a given finite set.
One more approach
will be considered below (under the name of ``backtracking'')
in section~\ref{traversal}.
But
sometimes it is much easier to count all the objects with
some property than it is
to generate them. The classic example is
$n \choose k$, which is the number of $k$-element subsets of an
$n$-element set.%
 \index{binomial coefficient}%
 \index{Pascal triangle}
These numbers form the ``Pascal triangle'' and
can be computed using the identities
$$
\begin{array}{rclcl}
   {\textstyle
    {n\choose 0}}&=&{\textstyle
        {n\choose n}}\;=\;1  &\qquad& (n \ge 1)\\[.5ex]
   {\textstyle
    {n\choose k}}&=&{\textstyle
{n-1\choose k-1}}+ {\textstyle {n-1 \choose k}}
&\qquad&(n > 1,\ 0 < k < n)
\end{array}
$$
or the formula
\[\textstyle{
  {n\choose k}}=\frac{n!}{k!\cdot (n-k)!}.
\]
(The first method is more efficient when many values of
${n \choose k}$ for different $n$ and $k$ are needed.)

    Let us give some other examples.

     \Problem{(Number of partitions)
Let $P(n)$ be the number of representations of a nonnegative integer
$n$ as a sum of positive integer summands (order is insignificant;
that is,
the representations $1+2$ and $2+1$ are identical).
We assume that $P(0) = 1$ (the only representation
has no summands at
all). Write a program that finds $P(n)$ for a given
$n$.}%
 \index{partitions!number of}%
 \index{number!of partitions}%

     \Solution. One can prove the following (nontrivial) formula
for $P(n)$:
\[
 P(n) = P(n-1)+P(n-2)-P(n-5)-P(n-7)+P(n-12)+P(n-15) +\cdots
\]
(terms are grouped in pairs, the signs before the pairs alternate,
arguments in $q$-th pair are
$n-(3q^2-q)/2$ and $n-(3q^2+q)/2$). We
assume $P(k)=0$ for $k\le 0$, so the sum is finite.

     Even if we did not know this formula, there is a way to
compute $P(n)$ that is much more efficient than counting all the
partitions one-by-one.

     By $R(n,k)$ (defined for $n\ge0$, $k\ge0$) we denote the
number of representations of
$n$ as a sum of positive integers not exceeding $k$.
(Let $R(0,k)$ be equal to $1$ for all $k\ge0$.)
Evidently, $P(n)=R(n,n)$. All the representations of $n$
are classified according to the maximal summand
(which is denoted by $i$ in the sequel).
The number $R(n,k)$ is the sum over all
$i$ in $\{1,\ldots,k\}$ of the number of partitions with elements
not exceeding $k$ and maximal element $i$.
The partitions of $n$ into a sum where all terms do not exceed $k$
and maximal term is equal to $i$ are in one-to-one correspondence
with the partitions of $n-i$ into terms not exceeding
$i$ (assuming that $i\le k$). Therefore,
\begin{eqnarray*}
    R(n,k)&=&\sum\limits_{i=1}^{k} R (n-i,i) \hbox{\ \ \ for $k\le n$};\\
    R(n,k)&=& R (n,n)  \hbox{\ \ \ for $k > n$}.
\end{eqnarray*}
These equations allows us  to construct
a table of values of the function $R$.\es

  \Problem{(Lucky numbers)
A sequence of $2n$ digits (each digit is in the $0,\ldots,9$ range)
is called ``lucky'' if the sum of the first $n$ digits is equal
to the sum of the last
$n$ digits. Find the number of all lucky sequences of a
given length.}

     \Solution.
Let us generalize the problem and find the number $T(n,k)$ of
sequences of length $2n$ where the difference between the sum of
first $n$ digits and the sum of the last $n$ digits is equal to $k$
(where $-9n\le k \le 9n$).

    We divide all these sequences
into classes according to the difference
between the first and last digit. If this difference is equal to
$t$, the difference between the remaining sums of $n-1$ digits is
$k-t$. Recall that there are $10-|t|$ pairs of decimal digits with
difference $t$, so we get the formula:
\[
   T(n,k) = \sum\limits_{t=-9}^{9}  (10-|t|) T(n-1, k-t).
\]
(Some terms may be missing if $k-t$ is too large.)\es

\medskip

In some cases, the answer may be given by an explicit formula.
For example, this is the case for Catalan numbers%
\index{Catalan numbers}

\label{catalan}
\Problem{Prove that the Catalan number, that is, the number of sequences
of length $2n$ composed of $n$ ones and $n$ minus ones such that
each initial segment has a nonnegative sum,
is equal to ${2n\choose n}/(n+1)$.}

\Hint{The Catalan number is the number of polygonal
paths going from
$(0,0)$ to $(2n,0)$ formed by vectors $(1,1)$ and
$(1,-1)$ that do not
intersect the half-plane $y<0$. Therefore, this number is
the difference
between the number of all polygonal paths of the
type described
(which is $2n\choose n$)
and the number of paths that intersect the half-plane $y<0$.
All paths of the type described
that intersect the half-plane $y<0$ intersect
the line $y=-1$. If we reflect the part of the polygonal
path that is on the right of the rightmost intersection point,
we get a one-to-one correspondence between the
polygonal paths in question and
all polygonal paths from
$(0,0)$ to $(2n,-2)$. It remains to check that
${2n\choose n}-{2n\choose n+1}={2n\choose n}/(n+1)$.}

