\section{Tree traversal (backtracking)}\label{traversal}

\subsection[Queens not attacking each other: position
tree traversal]{Queens not attacking
each other: position tree traversal}

     In the preceding section we considered several problems
of a similar type: ``generate all the elements of some set $A$''.
The scheme used to solve these problems was the following one:
A linear ordering on $A$ was imposed and a procedure to generate
the next element of $A$ (according to that order) was described.
Sometimes this scheme cannot be applied directly.
In this chapter,
we consider another useful approach that allows us to generate
all elements of some set. It is called ``backtracking'' or
``tree traversal''.%
\index{backtracking}\index{tree!traversal}

This approach is fairly general;
however, we prefer to start with a specific
example.

    \Problem{%
     Generate all the positions of $n$ chess queens on an
$n\times n$ board such that the queens are not attacking each other.
}%
\index{queens problem}%

     \Solution. Evidently, each of $n$ rows should
contain exactly one queen.
By $k$-position we mean a position where $k$ queens occupy
$k$ rows (starting from the bottom of the chessboard) containing
exactly one queen each. We do not
impose any restrictions as of yet and
allow positions where some queens are attacking other queens.
Arrange all positions into a tree, whose root is the empty
position ($k=0$). Each $k$-position has exactly $n$ descendants,
which have an additional queen in the $(k+1)$-th row (in one of the
columns $1,\ldots,n$). These $n$ descendants are ordered from
left to right according to the position of the last
(i.e., the uppermost) queen.\index{positions tree}%
\index{tree!of positions}

\setlength{\unitlength}{14pt}
\newlength{\unitlencor}
\setlength{\unitlencor}{13.2pt}%\unitlength-2*rulewidth
\newcommand{\rw}{\kern-0.4pt}% rulewidth
\def\bulbox{\boxit{%
               \hbox to \unitlencor{%
                 \hss
                 \vbox to \unitlencor{%
                   \vss
                     \hbox{$\bullet$}%
                   \vss
                 }%
                 \hss
               }%
            }%
}%
\def\empbox{\boxit{%
               \hbox to \unitlencor{%
                 \hss
                 \vbox to \unitlencor{%
                   \vss
                 }%
                 \hss
               }%
            }%
}%
\[
\begin{picture}(16,9)
\put(7,3){\vector(-3,1){3}}
\put(9,3){\vector(3,1){3}}
\put(3,6){\vector(-1,1){1}}
\put(5,6){\vector(1,1){1}}
\put(11,6){\vector(-1,1){1}}
\put(13,6){\vector(1,1){1}}

\put(1,7){%
  \vbox{\offinterlineskip
                 \hbox{\bulbox\rw\empbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox}%
  }%
}
\put(5,7){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\bulbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox}%
  }%
}
\put(9,7){%
  \vbox{\offinterlineskip
                 \hbox{\bulbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\bulbox}%
  }%
}
\put(13,7){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\bulbox}%
                 \rw%
                 \hbox{\empbox\rw\bulbox}%
  }%
}
\put(3,4){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox}%
  }%
}
\put(11,4){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\bulbox}%
  }%
}
\put(7,1){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox}%
  }%
}
\end{picture}
\]
We are to select (among the vertices of this tree) those
$n$-positions where queens are not attacking each other. To find them,
our program will traverse the positions tree. To avoid
unnecessary work, we make use of the following fact: If some tree
vertex
corresponds to a position where queens are attacking each other, all
descendants of this vertex have the same property and therefore
may be ignored safely. Therefore, this part of the position tree
may be discarded.
\clearpage
    Let us give some relevant definitions. A $k$-position is called
``admissible'' if {\em after the $k$-th queen is removed},
the remaining queens are not attacking each other. Our program will
consider only admissible positions.

\setlength{\unitlength}{8pt}
\setlength{\unitlencor}{7.21pt}%\unitlength-2*rulewidth
\def\bulbox{\boxit{%
               \hbox to \unitlencor{%
                 \hss
                 \vbox to \unitlencor{%
                   \vss
                     \hbox{$\bullet$}%
                   \vss
                 }%
                 \hss
               }%
            }%
}%
\def\empbox{\boxit{%
               \hbox to \unitlencor{%
                 \hss
                 \vbox to \unitlencor{%
                   \vss
                 }%
                 \hss
               }%
            }%
}%
\[
\begin{picture}(37,23)
\put(17,4){\vector(-3,1){9}}
\put(20,4){\vector(3,1){9}}
\put(18.5,4){\vector(0,1){3}}

\put(5,10){\vector(-1,3){1}}
\put(8,10){\vector(1,3){1}}
\put(6.5,10){\vector(0,1){3}}

\put(17,10){\vector(-1,3){1}}
\put(20,10){\vector(1,3){1}}
\put(18.5,10){\vector(0,1){3}}

\put(29,10){\vector(-1,3){1}}
\put(32,10){\vector(1,3){1}}
\put(30.5,10){\vector(0,1){3}}

\put(9,16){\vector(-1,3){1}}
\put(12,16){\vector(1,3){1}}
\put(10.5,16){\vector(0,1){3}}

\put(25,16){\vector(-1,3){1}}
\put(28,16){\vector(1,3){1}}
\put(26.5,16){\vector(0,1){3}}

\put(5,19){%
  \vbox{\offinterlineskip
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
       }%
  }%
\put(9,19){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\bulbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
       }%
  }%
\put(13,19){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
       }%
  }%
\put(21,19){%
  \vbox{\offinterlineskip
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
       }%
  }%
\put(25,19){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\bulbox\rw\empbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
       }%
  }%
\put(29,19){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
       }%
  }%
\put(1,13){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
       }%
  }%
\put(5,13){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\bulbox\rw\empbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
       }%
  }%
\put(9,13){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
       }%
  }%
\put(13,13){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\bulbox\rw\empbox}%
       }%
  }%
\put(17,13){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\bulbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\bulbox\rw\empbox}%
       }%
  }%
\put(21,13){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
                 \rw%
                 \hbox{\empbox\rw\bulbox\rw\empbox}%
       }%
  }%
\put(25,13){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
       }%
  }%
\put(29,13){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\bulbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
       }%
  }%
\put(33,13){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
       }%
  }%
\put(5,7){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\bulbox\rw\empbox\rw\empbox}%
       }%
  }%
\put(17,7){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\bulbox\rw\empbox}%
       }%
  }%
\put(29,7){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\bulbox}%
       }%
  }%
\put(17,1){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\empbox}%
       }%
  }%
\put(17,1){%
  \vbox{\offinterlineskip
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\empbox}%
                 \rw%
                 \hbox{\empbox\rw\empbox\rw\empbox}%
       }%
  }%
\end{picture}
\]
\[ \r{The tree of admissible positions for $n=3$} \]

\bigskip

