\Problem{%
     Prove that the program shown above
     terminates for any finite
tree.}

     \Solution. The procedure  \verb|go_up_and_process| terminates (since
the height of the
robot cannot increase indefinitely). Assume that the program
as a whole does not terminate. Leaves are never processed twice and
the number of leaves is finite. Therefore, there is a moment after
which leaves are not processed. This is possible only if the
robot
goes down at each step, but this is a contradiction. (The estimate for
the number of operations will be given below.)\es

\Problem{%
    Prove that the following program also processes all the leaves of
    a tree (one time each):
}
\begin{verbatim}
 var state: (LP, LAP);
 state := LP;
 while is_down or (state <> LAP) do begin
 | if (state = LP) and is_up then begin
 | | up_left;
 | end else if (state = LP) and not is_up then begin
 | | process; state := LAP;
 | end else if (state = LAP) and is_right then begin
 | |  right; state := LP;
 | end else begin {state = LAP, not is_right, is_down}
 | |  down;
 | end;
 end;
\end{verbatim}

     \Solution. The invariant relation: The value stored
in the variable \w{state} is correct, that is,
\begin{eqnarray*}
    \w{state} = \w{LP }\;&\Rightarrow&\; \r{LP is true}\\
    \w{state} = \w{LAP}\;&\Rightarrow&\; \r{LAP is true}
\end{eqnarray*}

The proof of termination: the change from \w{LP} to \w{LAP}
is possible only when a vertex is processed. Therefore,
if the
program does not terminate, the variable \w{state} achieves its
final value and does not change further, which is
impossible.\nobreak\es

\Problem{%
     Write a program that traverses the tree and processes all
vertices (not only leaves).}

    \Solution. Let $x$ be a vertex. Then all vertices of the tree
can be divided into four categories. Indeed, let $y$ be some
other
vertex. Consider the path from the root to $y$. Four cases are
possible:

\medskip

    (a) this path is a prefix of the path from the root to $x$
($y$ is {\it below}~$x$);

    (b) this path turns to the left before reaching $x$ ($y$ is
{\it on the left\/} of~$x$);

    (c) this path goes through $x$ ($y$ is {\it above}~$x$);

    (d) this path turns to the right before reaching $x$
($y$ is {\it on the right\/} of~$x$).

\medskip

\noindent
In particular, the vertex $x$ belongs to class (c). Now the
following conditions are used in our program:

\smallskip

   (ULP) all vertices under the current position and
on the left of it are processed;

\smallskip

   (ULAP) all vertices under the current position, on the
left of it, and above it are processed.

\smallskip

\noindent
Here is the program:
\begin{verbatim}
  procedure go_up_and_process;
  | {before: (ULP), after: (ULAP)}
  begin
  | {invariant: ULP}
  | while is_up do begin
  | | process;
  | | up_left;
  | end
  | {ULP, the current position is a leaf}
  | process;
  | {ULAP}
  end;
\end{verbatim}

The main algorithm:

\begin{Verbatim}
  before: robot is in the root, no vertices are processed
  after: robot is in the root, all vertices are processed

  {ULP}
  go_up_and_process;
  {invariant: ULAP}
  while is_down do begin
  | if is_right then begin {ULAP, is_right}
  | | right;
  | | {ULP}
  | | go_up_and_process;
  | end else begin
  | | {ULP, not is_right, is_down}
  | | down;
  | end;
  end;
  {ULAP, robot is in the root => all vertices are processed}
\end{Verbatim}

\Problem{%
     The program given in the solution of the preceding problem
processes any vertex before its descendants.
Modify the program in
such a way that any vertex will be processed twice, once
{\it before\/} and once {\it after\/} its descendants.
(The leaves should be processed once.)}

    \Solution. In the program below,
by ``Under-Left-Processed'' (ULP) we mean  ``all the vertices under
the current position of the robot are processed once; all the
vertices
on the left are processed completely (that is, leaves are processed
once, all other vertices are processed twice: once before and
once after their descendants)''.
By ``Under-Left-Above-Processed'' (ULAP) we mean
``all the vertices under
the current position of the robot are processed once; all vertices
on the left of and above the current position
are processed completely''.

Here is the auxiliary procedure:
\begin{verbatim}
  procedure go_up_and_process;
  | {before: (ULP), after: (ULAP)}
  begin
  | {invariant: ULP}
  | while is_up do begin
  | | process;
  | | up_left;
  | end
  | {ULP, the current position is a leaf}
  | process;
  | {ULAP}
  end;
\end{verbatim}

The main program:
\begin{Verbatim}
  before: robot is in the root, no vertices are processed
  after: robot is in the root, all vertices are processed

  {ULP}
  go_up_and_process;
  {invariant: ULAP}
  while is_down do begin
  | if is_right then begin {ULAP, is_right}
  | | right;
  | | {ULP}
  | | go_up_and_process;
  | end else begin
  | | {ULP, not is_right, is_down}
  | | down;
  | | process;
  | end;
  end;
  {ULAP, robot is in the root =>
     all vertices are processed completely}
\end{Verbatim}

\Problem{%
     Prove that the number of operations in this program is proportional
to the number of vertices. (Therefore, the same is true for
the programs
given above that differ from the last one only because some
\w{process} commands have been omitted.)}

    \Hint{Roughly speaking, each second operation is processing
some vertex, and any vertex is processed at most twice.}

