\bigskip

Let us return to the queens problem. In this problem,
we use
only the first and simplest of our tree traversal programs, which
processes each leaf once.%
 \index{tree!of positions, implementation}

We implement all the operations for the case of the positions tree.
Each position is represented by a variable
\w{k:0..n} (the number of queens) and an array
\w{c:~array[1..n]} \w{of} \w{1..n}. Here \w{c[i]} is the horizontal
coordinate of the
\w{i}-th queen (whose vertical coordinate is \w{i}).
If
$\w{i}>\w{k}$, the value of
\w{c[i]} is insignificant.
Only the admissible positions are included in the tree.
(According to our
definition, a position is admissible if after the uppermost
queen is removed, no queens are attacking each other.)

Now we are ready to present the program that solves queens'
problem:
{\samepage
\begin{verbatim}
  program queens;
  | const n = ...;
  | var
  |   k: 0..n;
  |   c: array [1..n] of 1..n;
  |
  | procedure begin_work; {initialize}
  | begin
  | | k := 0;
  | end;
  |
  | function danger: Boolean;
  | | {the uppermost queen is under attack}
  | | var b: Boolean; i: integer;
  | begin
  | | if k <= 1 then begin
  | | | danger := false;
  | | end else begin
  | | | b := false; i := 1;
  | | | {b <=> the uppermost queen is under attack of
  | | |   some queen with y-coordinate < i}
  | | | while i <> k do begin
  | | | | b := b or (c[i]=c[k]) {vertical}
  | | | |     or (abs(c[i]-c[k])=abs(i-k)); {diagonal}
  | | | | i := i+1;
  | | | end;
\end{verbatim}
}
\begin{Verbatim}
  | | | danger := b;
  | | end;
  | end;
  |
  | function is_up: Boolean;
  | begin
  | | is_up := (k < n) and not danger;
  | end;
  |
  | function is_right: Boolean;
  | begin
  | | is_right := (k > 0) and (c[k] < n);
  | end;
  | {danger: when k=0, the value c[k] is undefined}
  |
  | function is_down: Boolean;
  | begin
  | | is_down := (k > 0);
  | end;
  |
  | procedure up_left;
  | begin {k < n, not danger}
  | | k := k + 1;
  | | c [k] := 1;
  | end;
  |
  | procedure right;
  | begin {k > 0,  c[k] < n}
  | | c [k] := c [k] + 1;
  | end;
  |
  | procedure down;
  | begin {k > 0}
  | | k := k - 1;
  | end;
  |
  | procedure process;
  | | var i: integer;
  | begin
  | | if (k = n) and not danger then begin
  | | | for i := 1 to n do begin
  | | | | write ('<', i, ',' , c[i], '> ');
  | | | end;
  | | | writeln;
  | | end;
  | end;
  |
  | procedure go_up_and_process;
  | begin
  | | while is_up do begin
  | | | up_left;
  | | end
  | | process;
  | end;
  |
  begin
  | begin_work;
  | go_up_and_process;
  | while is_down do begin
  | | if is_right then begin
  | | | right;
  | | | go_up_and_process;
  | | end else begin
  | | | down;
  | | end;
  | end;
  end.
\end{Verbatim}

\setbox0\hbox{\protect\verb|is_up|}
\setbox1\hbox{\protect\verb|is_up|}

\Problem{%
     The program above spends a lot of time inside the procedure
\box0\ (to check if the uppermost queen is under attack,
we need $O(\w{n})$ operations). Modify the implementation of the
positions tree in such a way that all three tests
\box1/\verb|right|/\verb|down| and the corresponding three commands
require only $O(1)$ operations (that is, the number of operations for any of
them should be limited by a constant that does not depend on \w{n}).}

     \Solution. For any vertical and for any diagonal line (there are
two types of
diagonal lines
--- ascending and descending ones)  there is a Boolean
variable that indicates
if this line is occupied by some queen (except the
uppermost one, which is ignored). Recall that any of those lines may
be occupied by at most one queen (because the position is assumed
to be admissible).\es

