\subsection{Backtracking in other problems}
\label{traversalusage}

\Problem{%
     Use backtracking in the following problem:
An array of \w{n} positive integers
$\w{a[1]..a[n]}$ and a positive integer
\w{s} are given. Determine if
\w{s} can be represented as a sum of some of the elements of the
array
\w{a}. (Each element may be used at most once.)}
 \index{problem!knapsack}%
 \index{knapsack problem}%

     \Solution. Construct the position tree as follows: The
\w{k}-{\it position\/} is a sequence of \w{k} Boolean values that
determines which of the elements
$\w{a[1]..a[k]}$ are used as summands.
The position is {\it admissible\/} if the sum of the corresponding elements
does not exceed \w{s}.\es

\bigskip

     Remark. This approach is better than exhaustive search
(that considers all
$\w{2}^\w{n}$ subsets). We may also sort the array \w{a} in descending
order.
Also, we can change the definition of an admissible position to
exclude positions where the sum of rejected elements is larger than
the difference between \w{s} and the sum of all accepted elements.
However, this does not lead to a fundamental improvement; this problem
belongs to the category of the so-called ``NP-complete problems''.
 \index{problem!NP-complete}%
 \index{NP-completeness}%
See the book
by A.~Aho%
 \index{Aho, A.V.},
J.~Hopcroft%
 \index{Hopcroft, J.},
and J.~Ullman%
 \index{Ullman, J.D.}
[1]
and the book
by M.R.~Garey%
\index{Garey, M.R.}
and D.S.~Johnson%
\index{Johnson, D.S.}
[5].

This problem is traditionally called ``the  knapsack problem'':
\label{traversalknapsack}%
A knapsack that is capable of carrying \w{s} pounds
should be filled completely using only objects of weights
$\w{a[1]..a[n]}$. See section~%
\ref{delrec}, p.~\pageref{dynamicknapsack},
where a ``dynamic programming'' algorithm is given
whose running time
is polynomial in
$\w{n}+\w{s}$.

\Problem{%
     Generate all sequences of $n$ digits $0$, $1$ and $2$
that do not contain a substring of type $XX$.
(E.g., the sequence $210102$ is prohibited because it contains
$1010$.)\es
}

\Problem{%
       Repeat the previous
problem for binary strings of length
$n$ that do not contain a substring of type $XXX$.\es}

     Another problem of the same category: ``Is it possible to
compose a given polygon of `pentamino' blocks?'' The crucial
component of an effective algorithm for such a problem is a good
criterion that can (in some cases)
guarantee
that a given position cannot be extended
to a solution of the problem and therefore may be discarded.

