\section{Sorting}\label{sorting}

\subsection{Quadratic algorithms}

\Problem{%
     Let $\w{a[1]},\ldots,\w{a[n]}$ be an array of numbers
(say, integers).
Construct the array
$\w{b[1]},\ldots,\w{b[n]}$ that contains the same numbers in
increasing order:
$\w{b[1]}\le\cdots\le\w{b[n]}$.}

     Remark. The elements $\w{a[1]..a[n]}$ need not to be distinct.
In this case we require that the multiplicity of
each number in
$\w{b[1]..b[n]}$ should be equal to its multiplicity in
$\w{a[1]..a[n]}$.

     \Solution.%
\index{sorting!quadratic}
It is convenient to consider
$\w{a[1]..a[n]}$
and $\w{b[1]..b[n]}$ as the
initial and final values of
some array  \w{x}.
The requirement ``\w{a} and \w{b} contain the same numbers''
will be guaranteed if the only operation permitted on \w{x}
is the
exchange of two its elements. (Of course, we are also
allowed to read
elements of \w{x}, too.)
\begin{Verbatim}
  k := 0;
  {k minimal elements of x are in their places}
  while k <> n do begin
  | s := k + 1; t := k + 1;
  | {x[s] is minimal among x[k+1]...x[t] }
  | while t<>n do begin
  | | t := t + 1;
  | | if x[t] < x[s] then begin
  | | | s := t;
  | | end;
  | end;
  | {x[s] is minimal among x[k+1]..x[n] }
  | ... exchange x[s] and x[k+1];
  | k := k + 1;
  end;
\end{Verbatim}

\Problem{%
     Give another sorting algorithm which uses the following
invariant relation:
``first \w{k} elements are sorted''
($\w{x[1]}\le\cdots\le\w{x[k]}$).}

     \Solution.  (This algorithm is called {\it insertion sort}.)
\begin{verbatim}
  k:=1;
  {first k elements are sorted}
  while k <> n do begin
  | t := k + 1;
  | {k+1-th element moves to the left until it finds its
  |   place; t is its current position}
\end{verbatim}
\clearpage
\begin{verbatim}
  | while (t > 1) and (x[t] < x[t-1]) do begin
  | | ... exchange x[t-1] and x[t];
  | | t := t - 1;
  | end;
  end;
\end{verbatim}

     Remark. Danger: When \w{(t > 1)} is false, the test
$\w{x[t]}<\w{x[t-1]}$ refers to a non-existing value
\w{x[0]}.\es
 \index{error!index out of bounds}%

\medskip

     Both of the above
solutions require a number of operations
proportional to $\w{n}^2$. There are more efficient algorithms,
however,
as we shall see.

