\subsection{Sorting in $n\,\r{log}\,n$ operations}

\Problem{%
     Find a sorting algorithm that requires only
$O(n\,\log n)$ operations. (In other words,
the number of operations should not
exceed $Cn\log
n$ for some constant $C$ that does not depend on $n$.)}%
 \index{sorting! $n\log n$}%

     We give two solutions.

     \Solution\ 1 (merge sort).%
 \index{sorting!merge}

     Let \w{k} be a positive integer, and split the array
$\w{x[1]..x[n]}$ into segments of length
\w{k}. (The first segment is
$\w{x[1]..x[k]}$, the next is the segment
$\w{x[k+1]..x[2k]}$, etc.)
The last segment is incomplete if \w{n} is not a multiple of \w{k}.
We say that the array \w{x} is \w{k}-{\it sorted\/}
if each of these segments
(considered separately) is sorted. Of course,
any array is \w{1}-sorted. If an array of length \w{n} is \w{k}-sorted
for
$\w{k}\ge\w{n}$, it is sorted.

     Assume there is a procedure that transforms
any \w{k}-sorted array into a \w{2k}-sorted array (containing the same
elements). Using this procedure, we write down an algorithm as
follows:
\begin{verbatim}
 k:=1;
 {the array x is k-sorted}
 while k < n do begin
 | ..transform the k-sorted array into a 2k-sorted array;
 | k := 2 * k;
 end;
\end{verbatim}

How do we construct such a procedure?
It repeats the following step:
two sorted segments of length at most \w{k} are merged into
one sorted segment. Assume that the procedure
\[
    \w{merge (p,q,r: integer)}
\]
called with
$\w{p}\le\w{q}\le\w{r}$ merges the sorted segments
$\w{x[p+1]..x[q]}$ and $\w{x[q+1]..x[r]}$ into a
sorted segment $\w{x[p+1]..x[r]}$ (without changing other
parts of the array \w{x}):
%
%                 p               q               r
%           -------|---------------|---------------|-------
%                  |               |               |
%           -------|---------------|---------------|-------
%                                 |
%                                 |
%                                 V
%           -------|-------------------------------|-------
%                  |                               |
%           -------|-------------------------------|-------
%
\setlength{\unitlength}{1.2em}
\[
\begin{picture}(20,6)
\thinlines
\put(1,1){\line(1,0){18}}
\put(1,2){\line(1,0){18}}
\put(1,4){\line(1,0){18}}
\put(1,5){\line(1,0){18}}
\put(2,1){\line(0,1){1}}
\put(18,1){\line(0,1){1}}
\put(2,4){\line(0,1){1}}
\put(18,4){\line(0,1){1}}
\put(10,4){\line(0,1){1}}
\put(2,4){\makebox(8,1){\r{sorted}}}
\put(10,4){\makebox(8,1){\r{sorted}}}
\put(2,1){\makebox(16,1){\r{sorted}}}
\put(9.5,2.5){\makebox(1,1){$\downarrow$}}
\put(1,5){\makebox(1,1){\w{p}}}
\put(9,5){\makebox(1,1){\w{q}}}
\put(17,5){\makebox(1,1){\w{r}}}
\end{picture}
\]
The transformation of a
\w{k}-sorted array into a \w{2k}-sorted array is as follows:
\begin{verbatim}
  t:=0;
  {t is a multiple of 2k or t = n, x[1]..x[t] is
   2k-sorted; the rest of x is unchanged}
  while t + k < n do begin
  | p := t;
  | q := t+k;
  | ...r := min (t+2*k, n);
  |        {min(a,b) is the minimum of a and b}
  | merge (p,q,r);
  | t := r;
  end;
\end{verbatim}
The merge procedure uses an auxiliary array as temporary storage
for the result. This auxiliary array will be denoted by
\w{b}. Let \w{p0} and \w{q0} be the indices of the last elements
merged; \w{s0} is the index of the last element written to
\w{b}. At each step, one of the two following actions is performed:

\smallskip

\begin{verbatim}
        b[s0+1]:=x[p0+1];
        p0:=p0+1;
        s0:=s0+1;
\end{verbatim}
or
\begin{verbatim}
        b[s0+1]:=x[q0+1];
        q0:=q0+1;
        s0:=s0+1;
\end{verbatim}
(C fans will enjoy the shorthands \w{b[++s0]=x[++p0]}
and \w{b[++s0]=x[++q0]} here.)

The first action (where
the element is taken from the first segment)
may be performed if the following two conditions are fulfilled:

    (1) the first segment is not empty ($\w{p0}<\w{q}$); and

    (2) the second segment is empty ($\w{q0}=\w{r}$) or its
first element is greater than or equal to the first element of the
first segment\quad  [$(\w{q0} < \w{r}$) and
($\w{x[p0+1]}\le\w{x[q0+1]}$)].

The conditions that make the second action possible are similar.
We obtain
the
following program:
\begin{verbatim}
  p0 := p; q0 := q; s0 := p;
  while (p0 <> q) or (q0 <> r) do begin
  | if (p0 < q) and ((q0 = r) or ((q0 < r) and
  | |                (x[p0+1] <= x[q0+1]))) then begin
  | | b [s0+1] := x [p0+1];
  | | p0 := p0+1;
  | | s0 := s0+1;
  | end else begin
  | | {(q0 < r) and ((p0 = q) or ((p0<q) and
  | |   (x[p0+1] >= x[q0+1])))}
  | | b [s0+1] := x [q0+1];
  | | q0 := q0 + 1;
  | | s0 := s0 + 1;
  | end;
  end;
\end{verbatim}
(If both segments are nonempty and have equal first elements, both
actions are legal. In this case, the program chooses the first one.)

The only thing left
to do is copy the merged array back into
\w{x}.  (Warning: If you decide to perform copying outside
the merge procedure, please note that the last segment should
be copied even it was not merged with anything.)

     The program has a standard deficiency: the array
index can be out of
bounds when the Boolean expressions are evaluated (if
``short-circuit evaluation''
is not used).%
 \index{error!index out of bounds}%

