\bigskip
     \Solution\ 2 (heap sort).

     Draw a ``complete binary tree''. The root of this tree%
 \index{sorting!heapsort}
 \index{tree!binary}%
is drawn as a small circle at the bottom;
two arrows go from the root node to the
two nodes above it,
two arrows go from each of them, etc.
%
%                o  o o  o
%                 \/   \/
%                  o   o
%                   \ /
%                    o
%
\setlength{\unitlength}{2em}
\[
\begin{picture}(6,5)
\put(3,1){\circle*{.15}}
\put(2,2){\circle*{.15}}
\put(4,2){\circle*{.15}}
\put(1.5,3){\circle*{.15}}
\put(2.5,3){\circle*{.15}}
\put(3.5,3){\circle*{.15}}
\put(4.5,3){\circle*{.15}}
\put(1.166666,4){\circle*{.15}}
\put(1.833333,4){\circle*{.15}}
\put(2.166666,4){\circle*{.15}}
\put(2.833333,4){\circle*{.15}}
\put(3.166666,4){\circle*{.15}}
\put(3.833333,4){\circle*{.15}}
\put(4.166666,4){\circle*{.15}}
\put(4.833333,4){\circle*{.15}}
\put(3,1){\vector(-1,1){1}}
\put(3,1){\vector(1,1){1}}
\put(2,2){\vector(-1,2){.5}}
\put(2,2){\vector(1,2){.5}}
\put(4,2){\vector(-1,2){.5}}
\put(4,2){\vector(1,2){.5}}
\put(1.5,3){\vector(-1,3){.33333333}}
\put(1.5,3){\vector(1,3){.33333333}}
\put(2.5,3){\vector(-1,3){.33333333}}
\put(2.5,3){\vector(1,3){.33333333}}
\put(3.5,3){\vector(-1,3){.33333333}}
\put(3.5,3){\vector(1,3){.33333333}}
\put(4.5,3){\vector(-1,3){.33333333}}
\put(4.5,3){\vector(1,3){.33333333}}
\end{picture}
\]

     We say that arrows connect a ``father'' to its two ``sons''.
Each node has two sons and one father unless it is the the ``root''
(a node at the bottom) or the ``leaf'' (a node at the top).
For simplicity, we assume that the length of the array to be sorted
is a power of $2$ and the elements  fill completely
some level of the tree.
Fill the part of the tree below them using the following rule:
\[
  \r{father} = \min\,(\r{son}_1, \r{son}_2)
\]
According to this rule, the value at the root of the tree will be
the minimal element of the whole array.

     Take the minimal element out of the array. To do that, we
first locate it. It can be traced going from bottom to top,
traversing the son that has the same value as its father.
After the minimal element is removed, we replace it by the
symbol $\infty$ and modify its ancestors going from top to
bottom. (We assume that
$\min(t,\infty)=t$.)
Consequently,
the root of the tree contains the second minimal
element. We locate it, take it out (replacing it by $\infty$) and
modify the tree. This procedure is repeated until all the elements
are taken out and the root of the tree is occupied by~$\infty$.

     To write down the corresponding program the following
agreement is useful. Assume that the vertices of the tree are
numbered by
$\w{1},\w{2},\ldots$ in such a way that position
\w{n} has sons
\w{2n} and $\w{2n}+\w{1}$. We do not give the details, because
we will write down a more efficient algorithm that does not
use any additional memory (except for a fixed number of variables
and the array itself). Here it is:

     The elements to be sorted are placed at all levels of
the tree, not just the upper level. Suppose we want to sort
the array
$\w{x[1]..x[n]}$.
The tree has numbers $\w{1..n}$ as vertices. We assume that
\w{x[i]} is placed at vertex \w{i}. During execution, the
number of vertices in the tree will decrease. The current number
of vertices is stored in \w{k}. Therefore, at any time the
array
$\w{x[1]..x[n]}$ is divided into two parts. Its initial
segment $\w{x[1]..x[k]}$ represents a tree. The remaining
part $\w{x[k+1]..x[n]}$ contains the already sorted part of
the array; those elements have already reached their final destination.

     At each step, the algorithm extracts the maximal element from
the tree
and puts it into the sorted part (using the position
freed when the tree becomes smaller).

     Let us specify some terminology. The {\it vertices\/}
of the tree are
numbers from \w{1} up to the current value of
\w{k}. Each vertex \w{s} may have
{\it sons\/} \w{2s} and $\w{2s}+\w{1}$. If both numbers are larger than
\w{k}, the vertex \w{s} has no sons. Such a vertex is called a
{\it leaf}. If $\w{2s}=\w{k}$, the vertex \w{s} has exactly one
son (\w{2s}).

     For any \w{s} in $\w{1..k}$, we consider a
{\it subtree rooted at\/} \w{s} (or \w{s}-{\it subtree}\/).
It contains the vertex \w{s} and
all its descendants (sons, grandsons, etc.\ --- until we leave
the segment $\w{1..k}$).
The vertex \w{s} is called {\it regular\/} if the element placed in
it is the maximal element of the
\w{s}-subtree; the \w{s}-subtree is
called {\it regular\/} if all its vertices are regular. (In particular,
any leaf is a regular singleton subtree.)
Please note that the validity of the statement ``\w{s}-subtree
is regular'' depends not only upon \w{s} but also upon the current
value of \w{k}.)

     Remark. Modern textbooks (see, e.g., [3]) use terms ``child''
(``parent'', ``siebling'', etc.) instead of ``son'' (``father'',
``brother'', etc.) that are used in older textbooks (see, e.g., [1]).
When using this new terminology, you should have in mind that
each vertex has only one parent,
only one grandparent, etc.

\medskip

     The general structure of the algorithm is as follows:
\begin{verbatim}
 k:= n
 .. Make the 1-subtree regular;
 {x[1],..,x[k] <= x[k+1] <=..<= x[n]; 1-subtree is regular,
  therefore, x[1] is maximal among x[1]..x[k]}
 while k <> 1 do begin
 | .. exchange x[1] and x[k];
 | k := k - 1;
 | {x[1]..x[k-1] <= x[k] <=...<= x[n]; 1-subtree is
 |   regular everywhere except the root (may be)}
 | .. restore the regularity of 1-subtree
 end;
\end{verbatim}
As a tool, we use a procedure that restores the regularity of
the subtree which is regular everywhere except its root. Here it is:
\begin{verbatim}
 {s-subtree is regular everywhere except perhaps its root}
 t := s;
 {s-subtree is regular everywhere except perhaps t}
 while ((2*t+1 <= k) and (x[2*t+1] > x[t])) or
 |     ((2*t <= k) and (x[2*t] > x[t])) do begin
 | if (2*t+1 <= k) and (x[2*t+1] >= x[2*t]) then begin
 | | ... exchange x[t] and x[2*t+1];
 | | t := 2*t + 1;
 | end else begin
 | | ... exchange x[t] and x[2*t];
 | | t := 2*t;
 | end;
 end;
\end{verbatim}
  %
Let us look closely at this procedure to check its
correctness. Assume that all vertices of the
\w{s}-subtree are regular
except perhaps the vertex \w{t}. Consider the sons of \w{t}. They
are regular and therefore contain maximal elements of their subtrees.
Therefore, we have only three possibilities for the maximal element
of the \w{t}-subtree, namely, the vertex \w{t} and its sons.
If the vertex \w{t} contains the maximal element, this vertex
is regular, and
we are done. The while-construct can be
rewritten as follows:
\begin{verbatim}
while the maximal element is not t but one of its sons
|   do begin
| if it is the right son then begin
| | exchange t and its right son; t:= right son;
| end else begin {the maximal element is the left son of t}
| | exchange t and its left son; t:= left son;
| end
end
\end{verbatim}
After the exchange, the vertex
\w{t} becomes regular (since it contains the maximal element of
the
\w{t}-subtree). The son that does not take part in the exchange is
still regular. The other son may become non-regular.
Any other vertex \w{u} of the
\w{s}-subtree remains regular because the value
placed in \w{u} is unchanged, as well as the \w{u}-subtree
(though elements
of the subtree may be permuted).

   The same procedure may be used at the first stage of
our algorithm to make the
\w{1}-subtree regular:
\begin{verbatim}
  k := n;
  u := n;
  {s-subtrees are regular for all s>u}
  while u<>0 do begin
  | {u-subtree is regular everywhere
  |             except the root (may be)}
  | .. restore the regularity of u-subtree;
  | u:=u-1;
  end;
\end{verbatim}
This algorithm is called {\it heap sort}.

\smallskip

     Now we are ready to write down the heapsort program in Pascal.
We assume that  \w{n} is a constant and \w{x} is a variable of type
\w{arr} = \w{array [1..n]} \w{of} \w{integer}).
\begin{verbatim}
  procedure sort (var x: arr);
  | var u, k: integer;
  | procedure exchange(i, j: integer);
  | | var tmp: integer;
  | | begin
  | | tmp  := x[i];
  | | x[i] := x[j];
  | | x[j] := tmp;
  | end;
  | procedure restore (s: integer);
  | | var t: integer;
  | | begin
  | | t:=s;
  | | while ((2*t+1 <= k) and (x[2*t+1] > x[t])) or
  | | |     ((2*t <= k) and (x[2*t] > x[t])) do begin
  | | | if (2*t+1 <= k) and (x[2*t+1] >= x[2*t]) then begin
  | | | | exchange (t, 2*t+1);
  | | | | t := 2*t+1;
  | | | end else begin
  | | | | exchange (t, 2*t);
  | | | | t := 2*t;
  | | | end;
  | | end;
  | end;
  begin
  | k:=n;
  | u:=n;
  | while u <> 0 do begin
  | | restore (u);
  | | u := u - 1;
  | end;
  | while k <> 1 do begin
  | | exchange (1, k);
  | | k := k - 1;
  | | restore (1);
  | end;
  end;
\end{verbatim}

     Several remarks:
\begin{itemize}
\item
     The method used in the
heapsort algorithm has other applications.
One example is the priority queue implementation, see
section~\ref{priorityqueue}, p.~\pageref{priorityqueue}.
\item
    The advantage of the merge sorting algorithm is that
 \index{sorting!merge}%
it does not require the entire array to be placed into RAM.
We can sort portions of the array that
fit into available RAM (say, using
heapsort) and then merge the files obtained.
\item
     Another important sorting algorithm
(``Hoare quicksort'')
 \index{sorting!Hoare (quicksort)}%
 \index{Hoare sorting}%
 \index{quicksort algorithm}%
uses the following approach. To sort an array, choose a random
element $b$ and split the array into three parts:
elements smaller than $b$, equal to $b$ and greater than $b$.
(This problem is discussed in section~\ref{start}.) Now it remains to
sort the first and the third parts, which can be done recursively using the
same method.
The number of steps of this algorithm is a random variable. One can
prove that its expectation does not exceed $Cn\log n$.
It is one of the fastest algorithms in practice. (We shall discuss its
recursive and non-recursive implementations later.)
\item
     Finally, let us mention that sorting in $Cn\log n$ operations
may be performed using the technique of balanced trees (see
section~\ref{trees}), but the programs are rather complicated and
the constant $C$ is large enough to make this method impractical.\es
\end{itemize}
