\subsection{Lower bound for the number of comparisons}

Suppose we have $n$ objects (say, stones) of different
weights and a balance that can be used to find which of any
two given
stones is heavier. In programming terms, we have access to a
Boolean function \w{heavier(i,j:1..n)}. Our goal is to sort
all the stones
in increasing order
using the balance as few times as possible
(making the fewest calls to the function \w{heavier}).%
 \index{sorting!number of comparisons}%

    Of course, the number of comparisons depends not only on the
algorithm we choose but also on the initial order of the stones.
By {\it complexity\/} of the algorithm we mean the number of comparisons
{\it in the worst case}.%
 \index{sorting!lower complexity bound}%

\Problem{%
    Prove that any sorting algorithm for $n$ stones
    has complexity at least $\log_2 n!$. (Here
$n!=1\cdot2\cdots n$.)}

     \Solution. Assume that we have an algorithm of complexity $d$, that is,
an algorithm that makes at most $d$ comparisons (in all cases).
For any of $n!$ possible orderings of the
stones let us write down
the results of all the
comparisons (calls to the function \w{heavier}).
The protocol is a binary string of length
at most $d$.
If necessary, pad
it with trailing zeros to get a string of length $d$.
Now we have $n!$ binary strings of length $d$ (corresponding to
$n!$ permutations of input stones). All those strings are different,
otherwise our algorithm gives the same answer for two
different orderings
(and at least one of the answers is incorrect).
Therefore, $2^d \ge n!$.

    Another way to say the same thing
is to consider the tree of
possibilities that appear during the execution of the
algorithm.
Indeed, a tree of height $d$ has no more than
$2^d$ leaves.

     This argument shows that any algorithm that relies upon
comparisons and exchange operations only, requires at least $\log_2 n!$
comparisons.
A simple calculation shows that $\log_2 n! \ge \log_2 (n/2)^{n/2}$
(we omit the first half of the factors and replace the remaining factors
by $n/2$). Now $\log_2 (n/2)^{n/2} = (n/2)(\log_2 n -1) \ge Cn\log_2 n$
for some $C$. Therefore, our sorting algorithms are close to optimal
(improvement is limited to a constant factor).\es

However, a sorting algorithm that uses not only comparisons (but also
the
internal structure of the sorted objects)
may be faster. Here is an example:

\Problem{%
     An integer array $\w{a[1]..a[n]}$ is given; all the integers
are non-negative and do not exceed \w{m}. Sort this array using no more
than $C(\w{m}+\w{n})$ operations ($C$ is a constant that does not depend
on \w{m} and \w{n}).}

     \Solution. For each number in $\w{0..m}$, count how many times
it appears in the array. (These data can be collected during one pass
through the array.) Then erase the array and write down its elements
in increasing order using the information about the multiplicity of
each number.\es

Note that this algorithm does not exchange elements
of the array but puts ``fresh'' sorted numbers into the array.

\medskip

There exists another sorting method that sequentially performs
several ``partial sorts'' with respect to fixed bits.
Let us start with the following problem:

\Problem{%
     Rearrange the array $\w{a[1]..a[n]}$ in such a way that all
even elements precede all odd elements (not changing the order inside
each of the
two groups).}

     \Solution. Copy all the even elements into an auxiliary array.
Then append all the odd elements to that array and copy all elements
back.\es

\Problem{%
     An array of $n$ integers in the range $0,\ldots,2^k-1$ is given.
Each integer is written as a binary string of length $k$. Using the
tests ``$i$-th bit is $0$'' and ``$i$-th bit is $1$'' instead of
comparisons, sort all the integers. The number of operations should be
of order $nk$.}
 \index{sorting!radix}%

     \Solution. Sort all the numbers with respect to the
last bit as in the preceding problem. Then sort them with
respect to the bit next the last one, etc. After $k$ stages, the
numbers will be sorted. Indeed, by induction over $i$, we can
easily prove the following statement: ``after $i$ steps, any
two
numbers that differ only in the last $i$ bits, are in the correct
order''. (Or prove by induction the following statement:
``after $i$ steps the suffixes of length $i$ are in the
right order''.)\es

\smallskip

     A similar algorithm can be constructed using $m$-ary notation
instead of binary. The following problem is useful in this regard.

\Problem{%
     Assume that an array of $n$ elements and a function $f$
defined on those
elements are given. Assume that the possible values of $f$ are
$1,\ldots,m$. Rearrange the array in such a way that the values
of $f$ are in nondecreasing order and the elements with equal
values of $f$ are in the same order as before. The number of operations
should be of order
$m+n$.}

    \Hint{Create $m$ lists of total length $n$ using
``pointer implementation'' (see
section~\ref{types},
p.~\pageref{pointerlists}).
Put an element into the $i$-th list
if the value of $f$ is equal to $i$.
Another possibility: count how many elements have a given
value of $f$ (for all $m$ possible values); thereafter, we know
where the elements of any given $f$-value should be placed in the array.}

