\subsection{Problems related to sorting}

\Problem{%
      Find the minimal complexity (= the number of
comparisons in the worst case) for an algorithm that finds the
stone with minimal weight.}%
 \index{minimal element, search}%
 \index{search!of the minimal element}

     \Solution. The obvious algorithm with the invariant relation
``the minimal among the first $i$ stones is found'' requires
$n-1$ comparisons. No algorithm can have smaller complexity. This
is a corollary of a stronger statement, see the next problem.\es

\Problem{%
     An expert wants to convince a jury that a given stone has
minimal weight among $n$ given stones. The expert
wants to do this
using a balance less than $n-1$ times. Prove that this is
impossible. (The expert knows in advance the weights of all the
stones; the jury does not.)}

     \Solution. Consider stones as vertices of a graph. For any
comparison, draw an edge between the corresponding pair of
vertices. After $n-1$ measurements, the
graph is not connected; it has more than $1$ connected
component, because each edge decreases the number of connected
components by at most $1$. Therefore, the jury knows nothing
about the relation between weights of stones from different
components and may assume that the stone with minimal weight is
in any of the components.\es

     Let us stress the difference between this problem and the
preceding one. In this problem, we have to show that $n-2$
comparisons are not enough to prove that a given stone has
minimal weight even if we know the answer in advance --- not to mention
finding the answer. (The difference between the two settings is
clear in the case of sorting.
When a correct
answer is known, it can be confirmed by $n-1$ comparisons (each
stone should be compared with the next one), which is many fewer
comparisons than what was
needed to find the answer.)

\Problem{Prove that it is possible to find the stones with minimal
and maximal weights among $2n+1$ stones using only
$3n$ comparisons.}

\Hint{Divide all the stones into $n$ pairs (one
stone
remains) and
compare stones within each pair.}

\Problem{%
     Assume that $n$ stones of different weights are given.
Let $k$ be a number in the range $1,\ldots,k$. Find the $k$-th
stone (in the order of increasing weights) making not more than
$Cn$ comparisons, where
$C$ is some constant that does not depend on $k$ or $n$.}%
 \index{search!$k$-th element}%
 \index{median, search}%

     Remark. Using sorting, we can do this in $Cn\log n$ steps.
See the section~\ref{recursion}, p.~\pageref{median},
where a hint for this (rather
difficult) problem is given.\es

\smallskip

    The following problem has a surprisingly simple solution.

\Problem{%
     There are $n$ stones that look identical, but in fact, some of them
have different
weights. There is a device that can be applied to two stones and tells
whether
they are different or not (but it does not say which one is
heavier). It is known in advance that most of the stones (more than $50\%$)
are identical. Find one of those identical stones making no more than $n$
comparisons.
(Beware: it is possible that
two stones are identical but do not belong to the majority of
identical stones.)}%
 \index{search!majority representative}%

    \Hinta{If two stones are different, they may be discarded,
because one of them does not belong to the majority and the majority
remains.}

     \Solution. The program processes the stones one-by-one and keeps the
number of the stones processed in a variable
\w{i}. (Assume that stones are numbered $\w{1..n}$).
The program remembers the number of the
``current candidate'' \w{c} and its
``multiplicity'' \w{k}. The names are explained by the following
invariant relation \w{(I)}:
\begin{quote}
   If we add \w{k} copies of the
   \w{c}-th stone to the unprocessed
   stones ($\w{i+1..n}$), the majority stones in the
   initial array will remain the majority in the new array.
\end{quote}

Here is the program:
\begin{verbatim}
  k:=0; i:=0;
  {(I)}
  while i<>n do begin
  | if k=0 then begin
  | | k:=1; c:=i+1; i:=i+1;
  | end else if (i+1-th stone is the same as c-th)
  | |   then begin
  | | i:=i+1; k:=k+1;
  | |  {replace a physical stone by a virtual stone}
  | end else begin
  | | i:=i+1; k:=k-1;
  | | {discard one physical and one virtual stone}
  | end;
  end;
  c-th stone is the answer
\end{verbatim}

Remark. All three branches of the \w{if}-block
include the statement
\w{i:=i+1}, so it can be moved to the next level.\es

\smallskip

Let us mention that this program finds the most frequent stone only
if it forms the majority (more than $50\%$).

This problem can be found as problem 4-7
on page 75 of the book [3] in a completely
different setting (``VLSI chip testing'') where a recursive solution
is sketched.

\medskip

     At first glance,
the following problem seems unrelated to sorting.

\Problem{%
     There is a square array
\w{a[1..n,1..n]} filled by \w{0}s and \w{1}s.
It is known that for some
\w{i} the \w{i}-th row contains only \w{0}s and at the same time the
\w{i}-th column contains only \w{1}s (except the main diagonal entry \w{a[i,i]},
which may be arbitrary). Find this \w{i} (which is unique).
The number of operations should be of order \w{n}. (Please note
that the number of operations
should be much smaller than the total number of
elements in \w{a}.)}

    \Hint{Assume we get the Boolean value
\w{a[i][j]} when comparing two virtual stones with numbers
\w{i} and \w{j}. Recall that the maximal element among \w{n}
elements may be found using \w{n-1} comparisons. Take into account that
the array may not be ``transitive''; however, after two numbers are
compared, one of them may be discarded.}

