\section{Finite-state algorithms in text processing}\label{automata}

 \index{automaton!finite}%
 \index{finite automaton}%


\subsection{Compound symbols, comments, etc.}

\setbox0\hbox{\verb|^|}
\setbox1\hbox{\verb|x^y|}
\Problem{%
Throughout a program text
the operation $x^y$ was denoted by \w{x**y}.
It was decided that notation should be changed to
\box1. How do we do that?
The input text is read character-by-character; the output text
should be printed in the same manner.}%
 \index{compound symbols, replacement}%

     \Solution. At any time, the program is in one of two states:
``basic'' state and ``after'' state (after an asterisk):
\[
\begin{tabular}{|l|l|l|l|}
\hline
State     &  Next symbol   &  New state  &   Action \\[.2ex]
\hline
basic     &   \w{*}       &  after &   none  \\
\hline
basic     &  $x\ne\w{*}$  &  basic &   print $x$ \\
\hline
after     &    \w{*}      &  basic &   print \verb|^|\\
\hline
after     &  $x\ne\w{*}$  &  basic &   print \w{*}, $x$\\
\hline
\end{tabular}
\]
If after reading all the text, the program is in the ``after'' state, it
should print an asterisk (and quit).\es

\smallskip

Remark. Our program replaces
\w{***} by \verb|^*| (and not by
\verb|*^|). We did not specify the behavior of the program in this
case, assuming (as is  often  done)  that  some  ``reasonable''
behavior
is expected. In this example, the simplest way to describe
the required behavior is to list the states and the
corresponding
actions.

Please note also that if two asterisks appear in other
parts of the program (say, comments), they will be also
replaced.


\Problem{%
     Write a program that deletes all occurrences of the
substring
\w{abc}.\es}

\Problem{%
     In Pascal, comments are surrounded by curly braces like this:}
 \index{comments!removal}%
\begin{verbatim}
                begin {here a block begins}
                i:=i+1; {increase i by one}
\end{verbatim}
Write a program that removes all comments and puts a space
character
in the place
of a removed comment. (According to Pascal rules, \verb|1{one}2|
is equivalent to \verb|1 2|, not \verb|12|).

     \Solution. The program has two states: a  ``basic'' state and
an ``inside'' state (inside a comment).

\setbox0\hbox{\verb|{|}
\setbox1\hbox{\verb|}|}
$$
\begin{tabular}{|l|l|l|l|}
\hline
State     &  Next symbol  &  New state &   Action \\[0.2ex]
\hline
basic     & \verb|{|     &     inside  &   none   \\
\hline
basic     & $x\ne \box0$ &    basic   &  print $x$ \\
\hline
inside    & \verb|}|     &    basic   &  print a space \\
\hline
inside    & $x\ne \box1$ &   inside &    none  \\
\hline
\end{tabular}
$$
\nobreak
\hskip -2ex
\hbox to \textwidth{\hss\es}

\clearpage
     This program cannot deal with nested comments:
 \index{comments!nested}%
the string
\begin{verbatim}
       {{comment inside a} comment}
\end{verbatim}
is transformed into
\begin{verbatim}
         comment}
\end{verbatim}
(the latter string starts with two spaces).
It is impossible to deal with nested comments using a finite automaton
(a program that has finite number of internal states); roughly speaking,
we have to remember the number of opening braces and a finite automaton
cannot do that.

Please note that after reading all the text, the program may still be in
the ``inside'' state. Most probably, we would like to consider this as
an error.

\Problem{%
     Pascal programs also contain quoted strings. If a curly brace
appears inside a string, it does not mean the start of a comment.
Similarly, a quote symbol inside a comment does not signify
a string. How do we modify the above program to
take this into account?}

    \Hint{We need three states: ``basic'',
``inside a comment'', ``inside
a string''.}

(Note that actual Pascal conventions are more complicated allowing
a quote appear inside a quoted string, etc.)


\Problem{%
     One  more feature that  exists in   many   Pascal
implementations is
a comment of the type}
\begin{verbatim}
      i := i+1;   (* here i is increased by 1 *)
\end{verbatim}
A closing comment
symbol must be paired with an  opening comment
symbol of the same type
(e.g., \verb|{|$\ldots\w{*)}$ is not permitted). How do we deal with
these types of comments?\es

