\subsection{Numbers input}

     Assume that a program scans a decimal representation of
some number from left to right. The program should ``read'' this
number, that is,
 \index{decimal number!reading}%
put its value into
a variable of type \w{real}. Also, the program should complain if
the input is incorrect.

     Let us specify the problem in more detail. Assume that the
input string is divided into two parts: the part that is already processed
and the remaining part. We have access to a function \w{Next:char},
which returns the first symbol of the unprocessed part. Also, we have access
to a procedure \w{Move}, which moves the first unprocessed symbol
to the processed part.
%
%       ---------------------|--------------------------
%         processed part     | Next |  ?  |  ?  |  ?  |
%       ---------------------|--------------------------
%
\setlength{\unitlength}{1.4em}
\[
\begin{picture}(17,4)
\thinlines
\put(1,1){\line(1,0){15}}
\put(1,3){\line(1,0){15}}
\put(9,0.5){\line(0,1){3}}
\put(11,1){\line(0,1){2}}
\put(13,1){\line(0,1){2}}
\put(15,1){\line(0,1){2}}
\put(1,1){\makebox(8,2){\r{processed part}}}
\put(9,1){\makebox(2,2){\w{Next}}}
\put(11,1){\makebox(2,2){\r{?}}}
\put(13,1){\makebox(2,2){\r{?}}}
\end{picture}
\]

By a decimal number we mean a character string of the
type
\[
  \langle 0\ \r{or more spaces}\rangle\
  \langle 1\ \r{or more digits}\rangle
\]
or
\[
  \langle 0\ \r{or more spaces}\rangle\
  \langle 1\ \r{or more digits}\rangle\w{.}
  \langle 1\ \r{or more digits}\rangle
\]

\medskip

\noindent
Please note that this definition does not allow
the following strings:
\setbox0\hbox{\verb*|1. 1|}
\[
  \w{1.} \qquad \qquad \w{.1} \qquad \qquad
  \box0 \qquad  \qquad \w{-1.1}
\]
Let us now state the problem:

\Problem{%
     Read the maximal prefix of the input string that may be a
prefix of a decimal number. Determine
whether this prefix is a decimal
number or not.}

     \Solution.  Let  us  write a program using Pascal's
``enumeration
type'' for clarity. (The variable
\w{state} may have one of the listed values.)
\begin{verbatim}
    var state:
     (Accept, Error, Initial, IntPart, DecPoint, FracPart);

    state := Initial;
    while (state <> Accept) or (state <> Error) do begin
    | if state = Initial then begin
    | | if Next = ' ' then begin
    | | | state := Initial; Move;
    | | end else if Digit(Next) then begin
    | | | state := IntPart;
    | | |   {after the start of the integer part}
    | | | Move;
    | | end else begin
    | | | state := Error;
    | | end;
    | end else if state = IntPart then begin
    | | if Digit (Next) then begin
    | | | state := IntPart; Move;
    | | end else if Next = '.' then begin
    | | | state := DecPoint; {after the decimal point}
    | | | Move;
    | | end else begin
    | | | state := Accept;
    | | end;
\end{verbatim}
\clearpage
\begin{verbatim}
    | end else if state = DecPoint then begin
    | | if Digit (Next) then begin
    | | | state := FracPart; Move;
    | | end else begin
    | | | state := Error; {at least one digit is needed}
    | | end;
    | end else if state = FracPart then begin
    | | if Digit (Next) then begin
    | | | state := FracPart; Move;
    | | end else begin
    | | | state := Accept;
    | | end;
    | end else if
    | | {this cannot happen}
    | end;
    end;
\end{verbatim}
Please note that the
assignments $\w{state}\,\w{:=}\,\w{Accept}$ and
$\w{state}\,\w{:=}\,\w{Error}$ are not accompanied by a call to procedure
\w{Move}, so the symbol after the end of the decimal number
is left unprocessed.\es

\bigskip

     This program does not store the  value of the number.

\Problem{%
     Add the following requirement to the preceding program: If a
processed part is a decimal number, its value should be placed into the
variable
\w{val:real}.}

     \Solution. While reading the fractional part, we use the variable
\w{scale} which is a factor for the digit to come ($0.1, 0.01$ etc.).
\begin{verbatim}
    state := Initial; val:= 0;
    while (state <> Accept) or (state <> Error) do begin
    | if state = Initial then begin
    | | if Next = ' ' then begin
    | | | state := Initial; Move;
    | | end else if Digit(Next) then begin
    | | | state := IntPart;
    | | |   {after the start of the integer part}
    | | | val := DigitVal(Next);
    | | | Move;
    | | end else begin
    | | | state := Error;
    | | end;
    | end else if state = IntPart then begin
\end{verbatim}
\clearpage
\begin{Verbatim}
    | | if Digit (Next) then begin
    | | | state := IntPart; val := 10*val + DigitVal(Next);
    | | | Move;
    | | end else if Next = '.' then begin
    | | | state := DecPoint; {after the decimal point}
    | | | scale := 0.1;
    | | | Move;
    | | end else begin
    | | | state := Accept;
    | | end;
    | end else if state = DecPoint then begin
    | | if Digit (Next) then begin
    | | | state := FracPart;
    | | | val := val + DigitVal(Next)*scale;
    | | | scale := scale/10;
    | | | Move;
    | | end else begin
    | | | state := Error; {at least one digit is needed}
    | | end;
    | end else if state = FracPart then begin
    | | if Digit (Next) then begin
    | | | state := FracPart;
    | | | val := val + DigitVal(Next)*scale;
    | | | scale := scale/10;
    | | | Move;
    | | end else begin
    | | | state := Accept;
    | | end;
    | end else if
    | | {this cannot happen}
    | end;
    end;
\end{Verbatim}

\Problem{%
     Repeat the previous problem if
the number may be optionally
preceded by \w{-}
or \w{+}.\es}

     The format of numbers in this problem can be represented
as follows:
%
%   -----      ---------
%---| + |---->-|    |-------->--------------------->
% | -----  | | --------- | |                      |
% | -----  | |           | | -----     ---------  |
% |-| - |--| |----<------| |-| . |->---|    |--|
% | -----  |                 -----   | --------- |
% |        |                         |-----<-----|
% |--->----|
%
{\small
\setlength{\unitlength}{1.1em}
\[
\begin{picture}(20,6)
\thinlines
\put(-3,4.5){\line(1,0){6}}
\put(-1,2.5){\line(1,0){1}}
\put(-1,2.5){\line(0,1){1}}
\put(0,2.5){\line(0,1){1}}
\put(-1,3.5){\line(1,0){1}}
\put(-2,3){\line(1,0){1}}
\put(0,3){\line(1,0){1}}
\put(-2,3){\line(0,1){1.5}}
\put(1,3){\line(0,1){1.5}}
\put(1,3){\vector(-1,0){0.7}}
\put(-2,4.5){\vector(1,0){1.5}}
\put(4,4.5){\line(1,0){3}}
\put(10,4.5){\line(1,0){12}}
\put(6,3){\line(1,0){5}}
\put(2,2.5){\line(1,0){1}}
\put(4,2.5){\line(1,0){1}}
\put(12,2.5){\line(1,0){1}}
\put(14,2.5){\line(1,0){2}}
\put(19,2.5){\line(1,0){2}}
\put(2,1){\line(1,0){3}}
\put(15,1){\line(1,0){5}}
%
\put(2,1){\line(0,1){3.5}}
\put(5,1){\line(0,1){3.5}}
\put(6,3){\line(0,1){1.5}}
\put(11,3){\line(0,1){1.5}}
\put(12,2.5){\line(0,1){2}}
\put(15,1){\line(0,1){1.5}}
\put(20,1){\line(0,1){1.5}}
\put(21,2.5){\line(0,1){2}}
%
\put(3,4){\framebox(1,1){\w{+}}}
\put(3,2){\framebox(1,1){\w{-}}}
\put(7,4){\framebox(3,1){$\langle\r{digit}\rangle$}}
\put(13,2){\framebox(1,1){\w{.}}}
\put(16,2){\framebox(3,1){$\langle\r{digit}\rangle$}}
%
\put(2,1){\vector(1,0){1.5}}
\put(4,4.5){\vector(1,0){1.7}}
\put(16,4.5){\vector(1,0){1.5}}
\put(11,3){\vector(-1,0){2.5}}
\put(20,1){\vector(-1,0){2.5}}
\end{picture}
\]
}

\Problem{%
     The same problem if the number may be followed by an integer exponent,
as in \w{254E-4} ($=0.0254$) or \w{0.123E+9}
($=123\,000\,000$). Draw the corresponding picture.\es}

\Problem{%
     What changes in the above program above are necessary to allow
empty integer or fractional parts like in \w{1.}, \w{.1} or even
\w{.} (the latter number is considered to be equal to zero)?\es}

     We return to finite-state algorithms (also called
{\it finite automata\/}) in section~\ref{pattern}.

