\subsection{Queues}\label{queue}

  Values of type ``queue of elements of type \w{T}'' are sequences
\index{queue}%
values of type \w{T}. The same is true for stacks, but
the difference is
that of queue elements are added to the beginning of a
sequence  and are taken from the end of it. Therefore,
an element that arrived
first to a queue will be the first element taken from it.
Hence the name First In First Out (FIFO), which is used for queues.
The method used for stacks is called Last In First Out (LIFO).%
 \index{FIFO}\index{LIFO}

\medskip

Operations on queues:

\begin{itemize}
\item   Make\_empty (\w{var x:} queue of elements of type \w{T});
\item   Add (\w{t:T}, \w{var x:} queue of elements of type \w{T});
\item   Take (\w{var t:T}, \w{var x:} queue of elements of type \w{T});
\item   Is\_empty (\w{x:} queue of elements of type \w{T}): \w{Boolean};
\item   First\_element (\w{x:} queue of elements of type \w{T}): \w{T}.
\end{itemize}

\noindent
     The procedure ``Add'' adds the specified element
to the end
of the queue. The procedure ``Take'' is applicable if the queue is
not empty; it puts the first element of the queue into a
variable \w{t}, removing it from the queue.
(The first element is the longest-waiting
element.)

The procedures ``Add'' and ``Take'' are often called
``Enqueue'' and ``Dequeue''.

\subsubsection*{Queue: array implementation}

\Problem{%
     Implement a queue of limited size in such
a way that all operations run in $O(1)$ time (that is, in time not
exceeding some constant, which does not depend on length of the queue).}%
 \index{queue!array implementation}%

     \Solution. Assume that queue elements are stored
as consecutive
elements in an array. The queue grows to the right and is taken from the
left. A growing queue may reach the end of the array, so we assume
the array is ``wrapped around'' in circular fashion.

     Our implementation uses an array
\begin{verbatim}
        Content: array [0..n-1] of T
\end{verbatim}
and variables
\begin{verbatim}
        First: 0..n-1
        Length : 0..n
\end{verbatim}
The queue is formed by elements
\begin{center}
\leavevmode
\w{Content [First],} \w{Content [First + 1]},\,\ldots,
      \w{Content [First+Length-1]}
\end{center}

\noindent
where addition is performed modulo \w{n}. (Warning: If you instead
use variables \w{First} and \w{Last} whose values are residues modulo
\w{n}, be careful not to mix the empty queue with the
queue containing \w{n} elements.)

\medskip

     The queue operations are implemented as follows:

\medskip

          Make\_empty:
\begin{verbatim}
        Length := 0;
        First := 0;
\end{verbatim}

         Add an element \w{t}:
\begin{verbatim}
        {Length < n}
        Content [(First + Length) mod n] := t;
        Length := Length + 1;
\end{verbatim}


     Take element into variable \w{t}:
\begin{verbatim}
        {Length > 0}
        t := Content [First];
        First := (First + 1) mod n;
        Length := Length - 1;
\end{verbatim}

     Is\_empty:
\begin{verbatim}
        Length = 0
\end{verbatim}


     First\_element:
\begin{Verbatim}
        Content [First]
\end{Verbatim}

\Problem{%
      (Communicated by A.G.~Kushnirenko)  Implement a queue using two stacks
(and a fixed number of variables of type \w{T}).
For
$n$ queue operations starting with an empty queue, the
implementation should perform
not more than $Cn$ stack operations.}%
 \index{queue!made of two stacks}%
 \index{Kushnirenko, A.G.}

     \Solution. We maintain the following
invariant relation: {\it stacks whose bottoms
are put together, form the queue.} (In other words, listing all
elements of one stack from top to bottom and then
of
the other stack
from bottom to top, we list all the queue elements in the proper
order.) To add an element to the queue,
it is enough to add it to one of the stacks. To check if the
queue is empty, we must check that both stacks are empty.
When taking the first element from the queue, we should distinguish
between two cases. If the stack that contains
the first element
is not empty,
there is no problem.
If that stack is empty, the
required element is buried under all the elements of the second
stack. In this case, we move all the elements one-by-one onto the
first stack (their ordering is reversed) and return to
the first case.

The number of operations
for this step is not limited by any
constant. However, the requirement posed in the problem is still
met. Indeed, any element of the queue can participate in such
a process at most once during its presence in the queue.\es

\Problem{%
     Deque (double-ended queue) is a structure that  combines
the properties of a queue and a stack: we can add and remove
elements from both ends of a deque. Implement a deque using
an array in such a way that each deque operation runs in
$O(1)$ time.\es}%
 \index{deque!array implementation}%
 \label{deque}

\Problem{%
     (Communicated by A.G.~Kushnirenko.) A deque of elements
of type \w{T} is given. The deque contains several elements.
The program should detemine how many elements are in the
deque. Program may use variables
of type \w{T} and integer variables,
but arrows are not allowed.}%
 \index{Kushnirenko, A.G.}

    \Hint{(1) We can perform a cyclic shift on deque
elements taking an element from one end and
adding it to the other
end. After $n$ shifts in one direction, we return the
deque to its initial state by $n$ shifts in the other direction.
(2) How do we know that the cycle is complete?
If we know in advance
that
some
element is guaranteed not to appear in the deque, this is
easy. We put this element into the deque and wait until it
appears at the other end. But we do not have such an element.
Instead, we may perform (for any fixed $n$) a cyclic shift by
$n$ positions twice adding two different elements. If the
elements that appear after the shift are
different, we have made a complete cycle.}

\subsubsection*{Queue applications}

\Problem{(see E.W.~Dijkstra's book [4])
     Print in increasing order the first \w{n} positive integers
whose factorization contains only the factors $2$, $3$, and $5$.}

       \Solution. The program uses three queues \w{x2}, \w{x3},
\w{x5}. They are used to store elements which are $2$, $3$,
and $5$
times larger than already printed elements, but are not yet printed. We
use
the procedure
 %
\begin{verbatim}
        procedure Print_and_add (t: integer);
        begin
        | writeln (t);
        | Add (2*t, x2);
        | Add (3*t, x3);
        | Add (5*t, x5);
        end;
\end{verbatim}

The program is as follows:

\begin{verbatim}
  .. make queues x2, x3, x5 empty
  Print_and_add (1);
  k := 1;
  {invariant relation: k first elements of the required
    set are printed; the queues contain elements
    that are 2, 3 and 5 times bigger that the elements
    already printed, but are not printed yet
    (in increasing order)}
  while k <> n do begin
  | x := min (Next(x2), Next(x3), Next(x5));
  | Print_and_add (x);
  | k := k+1;
  | .. take x from the queues where it was present;
  end;
\end{verbatim}

    Let us check the correctness of the program. Assume that the
invariant relation is valid and we perform the operations as
prescribed. Let \w{x} be the smallest element of our set
that is not printed. Then \w{x} is larger than $1$, and
it is
divisible by
$2$, $3$, or $5$.
The quotient belongs to
the set, too. The quotient is smaller than \w{x} and is therefore
printed. Thus \w{x} is present in one of the queues. It
is the smallest element in any queue to which \w{x} belongs
(because all the elements less than \w{x} are already printed
and cannot appear in any queue). When \w{x} is printed,
we must delete \w{x} from the queues
and add the corresponding multiples of \w{x} to maintain the
invariant.

     It is easy to check that queue lengths do not exceed the
number of elements printed.\es

\medskip

     The next problem is related to graphs (see
section~\ref{graphs} for
other graph problems).

     Let $V$ be a finite set whose elements are called
{\it vertices}. Let $E$ be a subset of the set $V\times V$; the
elements of $E$ are called {\it edges}. The sets $E$ and $V$ define
 \index{graph!directed}%
 \index{directed graph}%
 \index{graph!vertex}%
 \index{vertex of a graph}%
 \index{graph!connected}%
 \index{connected graph}%
 \index{cycle, Euler (along all the edges)}%
a {\it directed graph}. A pair $\langle p, q\rangle\in E$ is
called an edge {\it going from $p$ to $q$}. One says that
this edge {\it leaves\/} $p$ and {\it enters\/} $q$. Usually
vertices are drawn as points and edges as arrows. According to
the above definition, there is at most one edge from $p$ to $q$;
edges that are loops (from $p$ to $p$) are allowed.

A (directed) {\it path\/} is a sequence of vertices connected
by edges (for example, path $pqrs$ contains four vertices $p$, $q$,
$r$, and $s$, connected by three edges $\langle p,q\rangle$,
$\langle q,r\rangle$, and
$\langle r,s\rangle$.

\Problem{%
Suppose
a directed graph satisfies two requirements: (1) it is
connected, that is, there is a path from any given vertex to any
other vertex; and (2) for any vertex
the number of incoming edges is equal to the number of outgoing
edges. Prove there exists an edge cycle that traverses
each edge exactly once. Give an algorithm to find this cycle.}

     \Solution. A ``worm'' is a nonempty queue of vertices such that
each pair of adjacent vertices is connected by a graph edge (going
in the direction from the first element to the last element).
The first element in the queue is the tail of the worm; the last
element in the queue is the worm's head. The worm can be drawn as a
chain of arrows; arrows lead from the tail to the head. When a
vertex is added, the worm grows near the head; when a vertex is
removed, the tail is cut off.

     Initially, the worm consists of a single vertex and
evolves according to the following rule:

\begin{verbatim}
while the worm includes not all the edges do begin
| if there is an unused edge leaving from the worm's head
| |   then begin
| | add this edge to the worm
| end else begin
| | {the head and tail of the worm are at the same vertex}
| | cut a piece of tail and add it to the head
| | {"the worm eats its own tail"}
| end;
end;
\end{verbatim}

     Let us prove that this algorithm terminates when the worm
spans all edges with its head and tail at the
same vertex.

     (1) Traversing the worm from tail to head, we enter each
vertex as many times as we leave it. We also know that each
vertex has as many incoming edges as it has outgoing edges.
Therefore, we fail to find an outgoing edge only if the head of
the worm is located at the same vertex as its tail.

     (2) The worm never becomes shorter. Therefore, it will
eventually reach some maximal length and never grow again.
In the latter case, the worm will slide over itself forever. This
is possible only if all the vertices visited do not have
outgoing edges. Since the graph is connected,
this
is possible
only if all the edges are included in the worm.

     Some remarks about the Pascal implementation. The vertices
are numbered $\w{1..n}$. For each vertex \w{i}, we store
the number \w{Out[i]} of outgoing edges, as well as the numbers
\w{Num[i][1]},\,\dots,\,\w{Num[i][Out[i]]} of vertices receiving
the outgoing edges. While constructing the worm, we always choose
the first unused edge. In this case, it is enough to keep (for
each vertex) only the {\it number\/} of used outgoing edges to
find the first unused edge.\es


\Problem{%
     Prove that for any $n$ there exists a bit string $x$  of length
$2^n$ with the following property:
any binary string of length $n$ is a substring
of the string
$xxx\ldots$\,.
Find an algorithm that constructs such a binary string in time
$C^n$ (for some constant $C$ that does not depend on $n$).}%
 \index{string!having all possible substrings of length $n$}%

    \Hint{Consider a graph whose vertices are binary strings
of length $n-1$. An edge leaving $x$ and entering $y$ exists if
and only if there is a string $z$ of length $n$ such that $x$ is
a prefix of $z$ and $y$ is a suffix of $z$. (In other words, if
$x$ minus its first bit is equal to $y$ minus its last bit.)
This graph is connected; each vertex has two incoming and two
outgoing edges. A cycle that traverses all edges provides a
string satistying the desired property.}

\Problem{%
     Implement $k$ queues with total length not exceeding
$n$, using memory of size $O(n+k)$ (that is,
not exceeding $C(n+k)$ for
some constant $C$). Each operation (except for initialization,
which makes all the queues empty) should run in time $O(1)$ (that is,
limited by a constant that does not depend on $n$).}%
 \index{queue!pointer implementation}%

     \Solution. We use the same method as for the pointer
implementation of stacks. For each queue, remember the element
that is first to be served; each element of the queue
remembers the next element (the one that came immediately after).
The last element believes that the next one is a special
element number $0$. We also have to remember the last element of
each queue (otherwise we would trace the queue each time when a
new element is added). As for stacks, all the free places
are linked
into
a chain. Please note that for an empty queue the information
about the last element makes no sense and is not used when adding
elements.

\begin{verbatim}
        Content: array [1..n] of T;
        Next: array [1..n] of 0..n;
        First: array [1..k] of 0..n;
        Last: array [1..k] of 0..n;
        Free: 0..n;
\end{verbatim}


\begin{verbatim}
  procedure Make_empty;
  | var i: integer;
  begin
  | for i := 1 to n-1 do begin
  | | Next [i] := i + 1;
  | end;
  | Next [n] := 0;
  | Free := 1;
  | for i := 1 to k do begin
  | | First [i]:=0;
  | end;
  end;
\end{verbatim}

\begin{verbatim}
  function Is_space: Boolean;
  begin
  | Is_space := Free  <> 0;
  end;
\end{verbatim}

\begin{verbatim}
  function Is_empty (queue_number: integer): Boolean;
  begin
  | Is_empty := First [queue_number] = 0;
  end;
\end{verbatim}

\begin{verbatim}
  procedure Take (var t: T; queue_number: integer);
  | var frst: integer;
  begin
  | {not Is_empty (queue_number)}
  | frst := First [queue_number];
  | t := Content [frst]
  | First [queue_number] := Next [frst];
  | Next [frst] := Free;
  | Free := frst;
  end;
\end{verbatim}

\begin{verbatim}
  procedure Add (t: T; queue_number: integer);
  | var new, lst: 1..n;
  begin
  | {Is_space}
  | new := Free; Free := Next [Free];
  | {location new is removed from free space list}
  | if Is_empty (queue__number) then begin
  | | First [queue_number] := new;
  | | Last [queue_number] := new;
  | | Next [new] := 0;
  | | Content [new] := t;
  | end else begin
  | | lst := Last [queue_number];
  | | {Next [lst] = 0 }
  | | Next [lst] := new;
  | | Next [new] := 0;
  | | Content [new] := t
  | | Last [queue_number] := new;
  | end;
  end;
\end{verbatim}

\begin{Verbatim}
  function First_element (queue_number: integer): T;
  begin
  | First_element := Contents [First [queue_number]];
  end;
\end{Verbatim}

\Problem{%
     The same problem for deques.}%
 \index{deque!pointer implementation}%

    \Hint{A deque is a symmetric structure, so we should
keep pointers to both the next and preceding elements. It is convenient
to tie the ends of each deque with a special
element that forms a ``ring''. Another ring can be constructed
from the
free locations.}

     In the following problem, the deque is used to store the
vertices of a convex polygon.%
 \index{convex hull}%

\Problem{%
     Assume that $n$ points in the plane are numbered from left
to right (and when the $x$-coordinates coincide, according to the
order of the $y$-coordinates). Write a program that finds the convex
hull of these $n$ points in time $O(n)$ (that is, the number of
operations should not exceed $Cn$ for some constant $C$). The
convex hull is a polygon, so the answer should be a list of all its
vertices.}

     \Solution. Consider the points one by one, each time adding
a new point to the existing convex hull. The ordering guarantees that
the new point becomes one of the vertices of the convex hull. We call
this vertex of the convex hull a ``marked'' vertex. At the next step
the marked vertex is visible from the point to be added. We extend
our polygon by a ``needle'', which goes from the marked vertex to the new
point and back. We obtain a degenerate polygon and eliminate
``concavities'' in that polygon.

$$
  \setlength{\unitlength}{.8cm}
  \begin{picture}(11, 7)
    %  \put(0,0){\framebox(10,7){}}
       \thicklines
       \put(2,1){\circle*{.2}}
       \put(1,4){\circle*{.2}}
       \put(3,6){\circle*{.2}}
       \put(6,5){\circle*{.2}}
       \put(7,3){\circle*{.2}}
       \put(10,3){\circle*{.2}}
       \put(6,2){\circle*{.2}}
       \put(4,1){\circle*{.2}}

       \put(1,4){\line(1,1){2}}
       \put(3,6){\line(3,-1){3}}
       \put(6,5){\line(1,-2){1}}
   %   \thinlines
       \put(7,3.04){\line(1,0){3}}
       \put(7,2.96){\line(1,0){3}}
   %   \thicklines
       \put(7,3){\line(-1,-1){1}}
       \put(6,2){\line(-2,-1){2}}
       \put(4,1){\line(-1,0){2}}
       \put(2,1){\line(-1,3){1}}

       \thinlines
       \put(6,5){\line(2,-1){4}}
       \put(4,1){\line(3,1){6}}
  \end{picture}
$$

     The program stores the vertices of a polygon in a deque
listed counter-clockwise from the ``head'' to the ``tail''.
The marked vertex is both
the head and the tail of the deque. Adding a ``needle'' means
that new vertex is added to both ends of the deque.
The elimination of
concavities is more difficult. Let us call
the elements nearest the head the ``subhead'' and ``subsubhead'',
respectively.
The elimination of concavities near the head is done as follows:
\begin{verbatim}
    while going from the head to the subsubhead we turn
    |   to the right near the subhead  do begin
    | remove the subhead from the deque
    end
\end{verbatim}
The concavity near the tail is eliminated in a similar way.

    Remark. Strictly speaking, operations involving
the subhead and subsubhead of a deque
are not allowed
by definition. However, they may be reduced to
a few legal operations (for example, we can
take three elements,
process them, and put back what remains).

    Another remark. Two degenerate cases are possible. The first
occurs when
we do not turn at all near the subhead (in this case, the
three vertices lie on the same
line); the second occurs when we make a $180^{\circ}$ turn
(this happens when we have a ``polygon'' with two vertices).   In the
first case, the subhead should be removed (to eliminate the redundant
vertices from the convex hull); in the second case, the deque
is left unchanged.\es

