\subsection{Sets}

     Let \w{T} be a type. There are several methods to store (finite)
sets of values of type \w{T}. There is no ``best'' method; the choice
depends on type \w{T} and on the operations needed.%
 \index{set!data types}%

\subsubsection*{Subsets of $\{1,\ldots,n\}$}

\Problem{%
     Using $O(n)$ space (that is, space
proportional to  $n$), store a subset
of
$\{1,\ldots,n\}$.}%
 \index{set!bit array implementation}%
$$
\begin{tabular}{lc}
        Operations               &        Time \\[.5ex]
\hline
        Make empty               &     $Cn$ \\
        Test membership         &    $C$\\
        Add                      &    $C$\\
        Delete                   &    $C$\\
        Minimal element          &    $Cn$\\
        Test if the set is empty  &    $Cn$
\end{tabular}
$$

     \Solution. Store the set as \w{array [1..n] of Boolean}.\es

\Problem{%
     The same problem with an additional requirement:
     test if the set is empty in constant (i.e., $O(1)$) time.}

     \Solution. Store the number of elements in an additional
     variable.\es

\Problem{%
     The same problem with the following restrictions:}
$$
\begin{tabular}{lc}
        Operations     &        Time \\[.5ex]
\hline
        Make empty             &     $Cn$ \\
        Test membership        &    $C$\\
        Add                     &    $C$\\
        Delete                  &    $Cn$\\
        Minimal element         &    $C$\\
        Test if the set is empty  &    $C$
\end{tabular}
$$

     \Solution. Maintain also the minimal element of the set.\es

\Problem{%
     The same problem with the following restrictions:}
$$
\begin{tabular}{lc}
         Operations    &        Time \\[.5ex]
\hline
        Make empty             &     $Cn$ \\
        Test membership        &    $C$\\
        Add                     &    $Cn$\\
        Delete                  &    $C$\\
        Minimal element         &    $C$\\
        Test if the set is empty &    $C$
\end{tabular}
$$

       \Solution. Store the minimal element of the set. Also,
for each
element we maintain pointers to the next and preceding elements
(in order determined by value).\es

\subsubsection*{Sets of integers}

    In the following problems, elements of the
set are integers (unbounded);
the number of elements does not exceed $n$.

\Problem{%
       The memory size is limited by $Cn$.}%
 \index{set!list implementation}%
$$
\begin{tabular}{lc}
        Operations                       &    Time\\[.5ex]
\hline
        Make empty                       &    $C$ \\
        Cardinality                      &    $C$ \\
        Test membership                  &    $Cn$\\
        Add element (known to be absent) &    $C$ \\
        Delete                           &    $Cn$\\
        Minimal element                  &    $Cn$\\
        Take some element                &    $C$
\end{tabular}
$$

     \Solution. The set is represented by the variables
\begin{verbatim}
         a:array [1..n] of integer, k: 0..n;
\end{verbatim}
The set contains \w{k} (distinct)
elements $\w{a[1]},\ldots,\w{a[k]}$.
In a sense, we keep the elements of the set in a stack. (We require
all elements in the stack to be different.) We may also use a queue
instead of a stack.\es

\Problem{%
       The memory size is limited by $Cn$.}%
$$
\begin{tabular}{lc}
        Operations                &        Time  \\[.5ex]
\hline
        Make empty                &        $C$ \\
        Test if the set is empty &        $C$ \\
        Test membership          &    $C\log n$\\
        Add                       &    $Cn$\\
        Delete                    &    $Cn$\\
        Minimal element           &    $C$
\end{tabular}
$$

     \Solution. We use the same representation as in the preceding problem,
with the additional restriction
$\w{a[1]}<\cdots<\w{a[k]}$. To test membership, we use binary search.\es

\medskip

     In the following problem, different methods are
combined.

\Problem{%
     Find all the vertices of a directed graph that can be reached from
a given vertex along the graph edges.
The program should run in time
$Cm$, where $m$ is the total number of edges leaving the
reachable vertices.}%
 \index{connected component!directed graph}%
 \index{graph!connected component}%

     \Solution. (A recursive solution is given in section~\ref{recursion}.)
Let
\w{num[i]} be the number of outgoing edges for vertex
\w{i} (assume that vertices are numbered $\w{1..n}$).
Let \w{out[i][1]},\,\ldots, \w{out[i][num[i]]} be the endpoints of the
edges starting from vertex
\w{i}.

\begin{verbatim}
  procedure Print_Reachable (i: integer);
  |   {print all the vertices reachable from i,
  |       including the vertex i itself}
  | var  X: subset of 1..n;
  |      P: subset of 1..n;
  |      q, v, w: 1..n;
  |      k: integer;
  begin
  | ...make X and P empty;
  | writeln (i);
  | ...add i to X, P;
  | {(1) P is the set of printed vertices; P contains i;
  |  (2) only vertices reachable from i are printed;
  |  (3) X is a subset of P;
  |  (4) all printed vertices which have an outgoing edge to
  |      a non-printed vertex, belong to X}
\end{verbatim}
\clearpage
\begin{verbatim}
  | while X is not empty do begin
  | | ...take some element of X into v;
  | | for k := 1 to num [v] do begin
  | | | w := out [v][k];
  | | | if w does not belong to P then begin
  | | | | writeln (w);
  | | | | add w to P;
  | | | | add w to X;
  | | | end;
  | | end;
  | end;
  end;
\end{verbatim}

     Let us check that the requirements (1)--(4) mentioned in the
program text,
are  satisfied.
(1)~We  print  a number and simultaneously
add it to $P$. (2)~Since \w{v} is in \w{X},
\w{v} is reachable; therefore, \w{w} is
reachable. (3)~Obvious. (4)~We have deleted  \w{v}  from  \w{X},
but all the endpoints of edges emanating from \w{v} are printed.

     Let us prove the upper bound for the number of operations.
If some element is removed from \w{X}, it
never
appears in
\w{X} again. Indeed, it was present in \w{P} when removed, and
only elements not in \w{P} can be added. Therefore the body of
the
while-loop is executed at most once for any reachable
vertex; the number of iterations of the
for-loop is equal to the
number of outgoing edges.

     For \w{X} we use a stack or queue representation (see
above); for \w{P} we use a Boolean array.\es

\Problem{%
     Solve the preceding problem if all the reachable vertices are
to be printed in the following order: first the given vertex,
then its neighbors, then (unprinted) neighbors of its
neighbors, etc.}%
\label{breadth}%
 \index{search!breadth-first}%

     \Hint{Use a queue for the representation of the
set \w{X} in the program above. By induction over $k$ we prove that
at some point all the vertices having distance not exceeding
$k$ (and no others) are printed, and all the vertices having distance
exactly $k$ (and no others) are in the queue. For the detailed solution
see
section~\ref{connectedcomponents}, p.~\ref{breadthfirst}
}

More elaborate data structures for sets are considered in
sections~\ref{hash}~(hash tables) and \ref{trees}~(trees).

\subsection{Priority queues}
\label{priorityqueue}

\Problem{%
     Implement a data structure that has the same set of
operations
as an array of length \w{n}, namely,}%
 \index{array!with minimal element}%
\begin{itemize}
\item   initialize;
\item   put \w{x} in the \w{i}-th cell;
\item   find the contents of the \w{i}-th cell;
\end{itemize}
as well as the operation
\begin{itemize}
\item find the index of the minimal element
\end{itemize}
(or one of the minimal elements). Any operation should
run in time $C\log \w{n}$ (except  initialization, which
should run in time
$C\w{n}$).\label{minarray}

     \Solution. We use the trick from the  heapsort algorithm. Assume
that the
array elements are positioned at the leaves of a binary
tree and each
non-leaf
vertex contains the minimum of its two sons. To
maintain this information and to trace the path from the root to
the minimal element, we need only $C\log \w{n}$ operations.\nobreak\es

\Problem{%
      A priority queue does not employ First In First Out (FIFO)
rule;
only an element's priority is important. An element
is added to the priority queue with some priority
(which is assumed to be
an
integer). When an element is taken from the queue, it is the
element with the greatest priority (or one of the elements with
greatest priority). Implement a priority queue in such a way
that adding and removing elements requires logarithmic (in the
size of the queue) time.}%
 \index{queue!priority}%
 \index{priority queue}%

    \Solution. Here we follow the idea of
the
heapsort algorithm%
 \index{sorting!heapsort}
in its final form. We place queue elements in an array
\w{x[1..k]} and maintain the following invariant
relation:
\w{x[i]} is higher
(has larger priority) than its sons \w{x[2i]} and \w{x[2i+1]},
if they exist. (Therefore, each element is higher than all its
descendants.) The priority information is maintained along with
the elements in the array, so we have an array of pairs
$\langle\r{element}, \r{priority}\rangle$. From the heapsort
algorithm, we know how to delete an element and maintain this relation.
Another thing we need to do
is restore this relation after adding
some element to the end of the array. This is done as follows:

\begin{verbatim}
    t:= the number of element added
    {invariant: any element is higher than any its
       descendant if the descendant is not t}
    while t is not root and t is higher than its father
    |   do begin
    | exchange t and its father
    end;
\end{verbatim}

Suppose the priority queue is formed by people standing
at the vertices of a tree (drawn on the ground); each person
has one
predecessor and at most two successors. The idea of the
algorithm is this:
A highly-ranked individual added to the queue begins to move
toward the head of the queue. If a predecessor has lower rank,
this new individual takes the predecessor's place. This
continues until a higher-ranked predecessor is encountered.\es

     Remark. The priority queue is an important data structure
in simulation. Indeed, events are taken to be queue elements
where the priority is determined by the time planned for the event.
 \index{simulation, event queue}%

