\section{Recursion}\label{recursion}

\subsection{Examples}

Up to now, we have not use recursion in our examples.
Instead we devote a special section to this important
programming technique. Recursion can be very useful and
convenient, and in some cases recursive solutions are
much shorter than nonrecursive ones.

When a recursive procedure is given, we have to show that:
 \index{recursive procedure}
 \index{recursion}
\begin{description}
\item[(a)] the procedure terminates;
\item[(b)] the procedure works properly (assuming it terminates).
\end{description}

     Let us start with (b). Here it is enough to check that a procedure
containing a recursive call works properly assuming that the called
program (with the same name) works properly. Indeed, in this case,
all the programs in the chain of recursive calls
(from the end of the chain to its beginning) work properly.
In other words, a
recursive procedure is proved correct by induction.

     To prove (a) we usually find a parameter that decreases
as the recursion depth increases
and prove that it cannot
decrease indefinitely.

\Problem{%
       Write a recursive program that computes the factorial
of a positive integer $n$ (i.e., the product
$n!=1\cdot2\cdots n$).}%
 \index{factorial!recursive program}

     \Solution. We apply the relations $1!=1$ and $n!= (n-1)!\cdot n$
for $n>1$.
\begin{Verbatim}
        procedure factorial (n: integer; var fact: integer);
        | {fact := n!}
        begin
        | if n=1 then begin
        | | fact:=1;
        | end else begin {n>1}
        | | factorial (n-1, fact);
        | | {fact = (n-1)!}
        | | fact:= fact*n;
        | end;
        end;
\end{Verbatim}
Using Pascal functions, we may write the above procedure as follows:
\begin{verbatim}
        function factorial (n: integer): integer;
        begin
        | if n=1 then begin
        | | factorial := 1;
        | end else begin {n>1}
        | | factorial := factorial(n-1)*n;
        | end;
        end;
\end{verbatim}
Please note that the identifier
\w{factorial}
has two different meanings.
It is a
local variable as well as a
function name. Fortunately, the difference
is clear because the function name has parentheses after it. However,
in the
case of a function without parameters we have to be careful.
(A common error occurs as the programmer
sees a variable whereas the
compiler sees a recursive call. This error is sometimes
difficult to find.)

\Problem{%
    The factorial of $0$
is defined as
$0!=1$. Modify the program accordingly.\nobreak\es}

\Problem{%
    Write a recursive program that
computes the nonnegative integral
power of a real number.\es}

\Problem{%
    Repeat the previous problem
with the requirement: the recursion
depth (number of recursion levels) should not exceed
$C\log n$, where $n$ is the exponent.}%
 \index{power!recursive computation}

    \Solution.
\begin{Verbatim}
        function power (a,n: integer): integer;
        begin
        | if n = 0 then begin
        | | power:= 1;
        | end else if n mod 2 = 0 then begin
        | | power:= power(a*a, n div 2);
        | end else begin
        | | power:= power(a, n-1)*a;
        | end;
        end;
\end{Verbatim}

\Problem{%
     What happens if we replace the line}
\begin{verbatim}
        power:= power(a*a, n div 2)
\end{verbatim}
in the above program
by the line
\begin{verbatim}
        power:= power(a, n div 2)* power(a, n div 2)
\end{verbatim}

     \Solution. The program is still correct, but becomes much
slower. In this case, one
call of the function \w{power} generates {\it
two\/} calls of the same function (with identical parameters).
Thus, the number of calls grows exponentially as a function of
the recursion depth. The program still has logarithmic
recursion depth, but the number of steps is now linear instead
of logarithmic.\es

     This difficulty can be avoided by writing
\begin{verbatim}
        t:= power(a, n div 2);
        power:= t*t;
\end{verbatim}
or by using Pascal's square function (\w{sqr}).

\Problem{%
     Using the procedure \w{write(x)} for $\w{x}=0\ldots9$,
write a recursive procedure that prints the decimal representation
of a positive integer
$n$.}%
 \index{decimal number!printing, recursive}

    \Solution. The recursive solution allows us to produce
digits
from right to left but
print them from left to right:
\begin{Verbatim}
     procedure print (n:integer); {n>0}
     begin
     | if n<10 then begin
     | | write (n);
     | end else begin
     | | print (n div 10);
     | | write (n mod 10);
     | end;
     end;
\end{Verbatim}

\Problem{%
     The ``Towers of Hanoi'' puzzle
consists of three vertical sticks
and $N$ rings of different sizes. The rings are put on one of the
sticks in such a way that larger rings are beneath smaller ones.
We are to move this tower onto another stick
one ring at a time.
While moving the rings
from one stick to another, we are not allowed to
put a larger ring
onto a smaller one. Write a program that shows the list of
movements required to solve the problem.}%
 \index{Hanoi, Towers of!recursive solution}%
 \index{Towers of Hanoi!recursive solution}%
 \label{hanoitowers}

     \Solution. The following recursive procedure moves \w{i}
upper rings from the \w{m}-th stick to the \w{n}-th stick
(we assume that the remaining rings are larger and remain
untouched).
\begin{verbatim}
  procedure move(i,m,n: integer);
  | var s: integer;
  begin
  | if i = 1 then begin
  | | writeln ('move ', m, '->', n);
  | end else begin
  | | s:=6-m-n; {s is the third stick; 1+2+3 = 6}
  | | move (i-1, m, s);
  | | writeln ('move ', m, '->', n);
  | | move (i-1, s, n);
  | end;
  end;
\end{verbatim}
(The first recursive call
moves a tower of $\w{i-1}$ rings onto the third
stick. After that the \w{i}-th ring is free and is moved to the
remaining stick.
The second recursive call
moves the tower onto the \w{i}-th ring.)\es

