\subsection{Trees: recursive processing}

     A binary tree is represented by
a picture like this:
$$
\setlength{\unitlength}{7mm}
\begin{picture}(5,8)

\put(2,1){\circle{.85}}
\put(1,3){\circle{.85}}
\put(3,3){\circle{.85}}
\put(2,5){\circle{.85}}
\put(4,5){\circle{.85}}
\put(1,7){\circle{.85}}

\put(1.8,1.4){\line(-1,2){.6}}
\put(2.8,3.4){\line(-1,2){.6}}
\put(1.8,5.4){\line(-1,2){.6}}
\put(2.2,1.4){\line(1,2){.6}}
\put(3.2,3.4){\line(1,2){.6}}

\end{picture}
$$
%                  o
%                   \
%                    o   o
%                     \ /
%                  o   o
%                   \ /
%                    o
%
The vertex at the bottom of the tree is called the
{\it root}.
Two lines may go up from any vertex:
one going
up-left and
one going
up-right.
These two vertices are called the
{\it left\/} and {\it right sons\/} of the given vertex.
 \index{tree!root}%
 \index{tree!vertex}%
 \index{descendant of a vertex}%
Any given
vertex may have
either
two sons, one son (which may be either the left
son or the right son),
or no sons at all. In the latter case, the vertex
is called a {\it leaf}.

     Let $x$ be a vertex of tree. Consider this vertex together
with its sons, grandsons, etc. This is a
{\it subtree rooted at\/} $x$,
the subtree of all {\it descendants} of the vertex $x$.

Please note that in most textbooks trees have root at the
top and grow downwards. In some books terms ``son'' (``father'',
``brother'', etc.) are replaced by ``child'' (``parent'',
``siebling'', etc.).

     In the following set of
problems tree vertices are
numbered by positive integers, and all numbers are different.
The number assigned to the tree root is kept in the
variable \w{root}. There exist two arrays
\begin{verbatim}
        l,r: array [1..N] of integer
\end{verbatim}
The left and right sons of the
vertex number \w{i} have numbers
\w{l[i]} and \w{r[i]}, respectively. If vertex $x$ has no left
(or right) son, the value of \w{l[i]} (resp., \w{r[i]}) is equal
to \w{0}. (Traditionally, we use the symbolic constant
 \index{tree!pointer implementation}
\w{nil} instead of the literal \w{0}.)
Numbers of
all vertices do not exceed \w{N}.

Let us stress that the vertex number
has no connection with its position in a tree and that some integers
in $\w{1}\ldots\w{N}$ are not assigned to
vertices at all. (Therefore,
some data in the arrays \w{l} and \w{r} are irrelevant.)

\Problem{%
    Assume that $\w{N}=7$, $\w{root}=3$, and the arrays
\w{l} and \w{r}
are as follows:}
$$
\begin{tabular}{r|ccccccc}
       \w{i}&   1& 2& 3& 4& 5& 6& 7\\
    \w{l[i]}&   0& 0& 1& 0& 6& 0& 7\\
    \w{r[i]}&   0& 0& 5& 3& 2& 0& 7
\end{tabular}
$$
Draw the corresponding tree.

%   Answer:          6   2
%                     \ /
%                  1   5
%                   \ /
%                    3

\medskip

    Answer:
$$
\def\cntr#1{\hbox%
             {%
                \hskip-.25em\lower.66ex\hbox{$#1$}%
             }
           }
\setlength{\unitlength}{7mm}
\begin{picture}(5,6)

\put(2,1){\circle{.85}}
\put(1,3){\circle{.85}}
\put(3,3){\circle{.85}}
\put(2,5){\circle{.85}}
\put(4,5){\circle{.85}}

\put(2,1){\cntr3}
\put(1,3){\cntr1}
\put(3,3){\cntr5}
\put(2,5){\cntr6}
\put(4,5){\cntr2}

\put(1.8,1.4){\line(-1,2){.6}}
\put(2.8,3.4){\line(-1,2){.6}}
\put(2.2,1.4){\line(1,2){.6}}
\put(3.2,3.4){\line(1,2){.6}}

\end{picture}
\eqno\es
$$

\Problem{%
     Write a program that counts all the vertices in a given tree.}

 \index{tree!number of vertices}
 \index{tree!recursive processing}
     \Solution. Consider a function \w{n(x)}, which is defined as the
number
of vertices in the subtree rooted at vertex number
\w{x}. We agree that
$\w{n(nil)}=\w{0}$ (and the corresponding subtree is empty) and ignore
the values \w{n(s)} for \w{s} not assigned to any vertex.
Here is a recursive program that computes
\w{n(x)}:
\begin{verbatim}
     function n(x:integer):integer;
     begin
     | if x = nil then begin
     | | n:= 0;
     | end else begin
     | | n:= n(l[x]) + n(r[x]) + 1;
     | end;
     end;
\end{verbatim}
(Vertices in the
\w{x}-subtree are vertices in the
subtrees rooted at its
sons plus the vertex \w{x} itself.) The procedure terminates because
the recursive calls refer to trees of smaller heights.\es

\Problem{%
     Write a program that counts the leaves in a tree.}%
 \index{tree!number of leaves}

    \Solution.
\begin{Verbatim}
   function n (x:integer):integer;
   begin
   | if x = nil then begin
   | | n:= 0;
   | end else if (l[x]=nil) and (r[x]=nil) then begin {leaf}
   | | n:= 1;
   | end else begin
   | | n:= n(l[x]) + n(r[x]);
   | end;
   end;
\end{Verbatim}

\Problem{%
     Write a program that finds the height of a tree. (The root
of a tree has depth $0$, its sons have depth $1$, its grandsons
have depth $2$, etc. The height of a tree is the maximal depth
of its vertices.}%
 \index{tree!height}

   \Hint{Let $\w{h(x)}$ be the height
of the
subtree rooted at \w{x}. The function \w{h} may be
defined recursively.}

\Problem{%
     Write a program which for a given \w{n} counts all the
vertices of depth \w{n} in a given tree.\es}

\smallskip

     Instead of counting vertices, we may ask to list them (in some
order).

\Problem{%
     Write a program that prints all vertices (one time each).}%
 \index{tree!traversal}

    \Solution. The procedure  \verb|print_subtree(x)|
prints all the vertices of the subtree rooted at
\w{x} (one time each). The main program consists of the call
\verb|print_subtree(root)|.
\begin{verbatim}
     procedure print_subtree (x:integer);
     begin
     | if x = nil then begin
     | | {nothing to do}
     | end else begin
     | | writeln (x);
     | | print_subtree (l[x]);
     | | print_subtree (r[x]);
     | end;
     end;
\end{verbatim}
This program uses the following ordering of tree vertices: first
the root, then the left subtree, and then the right subtree.
This order is determined by the order of the three lines in
the
\w{else}-part. Any of six possible permutations of these lines
gives a specific order of tree traversal.\es

