\subsection{The generation of combinatorial objects; search}

     Recursion is a convenient tool to write programs that generate
elements of some
finite
set.
As an example,
we now return to the problems
of section~\ref{generation}.

\Problem{%
     Write a program that prints all sequences of length
\w{n}  composed of the numbers
$\w{1..k}$. (Each sequence should be printed once, so
the program prints
$\w{k}^{\w{n}}$ sequences.)}%

     \Solution. The program employs an array
$\w{a[1]..a[n]}$ and an integer variable
\w{t}. The recursive procedure
\w{generate} prints all sequences with prefix
$\w{a[1]..a[t]}$; after it terminates, the value of
\w{t} and
$\w{a[1]..a[t]}$ are the same as before the call.
\begin{verbatim}
     procedure generate;
     | var i,j : integer;
     begin
     | if t = n then begin
     | | for i:=1 to n do begin
     | | | write(a[i]);
     | | end;
     | | writeln;
     | end else begin {t < n}
     | | for j:=1 to k do begin
     | | | t:=t+1;
     | | | a[t]:=j;
     | | | generate;
     | | | t:=t-1;
     | | end;
     | end;
     end;
\end{verbatim}
The main program body now consists of two lines:
\begin{verbatim}
     t:=0;
     generate;
\end{verbatim}

     Remark. For efficiency reasons we may move the commands
\w{t:=t+1} and \w{t:=t-1} out of the \w{for}-loop.\es

\Problem{%
     Write a program that prints all permutations of the
numbers
$\w{1..n}$ (each should be printed once).}%
 \index{permutations!generation}

     \Solution. The program utilizes an array
$\w{a[1]..a[n]}$ that contains a permutation of numbers
$\w{1..n}$. The recursive procedure \w{generate}
prints all permutations that have the same first \w{t}
elements as the permutation \w{a}. After the call, the values
of \w{t} and \w{a} are the same as before the call.
The main program is:
\begin{verbatim}
    for i:=1 to n do begin a[i]:=i; end;
    t:=0;
    generate;
\end{verbatim}
The procedure definition is as follows:
\begin{verbatim}
     procedure generate;
     | var i,j : integer;
\end{verbatim}
\clearpage
\begin{Verbatim}
     begin
     | if t = n then begin
     | | for i:=1 to n do begin
     | | | write(a[i]);
     | | end;
     | | writeln;
     | end else begin {t < n}
     | | for j:=t+1 to n do begin
     | | | ..exchange a[t+1] and a[j]
     | | | t:=t+1;
     | | | generate;
     | | | t:=t-1;
     | | | ..exchange a[t+1] and a[j]
     | | end;
     | end;
     end;
\end{Verbatim}

\Problem{%
     Print all increasing sequences of length \w{k} constructed from
the natural
numbers $\w{1..n}$.
(We assume that $\w{k}\le \w{n}$; otherwise the sequences
do not exist.)}%
 \index{monotone sequences!generation}

    \Solution. The program utilizes an array
$\w{a[1]..a[k]}$ and integer variable \w{t}. Assuming
that $\w{a[1]..a[t]}$ is an increasing sequence whose
terms are numbers in $\w{1..n}$, the recursive procedure
\w{generate} prints all its increasing extensions of length
\w{k}. (After the call, the values of \w{t} and
$\w{a[1]..a[t]}$ are the same as before the call.)
\begin{verbatim}
     procedure generate;
     | var i: integer;
     begin
     | if t = k then begin
     | | ...print a[1]..a[k]
     | end else begin
     | | t:=t+1;
     | | for i:=a[t-1]+1 to t-k+n do begin
     | | | a[t]:=i;
     | | | generate;
     | | end;
     | | t:=t-1;
     | end;
     end;
\end{verbatim}

     Remark. The \w{for}-loop may use \w{n} instead of
$\w{t}-\w{k}+\w{n}$. The above version is more efficient; we
use that the $\w{(k-1)}$-th term cannot exceed
$\w{n-1}$, the $\w{(k-2)}$-th term cannot exceed $\w{n-2}$, etc.

\clearpage

     The main program:
\begin{verbatim}
        t:=1;
        for j:=1 to 1-k+n do begin
        | a[1]:=j;
        | generate;
        end;
\end{verbatim}
(Another possibility is to add to \w{a}
an auxiliary
element
$\w{a[0]:=0}$, then let $\w{t:=0}$ and call the procedure
\w{generate} once.)\es

\Problem{%
      Generate all representations of a given positive integer
\w{n} as the sum of a nonincreasing sequence of positive
integers.}%
 \index{partitions!generation}

    \Solution. The program uses an array \w{a[1..n]}
(the maximal number of summands is \w{n}) and an integer variable
\w{t}. The procedure \w{generate} assumes that
$\w{a[1]}\ldots\w{a[t]}$ is a nonincreasing
sequence of positive integers whose sum does not exceed \w{n},
and prints all the representations that extend this sequence.
For efficiency reasons, the sum
$\w{a[1]}+\cdots+\w{a[t]}$ is kept in an auxiliary
variable~\w{s}.
\begin{verbatim}
     procedure generate;
     | var i: integer;
     begin
     | if s = n then begin
     | | ...print a[1]..a[t]
     | end else begin
     | | for i:=1 to min(a[t], n-s) do begin
     | | | t:=t+1;
     | | | a[t]:=i;
     | | | s:=s+i;
     | | | generate;
     | | | s:=s-i;
     | | | t:=t-1;
     | | end;
     | end;
     end;
\end{verbatim}
The main program looks like
\begin{verbatim}
     t:=1;
     for j:=1 to n do begin
     | a[1]:=j
     | s:=j;
     | generate;
     end;
\end{verbatim}

     Remark. A small improvement is possible, since we
may move the statements that
increase and decrease $\w{t}$ out of the loop. Also, instead of
setting the value of \w{s} each time (\w{s:=s+i})
and restoring it (\w{s:=s-i})
we may increase it by
\w{1} at each time through the loop
and restore the original value at the end of loop.
Finally, we may add an auxiliary element
$\w{a[0]}=\w{n}$ and simplify the main program:
\begin{Verbatim}
     t:=0; s:=0; a[0]:=n; generate;
\end{Verbatim}


\Problem{%
     Write a recursive program that traverses a tree (using the same
statements and conditions as in section~\ref{traversal}).}%
 \index{backtracking!recursive}
 \index{tree!traversal}

    \Solution. The procedure \verb|process_above| processes all the
leaves above the robot's position and returns the robot
to the start position. Here is the recursive definition:
\begin{Verbatim}
     procedure process_above;
     begin
     | if is_up then begin
     | | up_left;
     | | process_above;
     | | while is_right do begin
     | | | right;
     | | | process_above;
     | | end;
     | | down;
     | end else begin
     | | process;
     | end;
     end;
\end{Verbatim}

