\subsection{Other applications of recursion}

    {\bf Topological sorting}. Imagine $n$ government officials,
each of whom issues permissions of some type. We wish to obtain
all
the permissions (one from each official)
according to
the rules. The rules state (for each
official) a list of permissions that should be collected before
you can obtain
this permission. There is no hope of solving the problem
if the dependency
graph has a cycle (we cannot get permission from $A$
without having $B$'s permission
in advance, $B$ without $C$,$\,$\ldots, $Y$ without $Z$, and $Z$
without $A$). Assuming that such a cycle does not exist, we wish
to find a plan that secures one of the permitted
orders.

    Let us represent officials by points and dependencies by
arrows. (If permission $B$ should be obtained before $A$,
draw an arrow going from $A$ to $B$.)
We then have the following problem. There are $n$
points numbered from $1$ to $n$. From each point there are
several (maybe zero)
arrows that go to other points. (This
picture is called a {\it directed graph}.) The graph has
no cycles.
We want to put the graph vertices in such an order that the end
of any arrow precedes its beginning. This is the problem of {\it
topological sorting}.%
 \index{sorting!topological}%
 \index{topological sorting}

\Problem{%
     Prove that it is always possible to topologically sort
a directed graph without cycles.}

    \Solution. The absence of cycles implies that there exists a
vertex with no outgoing edges (otherwise, we can follow edges until
we come to the already visited vertex).  This vertex with no outgoing
edges gets number $1$. After we discard all the  edges  entering
vertex number $1$, we reduce our problem to the same problem
with a smaller number of edges.\es

\Problem{%
     Assume that a directed graph without cycles is stored in
the following manner:
Its vertices are numbered $\w{1..n}$.
For any \w{i} in $\w{1..n}$, the
value of \w{num[i]} is the number of edges leaving vertex \w{i},
and $\w{adr[i][1]},\ldots, \w{adr[i][num[i]]}$ are the numbers of
vertices those edges enter. Write a (recursive) algorithm
that performs a topological
sort
in time $C\cdot(\w{n}+\w{m})$,
where \w{m} is the number of edges (arrows) in the
graph.}\label{rectopsort}

     Remark. The solution to the preceding problem does not
provide
such an
algorithm directly; we need a more ingenious construction.

    \Solution. Our program prints the vertices
in question (their numbers).
It uses an array
\begin{verbatim}
            printed: array[1..n] of Boolean;
\end{verbatim}
\leavevmode
such that
\w{printed[i]} is true if and only if
vertex \w{i} is
already
printed (this information is updated each time
a vertex is printed).
We say that a sequence of printed
vertices is {\it correct\/} if ({\bf a}) no vertex
is printed twice, and ({\bf b}) for any printed vertex \w{i}
all the edges leaving \w{i} enter the vertices that
are printed before \w{i}.
\begin{verbatim}
   procedure add (i: 1..n);
   | {before: the sequence of printed vertices is correct}
   | {after: the sequence of printed vertices is correct
   |         and includes i}
   begin
   | if printed [i] then begin {i is printed already}
   | | {nothing to do}
   | end else begin {printed sequence is correct}
   | | for j:=1 to num[i] do begin
   | | | add(adr[i][j]);
   | | end;
   | | {printed sequence is correct; all the edges going out
   | |  of i are entering the vertices already printed; thus,
   | |  we may print i correctly if it is not printed yet}
   | | if not printed[i] then begin
   | | | writeln(i); printed [i]:= TRUE;
   | | end;
   | end;
   end;
\end{verbatim}
\clearpage
The main program is:
\begin{verbatim}
     for i:=1 to n do begin
     | printed[i]:= FALSE;
     end;
     for i:=1 to n do begin
     | add(i)
     end;
\end{verbatim}
\noindent The time bound will be proven shortly.

\Problem{%
     The program above can be simplified if we remove the test,
replacing}
\begin{verbatim}
          if not printed[i] then begin
          | writeln(i); printed [i]:= TRUE;
          end;
\end{verbatim}
by
\begin{verbatim}
          writeln(i); printed [i]:= TRUE;
\end{verbatim}
(Why?) How should we change the specification of the procedure?

    \Solution.  The specification of the procedure
is now as follows:
\begin{Verbatim}
    {before: the sequence of printed vertices is correct}
    {after: the sequence of printed vertices is correct
            and includes i; all newly printed vertices
            can be reached from i}
\end{Verbatim}

\Problem{%
The correctness of the program depends on the assumption
about the absence of cycles. However, we did not mention
this assumption in the solution of problem 7.4.2. Why?
}

     \Solution. We
omitted the proof that the program terminates.
Let us give it now. For any vertex we define its {\it level\/}
as the maximal length of a path going out of it along the
edges. The level is finite because there are no cycles.
Vertices of level $0$ have no outgoing edges. For any edge the
level of its endpoint is smaller than the level of its starting
point by at least $1$. When \w{add(i)} is executed, all
recursive calls refer to vertices whose levels are smaller.\es

\medskip

Now we return to the time bound. How many calls \w{add(i)} are
possible for some fixed \w{i}? The first call prints \w{i}; all
others check that \w{i} is printed and exit immediately. It is
also clear that
all the
calls \w{add(i)} are induced by the {\it
first\/} calls of \w{add(j)} for all \w{j} such that the edge
from \w{j} to \w{i} is present in the graph. Therefore, the
number of calls \w{add(i)} is equal to the number of incoming
edges for vertex \w{i}. All the calls except the first one
require $O(1)$ time. The first requires time proportional to the
number of outgoing edges (if we ignore the time
needed to perform \w{add(j)} for endpoints of outgoing edges).
Therefore the total time is proportional to the total number of
edges (plus the number of vertices).\es

\medskip

    {\bf Connected component of a graph.} An {\it undirected graph\/}
is a set of points (vertices) some of which are connected by lines (edges).
An undirected graph is a special case of a directed
graph where for each edge there is another edge going in the
reverse direction.%
 \index{graph!undirected}
 \index{undirected graph}
 \index{graph!vertex}
 \index{vertex of a graph}
 \index{graph!edge}
 \index{edge of a graph}

     The {\it connected component\/} of vertex \w{i} is the set of
all vertices that are reachable from \w{i} via graph edges.
Since the graph is undirected, the relation ``\w{j} belongs to the
connected component of \w{i}'' is an
equivalence relation.%
 \index{graph!connected component}
 \index{connected component!undirected graph}

\Problem{%
Suppose
an undirected graph is given (for each vertex its neighbors
are listed; see the
problem about topological sorting for details).
Write an algorithm
that for a given \w{i} prints all the vertices of the
connected component of \w{i} (each vertex is printed once; no
other vertices should be printed). The number of operations
should be proportional to the total number of vertices and
edges in the connected component.}

     \Solution. The program will ``blacken'' vertices of the
graph as they are printed. (Initially the vertices are assumed
to be white.) By ``white part'' of the graph we mean that part
of the graph
that
remains after we remove all black vertices and
all
edges adjacent to
black vertices. The procedure \w{add(i)} blackens the connected
component of \w{i} {\it in the white part of the graph\/} (and
does nothing if \w{i} is already black).
 %
\begin{verbatim}
     procedure  add (i:1..n);
     begin
     | if i is black then begin
     | | {nothing to do}
     | end else begin
     | | ..print i and mark i as black
     | | for all j that are neighbors of i do begin
     | | | add(j);
     | | end;
     | end;
     end;
\end{verbatim}
Let us prove that this procedure works properly (assuming that
all recursive calls work properly). Indeed, it cannot blacken
anything except the connected component of \w{i}. Let us check
that all vertices in the connected component are blackened
(and printed).
Let
\w{k} be a vertex that is reachable from \w{x} via path
$\w{i}\rightarrow\w{j}\rightarrow\cdots\rightarrow\w{k}$, which
includes only white vertices (and goes along graph edges). We
may assume that this path does not visit vertex \w{i} again.
Among all the paths, we consider the path with the smallest
\w{j} (in the order they are considered in the \w{for}-loop).
Then
after
the calls \w{add(m)} for preceding neighbors \w{m}, no one of
the vertices in the path
$\w{j}\rightarrow\cdots\rightarrow\w{k}$
becomes
black
(otherwise, \w{j} is not minimal). Therefore, \w{k} belongs to
the connected component of the white part at the time
when
\w{add(j)}
is called.

     To prove that the algorithm
terminates, it is enough to mention that
the number of white vertices decreases at each recursion level.

     Let us estimate the number of operations. Each vertex is
blackened at most once, during the first call \w{add(i)} (for a
given \w{i}). All subsequent calls occur when one of the
neighbors is blackened. Therefore, the number of those calls is
limited by the number of neighbors. And the only thing that
happens during those calls is the check that \w{i} is already
black. On the other hand, during the first call to \w{add(i)} all
neighbors are considered and corresponding recursive calls are
made. Therefore the total number of operations related to vertex
\w{i} (not including the operations performed during the
recursive calls \w{add(j)} for its neighbors \w{j}) is proportional
to the number of neighbors of \w{i}. The
upper bound
stated in
the problem follows.\es

\Problem{%
     Solve the same problem for a directed graph (that is,
print all the
vertices accessible from a given vertex). Note: the graph may contain
cycles.}%
 \index{connected component!directed graph}

   \Solution. Essentially the same program can be used. The line
``for all neighbors of a vertex'' should be replaced by ``for
all endpoints of outgoing edges''.\es

\medskip

     {\bf Hoare Quicksort}. A well-known sorting
 \index{Hoare, C.A.R.}
algorithm called ``quicksort''
is a recursive algorithm considered to be one of the fastest
 \index{sorting!Hoare (quicksort)}%
 \index{Hoare sorting}%
 \index{quicksort algorithm}%
algorithms available. Assume that an array
$\w{a[1]..a[n]}$ is given. The recursive procedure
\w{sort(l,r:integer)} sorts an interval of the array \w{a} with
indices in $(\w{l},\w{r}]$, that is, $\w{a[l+1]..a[r]}$
(leaving the remaining part of the array unchanged).

\begin{verbatim}
 procedure sort (l,r: integer);
 begin
 | if l = r then begin
 | | {nothing to do - the interval is empty}
 | end else begin
 | | ..find a random number s in the interval (l,r]
 | | b := a[s]
 | | ..rearrange the elements of the interval (l,r]
 | |        into three parts:
 | |   the elements smaller than b  - the interval (l,ll]
 | |   the elements equal to b      - the interval (ll,rr]
 | |   then elements greater than b - the interval (rr,r]
 | | sort (l,ll);
 | | sort (rr,r);
 | end;
 end;
\end{verbatim}
How do we rearrange the elements of the interval according to the
three
categories listed in the above algorithm?
As problem 1.2.32
(p.~\pageref{separation}) shows, it can be done in
time proportional to the length of the interval. Termination
is guaranteed because the length of the interval
decreases by at least $1$ for each recursion level.

\Problem{%
     (For readers familiar with probability theory) Prove that
the expected number of operations of the Hoare quicksort
algorithm does not exceed $Cn\log n$ where the constant $C$ does
not depend on the array to be sorted.}

    \Hint{Let $T(n)$ be the maximal value of the expected
number of operations
(the maximal value is computed over all possible
inputs of length $n$). The following inequality holds:
$$
  T(n) \le  Cn + \frac{1}{n}\sum_{k+l=n-1} \bigl( T(k)+T(l) \bigr)
$$
Indeed, the first summand corresponds to the stage where all
elements are rearranged (divided into ``less than'',
``equal to'', or
``greater than'' parts).
The second summand is an
average value taken over all possible choices of a random
number \w{s}.
(To be precise, we should note that
some of the elements may be
equal to the threshold,
so instead of $T(k)$ and $T(l)$ we should use the maximum
of $T(x)$ over all $x$ not exceeding $k$ (or $l$), but this
makes no difference.)
Now, we prove by induction over $n$ that $T(n) \le
C'n\ln n$. To compute the average value of $x\ln x$ for all
integer $x$ such that $1\le x\le n-1$,
we integrate $\int_1^n x \ln x dx$ by parts as
$\int \ln x \, d(x^2)$. When $C'$ is large enough, the term $Cn$
on the right-hand side is absorbed by the integral $\int x^2\,
d\ln x$ and the inductive step is finished.}

\Problem{%
      An array of $n$ different integers and a number $k$ is
given. Find the $k$-th element of the array (in increasing
order) using at most $Cn$ operations (where $C$ is some constant
that does not depend on $k$ and $n$).}%
 \index{search!of the $k$-th element}%
 \index{median, search}%
 \label{median}

     Remark. Sorting algorithms can be used, but the number of
operations ($Cn\log n$) is too big. The na\"\i{}ve algorithm
(find the minimal element, then the next one,$\,\ldots$, then the
$k$-th one) requires about $kn$ operations (which is not
allowed, because the constant $C$ must not depend on $k$).

     \Hint{An elegant method (though hardly practical since the
constants are rather big) goes as follows:

     A. Separate the array into $n/5$ groups each containing
$5$ elements. Sort each group.

     B. Consider the median (third) elements of each group.
This gives an array of $n/5$ elements. Calling our algorithm
recursively, find the median element of this array. Call it $M$.

     C. Compare all other elements of the initial array with $M$.
They are divided into two groups (elements less than
$M$ and elements greater than $M$). Count the
elements in both groups.
Then we know  which category the required
($k$-th) element belongs to
and what its number is inside that
part.

     D. Apply the algorithm recursively to that part to find
the required element.

     Let $T(n)$ be the
maximal possible number of operations when
this algorithm is applied to arrays of length not exceeding
$n$ (the number $k$ may be arbitrary). We have the following bound:
$$
     T(n) \le
        Cn + T(n/5) + T (0.7n).
$$
The last term may be explained as follows. Each
of the three categories
contains at
least $0.3n$ elements. Indeed, about half of all the median
elements (in $5$-element sets) are
smaller than $M$. And if a
median element of a $5$-element set is smaller
than $M$, then at
least two more elements are smaller than $M$. Therefore, $3/5$
of half of all elements are smaller than $M$.

  Now the bound $T(n) \le Cn$ can be proved by induction. The
crucial point is that $1/5 + 0.7 < 1$. }

