\subsection{Stack of postponed tasks}

We illustrate  another way to eliminate recursion using the
Towers of Hanoi (p.~\pageref{hanoitowers})
problem as an example.

\Problem{%
     Write a nonrecursive program that prints the sequence
of moves for Towers of Hanoi
problem.}%
 \index{Hanoi, Towers of!nonrecursive solution}
 \index{Towers of Hanoi!nonrecursive solution}

     \Solution. Recall the following
recursive program that moves
\w{i} upper rings from stick \w{m} to stick~\w{n}:

\begin{verbatim}
  procedure move(i,m,n: integer);
  | var s: integer;
  begin
  | if i = 1 then begin
  | | writeln ('move ', m, '->', n);
\end{verbatim}
\clearpage
\begin{verbatim}
  | end else begin
  | | s:=6-m-n; {s is the unused stick; 1+2+3=6}
  | | move (i-1, m, s);
  | | writeln ('move ', m, '->', n);
  | | move (i-1, s, n);
  | end;
  end;
\end{verbatim}
This program reduces the task ``move \w{i} rings from \w{m} to
\w{n}'' to three tasks of the same type. Two of
them deal with \w{i-1} rings; one of them deals with \w{1} ring.

Try to execute this program manually. You'll see that it is rather
difficult to remember which tasks are still to be done on different
recursion levels.

The nonrecursive program uses
a {\it stack of postponed tasks},%
 \index{stack!of postponed tasks}
whose elements are triples $\langle\w{i},\w{m},\w{n}\rangle$.
Each triple is interpreted as the request ``move \w{i} (upper)
rings from stick \w{m} to stick
\w{n}''. Tasks must be performed in
the order they appear on the stack (the
request on the top of the stack
is the most urgent one).
We obtain the following program:
\begin{verbatim}
  procedure move(i,m,n: integer);
  begin
  | make the stack empty
  | put <i,m,n> into the stack
  | {invariant: it remains to process
  |                all the requests in the stack}
  | while stack is not empty begin
  | | take the top of the stack into <j,p,q>
  | | if j = 1 then begin
  | | | writeln ('move ', p, '->', q);
  | | end else begin
  | | | s:=6-p-q; {s is the third stick; 1+2+3=6}
  | | | put the triple <j-1,s,q> into the stack
  | | | put the triple <1,p,q>   into the stack
  | | | put the triple <j-1,p,s> into the stack
  | | end;
  | end;
  end;
\end{verbatim}
(Please note that the triple put on the stack first
will be the last request
processed.) The stack of triples may be implemented
as three separate stacks or
one
stack of records containing three
integers (using a \w{record} type in Pascal).\es

\Problem{%
      (Communicated via A.K.~Zvonkin%
 \index{Zvonkin, A.K.}
by Andrzei Lissowski.)%
 \index{Lissowski, Andrzei}
There are other nonrecursive solutions of the Towers of Hanoi
problem. Prove the correctness of the following solution: the {\it
unused\/} stick (the stick that is neither the source nor the
target of the move) should alternate cyclicly. (Another
rule: alternately move the smallest ring and another ring,
always
moving the smallest one clockwise.)\es}

\Problem{%
In the recursive program
that prints a decimal number (7.1.6),%
\index{decimal number!printing}
replace the recursion by a stack.}

    \Solution. The digits are generated from right to left and
put onto the stack. They are taken from the stack (in the
reverse order) and printed.\es

\Problem{%
     Write a nonrecursive program that prints all the vertices of
a binary tree.}%
 \index{tree!traversal, nonrecursive}

    \Solution. In this case, the stack of postponed tasks will
contain requests of two types: ``print a vertex'' and ``print all
the vertices of a subtree rooted at a given vertex''. (We
consider \w{nil} to be the root of an empty tree.) Therefore, the
stack element is a pair $\langle$request type, vertex
number$\rangle$.

     When an element is taken off the stack, we either
process it immediately (if it is a request of the first type)
or put onto the stack the three requests caused by it (in one of
six possible orderings).\es

\Problem{%
     What if we only want to count the number of vertices but
not print them?}%
 \index{tree!number of vertices}

     \Solution. Instead of printing a vertex, we add $1$
to a counter. In other words, the invariant is the following:
(total number of vertices) = (counter) $+$ (the total number of
vertices in the subtrees
whose roots are in the stack).\es

\Problem{%
     For some orderings (among six possible), the program that
prints all vertices may be simplified. Show these
simplifications.}%
 \index{tree!traversal}

    \Solution. If the order required is
\begin{center}
        root, left subtree, right subtree,
\end{center}
then a request to print the root may be processed immediately; thus
we do not need to put it onto the stack.

     A more complicated construction is necessary for the case
\begin{center}
        left subtree, root, right subtree.
\end{center}
In this case, all the requests in the stack (except the first
one, which requests to print some subtree) are grouped into
pairs
\begin{center}
    print vertex \w{x},
print ``right subtree'' of \w{x}
\end{center}
(that is, the subtree rooted at the right son of \w{x}). We can combine
such pairs into requests of a special type
and use an additional variable
for the first request; in this way,
all requests on the
stack are homogenuous (have
the same type).

     For the symmetric case, similar simplifications are
possible. Thus, for at least four of six possible orderings
the program may be simplified.\es

     Remark. Another program that prints all the tree vertices
is based on a program constructed in section~\ref{traversal}.
That program uses the command ``down'', which is not currently
provided
in the representation of trees. Therefore, we
must keep a list of all vertices from the root to
the current position (this list behaves like a stack).%
\label{twotraversals}

\Problem{%
     Write a nonrecursive version of Hoare's quicksort algorithm.
How do we guarantee that the size of the stack does not exceed
$C\log n$, where $n$ is the number of elements to be sorted?}%
 \index{Hoare, C.A.R.}%
 \index{sorting!quicksort, nonrecursive}%
 \index{quicksort algorithm!nonrecursive}%
 \index{Hoare sorting!nonrecursive}%

     \Solution. The stack is filled with pairs $\langle
i,j\rangle$, which are interpreted as requests to sort the
corresponding intervals of the array. All such intervals are
disjoint, therefore the size of the stack does not exceed $n$.
To insure that the size of the stack is logarithmic, we follow
the rule: ``a larger request is pushed onto the stack first''.
Let $f(n)$ be the maximal size of the stack that may appear when
sorting some array of length $n$ using this rule. We desire
an upper bound for $f(n)$. Indeed, after the array is split into
two fragments, the shorter one is sorted first (whereas the
request to sort the longer one is kept on the stack); then the
longer fragment is sorted. At the first stage, the size of the stack
does not exceed $f(n/2)+1$, and at the second stage it does not
exceed $f(n-1)$; therefore
$$
      f(n) \le \max \{f(n/2)+1, f(n-1)\}
$$
A simple induction argument gives $f(n)=O(\log n)$.\es

