
\subsection{Difficult cases}

     Let $f$ be a function with nonnegative integer arguments
and values defined by the equations
\begin{eqnarray*}
        f(0)&=&a, \\
        f(x)&=&h(x, f(l(x))) \quad (x>0)
\end{eqnarray*}
Here $a$ is some number while
$h$ and $l$ are known functions.
In other words, the value of $f$ at $x$ is determined by the
value of $f$ at $l(x)$. We assume that for any $x$, the sequence
$$
        x,\: l(x),\: l(l(x)),\ldots
$$
reaches~$0$.
     If we know in addition that $l(x) < x$ for all $x$, the
computation of $f$ is trivial; just compute $f(0), f(1),
f(2),\ldots$ sequentially.

\Problem{%
    Write a nonrecursive program to compute $f$ in the general
case.}

     \Solution. To compute $f(x)$, compute the sequence
$$
        l(x),\: l(l(x)),\: l(l(l(x))), \ldots
$$
until $0$ appears. Now compute the values of $f$
for all terms of this sequence, going from right
to left.\es

\smallskip

The next example involves
a more complicated case of recursion.
(This example is hardly practical, and if it did appear
in practice, it would probably be better to leave the
recursion as is.)

Assume that a function $f$ with nonnegative
integer arguments and values is defined by the equations
\begin{eqnarray*}
        f(0)&=&a, \\
        f(x)&=&h(x, f(l(x)), f(r(x))) \quad  (x>0),
\end{eqnarray*}
where $a$ is a constant,
 and $l$, $r$, $h$ are (known) functions.
We assume that if one starts from any nonnegative integer and
applies functions $l$ and $r$ in some arbitrary order, one
eventually gets~$0$.

\Problem{%
    Write a nonrecursive program to compute $f$.}

    \Solution. It is possible to
construct a tree that has $x$ at the root,
and has
$l(i)$ and $r(i)$ as sons of vertex $i$
(unless $i$ is equal to $0$, in which case it is a leaf).
Then we may compute the values of $f$  from the leaves
to the
root.
However, we'll use another approach.

By a {\it reverse Polish notation\/}%
 \index{reverse Polish notation}
 (or {\it postfix%
 \index{postfix notation}
notation\/}) we mean an expression
where the
function symbol is placed after all the arguments; parentheses
are not used. Here are several examples:
$$
\begin{array}{llllllll}
          f(2)               &  2&f& & & & &  \\
          f(g(2))            &  2&g&f& & & &  \\
          s(2,t(7))          &  2&7&t&s& & &  \\
          s(2, u(2, s(5,3))\qquad &  2&2&5&3&s&u&s
\end{array}
$$
Postfix notation allows us to compute the value of an
expression easily using a so-called {\it stack calculator}.%
\label{stackcalc}%
 \index{calculator!stack}%
 \index{stack calculator}
This calculator has a stack that we assume to be horizontal
(the top of the stack is on the right), as well as
number and
function keys.
When a number key is pressed, the number in question is put onto
the stack. When a function key is pressed, the corresponding function
is applied to the several arguments
(according to its arity) taken
from the stack. For example, if the
stack contains the numbers
$$
        2\ 3\ 4\ 5\ 6
$$
and the function key $s$ is pressed (we assume that $s$ is a
function of two arguments), the new content of the stack is
$$
        2\ 3\ 4\ s(5,6)
$$

Now let us return to our problem. The program employs a stack
whose elements are nonnegative integers. It also uses a
sequence of numbers and the symbols \w{f}, \w{l}, \w{r}, \w{h}
(which we consider a sequence of keys on a stack calculator).
The invariant relation:
\begin{quote}
      If the number stack represents the current state of a stack
      calculator and we press all the keys in the sequence,
      the stack contains only one number that is the
      required answer.
\end{quote}
Suppose we want to compute $f(x)$. Put the number
$x$ onto a stack and consider a
sequence that contains only one symbol
\w{f}. (The invariant relation is true.) Then the stack and the
sequence are subjected to the following transformations:
$$
\setlength{\arraycolsep}{.8em}
\begin{array}{llll}

\r{old} & \r{old}        & \r{new}  & \r{new}               \\
\r{stack}   & \r{sequence}&\r{stack} &\r{sequence} \\[1.5ex]
%
  X\    & x\ P         & X\ x    &      P                \\
  X\ x  & \w{l}\ P     & X\ l(x) &      P                \\
  X\ x  & \w{r}\ P     & X\ r(x) &      P                 \\
  X\ x\ y\ z& \w{h}\ P & X\ h(x,y,z) &  P             \\
  X\ 0  & \w{f}\ P     & X\ a    &      P               \\
  X\ x  & \w{f}\ P     & X  & x\ x\ \w{l}\ \w{f}\ x\ \w{r}\ \w{f}\ \w{h}\ P
\end{array}
$$
Here $x$, $y$, $z$ are numbers, $X$ is a sequence of numbers,
and $P$ is a sequence of numbers and the symbols \w{f}, \w{l}, \w{r},
\w{h}. In the last line, we assume that $x\ne 0$. This line
corresponds to the equation
$$
        f(x) = h(x, f(l(x)), f(r(x)))
$$
in postfix notation.

The transformations are performed until the sequence is
empty. At that moment the invariant relation guarantees that the
stack contains only one number, and this number is
the answer required.

     Remark. The sequence may be considered as a stack of delayed
tasks (whose top is on the left).\es

