\section{Graph algorithms}\label{graphs}

\subsection{Shortest paths}

     This section is devoted to different versions of
one problem.
Suppose a country has \w{n} cities numbered
$\w{1..n}$. For each pair of cities \w{i} and \w{j},
an integer
\w{a[i][j]} is given that is the cost of a (nonstop)
plane ticket
from
\w{i} to \w{j}. We assume that flights exist between any two cities,
and
that $\w{a[k][k]}=\w{0}$ for any \w{k}. In general,
\w{a[i][j]} may be different from \w{a[j][i]}.
Our goal is to find the
minimal cost of a trip
from one city (\w{s}) to another one (\w{t}) that
takes into account all the possible travel plans
(nonstop, one stop, two stops etc.). This minimal cost does not
exceed
\w{a[s][t]} but may be smaller.%
 \index{graph!shortest paths}%
 \index{search!of a shortest path}
We allow \w{a[i][j]} to be negative for some \w{i}
and \w{j} (you are paid if you agree to use some flight).

    In the following problems, we compute the minimal cost
for some pairs of cities,
but first we have to check that our definition is correct.

\Problem{%
     Assume there is no cyclic travel plan
with negative
total cost. Prove that in this case a travel plan with minimal
cost exists.}

    \Solution. If a travel plan is long enough
(includes more than \w{n}
cities), it has a cycle, which may be omitted (because of our
assumption). Now there are only a finite number of travel plans
involving \w{n} or fewer cities.\es

\smallskip

Throughout the rest of this subsection,
we assume that this
condition (absence of negative cycles) is satisfied. (It is
evident if all edge costs
are nonnegative, but the latter condition is not always
imposed.)

\Problem{%
     Find the minimal travel cost from the first city to all other
cities in time $O(\w{n}^3)$.}%
 \index{Ford--Bellman algorithm}%
 \index{dynamic programming!shortest path}

    \Solution. By \w{MinCost(1,s,k)} we denote the minimal travel cost
from \w{1} to \w{s} with less than \w{k}
stops. It is easy to see that
$\w{MinCost}(\w{1},\w{s},\w{k+1})$ is equal to
$$
   \min \Bigl\{\w{MinCost(1,s,k)},
      \min_{\w{i=1..n}}\{\w{MinCost(1,i,k)}+\w{a[i]}\!\w{[s]}\}\Bigr\}
$$
The minimum on the right-hand side is taken over
all possible places of the last stop before the final
destination.

As we have seen in the solution of the preceding problem,
the cycles can be eliminated, so the
answer in question is \w{MinCost(1,i,n)} for all
$\w{i}=\w{1..n}$.
We get the following program:
\begin{Verbatim}
     k:= 1;
     for i := 1 to n do begin x[i] := a[1][i]; end;
     {invariant: x[i] = MinCost(1,i,k)}
     while k <> n do begin
     | for s := 1 to n do begin
     | | y[s] := x[s];
     | | for i := 1 to n do begin
     | | | if y[s] > x[i]+a[i][s] then begin
     | | | | y[s] := x[i]+a[i][s];
     | | | end;
     | | end
     | | {y[s] = MinCost(1,s,k+1)}
     | end;
     | for i := 1 to n do begin x[s] := y[s]; end;
     | k := k + 1;
     end;
\end{Verbatim}
This algorithm is called the dynamic programming algorithm, or
Ford--Bellman algorithm.

\Problem{%
Prove that the algorithm remains correct if the array \w{y} is not
used, that is, if all changes are made in array \w{x}
(just replace all \w{y}'s by \w{x}'s and delete redundant lines).}

     \Solution. In this case the invariant is
$$
   \w{MinCost}\,\w{(1,i,n)} \le \w{x[i]} \le \w{MinCost}\,\w{(1,i,k)}.
   \eqno\es
$$

\bigskip

This algorithm may be improved at least in two ways.
First, with
the same running time $O(\w{n}^3)$, we can find the minimal
travel cost $\w{i}\rightarrow\w{j}$ for {\it all\/} pairs
\w{i},\w{j} (not just $\w{i}=\w{1}$). Second, we can compute
all travel costs from a given vertex in time $O(\w{n}^2)$. (In
the latter case, however, we require all flight costs \w{a[i][j]}
to be nonnegative.)

\Problem{%
     Find the minimal travel costs $\w{i}\rightarrow\w{j}$
for all \w{i},\w{j}
in time $O(\w{n}^3)$.}%
 \index{Floyd algorithm}\label{floyd}

     \Solution. For any $\w{k} = \w{0..n}$ consider
the minimal travel cost from \w{i} to \w{j} {\it assuming
intermediate
stops are allowed only in cities $\w{1..k}$.}
This cost is denoted by
\w{A(i,j,k)}.
Then
\begin{eqnarray*}
    \w{A(i,j,0)} &=& \w{a[i][j]},\\
    \w{A(i,j,k+1)} &=&
    \min\Bigl\{\w{A(i,j,k)},\;\w{A(i,k+1,k)} +\w{A(k+1,j,k)}\Bigr\}
\end{eqnarray*}
(we either ignore city \w{k+1} or use it as an intermediate
stop;
there is no reason to visit it twice).\es

     This algorithm is called the Floyd algorithm.

\Problem{%
Assume
all costs \w{a[i][j]} are nonnegative. Find the minimal travel
cost $\w{1}\rightarrow\w{i}$ for all $\w{i}
=\w{1..n}$ in time $O(\w{n}^2)$.}%
 \index{Dijkstra algorithm (shortest path)}

     \Solution. Our algorithm will {\it mark\/} cities during its
operation. Initially, only city number \w{1} is marked. Finally,
all cities are marked. For all the cities, a ``current cost'' is
maintained. This cost is
a
number whose meaning is explained by the following
invariant relation:
\begin{itemize}

\item for any marked city \w{i}, the current cost is the minimal
cost of travel $\w{1}\rightarrow\w{i}$; it is guaranteed that
this minimal cost is obtained via a path through
marked cities only;

\item for any non-marked city \w{i}, the current cost is the
minimal cost among all travel plans $\w{1}\rightarrow\w{i}$
such that all intermediate stops are marked.

\end{itemize}

     The set of marked cities is extended using the following
observation:
\begin{quote}
For a non-marked city with minimal current cost (among
all non-marked cities), the current cost is the true cost and is
reached via a path going through marked cities only.
\end{quote}

Let us prove this. Assume that a shorter path exists. Consider
the first non-marked city along this path. Even if we stop the
trip in that city, the cost is already greater! (All
costs are nonnegative.)

     When a city is selected in this way, the algorithm marks it.
To maintain
the invariant, we update the current cost for
non-marked cities. It is enough to take into account only those
paths where the newly marked city is the last intermediate stop.
This is easy to do since the minimal travel cost from the
starting point to the newly marked city is already known.

     If we store the set of marked cities in a Boolean array,
we need $O(\w{n})$ operations per city.\es

     This algorithm is called the Dijkstra algorithm.

\medskip

     The problem of finding the shortest path has a natural
interpretation in terms of matrices.%
 \index{matrix product}%
 \index{cost matrix}
Assume that $A$ is the cost matrix for some carrier and $B$ is
the cost matrix for another carrier. Suppose we want to
make one stop along the way, using the first carrier (with matrix
$A$) for the first flight and the second carrier ($B$) for the
second flight. How much should we pay for the trip from \w{i}
to \w{j}?

\Problem{%
     Prove that the costs mentioned above form a matrix that can
be computed using a formula similar to the standard formula
for matrix multiplication. The only difference is
that
the sum is
replaced by a  $\min$ operation and
the
product is replaced by
a
sum:
 $$
C_{ij}=\min\limits_{k=1,\ldots,n} \{A_{ik} + B_{kj}\}\eqno\es
 $$
}


\Problem{%
     Prove that matrix ``multiplication'' defined by the
preceding
formula
is associative.\es}

\Problem{%
     Prove that finding the shortest
paths for all pairs of cities
is equivalent to
computation of $A^\infty$ for the cost matrix $A$ in the
following sense. For the sequence $A, A^2, A^3,\ldots$
there exists an $N$ such that all elements $A^N, A^{N+1}$, etc.
are equal to the matrix whose elements are minimal travel costs.
(We assume that there are no cycles with negative cost.)\es}

\Problem{%
     How large should $N$ be in the preceding problem?\es}

    The usual (unmodified) matrix multiplication may also be
applied, but in a different situation. Let us
assume that only some flights exist and let \w{a[i][j]} be equal
to \w{1} if there is a (direct) flight from \w{i} to \w{j};
otherwise, \w{a[i][j]=0}.
Compute the \w{k}-th power
of the matrix \w{a} (in the usual sense) and consider its
\w{(i,j)}-th element.

\Problem{%
     What is the meaning of this element?}

    \Solution. It is the number of different travel plans
from
\w{i} to \w{j} using \w{k} flights (and \w{k-1} intermediate
stops).\es
 \index{paths, number of}

\smallskip

     Let us return to our original problem (finding the shortest
path). We can easily extend our algorithms to the case
where not all pairs of cities are
connected by direct flights. Indeed, we may assume that
nonexisting flights are infinitely expensive (or just very
expensive), so our algorithms may be applied in this case too.
However, a new question arises. The number of actual flights may
be much smaller than $n^2$, so it is of interest to find
algorithms that are more effective in this special case.
First, we change the
representation of the initial data: for each
city we keep the number of outgoing flights and an array
containing
the destination points
and costs.

\Problem{%
     Prove that the Dijkstra algorithm may be modified in such a way
that if the number of cities is $n$ and the total number of
flights is $m$, then no more that $C(n+m)\log n$
operations are required.}%
 \index{Dijkstra algorithm (shortest path)}

    \Hint{What should we do at each step? We must choose
a non-marked city with minimal current cost and update the
data for
all cities that can be reached by direct flight from this city.
If there were an oracle to inform us which of the unmarked
cities has minimal current cost, $C(n+m)$ operations would
be enough. And an additional $\log n$-factor in the running time
allows us to
maintain the information needed to find the minimal value
in the array (see the problem on p.~\pageref{minarray}).}


